#ifndef lint
static const char Rcs_Id[] =
    "$Id: fields.cc,v 1.3 1998/10/23 23:03:22 geoff Exp $";
#endif

/*
 * $Log: fields.cc,v $
 * Revision 1.3  1998/10/23 23:03:22  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.2  1996/03/19  18:02:17  geoff
 * When parsing, make sure lineLength correctly reflects whether the
 * original line had a newline.  Declare the RCS ID as const.
 *
 * Revision 1.1  1996/02/14  19:16:38  geoff
 * Initial revision
 *
 */

#include <iostream.h>
#include <string.h>
#include "fields.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			ParsedFields::ParsedFields(istream& stream,
			  const char* delimiters, unsigned int flags,
			  unsigned int maxFields);
			ParsedFields::ParsedFields(const char* string,
			  const char* delimiters, unsigned int flags,
			  unsigned int maxFields);
ostream&		ParsedFields::write(ostream& stream, char delimiter)
			  const;
				// Write a line with fields to a stream
void			ParsedFields::parse(const char* delimiters,
			  unsigned int flags, unsigned int maxFields);
					// Parse buffer into fields
unsigned int		ParsedFields::processBackslash(const char* string,
			  char** outputPointer, bool strip);
					// Process backslash sequences
#endif // 0

unsigned int	ParsedFields::fieldIncrement = 20;
				// Increment for allocating fields
unsigned int	ParsedFields::lineIncrement = 512;
				// Increment for allocating lines

/*
 * Create a parsed-field data structure from an input stream.  Reads
 * one line of the stream into a buffer, breaks it up into fields, and
 * inserts them into the constructed class element.
 */
ParsedFields::ParsedFields(
    istream&		stream,		// Stream to read the line from
    const char*		delimiters,	// Characters to use for delimiters
    unsigned int	flags,		// Option flags;  see fields.hh
    unsigned int	maxFields)	// Maximum number of fields to parse
    : numFields(0),
      lineBuffer(NULL),
      fields(NULL),
      hadNewline(false),
      valid(false)
    {
    lineBuffer = new char[lineIncrement];
    if (lineBuffer == NULL)
	return;
    unsigned int maxLineSize = lineIncrement;
    unsigned int currentLineSize = 0;
    /*
     * Read in the line, expanding as necessary.
     */
    while (stream.get(&lineBuffer[currentLineSize],
      maxLineSize - currentLineSize))
	{
	currentLineSize += strlen (&lineBuffer[currentLineSize]);
	char terminator;
	if (stream  &&  stream.get(terminator))
	    {
	    if (terminator == '\n')
		{
		hadNewline = true;
		break;
		}
	    maxLineSize += lineIncrement;
	    char* newlineBuffer = new char[maxLineSize];
	    if (newlineBuffer == NULL)
		{
		delete lineBuffer;
		lineBuffer = NULL;
		return;
		}
	    (void)strcpy(newlineBuffer, lineBuffer);
	    delete lineBuffer;
	    lineBuffer = newlineBuffer;
	    lineBuffer[currentLineSize++] = terminator;
	    lineBuffer[currentLineSize] = '\0';
	    }
	else
	    break;
	}
    if (currentLineSize == 0)
	{
	delete lineBuffer;
	lineBuffer = NULL;
	return;
	}
    parse(delimiters, flags, maxFields);
    }

/*
 * Create a parsed-field data structure from a string.  Breaks the
 * string up into fields and inserts them into the constructed class
 * element.
 */
ParsedFields::ParsedFields(
    const char*		string,		// String to be parsed
    const char*		delimiters,	// Characters to use for delimiters
    unsigned int	flags,		// Option flags;  see fields.hh
    unsigned int	maxFields)	// Maximum number of fields to parse
    : numFields(0),
      fields(NULL),
      hadNewline(false),
      valid(false)
    {
    unsigned int lineSize = strlen(string);
    lineBuffer = new char[lineSize + 1];
    if (lineBuffer == NULL)
	return;
    (void)strcpy(lineBuffer, string);
    if (string[lineSize - 1] == '\n')
	{
	hadNewline = true;
	lineBuffer[--lineSize] = '\0';
	}
    parse(delimiters, flags, maxFields);
    }

/*
 * Write a parsed-fields structure back to a stream.
 */
ostream& ParsedFields::write(
    ostream&		stream,		// Stream to write to
    char		delimiter)	// Delimiter to put between fields
      const
    {
    for (unsigned int fieldNumber = 0;
      fieldNumber < numFields;
      fieldNumber++)
	{
	if (fieldNumber != 0)
	    stream << delimiter;
	stream << fields[fieldNumber];
	}
    if (hadNewline)
	stream << '\n';
    return stream;
    }

/*
 * Parse the lineBuffer into individual fields, as controlled by the
 * arguments.
 *
 * NEEDSWORK: document argument effects; steal from fields.3 man page.
 */
void ParsedFields::parse(
    const char*		delimiters,	// Delimiters that separate fields
    unsigned int	flags,		// Option flags; see fields.hh
    unsigned int	maxFields)	// Maximum number of fields to parse
					// ..or zero for no limit
    {
    lineLength = strlen(lineBuffer);
    if (hadNewline)
	lineLength++;
    numFields = 0;
    valid = false;
    unsigned int arraySize = (maxFields != 0  &&  maxFields < fieldIncrement)
      ? maxFields + 2 : fieldIncrement;
    fields = new char*[arraySize];
    if (fields == NULL)
	return;
    if ((flags & (SHELL_QUOTES | SINGLE_QUOTES | BACK_QUOTES | DOUBLE_QUOTES))
      == SHELL_QUOTES)
	flags |= SINGLE_QUOTES | BACK_QUOTES | DOUBLE_QUOTES;
    char* nextCharacter = lineBuffer;
    while (1)
	{
	if (flags & RUNS)
	    {
	    while (*nextCharacter != '\0'
	      &&  strchr(delimiters, *nextCharacter) != NULL)
		nextCharacter++;	// Skip runs of delimiters
	    if (*nextCharacter == '\0')
		break;
	    }
	char* outputPointer = nextCharacter;
	fields[numFields] = nextCharacter;
	/*
	 * Skip to the next delimiter.  At the end of skipping,
	 * "nextCharacter" will point to either a delimiter or a null
	 * byte.
	 */
	if (flags
	  & (SHELL_QUOTES | SINGLE_QUOTES | BACK_QUOTES
	    | DOUBLE_QUOTES | BACKSLASH))
	    {
	    while (*nextCharacter != '\0')
		{
		char quote;		// Character used for quoting, if any
		if (strchr(delimiters, *nextCharacter) != NULL)
		    break;
		else if (((flags & SINGLE_QUOTES)  &&  *nextCharacter == '\'')
		  ||  ((flags & BACK_QUOTES)  &&  *nextCharacter == '`')
		  ||  ((flags & DOUBLE_QUOTES)  &&  *nextCharacter == '"'))
		    {
		    if ((flags & SHELL_QUOTES) == 0
		      &&  nextCharacter != fields[numFields])
			quote = '\0';
		    else
			quote = *nextCharacter;
		    }
		else
		    quote = '\0';
		if (quote == '\0')
		    {
		    if (*nextCharacter == '\\'  &&  (flags & BACKSLASH))
			{
			nextCharacter++;
			if (*nextCharacter == '\0')
			    break;
			nextCharacter += processBackslash(nextCharacter,
			  &outputPointer, flags & STRIP_QUOTES);
			}
		    else
			*outputPointer++ = *nextCharacter++;
		    }
		else
		    {
		    // Process quoted string
		    if ((flags & STRIP_QUOTES) == 0)
			*outputPointer++ = quote;
		    ++nextCharacter;
		    while (*nextCharacter != '\0')
			{
			if (*nextCharacter == quote)
			    {
			    if ((flags & STRIP_QUOTES) == 0)
				*outputPointer++ = quote;
			    nextCharacter++;		// Go on past quote
			    if ((flags & SHELL_QUOTES) == 0)
				{
				while (*nextCharacter != '\0'
				  &&  strchr (delimiters, *nextCharacter)
				      == NULL)
				    nextCharacter++;	// Skip to delimiter
				}
			    break;
			    }
			else if (*nextCharacter == '\\')
			    {
			    if (flags & BACKSLASH)
				{
				nextCharacter++;
				if (*nextCharacter == '\0')
				    break;
				else
				    nextCharacter +=
				      processBackslash(nextCharacter,
					&outputPointer, flags & STRIP_QUOTES);
				}
			    else
				{
				*outputPointer++ = '\\';
				if (*++nextCharacter == '\0')
				    break;
				*outputPointer++ = *nextCharacter;
				}
			    }
			else
			    *outputPointer++ = *nextCharacter++;
			}
		    }
		}
	    }
	else
	    {
	    while (*nextCharacter != '\0'
	      &&  strchr(delimiters, *nextCharacter) == NULL)
		nextCharacter++;			// Skip to delimiter
	    outputPointer = nextCharacter;
	    }
	numFields++;
	if (*nextCharacter++ == '\0')
	    break;
	if (maxFields != 0  &&  numFields > maxFields)
	    break;
	*outputPointer = '\0';
	if (numFields >= arraySize)
	    {
	    arraySize += fieldIncrement;
	    char** newFields = new char*[arraySize];
	    if (newFields == NULL)
		return;
	    (void)memcpy((char*)newFields, (char*)fields,
	      numFields * sizeof(char*));
	    delete fields;
	    fields = newFields;
	    }
	}
    fields[numFields] = NULL;
    valid = true;
    return;
    }

/*
 * Process or copy a backslash sequence.  Called after the backslash
 * has already been processed.  "string" points to the first character
 * following the backslash.  "outputPointer" points to where to store
 * the result of processing, and is updated upon return.
 *
 * If "strip" is false, a backslash is stored at the beginning of the
 * output string (to compensate for the already-processed one), and
 * then the backslash-initiated sequence is copied verbatim.  The
 * routine always copies the correct number of characters, regardless
 * of whether the sequence is a backslash-escaped special character, a
 * C-style backslash sequence, or an octal or hexadecimal character
 * code.
 *
 * If "strip" is true, the backslash-initiated sequence is converted
 * to a single ASCII character which is stored in the output string.
 *
 * In any case, returns the number of characters in "string" which
 * have already been processed.
 */
unsigned int ParsedFields::processBackslash(
    const char*		string,		// String to be processed
    char**		outputPointer,	// IN OUT: Where to store result
    bool		strip)		// True to actually convert sequence
    {
    if (!strip)
	{
	*(*outputPointer)++ = '\\';
	if (*string != 'x'
	  &&  *string != 'X'  &&  (*string < '0'  ||  *string > '7'))
	    {
	    *(*outputPointer)++ = *string;
	    return *string != '\0';
	    }
	}
    const char* stringStart = string;
    char outputCharacter;
    switch (*string)
	{
	case '\0':
	    {
	    *(*outputPointer)++ = '\0';
	    return 0;
	    }
	case 'a':
	    {
	    *(*outputPointer)++ = '\007';
	    return 1;
	    }
	case 'b':
	    {
	    *(*outputPointer)++ = '\b';
	    return 1;
	    }
	case 'f':
	    {
	    *(*outputPointer)++ = '\f';
	    return 1;
	    }
	case 'n':
	    {
	    *(*outputPointer)++ = '\n';
	    return 1;
	    }
	case 'r':
	    {
	    *(*outputPointer)++ = '\r';
	    return 1;
	    }
	case 'v':
	    {
	    *(*outputPointer)++ = '\v';
	    return 1;
	    }
	case 'X':
	case 'x':
	    {
	    // Hexadecimal sequence
	    outputCharacter = 0;
	    if (*string >= '0'  &&  *string <= '9')
		outputCharacter = *string++ - '0';
	    else if (*string >= 'a'  &&  *string <= 'f')
		outputCharacter = *string++ - 'a' + 0xa;
	    else if (*string >= 'A'  &&  *string <= 'F')
		outputCharacter = *string++ - 'A' + 0xa;
	    if (*string >= '0'  &&  *string <= '9')
		outputCharacter = (outputCharacter << 4) | (*string++ - '0');
	    else if (*string >= 'a'  &&  *string <= 'f')
		outputCharacter =
		  (outputCharacter << 4) | (*string++ - 'a' + 0xa);
	    else if (*string >= 'A'  &&  *string <= 'F')
		outputCharacter =
		  (outputCharacter << 4) | (*string++ - 'A' + 0xa);
	    break;
	    }
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	    {
	    // Octal sequence
	    stringStart = string;
	    outputCharacter = *string++ - '0';
	    if (*string >= '0'  &&  *string <= '7')
		outputCharacter = (outputCharacter << 3) | (*string++ - '0');
	    if (*string >= '0'  &&  *string <= '7')
		outputCharacter = (outputCharacter << 3) | (*string++ - '0');
	    break;
	    }
	default:
	    *(*outputPointer)++ = *string;
	    return 1;
	}
    unsigned int size = string - stringStart;
    if (strip)
	{
	*(*outputPointer)++ = outputCharacter;
	return size;
	}
    else
	{
	(void)memcpy(*outputPointer, stringStart, size);
	*outputPointer += size;
	return size;
	}
    }
