#ifndef lint
static const char Rcs_Id[] =
    "$Id: hashtable.cc,v 1.13 1999/11/22 05:46:29 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: hashtable.cc,v $
 * Revision 1.13  1999/11/22 05:46:29  geoff
 * Get rid of small table sizes, and add one intermediate one
 *
 * Revision 1.12  1999/11/17 07:56:58  geoff
 * Expand the list of prime numbers to go up to around 2**31.
 *
 * Revision 1.11  1998/10/23 23:03:24  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.10  1998/10/13 05:57:29  geoff
 * Add HASH_CHECK_BUCKETS and HASH_CHECK_CHAIN.
 *
 * Revision 1.9  1998/03/05 03:41:30  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.8  1998/01/31 06:30:23  geoff
 * Make using the type allocator the default.  Panic on errors, rather
 * than just aborting.
 *
 * Revision 1.7  1997/07/11 17:50:53  geoff
 * Add the BUZ (Bob Uzgalis) hashing algorithm.
 *
 * Revision 1.6  1996/09/11 23:07:55  geoff
 * Use a type-specific allocator.
 *
 * Revision 1.5  1996/05/03  01:31:36  geoff
 * Update the copyright.
 *
 * Revision 1.4  1996/02/02  18:55:44  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.3  1995/10/10  19:39:44  geoff
 * Get rid of the obsolete memory-leak-tracking stuff.
 *
 * Revision 1.2  1995/09/15  00:12:57  geoff
 * Add abortive memory-leak-catching support.  Fix the iterator so that
 * removeCurrent doesn't require special caller behavior.
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

/*
 * Generalized hash-table package.
 */

#include <iostream.h>
#include "hashtable.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
		HashTable::HashTable(unsigned int initialSize,
		  unsigned int (*hashFunction)(const DATA_TYPE* key),
		  int (*compareFunction)(const DATA_TYPE* ent1,
		    const DATA_TYPE* ent2));
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
		HashTable::~HashTable();
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
bool		HashTable::expand(unsigned int newSize);
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
DATA_TYPE*	HashTable::find(DATA_TYPE* datum, action notFoundAction);
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
bool		HashTable::remove(DATA_TYPE* datum);
					// Remove entry, returns true if not
					// ..found
void		HashTable::showStatistics(ostream& stream) const;
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
void		HashTable::Iterator::removeCurrent();
#endif // 0
unsigned int	hashString(const char* key);
unsigned int	hashPJW(const char* key);
unsigned int	hashInteger(int key);
unsigned int	hashStringBUZ(const char* key);
unsigned int	hashIntegerBUZ(int key);
static unsigned int hashBUZ(const unsigned char* key, int keyLength);

const int	HASHSHIFT = 5;	// How much to shift hash by per char
const int	MAXPCT = 70;	// How full to let hash table get

/*
 * Hash table sizes.  Prime is probably a good idea, though in truth I
 * whipped the algorithm up on the spot rather than looking it up, so
 * who knows what's really best?  If we overflow the table, we just
 * use a double-and-add-1 algorithm.
 */
static unsigned int goodSizes[] =
    {
    223, 907, 3631, 14591, 29191, 58391, 116789, 233591, 467183,
    934387, 1868777, 3737557, 7475129, 14950261, 29900531, 59801111,
    119602237, 239204477, 478408967, 956817949, 1913635961,
    3827271937u
    };

/*
 * Constants for the BUZ algorithm.  BUZ_INIT is the initial value of
 * the hash function.  The lookup table is used to widen and randomize
 * the bytes of the key.  All values were gotten from /dev/random on a
 * Linux box.
 */
const unsigned int BUZINIT = 0x7b4402a2;
static unsigned int BUZ_Table[256] =
    {
    0xc70bb269, 0x13299943, 0xe9cee5e1, 0x32119a79,
    0xc2365851, 0x169ee8e3, 0x094c1ed8, 0x58e1d4e0,
    0x96eb1762, 0xfe296797, 0x89082f47, 0x27d5078d,
    0x8ebb9de0, 0x14df49e0, 0x38d21c63, 0x1f5b2770,
    0xaa6b0150, 0x7b0b656e, 0x98d37bc2, 0x4d7e85ec,
    0x987910e2, 0xb8cbac89, 0xa3f664a3, 0xeca3003c,
    0x7b364137, 0xb2a6edae, 0x5ef72906, 0x64a9e7b7,
    0x28cd6520, 0xca3c72df, 0x57398ce2, 0x8db893d4,
    0x0a5995cd, 0x2d109fb4, 0x0491162f, 0xb3488737,
    0x6fc4eb03, 0x9903cb21, 0xe82ff831, 0xb03ff8e5,
    0x054836ca, 0x5335e6f8, 0x01396c2a, 0xf9000899,
    0x03ed9d63, 0x2bf6946b, 0x9097fa8b, 0xacd8dfc7,
    0x8488b8a6, 0x0e39cd2e, 0xac1a4517, 0xcd49e035,
    0xe98b7e7b, 0xd3571502, 0xd602805e, 0xe7143cfe,
    0x46db0a6b, 0x0a4c9ebe, 0x4e2e1ca7, 0x3040fc62,
    0xe8818c02, 0x37155e7b, 0xe44ba138, 0x43cacdd4,
    0x53d986ba, 0xdd4dca35, 0x0f680f71, 0x6c1a551e,
    0x74263e95, 0xcfc4f5d5, 0x37b8ef45, 0xc00ac71d,
    0x3b059e0d, 0x208bc754, 0x41335fbe, 0x785a0ffc,
    0x189f024f, 0xd669c2d8, 0xe1b20f87, 0xba2550da,
    0x10167369, 0x85fad38f, 0x97d20e4e, 0x5bc0da5e,
    0x80799570, 0x93eb4058, 0x139042a6, 0x40b34bf6,
    0x15c21dfa, 0x8f852660, 0xa3d20fb3, 0x3d175cf9,
    0x792441a8, 0xdc5e71b5, 0x925f6350, 0x66e8d08b,
    0xc4606b59, 0x85d8b88c, 0x1ea4f459, 0x664f62bc,
    0x77407de3, 0x73d158ca, 0xb76ab172, 0xe9ed1aeb,
    0x93dc2009, 0xeb9da6ac, 0x3d26cf05, 0x675132bc,
    0xc29196fe, 0x2a62486f, 0x914e75e1, 0xa1c31883,
    0x1c28291c, 0xc73c668c, 0xf4ac07e6, 0x87c9a9ac,
    0xb7196ea7, 0x67cb7fa2, 0x55987797, 0x29ce38ea,
    0x427361b3, 0x5b5667a6, 0x68a72fb0, 0xcef8235a,
    0xd06e8f5b, 0x4d3633f5, 0x214d3a19, 0xbd09ec15,
    0x5c61c24b, 0x3928573c, 0x26083ab8, 0x857a5dee,
    0x3203e50c, 0x52a1a713, 0xa8270ee2, 0xdfb643a9,
    0x7797c1f3, 0x0f8ddc9f, 0x9368de21, 0x638ebd4e,
    0xd91808d7, 0x28ce69b8, 0xe424b0ce, 0xfe52fdef,
    0x89126c74, 0xdb5f3d91, 0xba488f47, 0x2b15cdb8,
    0xa517b0f9, 0x53950632, 0x1159546a, 0xe50f65a3,
    0x5f26b5d1, 0x68a3a955, 0xc2b78ea1, 0x49c33701,
    0x45457aee, 0xd49b550a, 0x244379b8, 0xec826af5,
    0x4fa6e0c9, 0xd4633425, 0x82f0bd85, 0xc23ccc2e,
    0xac73e11a, 0xdc94b283, 0x13e59bb2, 0x23b4880e,
    0x1d295c45, 0xef67488c, 0x6b74149e, 0xdf90d4ac,
    0xfc6e65a9, 0x406a3734, 0x86999303, 0xc73e7180,
    0x3c8a0b31, 0x75fa9249, 0xaca5e0e2, 0x4d0cc60d,
    0x4b174606, 0x836fb602, 0x4f9fc83a, 0xe16477a7,
    0xda1506a9, 0x905b28b7, 0x4229f5c2, 0xdf4c9144,
    0x731888e9, 0x2e37421f, 0x0c67c385, 0x44a2e520,
    0xfe3ee655, 0x92547582, 0x9525f4d4, 0xdaa8caf8,
    0xa25bd583, 0x0e315733, 0xd35fea29, 0xc9cfaa0f,
    0xc6bdf7e9, 0xa48b4e01, 0xfd30ffe0, 0xd0f63421,
    0x2b84803d, 0xe1b368a4, 0xbae5daa5, 0x4dd6336a,
    0xb60c4030, 0x7bb552f1, 0xf1b91481, 0xc8929b82,
    0xa1c22bf8, 0xd585aae4, 0x17fb4f6a, 0xa9c0d32e,
    0x2036f9b5, 0x3a95d611, 0xd554f25d, 0x3441153f,
    0x7fa89ff4, 0x8f91241c, 0x4b2cc5a9, 0xd3035a00,
    0x5c80707f, 0xe610fd47, 0xf60958c2, 0xb55a6fe7,
    0x3e1ba335, 0x2dead082, 0x1a8877e8, 0xd0791aad,
    0x9706ee52, 0xfb1dc525, 0x7fa1ba54, 0x6a3e9f81,
    0xa85a906a, 0x86ce1b46, 0x3b05833e, 0xc8d8fdcb,
    0x44e606c5, 0x8807beb2, 0xe46047d3, 0x85b9f5f8,
    0x56ed0cba, 0x3cf4e646, 0x970fd9dd, 0xd0600895,
    0x2d0a5f92, 0x891c4220, 0x017fbfe0, 0x4dde2016,
    0x3a6e421d, 0x7e3f2285, 0xdf7956e3, 0x52fdcf83
    };

TypeAllocator		HashTable::allocator(sizeof(HashTable::Entry), 5000);
#ifdef HASH_DEBUG_LEAKS
unsigned long		HashTable::globalElementCreations = 0;
					// No. of elements created, all tables
unsigned long		HashTable::globalElementDeletions = 0;
					// No. of elements destroyed, all tbls
#endif // HASH_DEBUG_LEAKS

#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
HashTable::HashTable(			// Create a hash table
    unsigned int	initialSize,	// Estimate of table size needed
    unsigned int	(*hashFunction)(const DATA_TYPE* key),
					// Function to hash an element
    int			(*compareFunction)(const DATA_TYPE* ent1,
			  const DATA_TYPE* ent2))
					// Function to compare two elements
    : entryCount(0),
      tableSize(0),
      cantExpand(false),
      expansions(0),
      totalCollisions(0),
      collisionsNow(0),
      recentCollisions(0),
      longestChain(0),
      recentChain(0),
      entries(NULL),
      hash(hashFunction),
#ifdef HASH_DEBUG_LEAKS
      compare(compareFunction),
      elementCreations(0),
      elementDeletions(0)
#else // HASH_DEBUG_LEAKS
      compare(compareFunction)
#endif // HASH_DEBUG_LEAKS
    {
    if (initialSize > 0)
	{
	if (expand(initialSize))
	    cantExpand = true;
	}
    }

#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
HashTable::~HashTable()			// Destroy a hash table
    {
    for (unsigned int hashValue = 0;  hashValue < tableSize;  hashValue++)
	{
	HashTable::Entry* current = &entries[hashValue];
	if (current->datum != NULL)
	    {
	    HashTable::Entry* nextent = current->next;
	    current->datum = NULL;
	    current->next = NULL;
	    for (current = nextent;  current != NULL;  current = nextent)
		{
		nextent = current->next;
		current->datum = NULL;
		current->next = NULL;
		delete current;
		}
	    }
	}
    delete[] entries;
#ifdef HASH_DEBUG_LEAKS
    if (elementDeletions != elementCreations)
	panic("leak in HashTable");
#endif // HASH_DEBUG_LEAKS
    }

/*
 * Expand a hash table.  A return value of 1 means the table couldn't
 * be expanded, but is still usable.  A return value of 2 means the
 * expansion failed and the table is no longer useful.
 */
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
bool HashTable::expand(			// Expand a hash table
    unsigned int	newSize)	// Size (elements) to expand to
    {
    if (tableSize >= newSize)
	return false;
    if (cantExpand)
	return true;
    unsigned int oldSize = tableSize;	// Save existing table parameters
    HashTable::Entry* oldEntries = entries; // ...
    unsigned int sizeIndex = 0;
    while (sizeIndex < sizeof goodSizes / sizeof(goodSizes[0]))
	{
	if (goodSizes[sizeIndex] > newSize)
	    break;
	sizeIndex++;
	}
    if (sizeIndex >= sizeof goodSizes / sizeof goodSizes[0])
	tableSize = 2 * newSize + 1;
    else
	tableSize = goodSizes[sizeIndex];
#ifdef HASH_DEBUG_LEAKS
    entries = new HashTable::Entry[tableSize](*this);
#else // HASH_DEBUG_LEAKS
    entries = new HashTable::Entry[tableSize];
#endif // HASH_DEBUG_LEAKS
    if (entries == NULL)
	{
	cantExpand = true;
	entries = oldEntries;
	return true;
	}
    else
	{
	/*
	 * Re-insert old entries into new table
	 */
	expansions++;
	collisionsNow = 0;
	recentCollisions = 0;
	recentChain = 0;
	HashTable::Entry* spareEntries = NULL;
	for (unsigned int i = 0;  i < oldSize;  i++)
	    {
	    HashTable::Entry* entry = &oldEntries[i];
	    if (entry->datum != NULL)
		{
		HashTable::Entry* where =
		  &entries[(*hash)((void *) entry->datum) % tableSize];
		if (where->datum == NULL)
		    {
		    where->datum = entry->datum;
		    where->next = NULL;
		    }
		else
		    {
		    totalCollisions++;
		    recentCollisions++;
		    collisionsNow++;
		    HashTable::Entry* newent;
		    if (spareEntries != NULL)
			{
			newent = spareEntries;
			spareEntries = spareEntries->next;
			}
		    else
			{
#ifdef HASH_DEBUG_LEAKS
			newent = new HashTable::Entry(*this);
#else // HASH_DEBUG_LEAKS
			newent = new HashTable::Entry;
#endif // HASH_DEBUG_LEAKS
			}
		    if (newent == NULL)
			{
			delete[] oldEntries;
			return true;
			}
		    newent->datum = where->datum;
		    newent->next = where->next;
		    where->next = newent;
		    where->datum = entry->datum;
		    unsigned long chainlength = 0;
		    for (HashTable::Entry* chainent = where;
		      chainent != NULL;
		      chainent = chainent->next)
			chainlength++;
		    if (chainlength > recentChain)
			{
			recentChain = chainlength;
			if (chainlength > longestChain)
			    longestChain = chainlength;
			}
		    }
		HashTable::Entry* nextent = entry->next;
		entry->next = NULL;
		entry->datum = NULL;
		for (entry = nextent;  entry != NULL;  entry = nextent)
		    {
		    nextent = entry->next;
		    entry->next = NULL;
		    where =
		      &entries[(*hash)((void *) entry->datum) % tableSize];
		    if (where->datum == NULL)
			{
			where->datum = entry->datum;
			entry->next = spareEntries;
			entry->datum = NULL;
			spareEntries = entry;
			}
		    else
			{
			totalCollisions++;
			recentCollisions++;
			collisionsNow++;
			DATA_TYPE* datum = entry->datum;
			entry->datum = where->datum;
			entry->next = where->next;
			where->datum = datum;
			where->next = entry;
			unsigned long chainlength = 0;
			for (HashTable::Entry* chainent = where;
			  chainent != NULL;
			  chainent = chainent->next)
			    chainlength++;
			if (chainlength > recentChain)
			    {
			    recentChain = chainlength;
			    if (chainlength > longestChain)
				longestChain = chainlength;
			    }
			}
		    }
		}
	    }
	delete spareEntries;
	}
    delete[] oldEntries;
    return false;
    }

#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
DATA_TYPE* HashTable::find(		// Find a datum in hash table
    DATA_TYPE*		datum,		// Datum to be located
    action		notFoundAction)	// Action to take on failure
    {
    unsigned int hashValue = (*hash)(datum);
					// Get hash value for datum
    if (tableSize == 0)
	expand(1);			// Prevent divide by zero
#ifdef HASH_CHECK_BUCKETS
    unsigned int bucket = hashValue % tableSize;
#endif // HASH_CHECK_BUCKETS
    HashTable::Entry* where = &entries[hashValue % tableSize];
					// Figure out where entry goes
    HashTable::Entry* current = where;	// Start search at that point
#ifdef HASH_CHECK_CHAIN
	DATA_TYPE* foundDatum = NULL;
#endif // HASH_CHECK_CHAIN
    if (current->datum != NULL)
	{
	for (  ;  current != NULL; current = current->next)
	    {
#ifdef HASH_CHECK_BUCKETS
	    if ((*hash)(current->datum) % tableSize != bucket)
		panic("datum in wrong bucket in HashTable::find");
#endif // HASH_CHECK_BUCKETS
	    if ((*compare)(current->datum, datum) == 0)
		{
#ifdef HASH_CHECK_CHAIN
		if (foundDatum != NULL)
		    panic("two matching entries in same hash chain");
		else
		    foundDatum = current->datum;
#else // HASH_CHECK_CHAIN
		return current->datum;
#endif // HASH_CHECK_CHAIN
		}
	    }
	}
#ifdef HASH_CHECK_CHAIN
    if (foundDatum != NULL)
	return foundDatum;
#endif // HASH_CHECK_CHAIN
    if (notFoundAction == FAIL)
	return NULL;
    if ((tableSize * MAXPCT) / 100 < entryCount)
	{
	if (expand(tableSize + 1))
	    return NULL;
	where = &entries[hashValue % tableSize];
	}
    entryCount++;
    if (where->datum == NULL)
	where->datum = datum;
    else
	{
	totalCollisions++;
	recentCollisions++;
	collisionsNow++;
#ifdef HASH_DEBUG_LEAKS
	current = new HashTable::Entry(*this);
#else // HASH_DEBUG_LEAKS
	current = new HashTable::Entry;
#endif // HASH_DEBUG_LEAKS
	if (current == NULL)
	    return NULL;
	current->next = where->next;
	current->datum = where->datum;
	where->next = current;
	where->datum = datum;
	unsigned long chainlength = 0;
	for (HashTable::Entry* chainent = where;
	  chainent != NULL;
	  chainent = chainent->next)
	    chainlength++;
	if (chainlength > recentChain)
	    {
	    recentChain = chainlength;
	    if (chainlength > longestChain)
		longestChain = chainlength;
	    }
	}
    return where->datum;
    }

/*
 * Remove an entry from a hash table.  Returns true if not found.
 */
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
bool HashTable::remove(
    register DATA_TYPE*	datum)		// What's to be freed
    {
    if (tableSize == 0)
	return true;
    HashTable::Entry* current = &entries[(*hash)(datum) % tableSize];
    if (current->datum != NULL)
	{
	for (HashTable::Entry* oldent = NULL;
	  current != NULL;
	  oldent = current, current = current->next)
	    {
	    if ((*compare)(current->datum, datum) == 0)
		{
		if (oldent == NULL)
		    {
		    // Remove first entry in hash table
		    if (current->next == NULL)
			current->datum = NULL;
		    else
			{
			collisionsNow--;
			oldent = current->next;
			current->next = oldent->next;
			current->datum = oldent->datum;
			oldent->datum = NULL;
			oldent->next = NULL;
			delete oldent;
			}
		    }
		else
		    {
		    // Remove an intermediate entry in hash table
		    collisionsNow--;
		    oldent->next = current->next;
		    current->datum = NULL;
		    current->next = NULL;
		    delete current;
		    }
		entryCount--;
		return false;
		}
	    }
	}
    return true;
    }

void HashTable::showStatistics(		// Display statistics on a stream
    ostream&		stream)		// Stream to display on
      const
    {
    stream << "size/entries/expansions = " << tableSize << ' '
      << entryCount << ' '
      << expansions << ", tot/now/recent collisions = "
      << totalCollisions << ' '
      << collisionsNow << ' ' << recentCollisions << ", max/recent chain = "
      << longestChain << ' ' << recentChain;
#ifdef HASH_DEBUG_LEAKS
    stream << " creations/deletions = " << elementCreations << ' '
      << elementDeletions;
#endif // HASH_DEBUG_LEAKS
    // Caller must provide a newline!
    }

void HashTable::Iterator::removeCurrent() // Delete current entry from table
    {
    HashTable::Entry* testent = &table->entries[hashValue];
    HashTable::Entry* oldent = NULL;
    for (  ;  testent != current;  oldent = testent, testent = testent->next)
	;
    if (oldent == NULL)
	{
	// Remove first entry in hash table
	if (testent->next == NULL)
	    {
	    testent->datum = NULL;
	    current = NULL;
	    ++(*this);
	    }
	else
	    {
	    table->collisionsNow--;
	    oldent = testent->next;
	    testent->next = oldent->next;
	    testent->datum = oldent->datum;
	    oldent->datum = NULL;
	    oldent->next = NULL;
	    delete oldent;
	    }
	}
    else
	{
	// Remove an intermediate entry in hash table
	table->collisionsNow--;
	oldent->next = testent->next;
	testent->datum = NULL;
	testent->next = NULL;
	current = oldent->next;
	delete testent;
	}
    table->entryCount--;
    suppressIncrement = true;
    }

unsigned int hashString(		// Hash a string
    const char*	     key)		// Key to be hashed
    {
    unsigned int hashValue = 0;
    while (*key != '\0')
	{
	hashValue = ((hashValue << HASHSHIFT)
	    | ((hashValue >> (32 - HASHSHIFT)) & ((1 << HASHSHIFT) - 1)))
	  ^ *key++;
	}
    return hashValue;
    }

unsigned int hashPJW(			// Hash a string, Aho/Sethi/Ullman
    const char*		key)		// Key to be hashed
    {
    /*
     * The following algorithm is taken from Aho, Sethi, and Ullman,
     * "Compilers: Principles, Techniques, and Tools," Addison-Wesley,
     * 1985, p. 436.
     *
     * Note that this code assumes 32-bit integers, and won't work
     * right on 16- or 64-bit machines.
     */
    unsigned int hashValue = 0;
    while (*key != '\0')
	{
	hashValue = (hashValue << 4) + *key++;
	register unsigned int rotate_bits = hashValue & 0xf0000000;
	if (rotate_bits)
	    hashValue ^= rotate_bits | (rotate_bits >> 24);
	}
    return hashValue;
    }

unsigned int hashInteger(		// Hash an integer
    int			key)		// Key to be hashed
    {

    return (unsigned int)key * (unsigned int)(key + 3);
    }

/*
 * Hash a string using the BUZ algorithm
 */
unsigned int hashStringBUZ (
    const char*		key)		// Key to be hashed
    {
    unsigned int hval = BUZINIT;	// Result
    const unsigned char* ukey = (unsigned char *) key;
					// Unsigned version of key
    unsigned int ch;
    while ((ch = *ukey++) != '\0')
	{
	if (hval & (1 << (sizeof hval * 8 - 1)))
	    hval = ((hval << 1) | 1) ^ BUZ_Table[ch];
	else
	    hval = (hval << 1) ^ BUZ_Table[ch];
	}
    hval = hashBUZ((unsigned char*)&hval, sizeof hval);
    return hashBUZ((unsigned char*)&hval, sizeof hval);
    }

/*
 * Hash an integer using the BUZ algorithm
 */
unsigned int hashIntegerBUZ (
    int			key)		/* Key to be hashed */
    {
    unsigned int hval = hashBUZ((unsigned char*)&key, sizeof key);
    hval = hashBUZ((unsigned char*)&hval, sizeof hval);
    return hashBUZ((unsigned char*)&hval, sizeof hval);
    }

/*
 * Do one pass of the BUZ algorithm on a fixed key.
 *
 * The BUZ algorithm was invented by Robert (Bob) Uzgalis.  I don't
 * have a reference for it yet, because I learned it from a lecture,
 * not a paper.  The basic idea is to use a (constant) lookup table of
 * random numbers to generate wider random bits for each input
 * character.  Also, the hash value from the key is rehashed twice for
 * cryptographic strength.
 */
static unsigned int hashBUZ (
    const unsigned char * key,		/* Key to be hashed */
    int			keylen)		/* Length of the key */
    {
    unsigned int hval = BUZINIT;
    while (--keylen >= 0)
	{
	unsigned int ch = *key++;
	if (hval & (1 << (sizeof hval * 8 - 1)))
	    hval = ((hval << 1) | 1) ^ BUZ_Table[ch];
	else
	    hval = (hval << 1) ^ BUZ_Table[ch];
	}
    return hval;
    }
