#ifndef lint
static const char Rcs_Id[] =
    "$Id: hashtest.cc,v 1.6 2000/11/30 06:34:47 geoff Exp $";
#endif

/*
 * $Log: hashtest.cc,v $
 * Revision 1.6  2000/11/30 06:34:47  geoff
 * Use typecasts to get around the need for a "const void*" type.
 *
 * Revision 1.5  1998/03/05 03:41:30  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.4  1997/07/11 17:50:54  geoff
 * Test with the BUZ (Bob Uzgalis) hashing algorithm.
 *
 * Revision 1.3  1996/03/19  18:02:18  geoff
 * Declare the RCS ID as const.  Fix an incorrect use of operator++.
 *
 * Revision 1.2  1995/09/15  00:13:02  geoff
 * Add a usage statement.  Fix the iterator usage to match the new
 * definition.
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

#include <iostream.h>
#include <string.h>
#include "hashtable.hh"

extern "C" int		atoi(const char*);
extern "C" int		sprintf(...);

int main (int argc, char* argv[])
    {
    HashTable htbl (1, (unsigned int (*)(const void *))hashStringBUZ,
      (int (*)(const void*, const void*))strcmp);
    if (argc != 2)
	{
	cerr << "Usage: hashtest test-size\n";
	return 1;
	}
    int tablesize = atoi(argv[1]);
    for (int tableno = 0;  tableno < tablesize;  tableno++)
	{
	char* value = new char[8];
	(void) sprintf(value, "%d", tableno);
	if ((char*)htbl.find((void*)value, HashTable::INSERT) != value)
	    cerr << "Argument " << tableno << " shouldn't have been found\n";
	}
    htbl.showStatistics(cout);
    cout << '\n';
    if (htbl.remove((void*)"0"))
	cerr << "Couldn't remove entry for '0'\n";
    for (HashTable::Iterator hi (htbl);  hi;  ++hi)
	{
	delete[] (char*) htbl[hi];
	hi.removeCurrent();
	}
    htbl.showStatistics(cout);
    cout << '\n';
    return 0;
    }
