#ifndef lint
static const char Rcs_Id[] =
    "$Id: listtest.cc,v 1.7 2000/11/30 06:34:47 geoff Exp $";
#endif

/*
 * $Log: listtest.cc,v $
 * Revision 1.7  2000/11/30 06:34:47  geoff
 * Use typecasts to get around the need for a "const void*" type.
 *
 * Revision 1.6  1998/03/05 03:41:31  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.5  1998/02/19 01:03:34  geoff
 * Test destructive and nondestructive list copying and assignment.
 *
 * Revision 1.4  1996/09/11 23:08:01  geoff
 * Test both dlists and slists.
 *
 * Revision 1.3  1996/05/03  01:31:44  geoff
 * Update the copyright.  Fix some compiler warnings.  Get rid of
 * listSize (use the count() member function instead).
 *
 * Revision 1.2  1995/09/15  00:13:10  geoff
 * Change the list constructor calls to provide the new argument.
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

#include <iostream.h>
#include <string.h>
#include "list.hh"

int		main(int argc, char* argv[]);
static void	djointest(DList& list1, DList& list2);
static void	sjointest(SList& list1, SList& list2);

int main(int, char*[])
    {
    DList dlista(false);
    DList dlistb(false);
    djointest(dlista, dlistb);
    dlista.pushTail((void*)"a1");
    djointest(dlista, dlistb);
    dlistb.pushTail((void*)"b1");
    djointest(dlista, dlistb);
    dlistb.pushTail((void*)"b2");
    djointest(dlista, dlistb);
    dlistb.pushTail((void*)"b3");
    dlistb.pushTail((void*)"b4");
    djointest(dlista, dlistb);
    dlistb.pushTail((void*)"b5");
    dlistb.pushTail((void*)"b6");
    dlistb.pushTail((void*)"b7");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlistb.pushTail((void*)"b8");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlistb.pushTail((void*)"b9");
    dlistb.pushTail((void*)"b10");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlistb.pushTail((void*)"b11");
    dlistb.pushTail((void*)"b12");
    dlistb.pushTail((void*)"b13");
    dlista.popAll();
    dlista.pushTail((void*)"a2");
    dlistb.pushTail((void*)"b14");
    dlistb.pushTail((void*)"b15");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlista.pushTail((void*)"a3");
    dlistb.pushTail((void*)"b16");
    dlistb.pushTail((void*)"b17");
    dlistb.pushTail((void*)"b18");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlista.pushTail((void*)"a4");
    dlista.pushTail((void*)"a5");
    dlistb.pushTail((void*)"b19");
    dlistb.pushTail((void*)"b20");
    djointest(dlista, dlistb);
    dlista.popAll();
    dlista.pushTail((void*)"a6");
    dlista.pushTail((void*)"a7");
    dlistb.pushTail((void*)"b19");
    dlistb.pushTail((void*)"b20");
    dlistb.pushTail((void*)"b21");
    djointest(dlista, dlistb);
    DList dlistc;
    unsigned int dlistasize = dlista.count();
    dlistc.destructiveCopyFrom(dlista);
    if (!dlista.isEmpty()  ||  dlistc.count() != dlistasize)
	cerr << "DList::destructiveCopyFrom failed\n";
    SList slista(false);
    SList slistb(false);
    sjointest(slista, slistb);
    slista.pushHead((void*)"a1");
    sjointest(slista, slistb);
    slistb.pushHead((void*)"b1");
    sjointest(slista, slistb);
    slistb.pushHead((void*)"b2");
    sjointest(slista, slistb);
    slistb.pushHead((void*)"b3");
    slistb.pushHead((void*)"b4");
    sjointest(slista, slistb);
    slistb.pushHead((void*)"b5");
    slistb.pushHead((void*)"b6");
    slistb.pushHead((void*)"b7");
    sjointest(slista, slistb);
    slista.popAll();
    slistb.pushHead((void*)"b8");
    sjointest(slista, slistb);
    slista.popAll();
    slistb.pushHead((void*)"b9");
    slistb.pushHead((void*)"b10");
    sjointest(slista, slistb);
    slista.popAll();
    slistb.pushHead((void*)"b11");
    slistb.pushHead((void*)"b12");
    slistb.pushHead((void*)"b13");
    slista.popAll();
    slista.pushHead((void*)"a2");
    slistb.pushHead((void*)"b14");
    slistb.pushHead((void*)"b15");
    sjointest(slista, slistb);
    slista.popAll();
    slista.pushHead((void*)"a3");
    slistb.pushHead((void*)"b16");
    slistb.pushHead((void*)"b17");
    slistb.pushHead((void*)"b18");
    sjointest(slista, slistb);
    slista.popAll();
    slista.pushHead((void*)"a4");
    slista.pushHead((void*)"a5");
    slistb.pushHead((void*)"b19");
    slistb.pushHead((void*)"b20");
    sjointest(slista, slistb);
    slista.popAll();
    slista.pushHead((void*)"a6");
    slista.pushHead((void*)"a7");
    slistb.pushHead((void*)"b19");
    slistb.pushHead((void*)"b20");
    slistb.pushHead((void*)"b21");
    sjointest(slista, slistb);
    SList slistc;
    unsigned int slistasize = slista.count();
    slistc.destructiveCopyFrom(slista);
    if (!slista.isEmpty()  ||  slistc.count() != slistasize)
	cerr << "SList::destructiveCopyFrom failed\n";
    return 0;
    }

static void djointest(DList& list1, DList& list2)
    {
    int list1size1 = list1.count();
    int list2size1 = list2.count();
    // copy both lists' elements onto copylist
    DList copyList(list1);
    for (DList::Iterator iter2(list2);  iter2;  ++iter2)
	copyList.pushTail(list2[iter2]);
    // join the lists and verify resulting sizes
    list1.join(list2);
    if (!list2.isEmpty())
	cerr << "djointest error:  list2 not empty\n";
    int list1size2 = list1.count();
    int list2size2 = list2.count();
    if (list2size2 != 0)
	cerr << "djointest error:  list2 has " << list2size2
	  << " elements after join\n";
    if (list1size2 != list1size1 + list2size1)
	cerr << "djointest error:  list1 size " << list1size2
	  << " doesn't match sum of original sizes " << list1size1
	  << " + " << list2size1 << " after join\n";
    // verify that copyList has the same elements as list1
    DList::Iterator iter1(list1);
    DList::Iterator copyIterator(copyList);
    for (  ;  iter1  &&  copyIterator;  ++iter1, ++copyIterator)
	{
	if (list1[iter1] != copyList[copyIterator])
	    {
	    cerr << "djointest error:  list1 and copyList don't match\n";
	    break;
	    }
	}
    if (iter1  ||  copyIterator)
	cerr << "djointest error:  iter1 and copyIterator"
	  " didn't terminate together\n";
    }

static void sjointest(SList& list1, SList& list2)
    {
    int list1size1 = list1.count();
    int list2size1 = list2.count();
    // copy both lists' elements onto copylist
    SList reverseCopyList(false);
    for (SList::Iterator iter1(list1);  iter1;  ++iter1)
	reverseCopyList.pushHead(list1[iter1]);
    SList copyList(list2);
    while (!reverseCopyList.isEmpty())
	copyList.pushHead(reverseCopyList.popHead());
    // join the lists and verify resulting sizes
    list1.join(list2);
    if (!list2.isEmpty())
	cerr << "sjointest error:  list2 not empty\n";
    int list1size2 = list1.count();
    int list2size2 = list2.count();
    if (list2size2 != 0)
	cerr << "sjointest error:  list2 has " << list2size2
	  << " elements after join\n";
    if (list1size2 != list1size1 + list2size1)
	cerr << "sjointest error:  list1 size " << list1size2
	  << " doesn't match sum of original sizes " << list1size1
	  << " + " << list2size1 << " after join\n";
    // verify that copyList has the same elements as list1
    SList::Iterator iter1(list1);
    SList::Iterator copyIterator(copyList);
    for (  ;  iter1  &&  copyIterator;  ++iter1, ++copyIterator)
	{
	if (list1[iter1] != copyList[copyIterator])
	    {
	    cerr << "sjointest error:  list1 and copyList don't match\n";
	    break;
	    }
	}
    if (iter1  ||  copyIterator)
	cerr << "sjointest error:  iter1 and copyIterator"
	  " didn't terminate together\n";
    }
