#ifndef _MEMLEAK_HH
#define _MEMLEAK_HH
/*
 * $Id: memleak.hh,v 1.1 1995/08/08 22:25:18 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: memleak.hh,v $
 * Revision 1.1  1995/08/08  22:25:18  geoff
 * Initial revision
 *
 */

/*
 * Classes used to help detect memory leaks.
 */

class ostream;

/*
 * The MemoryMarker class manages a table of bits which record
 * allocated memory space.  The idea is that the invoker will walk all
 * data structures, calling mark() on each bit of allocated memory.
 * After all structures have been marked, showHoles() can be used to
 * locate large areas of memory which are not known to be allocated.
 * This is less than perfect because it doesn't recognize areas that
 * were once allocated, but have since been freed.
 *
 * Note that in all cases, there are system dependencies on the way
 * the allocator works.  The current code is written to match libg++.
 */
class MemoryMarker
    {
    public:
			MemoryMarker();
			~MemoryMarker();
	void		clear();	// Clear all current markings
	void		mark(const void* area, unsigned long size);
					// Mark an area as reachable
	ostream&	showHoles(ostream& stream);
					// Show holes on an output stream
    private:
	static unsigned long maskSize(unsigned long bytesCovered);
					// Return size of marked-bit array
	void		expand(const void* area, unsigned long size);
					// Expand size of marked-bit array
	unsigned long*	markedBits;	// Array of marked bits
	unsigned long	markerSize;	// Size of area represented, bytes
	const void*	markerBase;	// Address of base of area rep'ed
	/*
	 * Machine- and compiler-specific definitions necessary to
	 * deal with the allocation system.
	 */
	const unsigned long BYTES_PER_MARKER_BIT = 8;
					// Allocated bytes covered by one bit
	const unsigned long BITS_PER_LONG = 32;
					// Bits covered by one mask word
	const unsigned long BLOCK_ROUNDING = 8192;
					// Block size to round things to
	const unsigned long MALLOC_EXTRA_BEFORE = 4;
					// Amount malloc adds before things
	const unsigned long MALLOC_EXTRA_AFTER = 4;
					// Amount malloc adds before things
    };

#include "memleak.icc"

#endif // _MEMLEAK_HH
