#ifndef lint
static const char Rcs_Id[] =
    "$Id: panic.cc,v 1.3 1998/01/31 06:30:26 geoff Exp $";
#endif

/*
 * $Id: panic.cc,v 1.3 1998/01/31 06:30:26 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: panic.cc,v $
 * Revision 1.3  1998/01/31 06:30:26  geoff
 * Add support for pausing or sleeping when panics happen, so that failed
 * processes can be attached with a debugger (idea borrowed from the Coda
 * project).
 *
 * Revision 1.2  1996/05/03  01:31:44  geoff
 * Update the copyright.
 *
 * Revision 1.1  1996/02/02  18:59:26  geoff
 * Initial revision
 *
 */

#include <iostream.h>
#include <stdlib.h>
#include <unistd.h>

void			panic(const char* message);
				// Panic the program
void			pauseOnPanics();
				// Cause panics to pause rather than aborting
void			sleepOnPanics(unsigned int time);
				// Cause panics to sleep a while before abort

static bool		pauseAfterPanic = false;
static unsigned int	sleepAfterPanic = 0;

void panic(			// Panic the program
    const char*		message) // Message to issue before exiting
    {
    cout.flush();
    cerr << message << '\n';
    cerr.flush();
    while (pauseAfterPanic)
	pause();		// Let somebody attach with debugger
    if (sleepAfterPanic != 0)
	sleep(sleepAfterPanic);
    abort();
    // NOTREACHED
    }

void pauseOnPanics()		// Cause panics to pause rather than aborting
    {
    pauseAfterPanic = true;
    }

void sleepOnPanics(		// Cause panics to sleep a while before abort
    unsigned int	time)	// Time to sleep after the panic
    {
    sleepAfterPanic = time;
    }
