#ifndef lint
static const char Rcs_Id[] =
    "$Id: simplestring.cc,v 1.3 2001/11/08 08:18:21 geoff Exp geoff $";
#endif

/*
 * $Id: simplestring.cc,v 1.3 2001/11/08 08:18:21 geoff Exp geoff $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: simplestring.cc,v $
 * Revision 1.3  2001/11/08 08:18:21  geoff
 * Add the non-assignment concatenation operators.
 *
 * Revision 1.2  1998/10/23 23:03:28  geoff
 * Fix things to compile with -pedantic.  Add some new functions.  Make a
 * number of things non-inline.
 *
 * Revision 1.1  1997/12/04 23:37:40  geoff
 * Initial revision
 *
 */

#include "simplestring.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			SimpleString::SimpleString(
			  const SimpleString& initialValue, unsigned int size);
SimpleString&		SimpleString::operator=(const SimpleString& source);
SimpleString&		SimpleString::operator=(const char* source);
SimpleString&		SimpleString::operator+=(const SimpleString& source);
SimpleString&		SimpleString::operator+=(const char* source);
SimpleString&		SimpleString::operator+=(char source);
SimpleString		SimpleString::operator+(const SimpleString& source);
SimpleString		SimpleString::operator+(const char* source);
SimpleString		SimpleString::operator+(char source);
SimpleString		SimpleString::subString(unsigned int start,
			    unsigned int subLength)
			  const;
					// Return substring
void			SimpleString::truncate(bool force);
					// Truncate and free up space
void			SimpleString::clone(unsigned int newSize) const;
					// Clone the value field
bool			SimpleString::underlyingGet(istream& stream,
			  char terminator,
			  SimpleStringTerminatorAction action);
					// Do get/gets
#endif // 0

TypeAllocator	SimpleString::allocator(sizeof(SimpleString), 1000);

SimpleString::SimpleString(
    const SimpleString&	initialValue,	// Initial value to give to string
    unsigned int	size)		// Amount of space to allocate
    {
    if (initialValue.value == NULL)
	{
	length = 0;
	maxSize = size;
	value = NULL;
	return;
	}
    length = initialValue.strlen();
    /*
     * If the size is defaulted and the initial value doesn't have the
     * maximum number of references, just generate a new reference.
     */
    if (size == 0
      &&  initialValue.value[initialValue.maxSize + REFERENCE_OFFSET]
	!= MAXIMUM_REFERENCES)
	{
	maxSize = initialValue.maxSize;
	value = initialValue.value;
	++value[maxSize + REFERENCE_OFFSET];
	}
    else
	{
	if (size < length)
	    maxSize = length;
	else
	    maxSize = size;
	value = new char[maxSize + OVERHEAD_SIZE];
	strcpy(value, initialValue.value);
	value[maxSize + REFERENCE_OFFSET] = 1;
	}
    }

SimpleString& SimpleString::operator=(
    const SimpleString& source)		// String to assign value from
    {
    if (this != &source)
	{
	if (source.value == NULL)
	    {
	    forgetValue(false);
	    return *this;
	    }
	length = source.strlen();
	/*
	 * Try to avoid copying
	 */
	if (source.value[source.maxSize + REFERENCE_OFFSET]
	  != MAXIMUM_REFERENCES)
	    {
	    forgetValue(true);
	    ++source.value[source.maxSize + REFERENCE_OFFSET];
	    value = source.value;
	    maxSize = source.maxSize;
	    }
	else
	    {
	    if (length > maxSize  ||  value == NULL
	      ||  value[maxSize + REFERENCE_OFFSET] != 1)
		{
		forgetValue(true);
		maxSize = length;
		value = new char[length + OVERHEAD_SIZE];
		value[maxSize + REFERENCE_OFFSET] = 1;
		}
	    strcpy(value, source.value);
	    }
	}
    return *this;
    }

SimpleString& SimpleString::operator=(
    const char*		source)		// C string to assign value from
    {
    if (source == NULL)
	forgetValue(false);
    else
	{
	length = ::strlen(source);
	if (length > maxSize  ||  value == NULL
	  ||  value[maxSize + REFERENCE_OFFSET] != 1)
	    {
	    forgetValue(true);
	    maxSize = length;
	    value = new char[length + OVERHEAD_SIZE];
	    value[maxSize + REFERENCE_OFFSET] = 1;
	    }
	strcpy(value, source);
	}
    return *this;
    }

SimpleString& SimpleString::operator+=(
    const SimpleString& source)		// String to concatenate at end
    {
    if (this == &source)
	panic("attempt to concatenate to self in SimpleString::operator+=");
    if (source.value != NULL)
	{
	length += source.length;
	if (length <= maxSize  &&  value != NULL)
	    {
	    clone(maxSize);		// Make sure we can write
	    strcat(value, source.value);
	    }
	else
	    {
	    bool zapOld;
	    if (value == NULL)
		zapOld = false;
	    else
		{
		zapOld = value[maxSize + REFERENCE_OFFSET] == 1;
		--value[maxSize + REFERENCE_OFFSET];
		}
	    maxSize = length;
	    char* newValue = new char[maxSize + OVERHEAD_SIZE];
	    if (value == NULL)
		strcpy(newValue, source.value);
	    else
		{
		strcpy(newValue, value);
		strcat(newValue, source.value);
		}
	    if (zapOld)
		delete[] value;
	    value = newValue;
	    value[maxSize + REFERENCE_OFFSET] = 1;
	    }
	}
    return *this;
    }

SimpleString& SimpleString::operator+=(
    const char*		source)		// C string to concatenate at end
    {
    if (source != NULL)
	{
	length += ::strlen(source);
	if (length <= maxSize  &&  value != NULL)
	    {
	    clone(maxSize);		// Make sure we can write
	    strcat(value, source);
	    }
	else
	    {
	    bool zapOld;
	    if (value == NULL)
		zapOld = false;
	    else
		{
		zapOld = value[maxSize + REFERENCE_OFFSET] == 1;
		--value[maxSize + REFERENCE_OFFSET];
		}
	    maxSize = length;
	    char* newValue = new char[maxSize + OVERHEAD_SIZE];
	    if (value == NULL)
		strcpy(newValue, source);
	    else
		{
		strcpy(newValue, value);
		strcat(newValue, source);
		}
	    if (zapOld)
		delete[] value;
	    value = newValue;
	    value[maxSize + REFERENCE_OFFSET] = 1;
	    }
	}
    return *this;
    }

SimpleString& SimpleString::operator+=(
    char		source)		// Single character to add at end
    {
    if (length < maxSize)
	{
	clone(maxSize);			// Make sure we can write
	value[length] = source;
	length++;
	value[length] = '\0';
	}
    else
	{
	bool zapOld;
	if (value == NULL)
	    zapOld = false;
	else
	    {
	    zapOld = value[maxSize + REFERENCE_OFFSET] == 1;
	    --value[maxSize + REFERENCE_OFFSET];
	    }
	maxSize = length + 1;
	char* newValue = new char[maxSize + OVERHEAD_SIZE];
	if (value != NULL)
	    strcpy(newValue, value);
	newValue[length] = source;
	++length;
	newValue[length] = '\0';
	if (zapOld)
	    delete[] value;
	value = newValue;
	value[maxSize + REFERENCE_OFFSET] = 1;
	}
    return *this;
    }

SimpleString SimpleString::operator+(
    const SimpleString& source)		// String to concatenate at end
    {
    SimpleString result(*this);
    result += source;
    return result;
    }

SimpleString SimpleString::operator+(
    const char*		source)		// C string to concatenate at end
    {
    SimpleString result(*this);
    result += source;
    return result;
    }

SimpleString SimpleString::operator+(
    char		source)		// Single character to add at end
    {
    SimpleString result(*this);
    result += source;
    return result;
    }

/*
 * Return a SimpleString containing a substring of the initial string.
 * "Start" is zero-indexed.
 */
SimpleString SimpleString::subString(
    unsigned int	start,		// Index of first desired character
    unsigned int	subLength)	// Length of substring wanted
    const
    {
    if (value == NULL)
	{
	SimpleString nullResult;
	return nullResult;
	}
    else if (start >= length)
	{
	SimpleString nullLengthResult("");
	return nullLengthResult;
	}
    if (start + subLength > length)
	subLength = length - start;
    SimpleString result(subLength);
    result.length = subLength;
    result.value = new char[subLength + OVERHEAD_SIZE];
    ::strncpy(result.value, value + start, subLength);
    result.value[subLength] = '\0';;
    result.value[subLength + REFERENCE_OFFSET] = 1;
    return result;
    }

/*
 * Truncate the space allocated for a string.  Normally, the space is
 * freed only if (a) the amount of space returned to the allocator
 * would be greater than or equal to TRUNCATION_THRESHOLD and (b)
 * there is only one reference to the string's value.  If "force" is
 * true, truncation is done even if it requires copying the string.
 */
void SimpleString::truncate(
    bool		force)		// True to force even if copy needed
    {
    if (value == NULL)
	return;
    if (!force
      &&  (value[maxSize + REFERENCE_OFFSET] != 1
	||  maxSize - length << TRUNCATION_THRESHOLD))
	return;				// Not worth truncating
    if (length == maxSize)
	return;				// Already truncated
    char* newValue = new char[length + OVERHEAD_SIZE];
    strcpy(newValue, value);
    newValue[length + REFERENCE_OFFSET] = 1;
    forgetValue(true);
    maxSize = length;
    value = newValue;
    }

/*
 * If necessary, clone the value field of a simple string so that it
 * can be modified.
 */
void SimpleString::clone(
    unsigned int	newSize)	// Size of copy, if nonzero
    const
    {
    if (value == NULL)
	{
	maxSize = newSize;
	value = new char[maxSize + OVERHEAD_SIZE];
	value[maxSize + REFERENCE_OFFSET] = 1;
	value[0] = '\0';
	return;
	}
    if (value[maxSize + REFERENCE_OFFSET] == 1)
	return;
    if (newSize > length)
	maxSize = newSize;
    char* newValue = new char[maxSize + OVERHEAD_SIZE];
    strcpy(newValue, value);
    --value[maxSize + REFERENCE_OFFSET];
    value = newValue;
    value[maxSize + REFERENCE_OFFSET] = 1;
    }

/*
 * Do the underlying work of get and gets.
 */
bool SimpleString::underlyingGet(
    istream&		stream,		// Stream to get from
    char		terminator,	// Terminator to use
    SimpleStringTerminatorAction
			action)		// What to do with terminator
    {
    char* result;
    if (stream.gets(&result, terminator))
	{
	length = ::strlen(result);
	switch (action)
	    {
	    case GET_PUT_BACK:
		{
		stream.putback(terminator);
		break;
		}
	    case GET_DISCARD:
	    default:
		{
		break;
		}
	    }
	if (maxSize < length  ||  value == NULL
	  ||  value[maxSize + REFERENCE_OFFSET] > 1)
	    {
	    forgetValue(true);
	    maxSize = length;
	    value = new char[maxSize + OVERHEAD_SIZE];
	    }
	::strcpy(value, result);
	value[maxSize + REFERENCE_OFFSET] = 1;
	delete[] result;
	}
    return stream;
    }
