#ifndef lint
static const char Rcs_Id[] =
    "$Id: simplestringtest.cc,v 1.4 2000/11/30 06:41:38 geoff Exp $";
#endif

/*
 * $Log: simplestringtest.cc,v $
 * Revision 1.4  2000/11/30 06:41:38  geoff
 * Fix a bug in the substring tests (trying to get a zero-length substring).
 *
 * Revision 1.3  2000/11/30 06:34:47  geoff
 * Fix a bug where I incorrectly assumed that NULL could be used for a
 * const char*.
 *
 * Revision 1.2  1998/10/23 23:03:29  geoff
 * Add more tests.
 *
 * Revision 1.1  1997/12/04 23:37:40  geoff
 * Initial revision
 *
 */

#include <iostream.h>
#include "simplestring.hh"

int		main(int argc, char* argv[]);

int main(int, char*[])
    {
    /*
     * Test compilation of all operations
     */
    char*		normalChar = NULL;
    const char*		constChar = NULL;
    SimpleString	simpleString;
    const SimpleString	constSimpleString("a");
    SimpleString*	allocatedSimpleString = new SimpleString(simpleString);
    constChar = (const char*)simpleString;
    simpleString = *allocatedSimpleString;
    simpleString = constSimpleString;
    simpleString = normalChar;
    simpleString = constChar;
    simpleString += *allocatedSimpleString;
    simpleString += constSimpleString;
    simpleString += normalChar;
    simpleString += constChar;
    if (simpleString == constSimpleString
      ||  constSimpleString == simpleString
      ||  simpleString == normalChar
      ||  simpleString == constChar
      ||  constSimpleString == normalChar
      ||  constSimpleString == constChar
      ||  simpleString != constSimpleString
      ||  constSimpleString != simpleString
      ||  simpleString != normalChar
      ||  simpleString != constChar
      ||  constSimpleString != normalChar
      ||  constSimpleString != constChar)
	simpleString = NULL;
    simpleString = "abc";
    int j = simpleString.strlen() + constSimpleString.strlen();
    j = 0;
    char c = simpleString[j];
    c = simpleString[(unsigned)j];
    c = constSimpleString[j];
    c = constSimpleString[(unsigned)j];
    simpleString.truncate();
    delete allocatedSimpleString;
    /*
     * Do the same for writeable simple strings.
     */
    WritableSimpleString	writableSimpleString;
    const WritableSimpleString	constWritableSimpleString("a");
    WritableSimpleString*	allocatedWritableSimpleString =
				  new WritableSimpleString(
				    writableSimpleString);
    constChar = (const char*)writableSimpleString;
    writableSimpleString = *allocatedWritableSimpleString;
    writableSimpleString = constWritableSimpleString;
    writableSimpleString = normalChar;
    writableSimpleString = constChar;
    writableSimpleString += *allocatedWritableSimpleString;
    writableSimpleString += constWritableSimpleString;
    writableSimpleString += normalChar;
    writableSimpleString += constChar;
    if (writableSimpleString == constWritableSimpleString
      ||  constWritableSimpleString == writableSimpleString
      ||  writableSimpleString == normalChar
      ||  writableSimpleString == constChar
      ||  constWritableSimpleString == normalChar
      ||  constWritableSimpleString == constChar
      ||  writableSimpleString != constWritableSimpleString
      ||  constWritableSimpleString != writableSimpleString
      ||  writableSimpleString != normalChar
      ||  writableSimpleString != constChar
      ||  constWritableSimpleString != normalChar
      ||  constWritableSimpleString != constChar)
	writableSimpleString = (const char*)NULL;
    writableSimpleString = "abc";
    j = writableSimpleString.strlen() + constWritableSimpleString.strlen();
    j = 0;
    c = writableSimpleString[j];
    c = writableSimpleString[(unsigned)j];
    c = constWritableSimpleString[j];
    c = constWritableSimpleString[(unsigned)j];
    writableSimpleString[j] = 'x';
    writableSimpleString[(unsigned)j] = 'x';
    writableSimpleString.truncate();
    delete allocatedWritableSimpleString;
    /*
     * Test execution of (some) operations
     */
    SimpleString	nullString;
    SimpleString	string1("a");
    SimpleString	string2 = "ab";
    SimpleString	string1a(string1);
    SimpleString	string2a = string2;
    if (string1 == "ab"  ||  string1 != "a")
	cerr << "string1 is wrong (1)\n";
    if (string1 == string2  ||  string1 != string1a)
	cerr << "string1 is wrong (2)\n";
    string1 += string2;
    if (string1 != "aab")
	cerr << "concatenation failed (1)\n";
    string1 += "c";
    if (string1 != "aabc")
	cerr << "concatenation failed (2)\n";
    if (string1.subString(1, 2) != "ab")
	cerr << "substring failed (1)\n";
    if (string1.subString(2, 2) != "bc")
	cerr << "substring failed (2)\n";
    int aCount = 0;
    for (unsigned int i = 0;  i < string1.strlen();  i++)
	{
	if (string1[i] == 'a')
	    aCount++;
	}
    if (aCount != 2)
	cerr << "subscripting failed: " << aCount << '\n';
    aCount = 0;
    const WritableSimpleString string1b(string1);
    for (unsigned int i = 0;  i < string1b.strlen();  i++)
	{
	if (string1b[i] == 'a')
	    aCount++;
	}
    if (aCount != 2)
	cerr << "subscripting failed: " << aCount << '\n';
    WritableSimpleString string1c(string1);
    for (unsigned int i = 0;  i < string1c.strlen();  i++)
	{
	if (((const WritableSimpleString&)string1c)[i] == 'a')
	    string1c[i] = 'd';
	}
    if (string1c != "ddbc")
	cerr << "subscripting failed: " << (const char*)string1c << '\n';
    SimpleString* string3 = new SimpleString("abc", 5);
    *string3 += "de";
    if (*string3 != "abcde")
	cerr << "concatenation failed (3)\n";
    *string3 = (const char*)string1;
    string3->truncate(true);
    if (*string3 != string1)
	cerr << "truncation failed\n";
    *string3 = "abcdef";
    if (*string3 != "abcdef")
	cerr << "assignment failed\n";
    delete string3;
    return 0;
    }
