#ifndef lint
static const char Rcs_Id[] =
    "$Id: typealloc.cc,v 1.4 1998/10/23 23:03:30 geoff Exp $";
#endif

/*
 * $Id: typealloc.cc,v 1.4 1998/10/23 23:03:30 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: typealloc.cc,v $
 * Revision 1.4  1998/10/23 23:03:30  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.3  1998/03/05 03:41:31  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.2  1998/01/31 06:30:27  geoff
 * Panic on errors, rather than just aborting.
 *
 * Revision 1.1  1997/07/16  22:16:45  geoff
 * Initial revision
 *
 */

#include "panic.hh"
#include "typealloc.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			TypeAllocator::TypeAllocator(
			  unsigned long normalElementSize,
			  unsigned long chunkSize);
			TypeAllocator::~TypeAllocator();
void*			TypeAllocator::allocElement();
					// Allocate an element
void			TypeAllocator::freeElement(void* element);
					// Return an element to the allocator
#endif // 0

TypeAllocator::TypeAllocator(
    unsigned long	normalElementSize,
					// Size of elements allocated,
    unsigned long	chunkSize)
					// Number of elements alloc'ed at once
					// ..zero means use 1-page sizes
    : elementSize((normalElementSize + ALIGNMENT - 1) / ALIGNMENT * ALIGNMENT),
      allocationIncrement(chunkSize),
      firstArea(NULL),
      lastArea(NULL),
      areaWithFirstFreeElement(NULL),
      areaWithLastFreedElement(NULL)
    {
    if (allocationIncrement == 0)
	allocationIncrement =
	  (PAGE_SIZE - sizeof(FreeArea) - MALLOC_OVERHEAD)
	    / elementSize;
    }

TypeAllocator::~TypeAllocator()
    {
    for (FreeArea* area(firstArea); area != NULL;  )
	{
	FreeArea* nextArea = area->next;
	delete[] (char*)area;
	area = nextArea;
	}
    }

void* TypeAllocator::allocElement()	// Allocate an element
    {
    if (areaWithFirstFreeElement == NULL)
	{
	/*
	 * There are no free elements.  Create a new blank area.  We
	 * assume that "new char*" will return an aligned object.
	 */
	unsigned long areaBytes(elementSize * allocationIncrement);
	FreeArea& area(
	  *(FreeArea*)new char[sizeof (FreeArea)
	    - sizeof firstArea->firstElement + areaBytes]);
	area.elementCount = area.freeCount = allocationIncrement;
	area.areaEnd = (void*)((char*)&area.firstElement + areaBytes);
	FreeElement* element((FreeElement*)&area.firstElement);
	area.elements = element;
	for (unsigned long initializedElements(1);
	  initializedElements < allocationIncrement;
	  initializedElements++)
	    {
	    FreeElement* nextElement(
	      (FreeElement*)((void*)((char*)element + elementSize)));
	    element->next = nextElement;
	    element = nextElement;
	    }
	/*
	 * We still need to initialize the last element.
	 */
	element->next = NULL;
	/*
	 * Put the new area into the area list in the proper place.
	 */
	areaWithFirstFreeElement = &area;
	FreeArea* testArea(firstArea);
	if (testArea == NULL)
	    {
	    firstArea = lastArea = &area;
	    area.previous = area.next = NULL;
	    }
	else if ((void*)&area > (void*)lastArea)
	    {
	    /*
	     * The new area goes at the end.
	     */
	    area.next = NULL;
	    area.previous = lastArea;
	    lastArea->next = &area;
	    lastArea = &area;
	    }
	else
	    {
	    while ((void*)&area > (void*)testArea)
		testArea = testArea->next;
	    /*
	     * The new area goes just before testArea.
	     */
	    area.previous = testArea->previous;
	    area.next = testArea;
	    testArea->previous = &area;
	    if (area.previous == NULL)
		firstArea = &area;
	    else
		area.previous->next = &area;
	    }
	}
    /*
     * There is at least one area with at least one free element.
     * Allocate it.
     */
    FreeArea* area(areaWithFirstFreeElement);
    if (area->freeCount == 0)
	panic("bad free count in TypeAllocator::allocElement");
    area->freeCount--;
    FreeElement* element(area->elements);
    area->elements = area->elements->next;
    if ((area->elements == NULL  &&  area->freeCount != 0)
      ||  (area->elements != NULL  &&  area->freeCount == 0))
	panic("bad free count in TypeAllocator::allocElement");
    while (area != NULL  &&  area->freeCount == 0)
	area = area->next;
    areaWithFirstFreeElement = area;
    return (void*)element;
    }

void TypeAllocator::freeElement(	// Return an element to the allocator
    void*		element)
    {
    FreeArea* area(firstArea);
    /*
     * As an optimization, see if we can jump a bit forward in the
     * area list.
     */
    if (areaWithFirstFreeElement != NULL
      &&  (void*)areaWithFirstFreeElement < element)
	area = areaWithFirstFreeElement;
    if (areaWithLastFreedElement != NULL
      &&  (void*)areaWithLastFreedElement < element)
	area = areaWithLastFreedElement;
    while ((void*)area->areaEnd < element)
	area = area->next;
    if ((void*)area > element)
	panic("bad element address in TypeAllocator::freeElement");
    /*
     * We've found the area containing this element.  Free it.
     */
    areaWithLastFreedElement = NULL;
    ((FreeElement*)element)->next = area->elements;
    area->elements = (FreeElement*)element;
    if (++area->freeCount > area->elementCount)
	panic("bad free count in TypeAllocator::freeElement");
    else if (area->freeCount == area->elementCount)
	{
	/*
	 * We've freed every element in this area.  Return the entire
	 * area to the allocator.
	 */
	if (area->previous == NULL)
	    firstArea = area->next;
	else
	    area->previous->next = area->next;
	if (area->next == NULL)
	    lastArea = area->previous;
	else
	    area->next->previous = area->previous;
	if (areaWithFirstFreeElement == area)
	    {
	    areaWithFirstFreeElement = area->next;
	    while (areaWithFirstFreeElement != NULL
	      &&  areaWithFirstFreeElement->freeCount == 0)
		areaWithFirstFreeElement = areaWithFirstFreeElement->next;
	    }
	areaWithLastFreedElement = NULL;
	delete[] (char*)area;
	}
    else if (areaWithFirstFreeElement == NULL
      ||  area < areaWithFirstFreeElement)
	{
	/*
	 * We're the new "lowest" area.  Record this fact so that the
	 * next allocation will be from us.
	 */
	if (area->freeCount != 1)
	    panic("bad free count in TypeAllocator::freeElement");
	areaWithFirstFreeElement = area;
	}
    }
