#ifndef lint
static char Rcs_Id[] =
    "$Id: hash.c,v 1.11 1999/11/22 05:46:13 geoff Exp $";
#endif

/*
 * $Log: hash.c,v $
 * Revision 1.11  1999/11/22 05:46:13  geoff
 * Fix a stupid cut-and-paste-from C++ problem
 *
 * Revision 1.10  1999/11/22 05:45:33  geoff
 * Get rid of small table sizes, and add one intermediate one
 *
 * Revision 1.9  1999/11/17 07:18:23  geoff
 * Make goodsizes be unsigned
 *
 * Revision 1.8  1999/11/17 07:17:14  geoff
 * Expand the primes table to 2**31 and a bit
 *
 * Revision 1.7  1997/07/11 17:47:48  geoff
 * Add the BUZ (Bob Uzalis) hashing algorithm.
 *
 * Revision 1.6  1997/02/03 23:47:56  geoff
 * Improvements from two years ago: supply hashpjw and keep performance
 * statistics.
 *
 * Revision 1.5  1994/11/04  19:05:31  geoff
 * Change hash actions and hash walk return values to be enums.  Add the
 * DEL_CONT and DEL_ABORT actions to hashwalk.
 *
 * Revision 1.4  1994/07/27  18:17:19  geoff
 * Change HASHSHIFT to hash strings better.  Remember to recalculate "where"
 * if the hash table is resized.
 *
 * Revision 1.3  1994/03/17  19:35:34  geoff
 * Make hashdestroy delete all active entries
 *
 * Revision 1.2  1994/01/04  23:48:59  geoff
 * Don't use the private malloc
 *
 * Revision 1.1  1994/01/04  20:18:49  geoff
 * Initial revision
 *
 */

/*
 * Generalized hash-table package.
 */

#include "hash.h"

hashtbl_t *	hashmake P ((unsigned int nel,
		  unsigned int (*hash) (VOID *),
		  int (*cmp) (VOID *, VOID *)));
static int	hashexpand P ((hashtbl_t * htbl, unsigned int size));
VOID *		hashfind P ((hashtbl_t * htbl, VOID * datum,
		  hashaction_t action));
int		hashdelete P ((hashtbl_t * htbl, VOID * datum));
void		hashdestroy P ((hashtbl_t * htbl, void (*zapfunc) (VOID *)));
hashwalk_t	hashwalk P ((hashtbl_t * htbl,
		  hashwalk_t (*func) (VOID *, VOID *), VOID * client_data));
unsigned int	strhash P ((char * key));
unsigned int	hashpjw P ((char * key));
unsigned int	inthash P ((int key));
unsigned int	strhashbuz P ((char * key));
unsigned int	inthashbuz P ((int key));
static unsigned int hashbuz P ((unsigned char * key, int keylen));

#define HASHSHIFT   5		/* How much to shift hash by per char */
#define MAXPCT	    70		/* How full to let hash table get */

#ifndef NULL
#define NULL	    0		/* Since we don't include stdio.h */
#endif /* NULL */

/*
 * Hash table sizes.  Prime is probably a good idea, though in truth I
 * whipped the algorithm up on the spot rather than looking it up, so
 * who knows what's really best?  If we overflow the table, we just
 * use a double-and-add-1 algorithm.
 */
static unsigned int goodsizes[] =
    {
    223, 907, 3631, 14591, 29191, 58391, 116789, 233591, 467183,
    934387, 1868777, 3737557, 7475129, 14950261, 29900531, 59801111,
    119602237, 239204477, 478408967, 956817949, 1913635961,
    3827271937u
    };

/*
 * Constants for the BUZ algorithm.  BUZINIT is the initial value of
 * the hash function.  The lookup table is used to widen and randomize
 * the bytes of the key.  All values were gotten from /dev/random on a
 * Linux box.
 */
#define BUZINIT		0x7b4402a2
static unsigned int buzTable[256] =
    {
    0xc70bb269, 0x13299943, 0xe9cee5e1, 0x32119a79,
    0xc2365851, 0x169ee8e3, 0x094c1ed8, 0x58e1d4e0,
    0x96eb1762, 0xfe296797, 0x89082f47, 0x27d5078d,
    0x8ebb9de0, 0x14df49e0, 0x38d21c63, 0x1f5b2770,
    0xaa6b0150, 0x7b0b656e, 0x98d37bc2, 0x4d7e85ec,
    0x987910e2, 0xb8cbac89, 0xa3f664a3, 0xeca3003c,
    0x7b364137, 0xb2a6edae, 0x5ef72906, 0x64a9e7b7,
    0x28cd6520, 0xca3c72df, 0x57398ce2, 0x8db893d4,
    0x0a5995cd, 0x2d109fb4, 0x0491162f, 0xb3488737,
    0x6fc4eb03, 0x9903cb21, 0xe82ff831, 0xb03ff8e5,
    0x054836ca, 0x5335e6f8, 0x01396c2a, 0xf9000899,
    0x03ed9d63, 0x2bf6946b, 0x9097fa8b, 0xacd8dfc7,
    0x8488b8a6, 0x0e39cd2e, 0xac1a4517, 0xcd49e035,
    0xe98b7e7b, 0xd3571502, 0xd602805e, 0xe7143cfe,
    0x46db0a6b, 0x0a4c9ebe, 0x4e2e1ca7, 0x3040fc62,
    0xe8818c02, 0x37155e7b, 0xe44ba138, 0x43cacdd4,
    0x53d986ba, 0xdd4dca35, 0x0f680f71, 0x6c1a551e,
    0x74263e95, 0xcfc4f5d5, 0x37b8ef45, 0xc00ac71d,
    0x3b059e0d, 0x208bc754, 0x41335fbe, 0x785a0ffc,
    0x189f024f, 0xd669c2d8, 0xe1b20f87, 0xba2550da,
    0x10167369, 0x85fad38f, 0x97d20e4e, 0x5bc0da5e,
    0x80799570, 0x93eb4058, 0x139042a6, 0x40b34bf6,
    0x15c21dfa, 0x8f852660, 0xa3d20fb3, 0x3d175cf9,
    0x792441a8, 0xdc5e71b5, 0x925f6350, 0x66e8d08b,
    0xc4606b59, 0x85d8b88c, 0x1ea4f459, 0x664f62bc,
    0x77407de3, 0x73d158ca, 0xb76ab172, 0xe9ed1aeb,
    0x93dc2009, 0xeb9da6ac, 0x3d26cf05, 0x675132bc,
    0xc29196fe, 0x2a62486f, 0x914e75e1, 0xa1c31883,
    0x1c28291c, 0xc73c668c, 0xf4ac07e6, 0x87c9a9ac,
    0xb7196ea7, 0x67cb7fa2, 0x55987797, 0x29ce38ea,
    0x427361b3, 0x5b5667a6, 0x68a72fb0, 0xcef8235a,
    0xd06e8f5b, 0x4d3633f5, 0x214d3a19, 0xbd09ec15,
    0x5c61c24b, 0x3928573c, 0x26083ab8, 0x857a5dee,
    0x3203e50c, 0x52a1a713, 0xa8270ee2, 0xdfb643a9,
    0x7797c1f3, 0x0f8ddc9f, 0x9368de21, 0x638ebd4e,
    0xd91808d7, 0x28ce69b8, 0xe424b0ce, 0xfe52fdef,
    0x89126c74, 0xdb5f3d91, 0xba488f47, 0x2b15cdb8,
    0xa517b0f9, 0x53950632, 0x1159546a, 0xe50f65a3,
    0x5f26b5d1, 0x68a3a955, 0xc2b78ea1, 0x49c33701,
    0x45457aee, 0xd49b550a, 0x244379b8, 0xec826af5,
    0x4fa6e0c9, 0xd4633425, 0x82f0bd85, 0xc23ccc2e,
    0xac73e11a, 0xdc94b283, 0x13e59bb2, 0x23b4880e,
    0x1d295c45, 0xef67488c, 0x6b74149e, 0xdf90d4ac,
    0xfc6e65a9, 0x406a3734, 0x86999303, 0xc73e7180,
    0x3c8a0b31, 0x75fa9249, 0xaca5e0e2, 0x4d0cc60d,
    0x4b174606, 0x836fb602, 0x4f9fc83a, 0xe16477a7,
    0xda1506a9, 0x905b28b7, 0x4229f5c2, 0xdf4c9144,
    0x731888e9, 0x2e37421f, 0x0c67c385, 0x44a2e520,
    0xfe3ee655, 0x92547582, 0x9525f4d4, 0xdaa8caf8,
    0xa25bd583, 0x0e315733, 0xd35fea29, 0xc9cfaa0f,
    0xc6bdf7e9, 0xa48b4e01, 0xfd30ffe0, 0xd0f63421,
    0x2b84803d, 0xe1b368a4, 0xbae5daa5, 0x4dd6336a,
    0xb60c4030, 0x7bb552f1, 0xf1b91481, 0xc8929b82,
    0xa1c22bf8, 0xd585aae4, 0x17fb4f6a, 0xa9c0d32e,
    0x2036f9b5, 0x3a95d611, 0xd554f25d, 0x3441153f,
    0x7fa89ff4, 0x8f91241c, 0x4b2cc5a9, 0xd3035a00,
    0x5c80707f, 0xe610fd47, 0xf60958c2, 0xb55a6fe7,
    0x3e1ba335, 0x2dead082, 0x1a8877e8, 0xd0791aad,
    0x9706ee52, 0xfb1dc525, 0x7fa1ba54, 0x6a3e9f81,
    0xa85a906a, 0x86ce1b46, 0x3b05833e, 0xc8d8fdcb,
    0x44e606c5, 0x8807beb2, 0xe46047d3, 0x85b9f5f8,
    0x56ed0cba, 0x3cf4e646, 0x970fd9dd, 0xd0600895,
    0x2d0a5f92, 0x891c4220, 0x017fbfe0, 0x4dde2016,
    0x3a6e421d, 0x7e3f2285, 0xdf7956e3, 0x52fdcf83
    };

extern VOID *	calloc P ((unsigned int, unsigned int));
extern void	free P ((VOID *));
extern VOID *	malloc P ((unsigned int));

hashtbl_t * hashmake (nel, hash, cmp) /* Create a hash table */
    unsigned int	nel;		/* Number of elements */
    unsigned int	(*hash) P ((VOID *)); /* Hash function */
    int			(*cmp) P ((VOID *, VOID *)); /* Compare function */
    {
    register hashtbl_t * htbl;	/* Table we're making */

    htbl = (hashtbl_t *) malloc (sizeof (hashtbl_t));
    if (htbl == NULL)
	return NULL;
    htbl->nents = 0;
    htbl->size = 0;
    htbl->entries = NULL;
    htbl->cantexpand = 0;
    htbl->hash = hash;
    htbl->cmp = cmp;
    htbl->expansions = 0;
    htbl->totcollisions = htbl->curcollisions = htbl->expcollisions = 0;
    if (nel > 0)
	{
	if (hashexpand (htbl, nel))
	    {
	    free ((VOID *) htbl);
	    return NULL;
	    }
	}
    return htbl;
    }

/*
 * Expand a hash table.  A return value of 1 means the table couldn't
 * be expanded, but is still usable.  A return value of 2 means the
 * expansion failed and the table is no longer useful.
 */
static int hashexpand (htbl, size) /* Expand a hash table */
    register hashtbl_t * htbl;	/* Hash table to be expanded */
    unsigned int	size;	/* Size to expand to */
    {
    VOID *		datum;	/* Datum from current entry */
    hashent_t *		entry;	/* Entry being re-inserted */
    unsigned int	i;	/* Useful index */
    hashent_t *		oldents; /* Old hash-entries list */
    unsigned int	oldsize; /* Old size of the table */
    hashent_t *		newent;	/* New entry just created */
    hashent_t *		nextent; /* Next entry to process */
    hashent_t *		spareents; /* Spare entries for use in rehashing */
    hashent_t *		where;	/* Where next entry goes */

    if (htbl->size >= size)
	return 0;
    if (htbl->cantexpand)
	return 1;
    oldsize = htbl->size;
    oldents = htbl->entries;
    for (i = 0;  i < sizeof goodsizes / sizeof (goodsizes[0]);  i++)
	{
	if (goodsizes[i] > size)
	    break;
	}
    if (i >= sizeof goodsizes / sizeof goodsizes[0])
	htbl->size = 2 * size + 1;
    else
	htbl->size = goodsizes[i];
    htbl->entries =
      (hashent_t *) calloc ((unsigned) htbl->size, sizeof (hashent_t));
    if (htbl->entries == NULL)
	{
	htbl->cantexpand = 1;
	htbl->entries = oldents;
	return 1;
	}
    else
	{
	/*
	 * Re-insert old entries into new table
	 */
	htbl->expansions++;
	htbl->expcollisions = 0;
	htbl->curcollisions = 0;
	for (i = 0, spareents = NULL;  i < oldsize;  i++)
	    {
	    entry = &oldents[i];
	    if (entry->datum != NULL)
		{
		where =
		  &htbl->entries[(*htbl->hash) ((VOID *) entry->datum)
		    % htbl->size];
		if (where->datum == NULL)
		    {
		    where->datum = entry->datum;
		    where->next = NULL;
		    }
		else
		    {
		    htbl->totcollisions++;
		    htbl->expcollisions++;
		    htbl->curcollisions++;
		    if (spareents != NULL)
			{
			newent = spareents;
			spareents = spareents->next;
			}
		    else
			newent = (hashent_t *) malloc (sizeof (hashent_t));
		    if (newent == NULL)
			return 2;
		    newent->datum = where->datum;
		    newent->next = where->next;
		    where->next = newent;
		    where->datum = entry->datum;
		    }
		for (entry = entry->next;  entry != NULL;  entry = nextent)
		    {
		    nextent = entry->next;
		    where =
		      &htbl->entries[(*htbl->hash) ((VOID *) entry->datum)
			% htbl->size];
		    if (where->datum == NULL)
			{
			where->datum = entry->datum;
			entry->next = spareents;
			spareents = entry;
			}
		    else
			{
			htbl->totcollisions++;
			htbl->expcollisions++;
			htbl->curcollisions++;
			datum = entry->datum;
			entry->datum = where->datum;
			entry->next = where->next;
			where->datum = datum;
			where->next = entry;
			}
		    }
		}
	    }
	if (oldents != NULL)
	    free ((VOID *) oldents);
	for (  ;  spareents != NULL;  spareents = nextent)
	    {
	    nextent = spareents->next;
	    free ((VOID *) spareents);
	    }
	}
    return 0;
    }

VOID * hashfind (htbl, datum, action)
    register hashtbl_t * htbl;		/* Hash table to search in */
    VOID *		datum;		/* Datum to be located */
    hashaction_t	action;		/* Action to take on failure */
    {
    register hashent_t * curent;	/* Current entry being searched */
    unsigned int	 hval;		/* Hash value of the datum */
    register hashent_t * where;		/* Where entry fits in hash table */

    hval = (*htbl->hash) (datum);
    where = curent = &htbl->entries[hval % htbl->size];
    if (curent->datum != NULL)
	{
	for (  ;  curent != NULL; curent = curent->next)
	    {
	    if ((*htbl->cmp) (curent->datum, datum) == 0)
		return curent->datum;
	    }
	}
    if (action == HASH_FAIL)
	return NULL;
    if ((htbl->nents * 100) / MAXPCT >= htbl->size)
	{
	if (hashexpand (htbl, htbl->nents + 1))
	    return NULL;
	where = &htbl->entries[hval % htbl->size];
	}
    htbl->nents++;
    if (where->datum == NULL)
	where->datum = datum;
    else
	{
	htbl->totcollisions++;
	htbl->expcollisions++;
	htbl->curcollisions++;
	curent = (hashent_t *) malloc (sizeof (hashent_t));
	if (curent == NULL)
	    return NULL;
	curent->next = where->next;
	curent->datum = where->datum;
	where->next = curent;
	where->datum = datum;
	}
    return where->datum;
    }

int hashdelete (htbl, datum)		/* Delete a hash entry */
    register hashtbl_t * htbl;		/* Hash table to delete from */
    register VOID *	datum;		/* What's to be freed */
    {
    register hashent_t * curent;	/* Current entry being searched */
    register hashent_t * oldent;	/* Previous entry */

    if (htbl->size == 0)
	return 1;
    curent = &htbl->entries[(*htbl->hash) (datum) % htbl->size];
    if (curent->datum != NULL)
	{
	for (oldent = NULL;
	  curent != NULL;
	  oldent = curent, curent = curent->next)
	    {
	    if ((*htbl->cmp) (curent->datum, datum) == 0)
		{
		if (oldent == NULL)
		    {
		    /* Delete first entry in hash table */
		    if (curent->next == NULL)
			curent->datum = NULL;
		    else
			{
			htbl->curcollisions--;
			oldent = curent->next;
			curent->next = oldent->next;
			curent->datum = oldent->datum;
			free ((VOID *) oldent);
			}
		    }
		else
		    {
		    /* Delete an intermediate entry in hash table */
		    htbl->curcollisions--;
		    oldent->next = curent->next;
		    free ((VOID *) curent);
		    }
		return 0;
		}
	    }
	}
    return 1;
    }

void hashdestroy (htbl, func)		/* Destroy a hash table */
    hashtbl_t *		htbl;		/* Table to be destroyed */
    void		(*func) P ((VOID * datum)); /* Func to destroy data */
    {
    register hashent_t * curent;	/* Current entry being searched */
    int			hval;		/* Hash value being destroyed */
    hashent_t *		nextent;	/* Next entry */

    if (htbl == NULL)
	return;
    for (hval = 0;  hval < htbl->size;  hval++)
	{
	curent = &htbl->entries[hval];
	if (curent->datum != NULL)
	    {
	    if (func != NULL)
		(*func) (curent->datum);
	    for (curent = curent->next;  curent != NULL;  curent = nextent)
		{
		if (func != NULL)
		    (*func) (curent->datum);
		nextent = curent->next;
		free ((VOID *) curent);
		}
	    }
	}
    if (htbl->entries != NULL)
	free ((VOID *) htbl->entries);
    free ((VOID *) htbl);
    }

hashwalk_t hashwalk (htbl, func, client_data) /* Walk all entries in table */
    register hashtbl_t * htbl;		/* Table to be walked */
    hashwalk_t		(*func) P ((VOID * datum, VOID * client_data));
					/* Function to call for each datum */
    VOID *		client_data;	/* Parameter to provide to function */
    {
    register hashent_t * curent;	/* Entry being walked */
    int			hval;		/* Hash value being walked */
    hashent_t *		oldent;		/* Previous entry */
    hashent_t *		nextent;	/* Next entry */
    int			retval;		/* Return value from function */

    for (hval = 0, retval = 0;  hval < htbl->size;  hval++)
	{
	curent = &htbl->entries[hval];
	if (curent->datum != NULL)
	    {
	    for (oldent = NULL;
	      curent != NULL;
	      oldent = curent, curent = nextent)
		{
		nextent = curent->next;
		retval = (*func) (curent->datum, client_data);
		if (retval == HW_DEL_CONT  ||  retval == HW_DEL_ABORT)
		    {
		    if (oldent == NULL)
			{
			/* Delete first entry in hash table */
			if (nextent == NULL)
			    curent->datum = NULL;
			else
			    {
			    htbl->curcollisions--;
			    curent->next = nextent->next;
			    curent->datum = nextent->datum;
			    free ((VOID *) nextent);
			    nextent = curent;
			    curent = NULL;
			    }
			}
		    else
			{
			/* Delete an intermediate entry in hash table */
			htbl->curcollisions--;
			oldent->next = nextent;
			free ((VOID *) curent);
			curent = oldent;
			}
		    }
		if (retval == HW_ABORT  ||  retval == HW_DEL_ABORT)
		    return retval;
		}
	    }
	}
    return HW_OK;
    }

unsigned int strhash (key)		/* Hash a string */
    register char *	key;		/* Key to be hashed */
    {
    register unsigned int hval;		/* Result */

    for (hval = 0;  *key != '\0';  )
	{
	hval = ((hval << HASHSHIFT)
	    | ((hval >> (32 - HASHSHIFT)) & ((1 << HASHSHIFT) - 1)))
	  ^ *key++;
	}
    return hval;
    }

unsigned int hashpjw (key)		/* Hash a string, Aho/Sethi/Ullman */
    register char *	key;		/* Key to be hashed */
    {
    register unsigned int hval;		/* Result */
    register unsigned int rotate_bits;	/* Bits used in rotating result */

    /*
     * The following algorithm is taken from Aho, Sethi, and Ullman,
     * "Compilers: Principles, Techniques, and Tools," Addison-Wesley,
     * 1985, p. 436.
     *
     * Note that this code assumes 32-bit integers, and won't work
     * right on 16- or 64-bit machines.
     */
    for (hval = 0;  *key != '\0';  )
	{
	hval = (hval << 4) + *key++;
	rotate_bits = hval & 0xf0000000;
	if (rotate_bits)
	    hval ^= rotate_bits | (rotate_bits >> 24);
	}
    return hval;
    }

unsigned int inthash (key)		/* Hash an integer */
    int			key;		/* Key to be hashed */
    {

    return (unsigned int) key;
    }

/*
 * Hash a string using the BUZ algorithm
 */
unsigned int strhashbuz (key)
    char *		key;		/* Key to be hashed */
    {
    register unsigned int ch;		/* Next character to be hashed */
    register unsigned int hval = BUZINIT; /* Result */
    unsigned int	hval2;		/* Copy of result for rehashing */
    unsigned char *	ukey = (unsigned char *) key;
					/* Unsigned version of key */

    while ((ch = *ukey++) != '\0')
	{
	if (hval & (1 << (sizeof hval * 8 - 1)))
	    hval = ((hval << 1) | 1) ^ buzTable[ch];
	else
	    hval = (hval << 1) ^ buzTable[ch];
	}
    hval2 = hval;
    hval2 = hashbuz ((unsigned char *) &hval2, sizeof hval2);
    return hashbuz ((unsigned char *) &hval2, sizeof hval2);
    }

/*
 * Hash an integer using the BUZ algorithm
 */
unsigned int inthashbuz (key)
    int			key;		/* Key to be hashed */
    {
    unsigned int	hval;		/* Result */

    hval = hashbuz ((unsigned char *) &key, sizeof key);
    hval = hashbuz ((unsigned char *) &hval, sizeof hval);
    return hashbuz ((unsigned char *) &hval, sizeof hval);
    }

/*
 * Do one pass of the BUZ algorithm on a fixed key.
 *
 * The BUZ algorithm was invented by Robert (Bob) Uzgalis.  I don't
 * have a reference for it yet, because I learned it from a lecture,
 * not a paper.  The basic idea is to use a (constant) lookup table of
 * random numbers to generate wider random bits for each input
 * character.  Also, the hash value from the key is rehashed twice for
 * cryptographic strength.
 */
static unsigned int hashbuz (key, keylen)
    register unsigned char * key;	/* Key to be hashed */
    register int	keylen;		/* Length of the key */
    {
    register unsigned int ch;		/* Next character to be hashed */
    register unsigned int hval = BUZINIT; /* Result */

    while (--keylen >= 0)
	{
	ch = *key++;
	if (hval & (1 << (sizeof hval * 8 - 1)))
	    hval = ((hval << 1) | 1) ^ buzTable[ch];
	else
	    hval = (hval << 1) ^ buzTable[ch];
	}
    return hval;
    }
