/*
 * $Id: hash.h,v 1.5 1997/07/11 17:47:49 geoff Exp $
 *
 * $Log: hash.h,v $
 * Revision 1.5  1997/07/11 17:47:49  geoff
 * Add the BUZ (Bob Uzalis) hashing algorithm.
 *
 * Revision 1.4  1997/02/03 23:47:56  geoff
 * Improvements from two years ago: use longs to store various sizes, and
 * support performance statistics.
 *
 * Revision 1.3  1994/11/04  19:05:32  geoff
 * Change hash actions and hash walk return values to be enums.
 *
 * Revision 1.2  1994/03/17  19:43:10  geoff
 * Change the declaration of hashdestroy to include the new zapfunc argument.
 *
 * Revision 1.1  1994/01/04  20:18:49  geoff
 * Initial revision
 *
 */

/*
 * Header file for the generalized hash package.
 */

#undef P
#undef VOID
#ifdef __STDC__
#define P(x)	x
#define VOID	void
#else /* __STDC__ */
#define P(x)	()
#define VOID	char
#endif /* __STDC__ */

/*
 * Structure used to describe an entry in a hash table
 */
typedef struct	hashent
    {
    struct hashent *
		next;		/* Next entry in collision chain */
    VOID *	datum;		/* Datum provided by caller */
    }
		hashent_t;

/*
 * Structure used to describe a hash table.
 */
typedef struct
    {
    long	nents;		/* Number of entries currently in table */
    long	size;		/* Size of the hash table */
    int		cantexpand;	/* NZ if can't expand further */
    int		expansions;	/* Number of expansions performed */
    long	totcollisions;	/* Total collisions ever experienced */
    long	curcollisions;	/* Current number of collisions in table */
    long	expcollisions;	/* Total collisions since last expansion */
    hashent_t *	entries;	/* Entries in the table */
    unsigned int (*hash) P ((VOID *));
				/* Routine to calculate a hash value */
    int		(*cmp) P ((VOID *, VOID *));
				/* Routine to compare two entries */
    }
		hashtbl_t;

/*
 * Actions to take in hashfind
 */
typedef enum
    {
    HASH_FAIL,			/* Fail if not found */
    HASH_INSERT			/* Insert if not found */
    }
		hashaction_t;

/*
 * Return values from hashwalk and functions called by it
 */
typedef enum
    {
    HW_OK,			/* Walk succeeded */
    HW_CONT,			/* Continue walking */
    HW_DEL_CONT,		/* Delete item and continue */
    HW_ABORT,			/* Abort hash walk */
    HW_DEL_ABORT		/* Delete item and abort walk */
    }
		hashwalk_t;

extern hashtbl_t * 
		hashmake P ((unsigned int nel,
		  unsigned int (*hash) (VOID *),
		  int (*cmp) (VOID *, VOID *)));
extern VOID *	hashfind P ((hashtbl_t * htbl, VOID * datum,
		  hashaction_t action));
extern int	hashdelete P ((hashtbl_t * htbl, VOID * datum));
extern hashwalk_t hashwalk P ((hashtbl_t * htbl,
		  hashwalk_t (*func) (VOID *, VOID *), VOID * client_data));
extern void	hashdestroy P ((hashtbl_t * htbl, void (*zapfunc) (VOID *)));
extern unsigned int
		strhash P ((char * key));
extern unsigned int
		hashpjw P ((char * key));
extern unsigned int
		inthash P ((int key));
extern unsigned int
		strhashbuz P ((char * key));
extern unsigned int
		inthashbuz P ((int key));
