#ifndef lint
static char Rcs_Id[] =
    "$Id: list.c,v 1.2 1997/02/03 23:47:57 geoff Exp $";
#endif

/*
 * $Log: list.c,v $
 * Revision 1.2  1997/02/03 23:47:57  geoff
 * Improvements from two years ago: add joinlists.
 *
 * Revision 1.1  1994/01/04  20:18:49  geoff
 * Initial revision
 *
 */

#include "list.h"

void		initlist P ((link_t * head));
void		inslist P ((link_t * where, link_t * item));
void		remlist P ((link_t * item));
void		appendlist P ((link_t * where, link_t * head));

#ifndef NULL
#define NULL	0
#endif /* NULL */

void initlist (head)			/* Initialize a list */
    link_t *			head;	/* Head of the new list */
    {

    head->next = head->prev = head;
    }

void inslist (where, item)		/* Insert an item on a list */
    register link_t *		where;	/* Item to insert after */
    register link_t *		item;	/* Item to be inserted */
    {
    
    item->next = where->next;
    item->prev = where;
    item->next->prev = item;
    where->next = item;
    }

void remlist (item)			/* Remove an item from a list */
    register link_t *		item;	/* Item to be removed */
    {

    item->prev->next = item->next;
    item->next->prev = item->prev;
    item->next = item->prev = NULL;	/* To catch errors */
    }

void appendlist (where, head)		/* Append one list to another */
    register link_t *		where;	/* Where to append the list */
    register link_t *		head;	/* Head of list to be appended */
    {

    /*
     * Appendlist appends the list headed at "head" into an existing
     * list just before "where".  The item at "head" is not included
     * in the append, and is set to a single-element list containing
     * only itself.  The assumption is that both "head" and "where"
     * are list headers which are not themselves elements, although
     * the code only requires this for "head".  The result is that the
     * list at "head" is appended to the end of the list at "where".
     *
     * Contrast with joinlists.
     */
    if (head->next != head)
	{
	head->next->prev = where->prev;
	head->prev->next = where;
	where->prev->next = head->next;
	where->prev = head->prev;
	head->next = head->prev = head;
	}
    }

void joinlists (where, other)		/* Join two lists */
    register link_t *		where;	/* List to be joined on */
    register link_t *		other;	/* List to be joined */
    {

    /*
     * Joinlist appends the item at "other", and all following items,
     * just after the item at "where".  The assumption is that both
     * "where" and "other" are list elements in lists which are not
     * headed by dummy elements, although the code only requires this
     * for "other".
     *
     * Contrast with appendlist.
     */
    other->prev->next = where->next;
    where->next->prev = other->prev;
    other->prev = where;
    where->next = other;
    }
