#ifndef lint
static char Rcs_Id[] =
    "$Id: mimeencode.c,v 1.1 1998/10/23 23:04:34 geoff Exp $";
#endif

/*
 * Simple routines to support MIME-style encoding and decoding of strings.
 */

/*
 * $Log: mimeencode.c,v $
 * Revision 1.1  1998/10/23 23:04:34  geoff
 * Initial revision
 *
 */

#include <ctype.h>

#undef P
#undef VOID
#ifdef __STDC__
#define P(x)	x
#define VOID	void
#else /* __STDC__ */
#define P(x)	()
#define VOID	char
#endif /* __STDC__ */

char *		mimeEncode P ((char * source, char * whatToEncode));
char *		mimeDecode P ((char * source));

#ifndef NULL
#define NULL	0
#endif /* NULL */

/*
 * This table is indexed by hex digit to produce an equivalent ASCII
 * character.
 */
static char hexify[16] =
    {
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
    };

/*
 * This table is indexed by ASCII character to decode the equivalent
 * 4-bit hex value.
 */
static char unhexify[256] =
    {
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /*  0-0F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 10-1F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 20-2F */
     0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0, /* 30-3F */
     0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 40-4F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 50-5F */
     0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 60-6F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 70-7F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 80-8F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* 90-9F */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* A0-AF */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* B0-BF */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* C0-CF */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* D0-DF */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* E0-EF */
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, /* F0-FF */
    };

extern char *	malloc P ((unsigned int));	/* Memory allocator */

/*
 * Convert an input string into a MIME equivalent, in which certain
 * characters are replaced by their hex encodings.  The characters to
 * be encoded can be specified by the caller, or can be defaulted to
 * whitespace and nonprinting characters by passing NULL.
 *
 * The result is returned in a malloc'ed buffer, which it is the
 * caller's responsibility to free.  NULL is returned on an allocation
 * failure.
 */
char * mimeEncode (source, whatToEncode)
    char *	source;		/* String to be encoded */
    char *	whatToEncode;	/* String with "forbidden" characters */
    {
    register int ch;		/* Character being processed */
    static char	encodeIt[256];	/* Table of what to encode */
    static int	initialized = 0; /* NZ if tables have been initialized */
    char *	result;		/* Space to hold the result */
    register char * resultp;	/* Pointer into the result */
    int		resultSize;	/* Size of the output string */
    register char * sourcep;	/* Pointer into the source */

    if (whatToEncode != NULL)
	{
	encodeIt[0] = 1;
	for (ch = 1;  ch < sizeof encodeIt / sizeof encodeIt[0];  ch++)
	    encodeIt[ch] = 0;
	while (*whatToEncode != '\0')
	    encodeIt[(unsigned)*whatToEncode++] = 1;
	initialized = 0;	/* Be sure next time reinitializes table */
	}
    else if (!initialized)
	{
	encodeIt[0] = 1;
	for (ch = 1;  ch < sizeof encodeIt / sizeof encodeIt[0];  ch++)
	    encodeIt[ch] = !isascii (ch)  ||  iscntrl (ch)  ||  isspace (ch);
	encodeIt['='] = 1;
	encodeIt['\''] = 1;
	encodeIt['"'] = 1;
	initialized = 1;
	}
    /*
     * Scan the source string to discover how much space we need.
     */
    for (sourcep = source, resultSize = 0;  *sourcep != '\0';  sourcep++)
	resultSize += encodeIt[(unsigned) *sourcep] ? 3 : 1;
    /*
     * Allocate space for the result and calculate it.
     */
    result = (char *) malloc (resultSize + 1);
    if (result == NULL)
	return NULL;
    for (sourcep = source, resultp = result;  *sourcep != '\0';  sourcep++)
	{
	if (encodeIt[(unsigned) *sourcep])
	    {
	    *resultp++ = '=';
	    *resultp++ = hexify[((unsigned) *sourcep >> 4) & 0xF];
	    *resultp++ = hexify[((unsigned) *sourcep) & 0xF];
	    }
	else
	    *resultp++ = *sourcep;
	*resultp = '\0';
	}
    return result;
    }

/*
 * Remove the MIME encoding from an input string.
 *
 * The result is returned in a malloc'ed buffer, which it is the
 * caller's responsibility to free.  NULL is returned on an allocation
 * failure.
 */
char * mimeDecode (source)
    char *	source;		/* String to be decoded */
    {
    char *	result;		/* Space to hold the result */
    register char * resultp;	/* Pointer into the result */
    int		resultSize;	/* Size of the output string */
    register char * sourcep;	/* Pointer into the source */

    for (sourcep = source, resultSize = 0;  *sourcep != '\0';  resultSize++)
	{
	if (*sourcep++ == '=')
	    sourcep += 2;
	}
    /*
     * Allocate space for the result and calculate it.
     */
    result = (char *) malloc (resultSize + 1);
    if (result == NULL)
	return NULL;
    for (sourcep = source, resultp = result;  *sourcep != '\0';  resultp++)
	{
	if (*sourcep != '=')
	    *resultp = *sourcep++;
	else
	    {
	    *resultp = (unhexify[(unsigned) sourcep[1]] << 4)
	      | unhexify[(unsigned) sourcep[2]];
	    sourcep += 3;
	    }
	}
    return result;
    }
