#ifndef lint
static char Rcs_Id[] =
    "$Id: qsort.c,v 1.3 1995/03/28 19:55:55 geoff Exp $";
#endif

/*
 * This file contains a complete reimplementation of qsort(1), as
 * given in the following reference:
 *
 *	Jon L. Bentley and M. Douglas McIlroy, "Engineering a Sort
 *	Function", Software--Practice and Experience, 23 (11),
 *	1249-1265 (November 1993).
 */

/*
 * $Log: qsort.c,v $
 * Revision 1.3  1995/03/28  19:55:55  geoff
 * Fix a very important conditional test
 *
 * Revision 1.2  1994/02/15  21:39:30  geoff
 * Improve parenthesization to get rid of some warnings.  Fix the
 * definition of the swap() macro to get around Sun CC problems.  Add
 * some comments, slightly improve the efficiency of the code, and fix a
 * bug in med3 calculations.
 *
 * Revision 1.1  1993/12/21  22:37:14  geoff
 * Initial revision
 *
 */

#include <stddef.h>

#ifdef __STDC__
#define P(x)	x
#else /* __STDC__ */
#define P(x)	()
#endif /* __STDC__ */

typedef long		WORD;
#define W		sizeof (WORD)	/* Must be a power of 2 */
#define SWAPINIT(a, es)	swaptype = \
			  ((a - (char *) 0) | es) % W \
			    ? 2 \
			    : es > W ? 1 : 0

#define exch(a, b, t)	(t = a, a = b, b = t)
#ifdef __STDC__
#define swap(a, b)	swaptype != 0 \
			  ? swapfunc (a, b, es, swaptype) \
			  : (void) exch (*(WORD *) (a), *(WORD *) (b), t)
#else /* __STDC__ */
/* Sun's CC doesn't like voids in ?: constructs */
#define swap(a, b)	if (swaptype != 0) \
			  swapfunc (a, b, es, swaptype); \
			  else exch (*(WORD *) (a), *(WORD *) (b), t)
#endif /* __STDC__ */

#define vecswap(a, b, n) \
			if (n > 0) swapfunc (a, b, n, swaptype)

#define PVINIT(pv, pm)	if (swaptype != 0) \
			    { \
			    pv = a; \
			    swap (pv, pm); \
			    } \
			else \
			    pv = (char *) &v, v = *(WORD *) pm

#define min(a, b)	((a) < (b) ? (a) : (b))

void		qsort P ((char * a, size_t n, size_t es, int (*cmp) ()));
static char *	med3 P ((char * a, char * b, char * c, int (*cmp) ()));
static void	swapfunc P ((char * a, char * b, size_t n, int swaptype));

void qsort (a, n, es, cmp)
    char *		a;
    size_t		n;
    size_t		es;
    int			(*cmp) ();
    {
    char *		pa;
    char *		pb;
    char *		pc;
    char *		pd;
    char *		pl;
    char *		pm;
    char *		pn;
    char *		pv;
    int			r;		/* Return code from cmp */
    int			swaptype;	/* 0 = words, 1 = by words, 2 = arb */
    WORD		t;		/* Temp for swapping */
    WORD		v;		/* Space for partition value */
    size_t		s;

    SWAPINIT (a, es);
    if (n < 7)
	{
	/* Insertion sort on smallest arrays */
	pv = a + n * es;
	for (pm = a + es;  pm < pv;  pm += es)
	    {
	    for (pl = pm;  pl > a  &&  cmp (pl - es, pl) > 0;  pl -= es)
		swap (pl, pl - es);
	    }
	return;
	}
    pm = a + (n / 2) * es;	/* Small arrays, middle element */
    if (n > 7)
	{
	pl = a;
	pn = a + (n - 1) * es;
	if (n > 40)
	    {
	    /* Big arrays, pseudomedian of 9 */
	    s = (n / 8) * es;
	    pl = med3 (pl, pl + s, pl + 2 * s, cmp);
	    pm = med3 (pm - s, pm, pm + s, cmp);
	    pn = med3 (pn - 2 * s, pn - s, pn, cmp);
	    }
	pm = med3 (pl, pm, pn, cmp); /* Mid-size, median of 3 */
	}
    PVINIT (pv, pm);		/* pv points to partition value */
    pa = pb = a;
    pc = pd = a + (n - 1) * es;
    for (  ;  ;  )
	{
	while (pb <= pc  &&  (r = cmp (pb, pv)) <= 0)
	    {
	    if (r == 0)
		{
		swap (pa, pb);
		pa += es;
		}
	    pb += es;
	    }
	while (pc >= pb  &&  (r = cmp (pc, pv)) >= 0)
	    {
	    if (r == 0)
		{
		swap (pc, pd);
		pd -= es;
		}
	    pc -= es;
	    }
	if (pb > pc)
	    break;
	swap (pb, pc);
	pb += es;
	pc -= es;
	}
    pn = a + n * es;
    s = min (pa - a, pb - pa);
    vecswap (a, pb - s, s);
    s = min (pd - pc, pn - pd - es);
    vecswap (pb, pn - s, s);
    if ((s = pb - pa) > es)
	qsort (a, s / es, es, cmp);
    if ((s = pd - pc) > es)
	qsort (pn - s, s / es, es, cmp);
    }

static char * med3 (a, b, c, cmp)
    char *		a;
    char *		b;
    char *		c;
    int			(*cmp) ();
    {

    return cmp (a, b) < 0
      ? (cmp (b, c) < 0 ? b : cmp (a, c) < 0 ? c : a)
      : (cmp (b, c) > 0 ? b : cmp (a, c) > 0 ? c : a);
    }

static void swapfunc (a, b, n, swaptype)
    char *		a;
    char *		b;
    size_t		n;
    int			swaptype;
    {

    if (swaptype <= 1)
	{
	WORD		t;

	for (  ;  n > 0;  a += W, b += W, n -= W)
	    exch (*(WORD *) a, *(WORD *) b, t);
	}
    else
	{
	char		t;

	for (  ;  n > 0;  a += 1, b += 1, n -= 1)
	    exch (*a, *b, t);
	}
    }
