(defun convert-to-stream nil
  "Find the next printf/fprintf function and convert it to use C++ streams.
Assumes that doing << from a format argument will produce the correct output.
Doesn't correctly lots of things, so be careful!"
  (interactive)
  (beginning-of-line)
  (if (search-forward "printf" nil t)
      (let ((stmt-start) (here) (stmt-end) (pctloc) (argstart)
	      (argval) (indent-column))
	(beginning-of-line)
	(setq stmt-start (point-marker))
	(search-forward ");")
	(setq stmt-end (point-marker))
	(replace-match ",;")
	(goto-char stmt-start)
	(if (search-forward "(void) " stmt-end t)
	    (replace-match "" nil t))
	(search-forward "printf" stmt-end)
	(forward-word -1)
	(if (looking-at "sprintf")
	    (end-of-line)
	  (if (looking-at "fprintf")
	      (progn
		(setq here (point))
		(search-forward "(")
		(delete-region here (point))
		(if (looking-at "stderr")
		    (progn (delete-char 6) (insert "cerr"))
		  (forward-word 1))
		(insert " << "))
	    (insert "cout << "))
	  (setq here (point))
	  (search-forward "\"" stmt-end)
	  (delete-region here (1- (point)))
	  ;; Convert arguments to embedded format
	  (while (search-forward "%" stmt-end t)
	    (backward-delete-char 1)
	    (setq pctloc (point))
	    (re-search-forward "[0-9]*.?[0-9]*[a-z]+" stmt-end)
	    (delete-region pctloc (point))
	    (search-forward "," stmt-end)
	    (setq argstart (point))
	    (search-forward "," stmt-end)
	    (setq argval (buffer-substring argstart (1- (point))))
	    (delete-region argstart (point))
	    (goto-char pctloc)
	    (insert argval)
	    (insert "<< \"")
	    (backward-char 4)
	    (just-one-space-no-newlines)
	    (goto-char pctloc)
	    (insert "\" <<")
	    (just-one-space-no-newlines))
	  (goto-char stmt-end)
	  (backward-char 2)
	  (if (looking-at ",") (delete-char 1))
	  (end-of-line)
	  (backward-char 5)
	  (if (looking-at "\"\\\\n\"")
	      (progn (delete-char 4) (insert "'\\n'")))
	  (end-of-line)
	  (backward-char 7)
	  (if (looking-at " << \"\"") (delete-char 6))
	  (goto-char stmt-start)
	  (while (re-search-forward "\"\\([^\\\\]\\)\"" stmt-end t)
	    (replace-match "'\\1'" nil nil))
	  ;; Correct indentation
	  (goto-char stmt-start)
	  (skip-chars-forward " \t")
	  (setq indent-column (+ (current-column) 2))
	  (move-to-column 79)
	  (while (= (current-column) 79)
	    (let ((here (point)) (line-start))
	      (move-to-column indent-column)
	      (setq line-start (point))
	      (goto-char here)
	      (if (search-backward " <<" line-start t)
		  (progn
		    (delete-char 1)
		    (insert "\n")
		    (indent-to indent-column)
		    (move-to-column 79))
		(move-to-column 0)))))	; give up trying to indent
	)))
(defun just-one-space-no-newlines nil
  "Delete all white space around point, leaving a single space in its place."
  (skip-chars-backward " \t\n")
  (let ((here (point)))
    (skip-chars-forward " \t\n")
    (delete-region here (point)))
  (insert " "))
(defvar c-block-comment-prefix " *"
  "Prefix given to lines after the first in a C block comment.")
(defvar c++-block-comment-prefix "//"
  "Prefix given to lines after the first in a C++ block comment.")
(defvar c-block-comment-terminator " */"
  "Last line of a C block comment (excluding indentation).")
(defvar convert-block-comments nil
  "If t, C-style block comments are converted to C++-style")
(defun convert-c++-comments nil
  "Convert /*-style comments to //-style comments.
Correctly handles block comments where each line is preceded by
c-block-comment-prefix."
  (interactive)
  (while (search-forward "/*" nil t)
      (progn
	(replace-match "//")
	(if (looking-at "\n")
	    (if convert-block-comments
		(let			; Block comment
		    ((comment-start (point)) (comment-end) (block-limit))
		  (search-forward "*/")
		  (setq comment-end (point-marker))
		  (beginning-of-line)
		  (setq block-limit (point-marker))
		  (goto-char comment-start)
		  (forward-line 1)
		  (while (search-forward c-block-comment-prefix block-limit t)
		    (replace-match c++-block-comment-prefix)
		    (forward-line 1))
		  (if (search-forward c-block-comment-terminator comment-end t)
		      (replace-match "//")))
	      (replace-match "/*"))
	  (end-of-line)			; One-line comment
	  (backward-char 3)
	  (if (looking-at " \\*/")
	      (delete-char 3)
	    (forward-char 3)
	    (if (looking-at "\\*/")
		(delete-char 2)))))))
(defun unconvert-block-comments nil
  "Convert //-style block comments back to /*-style comments.
Correctly handles block comments where each line is preceded by
c++-block-comment-prefix."
  (interactive)
  (while (search-forward "//" nil t)
    (if (looking-at "\n")
	(progn
	  (delete-backward-char 2)
	  (insert "/*")
	  (forward-line 1)
	  (skip-chars-forward " \t")
	  (while (looking-at c++-block-comment-prefix)
	    (replace-match c-block-comment-prefix)
	    (forward-line 1)
	    (skip-chars-forward " \t"))
	  (forward-line -1)
	  (search-forward c-block-comment-prefix)
	  (replace-match " */")
	  (forward-line 1)))))
