#ifndef _CDECLS_HH
#define _CDECLS_HH
/*
 * $Id: cdecls.hh,v 1.11 2005/04/13 21:03:01 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: cdecls.hh,v $
 * Revision 1.11  2005/04/13 21:03:01  geoff
 * Updates to compile on more recent Linux releases.
 *
 * Revision 1.10  2000/01/03 21:56:50  geoff
 * Remove a declaration that conflicts on the 386.
 *
 * Revision 1.9  1997/12/18 01:24:12  geoff
 * Use __sparc__ rather than __sun4c__ to detect Sparcs.
 *
 * Revision 1.8  1997/07/14 21:43:36  geoff
 * Change the declaration of bzero to be Sparc-only, instead of not-Linux.
 *
 * Revision 1.7  1997/07/11 17:40:54  geoff
 * Add declarations for drand48 and srand48 on non-Linux systems.
 *
 * Revision 1.6  1996/05/03 01:11:31  geoff
 * Update the copyright.  Add declarations of fsync, setpwent, and
 * endpwent.
 *
 * Revision 1.5  1996/04/10  01:28:22  geoff
 * Add a declaration of setsockopt.
 *
 * Revision 1.4  1996/02/02  18:28:43  geoff
 * Add some new declarations required by upgrading to a new C++ compiler.
 *
 * Revision 1.3  1995/05/05  22:41:01  geoff
 * Add some linux support.
 *
 * Revision 1.2  1995/04/20  21:56:54  geoff
 * Add a declaration of readlink.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#ifndef __i386__
#include <stdio.h>
#endif // __i386__
#include <sys/param.h>

/*
 * Don't ask me why these aren't defined in stdlib.h or unistd.h, but
 * they're not.
 */
extern "C"
    {
#ifndef __i386__
    extern int		accept(int s, struct sockaddr* addr, int* addrlen);
    extern int		bind(int s, const struct sockaddr* name,
			  int namelen);
#endif // __i386__
#ifdef __sparc__
    extern void		bzero(char* b, int length);
#endif // __sparc__
#ifndef __i386__
    extern int		connect(int s, const struct sockaddr* name,
			  int namelen);
#endif // __i386__
    extern char*	ctime(const time_t* clock);
#ifndef linux
    extern double	drand48(void);
    extern void		endpwent(void);
    extern int		fsync(int fd);
    extern int		getdtablesize(void);
    extern int		gethostname(char* name, int namelen);
#endif // linux
#ifndef __i386__
    extern int		getrlimit(int resource, const struct rlimit* rlp);
    extern int		gettimeofday(struct timeval* tp,
			  struct timezone* tzp);
    extern int		listen(int s, int backlog);
    extern int		lstat(const char* path, struct stat* statb);
#endif // __i386__
    extern int		open(const char* file, int mode, ...);
#ifndef __i386__
    extern int		pclose(FILE* file);
    extern int		readlink(const char*, char*, int);
    extern char*	realpath(const char* path,
			  char resolved_path[MAXPATHLEN]);
    extern int		select(int width, fd_set* readfds,
			  fd_set* writefds, fd_set* exceptfds,
			  struct timeval* timeout);
    extern void		setpwent(void);
    extern int		setsockopt(int s, int level, int optname, void* optval,
			  int optlen);
#endif // __i386__
    extern int		socket(int domain, int type, int protocol);
#ifndef linux
    extern void		srand48(long);
#endif // linux
    extern time_t	time(time_t* tloc);
#ifndef linux
    extern int		truncate(const char* path, const off_t length);
#endif // linux
    }

#endif // _CDECLS_HH
