#ifndef lint
static const char Rcs_Id[] =
    "$Id: cluster.cc,v 1.31 1998/10/23 23:14:56 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: cluster.cc,v $
 * Revision 1.31  1998/10/23 23:14:56  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.30  1998/10/13 06:23:13  geoff
 * Fix a number of problems in compareweights.
 *
 * Revision 1.29  1998/08/13 18:38:38  geoff
 * Remaining changes needed to support keeping overlapped cluster members
 * separately.  Somehow, these didn't get checked in (or didn't get
 * written) before I switched over to work on the bug-fix branch.
 *
 * Revision 1.28  1998/07/15 21:43:29  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.27  1998/07/08 20:55:23  geoff
 * Fix a bug in Cluster::combine that could cause duplicate files to
 * exist in a cluster.
 *
 * Revision 1.26  1998/05/14 01:15:23  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.25  1998/05/02 00:29:03  geoff
 * Put the element iterator in the cluster namespace, instead of using a
 * naming convention.
 *
 * Revision 1.24  1998/04/09 00:32:58  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.23  1998/03/05 03:18:14  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.22  1998/02/19 02:09:53  geoff
 * Add destructiveCopyFrom and emptySelf.  Take care that the cluster
 * size doesn't get reduced below zero when a member is removed.
 *
 * Revision 1.21  1997/12/18 01:24:12  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.20  1997/12/04 23:15:13  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.19  1997/01/02 06:48:15  geoff
 * Add a removeCurrent operator for the cluster iterator.
 *
 * Revision 1.18  1996/10/04 07:11:56  geoff
 * Add stability to the cluster sort by breaking ties based on cluster
 * size in blocks and then files (smaller preferred in both cases), and
 * on file name if all else fails.  This will reduce the phenomenon Mark
 * observed where successive hoardings alternate between a set of files
 * that had never been referenced.
 *
 * Revision 1.17  1996/09/11 22:59:31  geoff
 * Add combineDisjoint and findOverlap.  Use these to build combine().
 * Convert from dlists to slists.
 *
 * Revision 1.16  1996/06/11  06:13:02  geoff
 * Many changes to support investigated clusters.  Add a version
 * updateReferenceTime that scans the members.  Add compareWeights,
 * typeName, and name..  Don't automatically update the reference time
 * when files are added.  Display both total blocks and hoardable blocks
 * when showing clusters, and display the cluster type and name if they
 * are non-null.
 *
 * Revision 1.15  1996/05/03  01:11:33  geoff
 * Update the copyright.  Change countMinimumStorage to only count
 * unstored files.  Fix some hidden variables.
 *
 * Revision 1.14  1996/04/10  01:28:23  geoff
 * Pass the relevant manager name to countStorage.  In adjustStorage,
 * change the status for the proper manager.
 *
 * Revision 1.13  1996/02/02  20:09:39  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.12  1996/02/02  18:28:44  geoff
 * Fix some minor problems detected by the new C++ compiler.  Move
 * updateReferenceTime to the C code (this was probably a mistake).
 *
 * Revision 1.11  1995/11/10  02:58:20  geoff
 * Add zzinlinefunctions.hh.  Get rid of a spurious "const" declaration.
 *
 * Revision 1.10  1995/09/15  00:04:51  geoff
 * Globally change "cache" to "hoard".  Change the RCS ID to be a const.
 *
 * Revision 1.9  1995/08/07  20:02:21  geoff
 * Add the SO_EMPTY_CLUSTERS output option.
 *
 * Revision 1.8  1995/06/30  19:04:21  geoff
 * Change the cluster-storage code so that first all files must be set
 * unstored, and then desired files are set stored.  This allows proper
 * treatment of files that are in multiple clusters.  Add
 * countMinimumStorage and countMaximumStorage, and have adjustStorage
 * return the amount of space actually added to the hoard.  Make a number
 * of changes to the cluster-showing format.  When combining clusters, be
 * careful about duplicate files.
 *
 * Revision 1.7  1995/06/17  00:10:51  geoff
 * Make show be a parameterized routine, and operator<< just be a wrapper
 * around show.  Flush cerr/cout before all aborts.
 *
 * Revision 1.6  1995/05/27  00:15:15  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.5  1995/05/23  00:39:00  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.4  1995/05/05  22:41:03  geoff
 * Don't assume iostream.h has been been included.  Add a bunch of
 * PRIVATE_MANAGEMENT_DATA ifdefs.  Add operator<<.  Make show be a const
 * function.
 *
 * Revision 1.3  1995/03/29  23:57:07  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:03:43  geoff
 * Add a destructor which correctly removes files from the cluster (and
 * thus the cluster from the files).  Add removeFile.  Fix show() to
 * handle empty clusters.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "correlator.hh"
#include "cluster.hh"
#include "clustering_mgr.hh"
#include "file.hh"
#include "globalmanager.hh"
#include "managerbase.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <seer/seer.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			Cluster::~Cluster();
void			Cluster::destructiveCopyFrom(Cluster& source);
					// Copy efficiently, destroying source
void			Cluster::addNonoverlappedFile(File* newFile);
					// Add new file to cluster
void			Cluster::addOverlappedFile(File* newFile);
					// Add new overlapped file to cluster
void			Cluster::prepareNonoverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
void			Cluster::prepareOverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
void			Cluster::removeNonoverlappedFile(File* oldFile);
					// Remove old file from cluster
void			Cluster::removeOverlappedFile(File* oldFile);
					// Remove overlapped file from cluster
void			Cluster::emptyNonoverlaps();
					// Remove non-overlapped contents
void			Cluster::emptyOverlaps();
					// Remove overlapped cluster contents
void			Cluster::combineDisjoint(Cluster& that);
					// Combine two known-disjoint clusters,
					// emptying second
void			Cluster::combine(Cluster& that);
					// Combine two clusters that
					// ..may overlap, emptying second
unsigned long		Cluster::findOverlap(const Cluster& that)
			  const;	// Find overlap between two clusters
void			Cluster::updateReferenceTime();
					// Update last reference time
					// ..by scanning cluster members
void			Cluster::updateReferenceTime(ReferenceCount newTime);
					// Update last reference time
					// ..from a known value
void			Cluster::computeWeight();
					// Calculate hoarding weight
unsigned long		Cluster::countMinimumStorage(
			    const HoardingManagerBase& manager)
			  const;
					// Count min storage needed for cluster
unsigned long		Cluster::countMaximumStorage(
			    const HoardingManagerBase& manager)
			  const;
					// Count max storage needed for cluster
unsigned long		Cluster::adjustStorage(
			  const HoardingManagerBase& manager);
int			Cluster::compareWeights(const Cluster** cluster1,
			  const Cluster** cluster2);
			ClusteringManagerData::~ClusteringManagerData();
ostream&		operator<<(ostream& stream, const Cluster& cluster);
					// Display cluster on a stream
ostream&		Cluster::show(ostream& outFile,
			    ShowOptions options, const char* managerName)
			  const;
					// Show cluster in printable form,
					// ..with options
const char*		Cluster::typeName() const;
					// Return cluster type name (if any)
const char*		Cluster::encodedTypeName(ShowOptions options) const;
					// Return encoded cluster type name
const char*		Cluster::name() const;
					// Return cluster name (if any)
const char*		Cluster::encodedName(ShowOptions options) const;
					// Return encoded cluster name (if any)
void			Cluster::NonoverlappedIterator::removeCurrent();
					// Zap current w/o affecting iterator
void			Cluster::OverlappedIterator::removeCurrent();
					// Zap current w/o affecting iterator
void			Cluster::AllElementIterator::removeCurrent();
					// Zap current w/o affecting iterator
#endif // 0
				   
unsigned long		Cluster::nextID = 0;

extern void		correlatorPanic(const char* message);

Cluster::~Cluster()
    {
    if (!nonoverlappedMembers.isEmpty()  ||  !overlappedMembers.isEmpty())
	correlatorPanic("correlator:  attempt to destroy non-empty cluster");
    }

/*
 * Efficiently copy a cluster, destroying the source cluster in the
 * process.  This may seem like a strange operator, but it's useful in
 * the clustering manager, where some clusters must be transferred
 * from the old array of clusters to a new one.  Take a look at that
 * code if you're curious.
 *
 * Note that the cluster identification is generated anew, rather than
 * being copied from the source.  This is more useful for our
 * purposes.
 */
void Cluster::destructiveCopyFrom(
    Cluster&		source)
    {
    nonoverlappedMembers.destructiveCopyFrom(source.nonoverlappedMembers);
    overlappedMembers.destructiveCopyFrom(source.overlappedMembers);
    storage = source.storage;
    lastReference = source.lastReference;
    hoardWeight = source.hoardWeight;
    nonoverlappedBlocks = source.nonoverlappedBlocks;
    overlappedBlocks = source.overlappedBlocks;
    numNonoverlappedFiles = source.numNonoverlappedFiles;
    numOverlappedFiles = source.numOverlappedFiles;
    identification = nextID++;
    }

void Cluster::addNonoverlappedFile(	// Add non-overlapped file to a cluster
    File*		newFile)	// File to be added
    {
    numNonoverlappedFiles++;
    nonoverlappedBlocks += newFile->size();
    nonoverlappedMembers.pushHead((void*) newFile);
    }

void Cluster::addOverlappedFile(	// Add an overlapped file to a cluster
    File*		newFile)	// File to be added
    {
    numOverlappedFiles++;
    overlappedBlocks += newFile->size();
    overlappedMembers.pushHead((void*) newFile);
    }

/*
 * Prepare to remove a non-overlapped file from the cluster
 */
void Cluster::prepareNonoverlappedRemove(
    File*		oldFile)	// File to be removed
    {
    unsigned long size = oldFile->size();
    if (nonoverlappedBlocks > size)
	nonoverlappedBlocks -= size;
    else
	nonoverlappedBlocks = 0;
    numNonoverlappedFiles--;
    }

/*
 * Prepare to remove an overlapped file from the cluster
 */
void Cluster::prepareOverlappedRemove(
    File*		oldFile)	// File to be removed
    {
    unsigned long size = oldFile->size();
    if (overlappedBlocks > size)
	overlappedBlocks -= size;
    else
	overlappedBlocks = 0;
    numOverlappedFiles--;
    }

void Cluster::removeNonoverlappedFile(	// Remove a file from a cluster
    File*		oldFile)	// File to remove
    {
    this->prepareNonoverlappedRemove(oldFile);
    nonoverlappedMembers.remove(oldFile);
    }

void Cluster::removeOverlappedFile(	// Remove a file from a cluster
    File*		oldFile)	// File to remove
    {
    this->prepareOverlappedRemove(oldFile);
    overlappedMembers.remove(oldFile);
    }

void Cluster::emptyNonoverlaps()	// Remove non-overlapped contents
    {
    for (Cluster::NonoverlappedIterator element(*this);  element;  ++element)
	element.removeCurrent();
    }

void Cluster::emptyOverlaps()		// Remove overlapped cluster contents
    {
    for (Cluster::OverlappedIterator element(*this);  element;  ++element)
	element.removeCurrent();
    }

/*
 * Combine two clusters that are known to have no common members.
 * Unpleasant results may occur if common members actually exist.  The
 * second ("that") cluster is emptied by this operation.
 */
void Cluster::combineDisjoint(
    Cluster&		that)		// Cluster to combine with this one
    {
    if (this == &that)
	return;				// Don't combine cluster with itself!
    nonoverlappedBlocks += that.nonoverlappedBlocks;
    overlappedBlocks += that.overlappedBlocks;
    numNonoverlappedFiles += that.numNonoverlappedFiles;
    numOverlappedFiles += that.numOverlappedFiles;
    that.nonoverlappedBlocks = that.overlappedBlocks = 0;
    that.numNonoverlappedFiles = that.numOverlappedFiles = 0;
    nonoverlappedMembers.join(that.nonoverlappedMembers);
    overlappedMembers.join(that.overlappedMembers);
    }

/*
 * Combine two clusters that may have common members, eliminating
 * duplicates.  The second ("that") cluster is emptied by this
 * operation.
 */
void Cluster::combine(
    Cluster&		that)		// Cluster to combine with this one
    {
    if (this == &that)
	return;				// Don't combine cluster with itself!
    /*
     * Since a file can be a member of more than one cluster, we must
     * be careful that we don't put it in the same cluster twice when
     * combining.  Derived classes must also deal with this
     * possibility (see SortedCluster, for example).
     */
    (void)findOverlap(that);		// Unmark all overlapped files
    if (numOverlappedFiles <= that.numOverlappedFiles)
	{
	for (Cluster::OverlappedIterator element(that);  element;  ++element)
	    {
	    File& possibleOverlap(*element);
	    if (!possibleOverlap.isMarked())
		element.removeCurrent();
	    }
	}
    else
	{
	for (Cluster::OverlappedIterator element(*this);  element;  ++element)
	    {
	    File& possibleOverlap(*element);
	    if (!possibleOverlap.isMarked())
		element.removeCurrent();
	    }
	}
    nonoverlappedBlocks += that.nonoverlappedBlocks;
    overlappedBlocks += that.overlappedBlocks;
    numNonoverlappedFiles += that.numNonoverlappedFiles;
    numOverlappedFiles += that.numOverlappedFiles;
    that.nonoverlappedBlocks = that.overlappedBlocks = 0;
    that.numNonoverlappedFiles = that.numOverlappedFiles = 0;
    nonoverlappedMembers.join(that.nonoverlappedMembers);
    overlappedMembers.join(that.overlappedMembers);
    }

/*
 * Find the overlap between two clusters by counting the number of
 * files shared between them.  This can be used to detect
 * superset/subset relationships.
 *
 * As a side effect, we also leave all files in the smaller cluster
 * unmarked, and mark all files in the larger that are *not* in the
 * smaller.  If the clusters are the same size, "this" is considered
 * to be the smaller cluster.  The caller can depend upon this
 * behavior.  This allows easy calculation of the difference between
 * the clusters.
 *
 * If the return value is equal to the number of files in either
 * cluster, that cluster is a subset of the other.
 *
 * We require two passes through one cluster, and one through the
 * other, to develop the count.  For efficiency, we choose the smaller
 * cluster as the two-pass one.  In the first pass, we unmark all of
 * the files in the smaller cluster.  (By definition, we cannot assume
 * an initial state for file marks.)  We then mark the files in the
 * larger cluster, and then return to the smaller for a second pass,
 * counting marked files.
 */
unsigned long Cluster::findOverlap(
    const Cluster&	that)		// Second cluster to test
    const
    {
    const Cluster* smallCluster(this);	// Pointer to smaller cluster
    const Cluster* bigCluster(&that);	// Pointer to larger cluster
    if (numOverlappedFiles > that.numOverlappedFiles)
	{
	smallCluster = &that;
	bigCluster = this;
	}
    Cluster::OverlappedIterator smallIterator((Cluster&)*smallCluster);
    for (  ;  smallIterator;  ++smallIterator)
	(*smallIterator).unmark();
    for (Cluster::OverlappedIterator bigIterator((Cluster&)*bigCluster);
      bigIterator;
      ++bigIterator)
	(*bigIterator).mark();
    unsigned long overlap(0);
    for (smallIterator.reset();  smallIterator;  ++smallIterator)
	{
	File& file(*smallIterator);
	if (file.isMarked())
	    {
	    overlap++;
	    file.unmark();
	    }
	}
    return overlap;
    }

/*
 * Update the cluster's last reference time by scanning the members
 */
void Cluster::updateReferenceTime()
    {
    ReferenceCount latestTime(0);
    lastReference = 0;
    for (Cluster::AllElementIterator element(*this);  element;  ++element)
	{
	ReferenceCount elementTime(
	  clusteringManager.managerData(*element).lastReferenceTime());
	if (elementTime > latestTime)
	    latestTime = elementTime;
	}
    updateReferenceTime(latestTime);
    }

/*
 * Update the cluster's last reference time from a known value
 */
void Cluster::updateReferenceTime(
    ReferenceCount	newTime)	// New time to assign (conditionally)
    {
    if (lastReference < newTime)
	lastReference = newTime;
    }

void Cluster::computeWeight()		// Calculate hoarding weight
    {
    hoardWeight = clusteringManager.currentReferenceTime() - lastReference;
    }

/*
 * Calculate the maximum storage that would be needed if a cluster
 * were added to the hoard.  See File::countStorage for more comments.
 *
 * Bugs: this routine actually overestimates storage in many cases,
 * because it counts parent directories multiple times and potentially
 * can count link targets multiple times as well.
 */
unsigned long Cluster::countMaximumStorage(
    const HoardingManagerBase&
			manager)	// Manager we're counting for
    const
    {
    /*
     * Unfortunately, there is no simple way (that I know of) to
     * create a constant list iterator, so we'll do the following cast
     * to work around the fact that we're a const function.
     */
    unsigned long countedBlocks = 0;
    for (Cluster::AllElementIterator clusterMember((Cluster&)*this);
      clusterMember;
      ++clusterMember)
	countedBlocks
	  += (*clusterMember).countStorage(manager);
    return countedBlocks;
    }

/*
 * Calculate the minimum storage that would be needed if a cluster
 * were added to the hoard.  This routine assumes that all parent
 * directories and link targets outside the cluster are already
 * stored.  As such, it will underestimate slightly with respect to
 * directories, and drastically with respect to link targets.
 */
unsigned long Cluster::countMinimumStorage(
    const HoardingManagerBase&
			manager)	// Manager we're counting for
    const
    {
    /*
     * Unfortunately, there is no simple way (that I know of) to
     * create a constant list iterator, so we'll do the following cast
     * to work around the fact that we're a const function.
     */
    unsigned long countedBlocks = 0;
    for (Cluster::AllElementIterator clusterMember((Cluster&)*this);
      clusterMember;
      ++clusterMember)
	{
	File& file = *clusterMember;
	if (!manager.managerData(file).isStored())
	    countedBlocks += file.hoardableSize();
	}
    return countedBlocks;
    }

/*
 * Set the storage state of all files in a cluster so that it matches
 * the desired state.  Returns the number of blocks added to the
 * hoard.
 */
unsigned long Cluster::adjustStorage(
    const HoardingManagerBase&
			manager)	// Manager we're adjusting for
    {
    if (storage == NOT_STORED)		// If not-stored, files are correct
	return 0;
    unsigned long blocksAdded = 0;
    for (Cluster::AllElementIterator clustiter(*this);
      clustiter;
      ++clustiter)
	{
	File& file = *clustiter;
	blocksAdded += manager.managerData(file).store(file);
	}
    return blocksAdded;
    }

int Cluster::compareWeights(		// Compare cluster weights
    const Cluster**	cluster1,	// First cluster to compare
    const Cluster**	cluster2)	// Second cluster to compare
    {
    long		cmpval;		// Current comparison value

    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (*cluster1 == *cluster2)
	return 0;
    cmpval = 0;
    /*
     * The next sort key is frequency of use.  If the cluster contains
     * only one member and is accessed "often", it automatically gets
     * hoard priority.
     */
    int count1 = (*cluster1)->totalCount();
    File* sampleFile1 = NULL;
    if (count1 == 1)
	{
	Cluster::AllElementIterator grabFile((Cluster&)**cluster1);
	sampleFile1 = &(*grabFile);
	}
    int count2 = (*cluster2)->totalCount();
    File* sampleFile2 = NULL;
    if (count2 == 1)
	{
	Cluster::AllElementIterator grabFile((Cluster&)**cluster2);
	sampleFile2 = &(*grabFile);
	}
    if (sampleFile1 != NULL
      &&  clusteringManager.isOften(*sampleFile1,
	GlobalManager::getOftenPercent()))
	{
	if (sampleFile2 != NULL
	  &&  clusteringManager.isOften(*sampleFile2,
	    GlobalManager::getOftenPercent()))
	    cmpval =
	      clusteringManager.managerData(*sampleFile1).totalReferenceCount()
		- clusteringManager.managerData(*sampleFile2)
		  .totalReferenceCount();
	else
	    cmpval = -1;
	}
    else if (sampleFile2 != NULL
      &&  clusteringManager.isOften(*sampleFile2,
	GlobalManager::getOftenPercent()))
	cmpval = 1;
    if (cmpval != 0)
	return cmpval;
    /*
     * The next sort key is the cluster weight.  Smaller weights
     * indicate more desirable clusters.
     */
    cmpval = (*cluster1)->weight() - (*cluster2)->weight();
    if (cmpval != 0)
	return cmpval;
    /*
     * The next sort key is the cluster size in blocks.  Smaller
     * clusters are more desirable.
     */
    cmpval = (*cluster1)->totalSize() - (*cluster2)->totalSize();
    if (cmpval != 0)
	return cmpval;
    /*
     * The next sort key is the cluster size in files.  Smaller
     * clusters are more desirable.
     */
    cmpval = count1 - count2;
    if (cmpval != 0)
	return cmpval;
    /*
     * The following sort keys are simply an attempt to make the sort
     * stable, which produces better hoarding behavior because every
     * hoarding run produces consistent results.
     */
    if (count1 == 0  &&  count2 == 0)
	return cmpval;
    /*
     * The next sort key is the name of the sample file.  If either
     * cluster size is greater than 1, we need to set the
     * corresponding sample file because it wasn't set above.  (If the
     * size is zero, we've got big problems and we'll wind up bombing
     * on a null-pointer reference.)
     */
    if (count1 > 1)
	{
	Cluster::AllElementIterator grabFile((Cluster&)**cluster1);
	sampleFile1 = &(*grabFile);
	}
    if (count2 > 1)
	{
	Cluster::AllElementIterator grabFile((Cluster&)**cluster2);
	sampleFile2 = &(*grabFile);
	}
    cmpval = strcmp(sampleFile1->name(), sampleFile2->name());
    return cmpval;
    }

ostream& operator<<(			// Display cluster on a stream
    ostream&		stream,		// Stream to display on
    const Cluster&	cluster)	// Cluster to display
    {
    return cluster.show(stream, SO_EVERYTHING, NULL);
					// Just use virtual function
    }

/*
 * Show cluster in printable form, with options
 *
 * Show options honored: SO_EMPTY_CLUSTERS, SO_FULL_FILE_INFORMATION
 */
ostream& Cluster::show(
    ostream&		outFile,	// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    if (numNonoverlappedFiles + numOverlappedFiles == 0)
	{
	if (options & SO_EMPTY_CLUSTERS)
	    outFile << "CH\tCluster " << id()
	      << " lastReference 0 0 files 0 blocks, "
	      << (storage == STORED ? "STORED\n" : "NOT_STORED\n")
	      << "CE\n";
	return outFile;
	}
    /*
     * Unfortunately, there is no simple way (that I know of) to
     * create a constant list iterator, so we'll do the following cast
     * to work around the fact that we're a const function.
     */
    Cluster::AllElementIterator clusterIterator((Cluster&)*this);
    unsigned long hoardableBlocks = 0;
    for (  ;  clusterIterator;  ++clusterIterator)
	hoardableBlocks += (*clusterIterator).hoardableSize();
    outFile << "CH\tCluster " << id() << " lastReference "
      << lastReference << ' ' << totalCount() << " file"
      << (totalCount() == 1 ? " " : "s ")
      << hoardableBlocks << ' '
      << totalSize() << " block" << (totalSize() == 1 ? "" : "s")
      << (storage == STORED ? " STORED" : " NOT_STORED");
    const char* clusterTypeName = encodedTypeName(options);
    if (clusterTypeName != NULL)
	outFile << ' ' << clusterTypeName;
    const char* clusterName = encodedName(options);
    if (clusterName != NULL)
	outFile << (clusterTypeName == NULL ? ' ' : ':') << clusterName;
    outFile << '\n';
    for (clusterIterator.reset();  clusterIterator;  ++clusterIterator)
	{
	if ((options & SO_FULL_FILE_INFORMATION) == 0)
	    outFile << "CD\t    ";
	(*clusterIterator).show(outFile, options, managerName);
	outFile << '\n';
	}
    outFile << "CE\n";
    return outFile;
    }

const char* Cluster::typeName()		// Return cluster type name (if any)
    const
    {
    return NULL;
    }

const char* Cluster::encodedTypeName(	// Return encoded cluster type name
    ShowOptions			)	// Encoding options
    const
    {
    return NULL;
    }

const char* Cluster::name()		// Return cluster name (if any)
    const
    {
    return NULL;
    }

const char* Cluster::encodedName(	// Return encoded cluster name (if any)
    ShowOptions			)	// Encoding options
    const
    {
    return NULL;
    }

/*
 * Delete the current element without affecting the surrounding iterator.
 */
void Cluster::NonoverlappedIterator::removeCurrent()
    {
    cluster.prepareNonoverlappedRemove(&(operator*()));
    SList::Iterator::removeCurrent();
    }

/*
 * Delete the current element without affecting the surrounding iterator.
 */
void Cluster::OverlappedIterator::removeCurrent()
    {
    cluster.prepareOverlappedRemove(&(operator*()));
    SList::Iterator::removeCurrent();
    }

/*
 * Delete the current element without affecting the surrounding iterator.
 */
void Cluster::AllElementIterator::removeCurrent()
    {
    if (nonoverlapIterator)
	nonoverlapIterator.removeCurrent();
    else
	overlapIterator.removeCurrent();
    }
