#ifndef _CLUSTERING_MGR_HH
#define _CLUSTERING_MGR_HH
/*
 * $Id: clustering_mgr.hh,v 1.35 2005/04/13 21:03:02 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: clustering_mgr.hh,v $
 * Revision 1.35  2005/04/13 21:03:02  geoff
 * Make markDirty non-inline.
 *
 * Revision 1.34  1998/10/23 23:15:03  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.33  1998/08/13 18:38:41  geoff
 * Integrate changes from the bug-fix branch:
 *     Add meaningless parameters to the reference-recording routines.  Get
 *     prepareFile and fillHoard involved in deciding the hoard space.
 *
 *     Many changes to make the last-reference time and similar things be
 *     manager-specific.
 *
 * Revision 1.32  1998/06/30 20:20:44  geoff
 * Preliminary changes to support separate tracking of overlapped and
 * nonoverlapped cluster members.  This code is only partially debugged
 * at this point.
 *
 * Revision 1.31  1998/06/05 18:06:44  geoff
 * Make cleanReferences and relationsChanged virtual.  Add functions to
 * manipulate the dirty bit.  Add usingSharedNeighbors.
 *
 * Revision 1.30  1998/05/14 01:15:25  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.29  1998/04/09 00:32:59  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.28  1998/03/05 03:18:15  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.27  1998/02/19 02:09:57  geoff
 * Add new routines and declarations to support incremental clustering.
 *
 * Revision 1.26  1997/12/18 01:24:27  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.25  1997/05/16 17:10:34  geoff
 * Add wasOften support and the file argument to the clear routine.  This
 * is the version used to run the simulations in the dissertation.
 *
 * Revision 1.24  1997/01/02 06:48:17  geoff
 * Add forgetCluster,requiredSize, and support for the cluster iterator's
 * removeCurrent operator.
 *
 * Revision 1.23  1996/12/03 02:57:52  geoff
 * Add support for reading and writing parameters and calculating the
 * equivalent LRU depth needed.
 *
 * Revision 1.22  1996/09/11 22:59:36  geoff
 * Convert from dlists to slists.  Add combineDisjoint, computeWeight,
 * and suppressDuplicates.
 *
 * Revision 1.21  1996/06/11  06:13:05  geoff
 * Get rid of compareWeights (it's now in the Cluster base class).
 * Change a couple of other declarations to reflect the new inheritance
 * setup for clusters.
 *
 * Revision 1.20  1996/05/03  01:11:43  geoff
 * Update the copyright.  Rename the CD_xxx flags to include strategic
 * underscores.
 *
 * Revision 1.19  1996/04/10  06:09:28  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.18  1996/04/10  01:28:45  geoff
 * Add the new ManagerData::manager() routine.
 *
 * Revision 1.17  1996/03/19  18:43:16  geoff
 * Inherit from HoardingManagerBase.  Add support for the
 * DERIVED_CLASS_BUG ifdef.
 *
 * Revision 1.16  1996/02/02  18:29:23  geoff
 * Add the new derived data accessors.  Declare inline functions as such
 * in the class definition.  Declare all virtual functions as such in
 * derived classes.  Add clusterCount().
 *
 * Revision 1.15  1995/12/19  00:35:00  geoff
 * Add some missing comments.  Add support for the
 * shared-nearest-neighbor nearness metric.  Make the cluster iterator be
 * a derived class.
 *
 * Revision 1.14  1995/11/22  00:22:44  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.13  1995/11/21  21:39:31  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support.
 *
 * Revision 1.12  1995/11/10  03:00:07  geoff
 * Add NEW_FILE_RELATIONS support.  Get rid of a spurious "const".  Don't
 * include the icc file, since zzinlinefunctions now does that.
 *
 * Revision 1.11  1995/10/10  20:38:22  geoff
 * Get rid of the COMPARISON_SUPPORT, SIMULATION_SUPPORT, and
 * TRACK_MEMORY_LEAKS ifdefs.
 *
 * Revision 1.10  1995/09/15  00:06:28  geoff
 * Add InternalTimer support.  Add abortive TRACK_MEMORY_LEAKS support.
 * Add statistics support.
 *
 * Revision 1.9  1995/08/07  20:03:29  geoff
 * Change some declarations as part of adding investigator support.
 *
 * Revision 1.8  1995/06/30  19:05:03  geoff
 * Get rid of adjustStorage.
 *
 * Revision 1.7  1995/06/17  00:12:19  geoff
 * Add show options and clearManagerData.  Add some "const" keywords, and
 * get rid of some "class" ones.  Reduce the flags to an unsigned char.
 *
 * Revision 1.6  1995/05/27  00:16:44  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.5  1995/05/23  00:40:28  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.4  1995/05/05  22:42:15  geoff
 * Add a lot of preliminary PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.3  1995/04/20  21:58:31  geoff
 * Add support for dumping management results to a stream.
 *
 * Revision 1.2  1995/03/29  21:05:09  geoff
 * Support the new parameter-setting style.  Declare some things properly
 * const.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cluster.hh"
#include "communication.hh"
#include "filerelation.hh"
#include "managerbase.hh"
#include "showoptions.hh"

class ByteSwappedStream;
class ClusteringAllClusterIterator;
class ClusteringManager;
class ClusteringOverlapIterator;
class ClusteringNonoverlapIterator;
class File;
class FileReference;
class InternalTimer;
class SortedCluster;
class ostream;

/*
 * Class to hold management data for files
 */
class ClusteringManagerData : public HoardingManagerDataBase
    {
    friend ClusteringManager;
    friend ClusteringAllClusterIterator;
    friend ClusteringOverlapIterator;
    friend ClusteringNonoverlapIterator;
    public:
			ClusteringManagerData(File& containingFile);
	virtual		~ClusteringManagerData();
	/*
	 * This should return ClusteringManager&, but that
	 * requires a cross-reference to a derived class, which at
	 * least g++ can't handle.
	 */
	virtual HoardingManagerBase&
			manager() const; // Return pointer to our manager
	virtual void	recordMissedReference(File& file,
			  ReferenceType referenceType,
			  bool isMeaningless);
					// Record that a file miss happened
	virtual void	recordReference(File& file, ReferenceType type,
			  bool isMeaningless);
					// Keep track of a file reference
	virtual ByteSwappedStream&	// Read binary version from stream
			read(ByteSwappedStream& stream, File& file);
	virtual ByteSwappedStream&	// Write binary version to stream
			write(ByteSwappedStream& stream,
			    const File& file)
			  const;
	virtual void	clear(const File& file);
					// Clear management data
	virtual ostream& show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
	inline bool	willCluster() const; // Test if file will cluster
	inline void	okToCluster();	// Mark file as OK to cluster
	void		joinNonoverlappedCluster(SortedCluster& newCluster);
					// Add cluster to file's cluster list
	void		joinOverlappedCluster(SortedCluster& newCluster);
					// Add cluster to file's cluster list
	void		leaveNonoverlappedCluster(SortedCluster& oldCluster);
					// Remove cluster from file's list
	void		leaveOverlappedCluster(SortedCluster& oldCluster);
					// Remove cluster from file's list
	bool		isInCluster(const SortedCluster* cluster) const;
					// See if file is in particular cluster
	bool		inSameCluster(
			  const ClusteringManagerData& that) const;
					// See if two files are in same cluster
	inline SortedCluster*
			primaryCluster() const;
					// Return file's primary cluster
	inline unsigned long
			clusterCount() const;
					// Return count of clusters this file
					// ..is a member of
	void		prepare();	// Prepare for clustering
	virtual void	cleanReferences();
					// Get rid of obsolete references
	inline void	markClean();	// Clear the dirty bit
	void		markDirty();	// Set the dirty bit
	inline bool	isDirty();	// Test the dirty bit
	virtual void	relationsChanged();
					// Note that relations have changed
	inline ReferenceCount totalReferenceCount() const;
					// Return total no. of references
	ByteSwappedStream& readTotalReferences(ByteSwappedStream& stream);
					// Read total refs from a stream
    private:
			ClusteringManagerData(
			  ClusteringManagerData& that);
					// Disable assignment
	ClusteringManager&
			operator=(const ClusteringManagerData& that);
					// Disable copy
	virtual ClusteringManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const ClusteringManagerData&
			    managerData(const File& file)
			  const;
					// Provide data access
	inline void	setWasOften();	// Set the was-often flag
	inline bool	wasOften() const; // See if file was isOften on last
					// ..run
	/*
	 * Private data
	 */
	File&		file;		// File we are a member of
	unsigned char	flags;		// Flags for this file
	SortedCluster*	cluster;	// Cluster this file is a member of
	SList		overlappingClusters; // Clusters this file overlaps
	ReferenceCount	totalReferences; // Total refs this file has seen
	/*
	 * Values for the flags field.  These should be defined using
	 * 'const int', but ANSI C++ forbids it, so I'll use an enum
	 * to kludge it.
	 */
	enum
	    {
	    CD_WEIGHTED = 0x01,		// Weighting has been done on file
	    CD_WILL_CLUSTER = 0x02,	// This file can be clustered
	    CD_HAS_CLUSTERED = 0x04,	// This file's been clustered
	    CD_IS_NEIGHBOR = 0x08,	// This file is a nearest
					// ..neighbor (see isSharedNearby)
	    CD_WAS_OFTEN = 0x10,	// This file was isOften at last
					// ..clustering (see requiredSize)
	    CD_DIRTY = 0x20		// This file will cluster differently
					// ..the next time clustering is run
	    };
    };

class ClusteringManager : public HoardingManagerBase
    {
    friend SortedCluster;
    friend ClusteringManagerData;
    public:
			ClusteringManager(const char* managerName,
			  const char* managerDefaults);
	virtual		~ClusteringManager();
	virtual ClusteringManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const ClusteringManagerData&
			managerData(const File& file)
			  const;
					// Provide data access
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
	virtual void	showParameters(ostream& stream) const;
					// Display management parameters
	virtual long	writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
	virtual long	readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
	virtual void	cleanGlobalData(); // Clean out global file references
	void		forgetCluster(Cluster* cluster);
					// Get rid of pointer to a disappearing
					// ..cluster
	virtual ostream& showStatistics(ostream& stream) const;
					// Display collected quality statistics
	virtual ostream& requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const;
					// Display hoard size needed for
					// ..no misses
	static inline bool
			usingSharedNeighbors();
					// Test value of sharedNeighbors
	static inline bool
			isLRU_Equivalent(const File& file);
					// Decide if file is in equiv. hoard
	static inline ReferenceCount
			equivalentLRU_Depth();
					// Return LRU depth for no misses
	inline bool	isOften(const File& file, double oftenPercent) const;
					// Test for often-referenced file
	virtual void	stateChanged(ClusteringManagerData* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Note that state has changed
    private:
	ClusteringManager&	// Disable copy
			operator=(const ClusteringManager& that);
	unsigned long	prepareFile(File& fileent);
	static int	compareReferences(const FileReference* reference1,
			  const FileReference* reference2);
	void		clusterTightFiles(File& fileent) const;
					// Combine tightly-related files into
					// ..clusters
	void		clusterLooseFiles(File& fileent) const;
					// Add loosely-related files to
					// ..clusters
	void		fillHoard(unsigned long availableSpace);
					// Decide what goes in the hoard
	bool		isNearby(const File& referencingFile,
			    const RelatedFile& relatedFile)
			  const;
					// Decide if two files are close
	bool		isVeryNearby(const File& referencingFile,
			    const RelatedFile& relatedFile)
			  const;
					// Decide if two files are very close
	bool		isSharedNearby(const File& referencingFile,
			    const RelatedFile& relatedFile,
			    int sharedThreshold)
			  const;
					// Decide if two files are close, using
					// ..the shared-nearest-neighbor method
	/*
	 * Private data
	 */
	double		alwaysClusterDistance;
					// Files this close always cluster
	int		closestK;	// Max closest files to cluster
	double		directoryDistanceWeight;
					// Use directory distances in nearness
					// ..measurements
	bool		incremental;	// Use incremental clustering algorithm
	double		investigatedRelationWeight;
					// Weight to give to investigated
					// ..relationships
	unsigned int	minimumReferences; // Min refs needed for valid dist
	double		sometimesClusterDistance;
					// These files cluster if closest
	int		tightPercent;	// Percent of total references needed
					// ..for "tight" binding
	int		sharedLooseThreshold;
					// No. of shared neighbors needed
					// ..for loose clustering
	int		sharedTightThreshold;
					// No. of shared neighbors needed
					// ..for tight clustering
	SortedCluster*	clusterTable;	// Table of known clusters
	unsigned long	clusterNumber;	// Index into clustertab
	unsigned long	clusterCount;	// Number of file clusters known
	Cluster**	sortedClusters;	// Table of clusters, priority-sorted
	unsigned long	sortedClusterCount; // Number of sorted clusters known
	/*
	 * The following parameters must be static because they are
	 * referred to from static member functions.
	 */
	static bool	sharedNeighbors; // True to use shared-neighbors
					// ..similarity measure
	static unsigned int
			mediumReferences; // Refs needed for reliable distance
	static ReferenceCount
			oldestReference; // Oldest file referred to recently
					// ..used to find equivalent LRU depth
    };

/*
 * Derived class to hold sorted clusters
 */
class SortedCluster : public Cluster
    {
    public:
	inline		SortedCluster();
	virtual		~SortedCluster();
	virtual void	destructiveCopyFrom(SortedCluster& source);
					// Copy efficiently, destroying source
	virtual void	addNonoverlappedFile(File* newFile);
					// Add new file to cluster
	virtual void	addOverlappedFile(File* newFile);
					// Add new file to cluster
	virtual void	prepareNonoverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
	virtual void	prepareOverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
	virtual void	combineDisjoint(Cluster& that);
					// Combine two known-disjoint clusters,
					// emptying second
	virtual void	combine(Cluster& that); // Combine two clusters that
					// ..may overlap, emptying second
	virtual void	computeWeight(); // Calculate hoarding weight
	virtual unsigned long
			adjustStorage(	// Cause files to be stored/unstored
			  const HoardingManagerBase& manager);
	void		suppressDuplicates()
			  const;	// Get rid of any clusters that
					// ..duplicate this one
    };

/*
 * Iterator class for walking a file's overlapping cluster list.
 */
class ClusteringOverlapIterator : public SList::Iterator
    {
    public:
	inline		ClusteringOverlapIterator(
			  ClusteringManagerData& data);
	inline SortedCluster&
			operator*() const; // Return element referred to
    private:
	ClusteringManagerData&
			target;		// Data whose list we're walking
    };

/*
 * Iterator class for walking a file's non-overlapping cluster list.
 * (Since there can only be one cluster, this is currently a fairly
 * fake iterator.)
 */
class ClusteringNonoverlapIterator
    {
    public:
	inline		ClusteringNonoverlapIterator(
			  ClusteringManagerData& data);
	inline		~ClusteringNonoverlapIterator();
	inline		operator bool() const; // For completion testing
	inline SortedCluster&
			operator*() const; // Return element referred to
	inline ClusteringNonoverlapIterator&
			operator++();	// Step to next entry
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			ClusteringNonoverlapIterator(
			  const ClusteringNonoverlapIterator&);
	ClusteringNonoverlapIterator&
			operator=(const ClusteringNonoverlapIterator&);
	/*
	 * Private data
	 */
	ClusteringManagerData*
			target;		// Data whose list we're walking
    };

/*
 * Iterator class for walking a file's entire cluster list.
 */
class ClusteringAllClusterIterator
    {
    public:
	inline		ClusteringAllClusterIterator(
			  ClusteringManagerData& data);
	inline		~ClusteringAllClusterIterator();
	inline		operator bool() const; // For completion testing
	inline SortedCluster&
			operator*() const; // Return element referred to
	inline ClusteringAllClusterIterator&
			operator++();	// Step to next entry
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			ClusteringAllClusterIterator(
			  const ClusteringAllClusterIterator&);
	ClusteringAllClusterIterator&
			operator=(const ClusteringAllClusterIterator&);
	/*
	 * Private data
	 */
	ClusteringNonoverlapIterator
			nonoverlapIterator;
	ClusteringOverlapIterator
			overlapIterator;
    };

extern ClusteringManager clusteringManager; // Master manager data structure

#define _CLUSTERING_MGR_ICC_NEEDED

#endif // _CLUSTERING_MGR_HH
