#ifndef _COMPILEOPTIONS_HH
#define _COMPILEOPTIONS_HH
/*
 * $Id: compileoptions.hh,v 1.27 1998/03/05 03:18:15 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: compileoptions.hh,v $
 * Revision 1.27  1998/03/05 03:18:15  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.26  1997/12/18 01:24:29  geoff
 * Get rid of some obsolete development ifdefs.  Move some ifdefs from
 * the observer to here.
 *
 * Revision 1.25  1997/01/02 06:48:18  geoff
 * Add debugging options to get the LRU manager to dump useful
 * information related to the required-size calculations.
 *
 * Revision 1.24  1996/10/11 22:08:55  geoff
 * Get rid of the workaround for the floating-output bug.
 *
 * Revision 1.23  1996/10/08 19:50:03  geoff
 * Add FLOATING_OUTPUT_BUG.
 *
 * Revision 1.22  1996/09/11 22:59:38  geoff
 * Add COMPLEX_CLUSTER_WEIGHTS, LINEAR_CLUSTER_COMBINE,
 * SUPPRESS_DUPLICATE_CLUSTERS, REPORT_SELECTED_FAILURES, and
 * USE_TYPE_ALLOCATOR.  Turn off DERIVED_CLASS bug for all OS versions.
 * Add TIMEVALS_ARE_INTS to work around Linux idiocy.
 *
 * Revision 1.21  1996/06/11  07:04:51  geoff
 * Turn on MISSING_BECOMES_UNSTORED.  Get rid of READ/WRITE EXTRAS and
 * PROBABLY_MEANINGLESS.
 *
 * Revision 1.20  1996/06/11  06:13:06  geoff
 * Add PROBABLY_MEANINGLESS (but disabled) and FICUS_DELETE_WORKAROUND.
 * Turn off FICUS_MISSING_DIR_BUG.
 *
 * Revision 1.19  1996/05/03  01:11:48  geoff
 * Update the copyright.  Add MISSING_BECOMES_UNSTORED (disabled) and
 * READ/WRITE_EXTRAS_FROM/TO_SAVE_FILE.
 *
 * Revision 1.18  1996/03/19  18:42:48  geoff
 * Add DERIVED_CLASS_BUG, and re-enable FICUS_MISSING_DIR_BUG.
 *
 * Revision 1.17  1996/02/02  18:28:45  geoff
 * Add CHECK_PROCESS_EXITS as a compilation option so I can turn it off
 * someday.
 *
 * Revision 1.16  1995/12/19  01:02:33  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.15  1995/12/19  00:34:07  geoff
 * Add TRACK_PROCESSES.  Turn off FICUS_MISSING_DIR_BUG; I don't think
 * the bug ever existed.  Add ifdefed-out new/delete macros that attempt
 * to improve the dmalloc support, but fail because the break operator
 * new and operator delete declarations.
 *
 * Revision 1.14  1995/11/22  00:19:52  geoff
 * Get rid of the TIME_SORTED_INPUT, HAND_BYTE_SWAPPING, and
 * NEW_FILE_RELATIONS ifdefs.
 *
 * Revision 1.13  1995/11/10  02:58:27  geoff
 * Add LENGTH_IN_STAT_RECORD, TIME_SORTED_INPUT, FICUS_STAT_BUG,
 * FICUS_MISSING_DIR_BUG, and NEW_FILE_RELATIONS (whew!).
 *
 * Revision 1.12  1995/10/10  20:37:44  geoff
 * Get rid of the obsolete TRACK_MEMORY_LEAKS, COMPARISON_SUPPORT, and
 * SIMULATION_SUPPORT stuff, and turn off the dmalloc option.
 *
 * Revision 1.11  1995/09/15  00:05:02  geoff
 * Globally change "cache" to "hoard".  Add TRACK_MEMORY_LEAKS and
 * USE_DMALLOC_LIBRARY.  Add SIMULATION_SUPPORT and COMPARISON_SUPPORT.
 *
 * Revision 1.10  1995/08/07  20:02:27  geoff
 * Add the ASSIGNABLE_FILE_IDS and TRACK_CWD options.
 *
 * Revision 1.9  1995/06/17  00:10:56  geoff
 * Add LRU_SHOW_ONLY_CACHEABLE.
 *
 * Revision 1.8  1995/05/27  00:15:42  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.7  1995/05/23  00:39:15  geoff
 * Turn off KEEP_CLUSTER_TABLE and turn on PRIVATE_MANAGEMENT_DATA.  Add
 * UNBUFFERED_STREAMS and FICUS_KNOWLEDGE.
 *
 * Revision 1.6  1995/05/05  22:41:18  geoff
 * Get rid of some obsolete ifdefs, and add a few new ones.  Move MAX_K
 * to this file.
 *
 * Revision 1.5  1995/04/20  21:57:21  geoff
 * Remove AFTER_EXEC and TRUST_INHERITANCE, and add USE_TIGHT_FRACTION.
 *
 * Revision 1.4  1995/03/30  00:44:18  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.3  1995/03/29  23:57:12  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:03:54  geoff
 * Switch to the new file table and parameter setting, and add a couple
 * of new temporary options.  Add the NAME_IN_STAT_RECORD debugging
 * option.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

/*
 * This file contains definitions of compilation options that are
 * tested in various files with #ifdef/#ifndef.
 */

/*
 * Things that I'm not sure should work this way
 */
#define LRU_SHOW_ONLY_HOARDABLE		// LRU mgr shows only hoardable files

/*
 * Debugging options
 */
#define NAME_IN_STAT_RECORD		// Put file names in ST_STAT records
#define ASSIGNABLE_FILE_IDS		// Allow file ID's to be assigned
					// ..according to a sensible scheme
#undef USE_DMALLOC_LIBRARY		// Don't expect malloc-debugging lib
#define LENGTH_IN_STAT_RECORD		// Put file size in ST_STAT records
#define CHECK_PROCESS_EXITS		// Make sure processExits gets run
#undef LRU_DUMP_TRIGGER_FILES		// Don't show files that trigger
					// ..required hoard size
#undef LRU_DUMP_EXTRA_HOARD		// Don't show files that are hoarded
					// ..unnecessarily

#ifdef USE_DMALLOC_LIBRARY
#include "dmalloc.h"
#if 0
/*
 * The following excessive cleverness doesn't work because it causes
 * parse errors in declarations of operator new.  Oh, well.
 */
#define new (_dmalloc_file = __FILE__, _dmalloc_line = __LINE__), new
#define delete (_dmalloc_file = __FILE__, _dmalloc_line = __LINE__), delete
#endif // 0
#define MALLOC_0_RETURNS_NULL
#endif // USE_DMALLOC_LIBRARY

/*
 * Prudence options
 */
#undef UNBUFFERED_STREAMS		// Don't buffer streams used w/ select

/*
 * Temporary development options
 */
#define USE_TYPE_ALLOCATOR		// Use TypeAllocator for some
					// ..allocation

/*
 * Configuration options
 */

/*
 * System and performance options
 */
#undef FICUS_KNOWLEDGE			// Don't be aware of Ficus filesystems
#ifdef __i386__
#undef FICUS_KNOWLEDGE			// Ficus doesn't run on 386's
#endif // __i386__
#ifdef __linux__
#define TRACK_CWD			// Observer has code to track CWD
#define TIMEVALS_ARE_INTS		// Idiots made tv_xxx ints!
#undef TRIM_DOUBLE_SLASH		// Kernel gives OK paths
#endif // __linux__
#ifndef __linux__
#define FICUS_STAT_BUG			// Ficus may report st_blocks wrong
#undef FICUS_MISSING_DIR_BUG		// Ficus doesn't produce missing dirs
#define FICUS_DELETE_WORKAROUND		// Work around Ficus deletion delays
#undef TIMEVALS_ARE_INTS		// tv_xxx is unsigned
#define TRIM_DOUBLE_SLASH		// Kernel gives paths with // in them
#endif // __linux__

/*
 * Define the following to resolve symlinks properly, at the expense
 * of more system load.
 */
#define USE_REALPATH			// Use realpath() from libc

/*
 * Define the following to strip have the observer strip /tmp_mnt from
 * the head of a path, so that automounter-mounted directories are
 * processed correctly.
 */
#undef STRIP_TMPMNT			// Strip /tmp_mnt from path heads
#define TMPMNT_PATH	"/tmp_mnt/"

/*
 * Define the following to have the observer strip extra ":gp" values
 * from the tail of a path.
 */
#ifdef __sparc__
#define STRIP_COLONGP			// Strip :gp from path tails
#endif // __sparc__

/*
 * Define the following to have the observer suppress references to
 * files that are identified only by their ffsh.
 */
#ifdef __sparc__
#define SUPPRESS_FFSH
#endif // __sparc__

/*
 * Define the following to have the observer strip all extra Ficus
 * values (:*) from within pathnames.
 */
#ifdef __sparc__
#define STRIP_FICUS_QUALIFIERS
#endif // __sparc__

/*
 * Maximum values certain system parameters can take.
 */

#endif // _COMPILEOPTIONS_HH
