#ifndef lint
static const char Rcs_Id[] =
    "$Id: connectserver.cc,v 1.11 1996/09/11 22:59:39 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: connectserver.cc,v $
 * Revision 1.11  1996/09/11 22:59:39  geoff
 * Add a missing gethostbyname that prevented connecting to remote
 * servers.
 *
 * Revision 1.10  1996/05/03  01:11:49  geoff
 * Update the copyright.
 *
 * Revision 1.9  1996/04/10  01:28:25  geoff
 * Get rid of the debugging code that detected excess fd use.  Set
 * SO_LINGER on all sockets.
 *
 * Revision 1.8  1996/02/02  18:28:46  geoff
 * Correct a minor type mismatch.
 *
 * Revision 1.7  1995/11/10  02:58:28  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.6  1995/10/10  20:37:45  geoff
 * Move an include to get rid of warnings under Linux.
 *
 * Revision 1.5  1995/09/15  00:05:04  geoff
 * Change the RCS ID to be a const.
 *
 * Revision 1.4  1995/06/17  00:10:58  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.
 *
 * Revision 1.3  1995/04/20  21:57:23  geoff
 * Make the arguments to connectServer be enums.  Get rid of a spurious
 * gethostbyname.
 *
 * Revision 1.2  1995/03/29  21:03:57  geoff
 * Add the reportFailures option.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
extern "C"
    {
#include <seer/seer.h>
    }

int			connectServer(char* serverid, int type,
			  ConnectionBlocking blockingType,
			  ConnectionFailureReporting reportingMethod);
void			setnonblock(int fd);

#define INVALID_INETADDR -1		// Error return from inet_addr

/*
 * Temporary stuff
 */
#define SERVER_BASE_PORT	5000
#define LOCALPATH	"/tmp/.seer_"

extern "C" char*	strerr(int errnum);

int connectServer(		// Connect to the correlator server
    char*		serverid, // String identifying which server
    int			type,	// Connection-type ID
    ConnectionBlocking	blockingType, // Type of busy-blocking desired
    ConnectionFailureReporting	// How to report failures
			reportingMethod)
    {
    struct sockaddr*	addr;	// Properly-typed pointer to socket address
    int			addrlen; // Length of the address
    char		conntype; // Byte for sending connection type
    int			fd;	// File descriptor of opened socket
    struct hostent*	hostp;	// Host-table entry for server host
    long		hostinetaddr; // Internet address of server host
    struct sockaddr_in	inaddr;	// IP socket data block
    int			nbytes;	// Number of bytes written to server
    char*		portp;	// Pointer to port portion of server ID
    struct sockaddr_un	unaddr;	// UNIX socket data block

    if (serverid == NULL)
	portp = NULL;
    else
	{
	portp = strchr(serverid, ':');
	if (portp != NULL)
	    *portp++ = '\0';
	}
    /*
     * This stuff is based on the X library.
     *
     * If the server ID is null, we'll use an AF_UNIX socket to a
     * local server, else we'll use TCP streams.
     */
    if (serverid == NULL  ||  serverid[0] == '\0')
	{
	// AF_UNIX socket
	serverid = "local server";
	unaddr.sun_family = AF_UNIX;
	(void) strcpy(unaddr.sun_path, LOCALPATH);
	(void) strncat(unaddr.sun_path, portp == NULL ? "0" : portp,
	  sizeof unaddr.sun_path);
	addrlen = strlen(unaddr.sun_path) + sizeof(unaddr.sun_family);
	addr = (struct sockaddr*) &unaddr;
	}
    else
	{
	/*
	 * TCP socket
	 *
	 * If numeric host name, then try to parse it as such.  Do the
	 * number first because some systems return garbage instead of
	 * INVALID_INETADDR.
	 */
	if (isascii(serverid[0])  &&  isdigit(serverid[0]))
	    hostinetaddr = inet_addr(serverid);
	else
	    hostinetaddr = (unsigned long) INVALID_INETADDR;
	if (hostinetaddr != INVALID_INETADDR)
	    {
	    inaddr.sin_family = AF_INET;
	    inaddr.sin_addr.s_addr = hostinetaddr;
	    }
	else
	    {
	    // Try by name
	    hostp = gethostbyname(serverid);
	    if (hostp == NULL)
		{
		cerr << "connectserver:  Couldn't find host " << serverid
		  << '\n';
		return -1;
		}
	    else if (hostp->h_addrtype != AF_INET)
		{
		cerr << "connectserver:  " << serverid
		  << " isn't an IP host\n";
		return -1;
		}
	    // Set up socket data
	    inaddr.sin_family = hostp->h_addrtype;
	    memcpy((char*) &inaddr.sin_addr, (char*) hostp->h_addr,
	      sizeof (inaddr.sin_addr));
	    }
	addr = (struct sockaddr*) &inaddr;
	addrlen = sizeof inaddr;
	inaddr.sin_port = SERVER_BASE_PORT;
	if (portp != NULL)
	    inaddr.sin_port += atoi(portp);
	inaddr.sin_port = htons(inaddr.sin_port);
	}
    if (portp != NULL)
	portp[-1] = ':';
    fd = socket((int) addr->sa_family, SOCK_STREAM, 0);
    if (fd < 0)
	{
	if (reportingMethod == reportFailures)
	    {
	    int saveErrno = errno;
	    cerr << "connectserver:  Couldn't open socket to " << serverid
	      << ": " << strerr(saveErrno) << '\n';
	    }
	return -1;
	}
    if (connect(fd, addr, addrlen) < 0)
	{
	if (reportingMethod == reportFailures)
	    {
	    int saveErrno = errno;
	    cerr << "connectserver:  Couldn't connect to server " << serverid
	      << ": " << strerr(saveErrno) << '\n';
	    }
	(void) close(fd);
	return -1;
	}
    struct linger lingerBuf;
    lingerBuf.l_onoff = 1;
    lingerBuf.l_linger = 0x7fffffff;	// Linger essentially forever
    (void)setsockopt(fd, SOL_SOCKET, SO_LINGER, &lingerBuf, sizeof lingerBuf);
    conntype = type;
    nbytes = write(fd, &conntype, sizeof conntype);
    if (nbytes < 0)
	{
	int saveErrno = errno;
	cerr << "connectserver:  Couldn't write type byte: "
	  << strerr(saveErrno) << '\n';
	(void) close(fd);
	return -1;
	}
    else if (nbytes != sizeof conntype)
	{
	cerr << "connectserver:  type-byte write failed: " << nbytes
	  << " bytes written\n";
	(void) close(fd);
	return -1;
	}
    if (blockingType == nonBlockingConnection)
	setnonblock(fd);
    return fd;
    }

void setnonblock(		// Make an fd nonblocking
    int			fd)	// FD to modify
    {

#if defined(O_NONBLOCK) && (!defined(ultrix) && !defined(hpux) && !defined(AIXV3))
    (void) fcntl(fd, F_SETFL, O_NONBLOCK);
#else
#ifdef FIOSNBIO
    int		arg;
    arg = 1;
    (void) ioctl(fd, FIOSNBIO, &arg);
#else
#if defined(AIXV3) && defined(FIONBIO)
    int		arg;
    arg = 1;
    (void) ioctl(fd, FIONBIO, &arg);
#else
    (void) fcntl(fd, F_SETFL, FNDELAY);
#endif
#endif
#endif
    }
