#ifndef lint
static const char Rcs_Id[] =
    "$Id: controller.cc,v 1.26 1998/07/23 01:55:42 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 * 
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

/*
 * $Log: controller.cc,v $
 * Revision 1.26  1998/07/23 01:55:42  geoff
 * Do a bit of format error-checking in the storage-setting code.
 *
 * Revision 1.25  1998/07/15 21:43:32  geoff
 * Get rid of CONTROL_RESTART; it's too painful to keep supporting.
 *
 * Revision 1.24  1998/06/02 23:59:38  geoff
 * Add support for symlink initialization.
 *
 * Revision 1.23  1998/05/14 01:15:26  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.22  1996/10/29 18:42:29  geoff
 * When setting reference times, skip trailing whitespace on each line.
 *
 * Revision 1.21  1996/10/11 22:08:55  geoff
 * Update the usage message.
 *
 * Revision 1.20  1996/10/04 07:11:58  geoff
 * Add support for reference-time setting.
 *
 * Revision 1.19  1996/09/16 22:16:42  geoff
 * Read a byte back from the correlator when sending it an exit command.
 * This allows us to wait until the actual exit, which is useful in
 * shutdown scripts.
 *
 * Revision 1.18  1996/09/11 22:59:40  geoff
 * Send investigator name when setting investigation information.  Add
 * fake-reference and force-reference options to setStorage.  Use an
 * enum, rather than ad-hoc values, to communicate storage settings to
 * the correlator.  Allow paths to be separated by any whitespace, rather
 * than just a few selected characters.
 *
 * Revision 1.17  1996/06/11  06:13:07  geoff
 * Add support for cluster investigators.  Change the setStorage code to
 * allow a length when files are marked as unstored.
 *
 * Revision 1.16  1996/05/03  01:11:50  geoff
 * Update the copyright.  Add read-instructions support.
 *
 * Revision 1.15  1996/04/10  01:28:26  geoff
 * Add setStorage.  Fix a couple of cosmetic typos.
 *
 * Revision 1.14  1996/02/02  20:09:39  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.13  1995/11/21  21:37:40  geoff
 * Change '-R' to '-i'.
 *
 * Revision 1.12  1995/11/10  02:58:30  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.11  1995/09/15  00:05:06  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.
 *
 * Revision 1.10  1995/08/07  20:02:29  geoff
 * Add investigator support and the getPath routine.  Comment the cases
 * in option processing.
 *
 * Revision 1.9  1995/06/17  00:11:00  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.  Use a switch to specify the server port.  Make sure cout
 * is buffered, for efficiency.  Flush cerr/cout before all aborts.
 *
 * Revision 1.8  1995/05/27  00:15:44  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.7  1995/05/23  00:39:18  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.  Be careful about flushing
 * before reading server input.  Change a bunch of exits into aborts so
 * I'll get debugging dumps.
 *
 * Revision 1.6  1995/05/05  22:41:21  geoff
 * Make use of ByteSwappedStreams for all communications with the
 * correlator.  Add support for NEW_MANAGEMENT_DISPLAYS.
 *
 * Revision 1.5  1995/04/20  21:57:27  geoff
 * Add support for the manage-and-dump command.  Use the new enums in
 * calls to connectServer.  Add support for more file types in dumps.
 *
 * Revision 1.4  1995/03/30  00:44:21  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.3  1995/03/29  23:57:15  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:03:59  geoff
 * Use byte-swapped streams to communicate with the correlator.  Add much
 * support for new parameter setting.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "correlator.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <ctype.h>
#include <errno.h>
#include <iomanip.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
extern "C"
    {
#include <seer/seer.h>
    }

int			main(int argc, char* argv[]);
static void		usage(void);
void			nomem(char* where, int line, unsigned int size);
static void		getparameters(ByteSwappedStream& stream,
			  char* managerName);
static void		setparameters(ByteSwappedStream& stream,
			  int argc, char* argv[]);
static void		setInvestigation(ByteSwappedStream& stream,
			  const char* investigatorName);
static void		setRelationInvestigation(ByteSwappedStream& stream,
			  const char* investigatorName);
static void		setClusterInvestigation(ByteSwappedStream& stream,
			  const char* investigatorName);
static void		setReferenceTimes(ByteSwappedStream& stream);
static void		setStorage(ByteSwappedStream& stream);
static bool		getPath(istream& stream, char* path);
				// Get a file path from a stream
static void		sendcommand(ByteSwappedStream& stream, int cmd);

static int		encodeFilenames = 0;
				// NZ to MIME-encode filenames

extern "C" char*	mimeDecode(const char* source);
				// Decode a MIME-encoded string
extern void		showent(seerbuf_t* seerp);
extern "C" char*	strerr(int errnum);

int main(			// Control correlator system
    int			argc,	// Argument count
    char*		argv[])	// Argument vector
    {
    int			doget = 0;  // NZ to get parameters
    int			doInvestigate = 0; // NZ to set investigation
    int			dokill = 0; // NZ to stop server
    int			domanage = 0; // NZ to run management
    int			doReadInstructions = 0; // NZ to read instructions
    int			dosave = 0; // NZ to save
    int			doset = 0;  // NZ to set parameters
    int			doSetReference = 0; // NZ to set reference times
    int			doSetStorage = 0; // NZ to set storage
    const char*		instructionName = NULL; // Instruction file name
    const char*		investigatorName = NULL; // Investigator to set for
    char*		managerName = NULL; // Manager to talk to
    int			serverfd; // FD to use to talk to server
    char*		serverPort = NULL; // Server port to use

    /*
     * Before doing anything, make sure that cout is buffered.  This
     * makes a huge difference in performance.  Without this, cout
     * appears to be unbuffered even if it's directed to a pipe or
     * file.
     */
    cout.rdbuf()->unbuffered(false);
    if (open(NOTRACE_DEV, 0) < 0)
	{
	int saveErrno = errno;
	cerr << "controller:  warning:  couldn't open " << NOTRACE_DEV
	  << ": " << strerr(saveErrno) << ", continuing anyway\n";
	}
    for (  ;  argc > 1  &&  argv[1][0] == '-';  argc--, argv++)
	{
	switch (argv[1][1])
	    {
	    case 'e':		// Use MIME encoding of filenames
		{
		encodeFilenames = 1;
		break;
		}
	    case 'g':		// Get management parameters
		{
		doget = 1;
		if (--argc <= 1)
		    usage();
		++argv;
		managerName = argv[1];
		break;
		}
	    case 'I':		// (Re)-read an instruction file
		{
		doReadInstructions = 1;
		if (--argc <= 1)
		    usage();
		++argv;
		instructionName = argv[1];
		break;
		}
	    case 'i':		// Set investigation information
		{
		doInvestigate = 1;
		if (--argc <= 1)
		    usage();
		++argv;
		investigatorName = argv[1];
		if (strncmp(investigatorName, "Relation:", 9) != 0
		  &&  strncmp(investigatorName, "Cluster:", 8) != 0)
		    {
		    cerr << "Investigator name '"
		      << investigatorName
		      << "' must begin with 'Relation:' or 'Cluster:'\n";
		    exit(1);
		    }
		break;
		}
	    case 'k':		// Terminate correlator
		{
		dokill = 1;
		break;
		}
	    case 'M':		// Run specific manager, returning results
		{
		domanage = 1;
		if (--argc <= 1)
		    usage();
		++argv;
		managerName = argv[1];
		break;
		}
	    case 'm':		// Run all managers, returning no results
		{
		domanage = 1;
		break;
		}
	    case 'P':		// Specify port to use to talk to server
		{
		if (--argc <= 1)
		    usage();
		++argv;
		serverPort = argv[1];
		break;
		}
	    case 'p':		// Set correlator parameters
		{
		doset = 1;
		break;
		}
	    case 'R':		// Set file reference times
		{
		doSetReference = 1;
		break;
		}
	    case 'S':		// Set file storage statuses
		{
		doSetStorage = 1;
		break;
		}
	    case 's':		// Save current database contents
		{
		dosave = 1;
		break;
		}
	    default:
		{
		usage();
		break;
		}
	    }
	}
    if (doset)
	{
	if (argc < 2)
	    usage();
	}
    else if (argc != 1  &&  argc != 2)
	usage();
    // The following test depends on flag values being 1 or 0.
    if (doget + doInvestigate + dokill + domanage + doReadInstructions
	+ dosave + doset + doSetReference + doSetStorage
      != 1)
	usage();
    serverfd = connectServer(serverPort, CONTROL_CONNECTION,
      blockingConnection, reportFailures);
    if (serverfd < 0)
	return 1;
    ByteSwappedStream serverStream(serverfd, ByteSwappedStream::bigEnded);
    if (!serverStream)
	{
	int saveErrno = errno;
	cerr << "controller:  couldn't convert server fd into stream: %s"
	  << strerr(saveErrno) << '\n';
	}
    serverStream.writeEndedness();
    serverStream.flush();
    if (doget)
	getparameters(serverStream, managerName);
    else if (doReadInstructions)
	{
	sendcommand(serverStream, CONTROL_READ_INSTRUCTIONS);
	int nameLength = strlen(instructionName);
	serverStream.write(&nameLength, sizeof nameLength);
	serverStream.write(instructionName, nameLength + 1);
	serverStream.flush();
	}
    else if (doInvestigate)
	setInvestigation(serverStream, investigatorName);
    else if (dokill)
	{
	sendcommand(serverStream, CONTROL_EXIT);
	/*
	 * We read a byte back from the server to determine when it
	 * exits.  We don't really care what's in the byte, or indeed
	 * whether the read succeeds, just so long as it doesn't
	 * terminate until the server has saved the database.
	 */
	char doneByte;
	serverStream.read(&doneByte, sizeof doneByte);
	}
    else if (domanage)
	{
	if (managerName == NULL)
	    sendcommand(serverStream, CONTROL_MANAGE);
	else
	    {
	    sendcommand(serverStream, CONTROL_MANAGE_DUMP);
	    int nameLength = strlen(managerName);
	    serverStream.write(&nameLength, sizeof nameLength);
	    serverStream.write(managerName, nameLength + 1);
	    char ch = encodeFilenames != 0;
	    serverStream.write(&ch, sizeof ch);
	    serverStream.flush();
	    while (serverStream.read(&ch, sizeof(ch))  &&  ch != '\0')
		cout << ch;
	    }
	}
    else if (dosave)
	sendcommand(serverStream, CONTROL_SAVE);
    else if (doset)
	setparameters(serverStream, argc, argv);
    else if (doSetReference)
	setReferenceTimes(serverStream);
    else if (doSetStorage)
	setStorage(serverStream);
    return 0;
    }

static void usage()		// Issue a usage message
    {

    cerr <<
      "Usage:  controller"
      " [-e] {-g name|-I name|-i name|-m|-M name|-k|-p|-R|-S|-s}"
      " [-P name:port] [parameters]\n";
    exit(1);
    }

void nomem(			// Issue an out-of-memory message
    char*		where,	// Routine that failed to allocate
    int			line,	// Line number of failure
    unsigned int	size)	// Size that was wanted
    {

    cerr << "observer:  out of memory at " << line << " in " << where;
    if (size)
	cerr << ", couldn't get " << size << " bytes";
    cerr << "!\n";
    exit(1);
    }

static void getparameters(	// Get correlator parameters
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    char*		managerName) // Name of manager to query
    {
    sendcommand(stream, CONTROL_GETPARMS);
    int nameLength = strlen(managerName);
    stream.write(&nameLength, sizeof nameLength);
    stream.write(managerName, nameLength + 1);
    char ch;
    while (stream.read(&ch, sizeof(ch))  &&  ch != '\0')
	cout << ch;
    }

static void setparameters(	// Set correlator parameters
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    int			argc,	// Count of arguments to parse
    char*		argv[])	// Arguments to parse
    {
    while (--argc > 0)
	{
	++argv;
	sendcommand(stream, CONTROL_SETPARMS);
	int length = strlen(argv[0]);
	stream.write(&length, sizeof length);
	stream.write(argv[0], length + 1);
	}
    }

/*
 * Set investigation information.  The information is provided to us
 * on standard input; we convert it into internal format and pass it
 * on to the correlator.
 */
static void setInvestigation(	// Set investigation parameters
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    const char*		investigatorName) // Investigator generating data
    {
    if (strncmp(investigatorName, "Relation:", 9) == 0)
	setRelationInvestigation(stream, investigatorName);
    else
	setClusterInvestigation(stream, investigatorName);
    }

/*
 * Send relationship-investigation information to the server
 *
 * Relationships are fed to the controller as:
 *
 *	weight source target1 target2 target3 ...
 *
 * where the weight is a floating-point number and other fields are
 * paths.  Relations are created with the given weight from source to
 * each of the targets.  The relationship name is given on the command
 * line.
 *
 * Relationships are fed to the correlator by sending a
 * SET_INVESTIGATION command with the given relationship name,
 * followed by the source path, and a number of target/weight pairs.
 * The list is terminated by a null target length field.
 */
static void setRelationInvestigation(
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    const char*		investigatorName) // Investigator generating data
    {
    int nameLength = strlen(investigatorName);
    static int nullLength = 0;
    double weight;
    while (cin >> weight)
	{
	char sourcePath[MAXPATHLEN];
	if (!getPath(cin, sourcePath))
	    {
	    cerr << "controller:  unexpected EOF after weight\n";
	    return;
	    }
	sendcommand(stream, CONTROL_SET_INVESTIGATION);
	if (!stream.write(&nameLength, sizeof nameLength)
	  ||  !stream.write(investigatorName, nameLength + 1))
	    {
	    int saveErrno = errno;
	    cerr << "controller:  error writing investigator name to server: "
	      << strerr(saveErrno);
	    panic("");
	    }
	int sourceLength = strlen(sourcePath);
	if (!stream.write(&sourceLength, sizeof sourceLength)
	  ||  !stream.write(sourcePath, sourceLength + 1))
	    {
	    int saveErrno = errno;
	    cerr << "controller:  trouble writing investigated source path: "
	      << strerr(saveErrno);
	    panic("");
	    }
	char targetPath[MAXPATHLEN];
	while (getPath(cin, targetPath))
	    {
	    int targetLength = strlen(targetPath);
	    if (!stream.write(&targetLength, sizeof targetLength)
	      ||  !stream.write(targetPath, targetLength + 1)
	      ||  !stream.write(&weight, sizeof weight))
		{
		int saveErrno = errno;
		cerr
		  << "controller:  trouble writing investigated target path:"
		  << strerr(saveErrno);
		panic("");
		}
	    }
        if (!stream.write(&nullLength, sizeof nullLength))
	    {
	    int saveErrno = errno;
	    cerr
	      << "controller:  trouble writing investigator line terminator: "
	      << strerr(saveErrno);
	    panic("");
	    }
	stream.flush();
	}
    }

/*
 * Send cluster-investigation information to the server
 *
 * Clusters are fed to the controller as:
 *
 *	name path1 path2 ...
 *
 * where "name" is the name to give to the cluster, and the various
 * paths are the members of the cluster.
 *
 * Clusters are sent to the correlator as a SET_INVESTIGATION command,
 * followed by the cluster type name (given on the command line), the
 * cluster name, and the cluster members.  The cluster is terminated
 * by a null member length field.
 */
static void setClusterInvestigation(
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    const char*		investigatorName) // Investigator generating data
    {
    int nameLength = strlen(investigatorName);
    static int nullLength = 0;
    char clusterName[MAXPATHLEN];
    while (getPath(cin, clusterName))
	{
	sendcommand(stream, CONTROL_SET_INVESTIGATION);
	if (!stream.write(&nameLength, sizeof nameLength)
	  ||  !stream.write(investigatorName, nameLength + 1))
	    {
	    int saveErrno = errno;
	    cerr << "controller:  error writing investigator name to server: "
	      << strerr(saveErrno);
	    panic("");
	    }
	int nameLength = strlen(clusterName);
	if (!stream.write(&nameLength, sizeof nameLength)
	  ||  !stream.write(clusterName, nameLength + 1))
	    {
	    int saveErrno = errno;
	    cerr << "controller:  trouble writing clustered name: "
	      << strerr(saveErrno);
	    panic("");
	    }
	char filePath[MAXPATHLEN];
	while (getPath(cin, filePath))
	    {
	    int pathLength = strlen(filePath);
	    if (!stream.write(&pathLength, sizeof pathLength)
	      ||  !stream.write(filePath, pathLength + 1))
		{
		int saveErrno = errno;
		cerr << "controller:  trouble writing clustered file path: "
		  << strerr(saveErrno);
		panic("");
		}
	    }
	if (!stream.write(&nullLength, sizeof nullLength))
	    {
	    int saveErrno = errno;
	    cerr
	      << "controller:  trouble writing cluster terminator: "
	      << strerr(saveErrno);
	    panic("");
	    }
	stream.flush();
	}
    }

/*
 * Set fake reference times for various files.  The file names and
 * reference time changes are read from the standard input.
 */
static void setReferenceTimes(
    ByteSwappedStream&	stream)	// Stream to read commands from
    {
    sendcommand(stream, CONTROL_SET_REFERENCE);
    char changedPath[MAXPATHLEN];
    long referenceChange;
    int whence;
    static int nullLength = 0;
    while (getPath(cin, changedPath)  &&  (cin >> referenceChange)
      &&  (cin >> whence))
	{
	if (whence < 0  ||  whence > 2)
	    {
	    /*
	     * A group of reference changes is finished.  Flush the
	     * output so that the correlator won't hang if it's a
	     * while before the next group comes in.
	     */
	    if (!stream.write(&nullLength, sizeof nullLength))
		{
		int saveErrno = errno;
		cerr << "controller:  trouble writing reference change"
		  << " terminator: ";
		panic(strerr(saveErrno));
		}
	    stream.flush();
	    sendcommand(stream, CONTROL_SET_REFERENCE);
	    break;
	    }
	else
	    {
	    int changedLength = strlen(changedPath);
	    if (!stream.write(&changedLength, sizeof changedLength)
	      ||  !stream.write(changedPath, changedLength + 1)
	      ||  !stream.write(&referenceChange, sizeof referenceChange)
	      ||  !stream.write(&whence, sizeof whence))
		{
		int saveErrno = errno;
		cerr
		  << "controller:  trouble writing reference change: ";
		panic(strerr(saveErrno));
		}
	    }
	/*
	 * Skip trailing whitespace.
	 */
	char nextCharacter;
	while ((cin >> nextCharacter)  &&  isspace(nextCharacter))
	    ;
	cin.putback(nextCharacter);
	}
    if (!stream.write(&nullLength, sizeof nullLength))
	{
	int saveErrno = errno;
	cerr << "controller:  trouble writing reference change terminator: ";
	panic(strerr(saveErrno));
	}
    stream.flush();
    }

/*
 * Report changed storage status for various files.  The file names
 * and new storage statuses are read from the standard input.
 */
static void setStorage(
    ByteSwappedStream&	stream)	// Stream to read commands from
    {
    sendcommand(stream, CONTROL_SET_STORAGE);
    char storageFlag;
    static int nullLength = 0;
    bool ok = true;
    while (ok  &&  (cin >> storageFlag))
	{
	switch (storageFlag)
	    {
	    case '.':
		{
		/*
		 * A group of storage settings is finished.  Flush the
		 * output so that the correlator won't hang if it's a
		 * while before the next group comes in.
		 */
		if (!stream.write(&nullLength, sizeof nullLength))
		    {
		    int saveErrno = errno;
		    cerr << "controller:  trouble writing storage setting"
		      << " terminator: ";
		    panic(strerr(saveErrno));
		    }
		stream.flush();
		sendcommand(stream, CONTROL_SET_STORAGE);
		break;
		}
	    case 'f':		// Mark file as stored and force a reference
	    case 'F':
		{
		static unsigned char setting =
		  (unsigned char)SS_STORE_AND_REFERENCE;
		char changedPath[MAXPATHLEN];
		if (getPath(cin, changedPath))
		    {
		    int changedLength = strlen(changedPath);
		    if (!stream.write(&changedLength, sizeof changedLength)
		      ||  !stream.write(changedPath, changedLength + 1)
		      ||  !stream.write(&setting, sizeof setting))
			{
			int saveErrno = errno;
			cerr
			  << "controller:  trouble writing storage setting: ";
			panic(strerr(saveErrno));
			}
		    }
		else if (!cin.eof())
		    {
		    cerr << "controller:  input formatting error\n";
		    ok = false;
		    }
		break;
		}
	    case 'l':		// Set the file's symlink target
	    case 'L':
		{
		static unsigned char setting = (unsigned char)SS_SYMLINK;
		char sourcePath[MAXPATHLEN];
		char targetPath[MAXPATHLEN];
		if (getPath(cin, sourcePath)  &&  getPath(cin, targetPath))
		    {
		    int sourceLength = strlen(sourcePath);
		    char trueTarget[MAXPATHLEN];
		    if (targetPath[0] == '/')
			(void)strcpy(trueTarget, targetPath);
		    else
			{
			/*
			 * Make relative symlinks into absolute ones
			 * before sending them to the correlator.
			 */
			(void)strcpy(trueTarget, sourcePath);
			char* lastSlash = strrchr(trueTarget, '/');
			if (lastSlash == NULL)
			    {
			    lastSlash = trueTarget;
			    trueTarget[0] = '/';
			    }
			lastSlash++;
			(void)strncpy(lastSlash, targetPath,
			  sizeof trueTarget - (lastSlash - trueTarget) - 1);
			trueTarget[sizeof trueTarget - 1] = '\0';
			}
		    int targetLength = strlen(trueTarget);
		    if (!stream.write(&sourceLength, sizeof sourceLength)
		      ||  !stream.write(sourcePath, sourceLength + 1)
		      ||  !stream.write(&setting, sizeof setting)
		      ||  !stream.write(&targetLength, sizeof targetLength)
		      ||  !stream.write(trueTarget, targetLength + 1))
			{
			int saveErrno = errno;
			cerr
			  << "controller:  trouble writing storage setting: ";
			panic(strerr(saveErrno));
			}
		    }
		else if (!cin.eof())
		    {
		    cerr << "controller:  input formatting error\n";
		    ok = false;
		    }
		break;
		}
	    case 'r':		// Fake a reference
	    case 'R':
		{
		static unsigned char setting = (unsigned char)SS_REFERENCE;
		char changedPath[MAXPATHLEN];
		if (getPath(cin, changedPath))
		    {
		    int changedLength = strlen(changedPath);
		    if (!stream.write(&changedLength, sizeof changedLength)
		      ||  !stream.write(changedPath, changedLength + 1)
		      ||  !stream.write(&setting, sizeof setting))
			{
			int saveErrno = errno;
			cerr
			  << "controller:  trouble writing storage setting: ";
			panic(strerr(saveErrno));
			}
		    }
		else if (!cin.eof())
		    {
		    cerr << "controller:  input formatting error\n";
		    ok = false;
		    }
		break;
		}
	    case 's':		// Mark file as stored
	    case 'S':
		{
		static unsigned char setting = (unsigned char)SS_STORE;
		char changedPath[MAXPATHLEN];
		if (getPath(cin, changedPath))
		    {
		    int changedLength = strlen(changedPath);
		    if (!stream.write(&changedLength, sizeof changedLength)
		      ||  !stream.write(changedPath, changedLength + 1)
		      ||  !stream.write(&setting, sizeof setting))
			{
			int saveErrno = errno;
			cerr
			  << "controller:  trouble writing storage setting: ";
			panic(strerr(saveErrno));
			}
		    }
		else if (!cin.eof())
		    {
		    cerr << "controller:  input formatting error\n";
		    ok = false;
		    }
		break;
		}
	    case 'u':		// Mark file(s) as unstored
	    case 'U':
		{
		static unsigned char setting = (unsigned char)SS_UNSTORE;
		char changedPath[MAXPATHLEN];
		unsigned long byteSize;
		if (getPath(cin, changedPath)  &&  (cin >> byteSize))
		    {
		    int changedLength = strlen(changedPath);
		    if (!stream.write(&changedLength, sizeof changedLength)
		      ||  !stream.write(changedPath, changedLength + 1)
		      ||  !stream.write(&setting, sizeof setting)
		      ||  !stream.write(&byteSize, sizeof byteSize))
			{
			int saveErrno = errno;
			cerr
			  << "controller:  trouble writing storage setting: ";
			panic(strerr(saveErrno));
			}
		    }
		else if (!cin.eof())
		    {
		    cerr << "controller:  input formatting error\n";
		    ok = false;
		    }
		break;
		}
	    default:
		{
		cerr << "controller:  bad storage flag '" << storageFlag
		  << "'\n";
		ok = false;
		break;
		}
	    }
	}
    if (!stream.write(&nullLength, sizeof nullLength))
	{
	int saveErrno = errno;
	cerr << "controller:  trouble writing storage setting"
	  << " terminator: ";
	panic(strerr(saveErrno));
	}
    stream.flush();
    }

/*
 * Get a file path from a stream.  First blank characters are skipped,
 * then non-blank characters are read, up to MAXPATHLEN, being stored
 * into "path".  Reading stops when a blank or newline is encountered.
 * The terminating character is consumed by the routine.
 *
 * Returns true if a path was successfully collected from the current
 * input line.
 */
static bool getPath(		// Get a file path from a stream
    istream&		stream,	// Stream to read from
    char*		path)	// Where to store result
    {
    char nextCharacter;
    char* pathPointer = path;
    /*
     * Skip blanks.  If we hit a newline, exit.
     */
    while (stream.get(nextCharacter))
	{
	if (nextCharacter == '\n')
	    return false;
	else if (!isspace(nextCharacter))
	    {
	    *pathPointer++ = nextCharacter;
	    break;
	    }
	}
    if (pathPointer == path)
	return false;
    while (pathPointer - path < MAXPATHLEN - 1  &&  stream.get(nextCharacter))
	{
	if (isspace(nextCharacter))
	    break;
	*pathPointer++ = nextCharacter;
	}
    *pathPointer++ = '\0';
    stream.putback(nextCharacter);
    if (encodeFilenames)
	{
	char* decodedPath = mimeDecode(path);
	(void)strcpy(path, decodedPath);
	free((char*)decodedPath);
	}
    return true;
    }

static void sendcommand(	// Send a no-response command to server
    ByteSwappedStream&	stream,	// Stream to use to talk to server
    int			cmd)	// Command to send
    {
    char		command; // Command byte for server

    command = cmd;
    if (!stream.write(&command, sizeof command))
	{
	int saveErrno = errno;
	cerr << "controller:  error writing command " << cmd
	  << " to server: " << strerr(saveErrno);
	panic("");
	}
    }
