#ifndef lint
static const char Rcs_Id[] =
    "$Id: correlator.cc,v 1.44 1999/02/04 21:05:13 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: correlator.cc,v $
 * Revision 1.44  1999/02/04 21:05:13  geoff
 * During a panic, remember to close all the readFiles fds as well as
 * control and listen files.
 *
 * Revision 1.43  1998/10/23 23:15:05  geoff
 * Don't declare growmalloc; it's obsolete.
 *
 * Revision 1.42  1998/07/28 23:47:38  geoff
 * Add support for a separate distribution for directory sizes.
 *
 * Revision 1.41  1998/07/15 21:43:33  geoff
 * Get rid of CONTROL_RESTART; it's too painful to keep supporting.
 * Report failures to open the save file.  Many changes to make the
 * last-reference time and similar things be manager-specific.
 *
 * Revision 1.40  1998/06/02 23:59:39  geoff
 * Add support for symlink initialization.
 *
 * Revision 1.39  1998/05/14 01:15:26  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.38  1998/05/02 00:29:06  geoff
 * Insist on exactly one database argument, since that's all we accept.
 *
 * Revision 1.37  1997/12/18 01:24:31  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.36  1997/12/04 23:15:28  geoff
 * Add correlatorPanic, which closes all incoming fds before panicking
 * (so that the observer won't hang), and use it to do all aborts.
 * Arrange to pause on panics.
 *
 * Revision 1.35  1997/02/12 02:52:03  geoff
 * Report when the -a or -g switch is specified, and report whenever a -M
 * switch appears.  This makes it easy to track simulation conditions by
 * examining the log file.
 *
 * Revision 1.34  1997/01/02 17:48:31  geoff
 * Report when storage settings are done
 *
 * Revision 1.33  1997/01/02 06:48:18  geoff
 * Add the -a option so that -g can be the default.  Add -f to support
 * size fakery.  Report important simulation parameters to the output
 * file.  If the save file is /dev/null, suppress creating and renaming
 * it so that /dev/null doesn't get destroyed accidentally.
 *
 * Revision 1.32  1996/12/03 02:57:54  geoff
 * Change the argument parsing to be two-pass.  Most arguments are parsed
 * in the first pass, but -M (manager parameters) is done in the second
 * pass; this is so that the -M switch can override parameters stored in
 * the saved database.  Add code to call the read/write parameters
 * routines when saving and restoring the database.
 *
 * Revision 1.31  1996/10/11 22:08:56  geoff
 * Add geometric-mean support.  Add code to write a proper magic number
 * to the saved database.
 *
 * Revision 1.30  1996/10/04 07:12:00  geoff
 * Add support for reference-time setting.  Put (and assume) a magic
 * number in the log file, but there isn't one, assume the older format.
 * Add support for command-interpreter references.
 *
 * Revision 1.29  1996/09/23 23:41:21  geoff
 * Add code to report the signal that terminated when we exit due to a
 * signal.
 *
 * Revision 1.28  1996/09/16 22:16:43  geoff
 * In non-simulation mode, reset the process group so that /etc/init
 * can't easily kill us.  When exiting, write a byte back to the
 * controller to tell it that the table save is complete.
 *
 * Revision 1.27  1996/09/13 18:01:24  geoff
 * Add a signal catcher so that simply killing the correlator will save
 * the table.
 *
 * Revision 1.26  1996/09/11 22:59:41  geoff
 * Don't ever go to actual time in simulation mode.  Add special code to
 * try to suppress management runs if there is pending input.  Fix an
 * off-by-one malloc error in savetable, and remember to free the name
 * before exit.  When re-executing, trim the defaults and description
 * from parameter values.  Add force and reference options to setStorage,
 * and use an enum for communication.  Add REPORT_SELECTED_FAILURES and
 * TIMEVALS_ARE_INTS support.  Maintain an observer replay count so that
 * certain things can be done differently during replays.  In simulation
 * mode, exit when the last replaying observer exited.
 *
 * Revision 1.25  1996/06/11  06:13:08  geoff
 * Add FICUS_DELETE_WORKAROUND to truncate the database file before
 * renaming on top of it.  Modify doControl (and its called functions) to
 * return a bool indicating whether the stream failed, so that it can be
 * closed by the caller.  Flush before exiting (pure paranoia!).  Add
 * support for cluster investigators (by letting the chosen investigator
 * do the reading from the stream), and for file lengths in the unstore
 * command.
 *
 * Revision 1.24  1996/05/03  01:11:51  geoff
 * Update the copyright.  Get rid of believeSavedState.  Make savetable
 * global (and accept a const).  Add instruction-reading support.  Flush
 * and fsync the database after saving it, and use renames to atomically
 * replace the old one.
 *
 * Revision 1.23  1996/04/10  01:28:26  geoff
 * Add setStorage.  Fix some places where an array deletion was needed.
 * Truncate the miss-log and table-save files when opening them.
 *
 * Revision 1.22  1996/03/19  18:42:49  geoff
 * Make a couple of error messages slightly more readable.
 *
 * Revision 1.21  1996/02/02  20:09:40  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.20  1996/02/02  18:28:47  geoff
 * Fix some minor problems detected by the new C++ compiler.  Clean up
 * the replayPacket when replay is finished, and let it be global.  Be
 * more lenient about where STAT records may appear in simulation mode.
 * Handle the possibility of an invalid next management time.  Use
 * count() routines rather than counting loops.
 *
 * Revision 1.19  1995/12/19  01:02:44  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.18  1995/12/19  00:34:08  geoff
 * Get rid of HANDLE_OLD_SAVE_FILES.  Add some cleanup code so dmalloc
 * doesn't find unfreed memory.  Add a lot of process-tracking support.
 *
 * Revision 1.17  1995/11/22  00:19:55  geoff
 * Get rid of the TIME_SORTED_INPUT and NEW_FILE_RELATIONS ifdefs.
 *
 * Revision 1.16  1995/11/21  21:37:44  geoff
 * Add support for doing investigations through NEW_FILE_RELATIONS.
 *
 * Revision 1.15  1995/11/10  02:58:31  geoff
 * Add zzinlinefunctions.hh.  Add extensive TIME_SORTED_INPUT support.
 * Store the save/replay file in its natural endedness.  Set the umask
 * before creating the correlator socket.  Add NEW_FILE_RELATIONS and
 * LENGTH_IN_STAT_RECORD support.  Conditionalize table destruction on
 * the dmalloc library, so we don't waste a lot of time while exiting in
 * the non-debug case.
 *
 * Revision 1.14  1995/10/10  20:37:47  geoff
 * Get rid of the COMPARISON_SUPPORT and SIMULATION_SUPPORT ifdefs; it's
 * easier to always include the code.  Move simulationMode to a header
 * file so it can be global.
 *
 * Revision 1.13  1995/09/15  00:05:08  geoff
 * Change the RCS ID to be a const.  Make use of the new internal timers.
 * Fix some include orderings that caused warning messages.  Add
 * comparison and simulation support.  Add proper bracketing to all
 * switch statements.  Fix an underspecified argument in the opening of
 * the save file.  Destroy the file table on exit, to help in detecting
 * memory leaks.  Get the system's file-descriptor table size properly.
 *
 * Revision 1.12  1995/08/07  20:02:32  geoff
 * Add investigator support.  Add DUP support.  Fix a tiny sprintf lint
 * error.
 *
 * Revision 1.11  1995/06/30  19:04:27  geoff
 * Add fresh believeSavedState support.  Change many error messages to
 * report the file offset where the error occurred.  Change maxfd to be
 * one more than the maximum active fd.  Fix the select-hanging bug.
 * Give time stamps when the database save starts and completes.  Accept
 * a time stamp from the observer with each packet (but don't use it for
 * anything).  Treat symlinks specially when necessary.
 *
 * Revision 1.10  1995/06/17  00:11:03  geoff
 * Add MANAGE_COUNT to allow management to run periodically during
 * catchup replays.  Save the errno before reporting errors, to be sure
 * it doesn't get clobbered.  Flush cerr/cout before all aborts.  Improve
 * the buffered-input kludge to work more reliably.  Don't record destroy
 * references when renaming files.
 *
 * Revision 1.9  1995/05/27  00:15:47  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.8  1995/05/26  23:02:55  geoff
 * Allow a null parent-file pointer.
 *
 * Revision 1.7  1995/05/23  00:39:20  geoff
 * Make processPacket return an indication of whether a stream error
 * occurred, so that the caller can do the removeFd.  This is needed to
 * prevent aborts in replayInput, where there is no fd to remove.  Put
 * the contents of the instruction file into the save file.  Change a lot
 * of exits into aborts so I can have debugging dumps.  Add code to use
 * the TestableStreambuf abstraction to test whether input is buffered on
 * sockets before calling select.  Add more PRIVATE_MANAGEMENT_DATA
 * support.
 *
 * Revision 1.6  1995/05/05  22:41:24  geoff
 * Add much initial PRIVATE_MANAGEMENT_DATA support.  Put the endedness
 * of the replay file directly into the file.  Fix ByteSwappedStream
 * constructors to use the new argument order.  Add support for
 * ST_READLINK in the save file.  Take out findParent.
 *
 * Revision 1.5  1995/04/20  21:57:31  geoff
 * Move readinstr to file.cc.  Add ongoing reporting of replay status.
 * Open the replay file after reading instructions, so that it matches
 * when the save file is opened.  When management is replayed, only run
 * the named manager.  Add support for the manage-and-dump command.  Use
 * findCanonical to find files.
 *
 * Revision 1.4  1995/03/30  00:44:23  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.3  1995/03/29  23:57:18  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:04:02  geoff
 * Use byte-swapped streams for save files and controller communication.
 * Fix many new-file-table bugs.  Add working new parameter-setting
 * support.  Log restarts into save file, and handle them properly on
 * replay.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#define MAIN
#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "communication.hh"
#include "correlator.hh"
#include "file.hh"
#include "filereference.hh"
#include "investigator.hh"
#include "internaltimer.hh"
#include "managerbase.hh"
#include "packet.hh"
#include "panic.hh"
#include "streamset.hh"
#include "zzinlinefunctions.hh"
#include <ctype.h>
#include <errno.h>
#ifdef FICUS_DELETE_WORKAROUND
#include <fcntl.h>
#endif // FICUS_DELETE_WORKAROUND
#include <fstream.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strstream.h>
#include <unistd.h>
#include <seer/seer.h>
extern "C"
    {
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/stat.h>
    }

/*
 * Define this to avoid storing information about any graft point
 */
#define SUPPRESS_GRAFT_POINTS

int			main(int argc, char* argv[]);
static void		usage(void);
void			nomem(const char* where, int line, unsigned int size);
void			correlatorPanic(const char* message);
static void		catchdeath(int sig);
static int		setupsockets(int port);
static int		open_tcp_socket(int port);
static int		open_unix_socket(int port);
static void		readlog(const char* logname);
static void		replayInput();
static void		watchInput();
static void		newConnection(ByteSwappedStream& listenFile);
static bool		doControl(ByteSwappedStream& controlStream);
void			saveTable();
static void		dumpTable(ByteSwappedStream& dumpFile);
static bool		setInvestigation(ByteSwappedStream& controlStream);
					// Set investigator info on files
static bool		setReferenceTimes(ByteSwappedStream& controlStream);
					// Set file reference times
static bool		setStorage(ByteSwappedStream& controlStream);
					// Set file storage statuses
static bool		processPacket(InputPacket& packet);

#define INVALID_INETADDR -1	// Error return from inet_addr

/*
 * NEEDSWORK. SunOS 4.1.1 declares signal and SIG_IGN in a fashion
 * that causes warning messages from g++.  The following casts
 * suppress the warnings.
 */
#ifdef __sparc__
#define SIGNAL_CLEANUP_CAST(x)	((void*)(x))
#else // __sparc__
#define SIGNAL_CLEANUP_CAST(x)	x
#endif // __sparc__

/*
 * Temporary stuff
 */
#define SERVER_BASE_PORT 5000
#define LOCALPATH	"/tmp/.seer_"
#define SERVER_BACKLOG	5	// Maximum connection backlog allowed

const int		REPORT_COUNT = 50000;
				// Report after every 50K records replayed
const int		MANAGE_COUNT = REPORT_COUNT * 4;
				// Manage after every 200K records replayed
				// ..must be multiple of REPORT_COUNT

fd_set			allfds;	// List of all FD's to select on
static StreamSet	controlFiles; // List of files controlling us
static char*		instrfname = NULL; // Instruction file name
static char*		logname; // Name of correlation log file to use
int			maxfd = -1; // Maximum FD to select on
static char*		missLogFileName = NULL; // Name of miss-log file
static StreamSet	listenFiles; // List of files we did a listen on
static InternalTimer	nextStatisticsTime(InternalTimer::VIRTUAL);
				// When to next display simulation statistics
static char*		progname; // Name we ran under
InputList		readFiles; // List of observers providing us input
static unsigned long	statisticsInterval; // Interval at which to report
				// ..simulation statistics
static unsigned long	statisticsOffset; // Offset to apply to stats interval
static int		terminate_correlator = 0; // Set NZ to exit cleanly
				// ..value is signal that killed us

extern void		showent(seerbuf_t* seerp);
extern char*		strdup(const char* string, int size);
extern "C" char*	strerr(int errnum);

int main(			// Correlate file references
    int			argc,	// Argument count
    char*		argv[])	// Argument vector
    {
    char*		replayfname = NULL; // File name to replay from
    char*		savefname = NULL; // File name to save input to
    int			serverport = 0; // Server port to use

    FD_ZERO(&allfds);
    progname = argv[0];
    pauseOnPanics();
    if (open(NOTRACE_DEV, 0) < 0) {
	{
	int saveErrno = errno;
	cerr << "correlator:  warning:  couldn't open " << NOTRACE_DEV
	  << ": " << strerr(saveErrno) << ", continuing anyway\n";
	}
    }
    int argNo;
    for (argNo = 1;  argNo < argc  &&  argv[argNo][0] == '-';  argNo++)
	{
	switch (argv[argNo][1])
	    {
	    case 'a':		// Use arithmetic means for semantic distance
		{
		cerr << "correlator: using arithmetic means\n";
		FileReference::setGeometricMeans(false);
		break;
		}
	    case 'c':		// Catchup from save file (continue saving)
		{
		if (++argNo >= argc)
		    usage();
		replayfname = savefname = argv[argNo];
		replaymode = 0;
		break;
		}
	    case 'f':		// Specify parameters for faking file sizes
		{
		if (++argNo >= argc)
		    usage();
		double geometricParameter = atof(argv[argNo]);
		if (geometricParameter <= 0.0  ||  geometricParameter >= 1.0)
		    {
		    cerr << "correlator: parameter of -f must be in (0,1)\n";
		    usage();
		    }
		cerr << "correlator: fake file size parameter 'geometric' "
		  << geometricParameter << '\n';
		File::setFileSizeDistribution("geometric", 1,
		  &geometricParameter);
		char* directoryParameterSlash = strchr(argv[argNo], '/');
		char* randomSeedSlash = NULL;
		if (directoryParameterSlash != NULL)
		    {
		    randomSeedSlash = strchr(directoryParameterSlash + 1, '/');
		    /*
		     * Backwards compatibility: if there is only one
		     * slash, it gives the random seed, not the
		     * directory-size parameter.
		     */
		    if (randomSeedSlash == NULL)
			{
			randomSeedSlash = directoryParameterSlash;
			directoryParameterSlash = NULL;
			}
		    }
		if (directoryParameterSlash == NULL)
		    {
		    /*
		     * Backwards-compatibility kludge: if there is no
		     * directory-size parameter, arbitrarily set it to
		     * 10 times the file-size parameter.
		     */
		    geometricParameter *= 10.0;
		    }
		else
		    geometricParameter = atof(directoryParameterSlash + 1);
		if (geometricParameter <= 0.0  ||  geometricParameter >= 1.0)
		    {
		    cerr << "correlator: parameter of -f must be in"
		      " (0,1)\n";
		    usage();
		    }
		cerr
		  << "correlator: fake directory size parameter 'geometric' "
		  << geometricParameter << '\n';
		File::setDirectorySizeDistribution("geometric", 1,
		  &geometricParameter);
		if (randomSeedSlash != NULL)
		    {
		    long randomSeed = atol(randomSeedSlash + 1);
		    srand48(randomSeed);
		    cerr << "correlator: random seed " << randomSeed << '\n';
		    }
		break;
		}
	    case 'g':		// Use geometric means for semantic distance
		{
		cerr << "correlator: using geometric means\n";
		FileReference::setGeometricMeans(true);
		break;
		}
	    case 'i':		// Specify file of caching instructions
		{
		if (++argNo >= argc)
		    usage();
		instrfname = argv[argNo];
		break;
		}
	    case 'L':		// Specify miss-logging file
		{
		if (++argNo >= argc)
		    usage();
		missLogFileName = argv[argNo];
		break;
		}
	    case 'M':		// Specify arguments to a manager
		{
		if (++argNo >= argc)
		    usage();
		cerr << "correlator: manager parameter '" << argv[argNo]
		  << "'\n";
		ManagerBase::setManagerParameters(argv[argNo]);
		break;
		}
	    case 'p':		// Specify server port to use
		{
		if (++argNo >= argc)
		    usage();
		serverport = atoi(argv[argNo]);
		break;
		}
	    case 'r':		// Replay from save file
		{
		if (++argNo >= argc)
		    usage();
		replayfname = argv[argNo];
		replaymode = 1;
		break;
		}
	    case 'S':		// Run in simulation mode
		{
		if (++argNo >= argc)
		    usage();
		simulationMode = true;
		statisticsInterval = atol(argv[argNo]);
		char* slash = strchr(argv[argNo], '/');
		if (slash == NULL)
		    statisticsOffset = 0;
		else
		    statisticsOffset = atol(slash + 1);
		cerr << "correlator: simulation parameters "
		  << statisticsInterval << '/' << statisticsOffset << '\n';
		statisticsInterval *= 60 * 60;
		statisticsOffset *= 60 * 60;
		break;
		}
	    case 's':		// Append all input to save file
		{
		if (++argNo >= argc)
		    usage();
		savefname = argv[argNo];
		break;
		}
	    case 'v':		// Verbose mode.  May be repeated
		{
		if (argv[argNo][2] == '\0')
		    verbose++;
		else
		    verbose = atoi(&argv[argNo][2]);
		break;
		}
	    default:
		{
		usage();
		break;
		}
	    }
	}
    if (argNo != argc - 1)
	usage();
    logname = argv[argNo];
    argNo++;
    (void) signal(SIGPIPE, SIG_IGN);
    if (replayfname == NULL  &&  savefname != NULL)
	{
	saveFile = new ByteSwappedStream(savefname, ios::out | ios::app,
	  ByteSwappedStream::naturalEndedness);
	if (!*saveFile)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  Couldn't open '" << savefname << "': "
	      << strerr(saveErrno) << '\n';
	    return 1;
	    }
	saveFile->writeEndedness();
	char saveType = ST_RESTART;
	if (!saveFile->write(&saveType, sizeof(saveType)))
	    {
	    int saveErrno = errno;
	    InternalTimer failTime(InternalTimer::ACTUAL);
	    cerr << "correlator:  difficulty writing restart to save file: "
	      << strerr(saveErrno) << " at " << failTime.ctime() << '\n';
	    }
	}
    if (missLogFileName != NULL)
	{
	missLogFile = new fstream(missLogFileName, ios::out | ios::trunc);
	if (!*missLogFile)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  Couldn't open '" << missLogFileName << "': "
	      << strerr(saveErrno) << '\n';
	    return 1;
	    }
	}
    if (replayfname != NULL)
	{
	replayFile = new ByteSwappedStream(replayfname, ios::in,
	  ByteSwappedStream::naturalEndedness);
	if (replayFile == NULL  ||  !*replayFile)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  Couldn't open '" << replayfname
	     << "': " << strerr(saveErrno) << '\n';
	    return 1;
	    }
	replayFile->readEndedness();
	/*
	 * The first item in a replay file should always be a restart.
	 */
	char saveType;
	if (!replayFile->read(&saveType, sizeof saveType))
	    {
	    cerr << "correlator:  zero-length replay file\n";
	    delete replayFile;
	    replayFile = NULL;
	    }
	else if (saveType != ST_RESTART)
	    {
	    cerr << "correlator:  replay file didn't begin with restart,"
	      " replay aborted\n";
	    delete replayFile;
	    replayFile = NULL;
	    }
	}
    File::readInstructions(instrfname);
    readlog(logname);		// Read current correlation tables
    /*
     * Re-parse manager parameters from the arguments, in case any of
     * them were clobbered by the values stored in the log file.
     */
    for (argNo = 1;  argNo < argc  &&  argv[argNo][0] == '-';  argNo++)
	{
	switch (argv[argNo][1])
	    {
	    case 'a':		// Use arithmetic means for semantic distance
		{
		break;
		}
	    case 'c':		// Catchup from save file (continue saving)
	    case 'f':		// Specify parameters for faking file sizes
	    case 'i':		// Specify file of caching instructions
	    case 'L':		// Specify miss-logging file
	    case 'p':		// Specify server port to use
	    case 'r':		// Replay from save file
	    case 'S':		// Run in simulation mode
	    case 's':		// Append all input to save file
		{
		++argNo;
		break;
		}
	    case 'g':		// Use geometric means for semantic distance
	    case 'v':		// Verbose mode.  May be repeated
		{
		break;
		}
	    case 'M':		// Specify arguments to a manager
		{
		++argNo;
		ManagerBase::setManagerParameters(argv[argNo]);
		break;
		}
	    default:
		{
		break;
		}
	    }
	}
    if (replayFile != NULL)
	{
	int saveReplay = replaymode;
	if (simulationMode)
	    replaymode = 0;
	replayInput();		// Replay previously-saved input
	delete replayFile;
	replayFile = NULL;
	if (saveReplay)
	    {
	    /*
	     * For memory-leak detection, it helps a lot if we clear
	     * out as many hunks of allocated memory as possible.
	     * Since the greatest proportion of allocated memory comes
	     * from the File class, we'll ask it to clean itself up.
	     */
	    File::destroyTable();
	    return 0;
	    }
	}
    if (!simulationMode)
	{
	InternalTimer::timerMode(InternalTimer::ACTUAL);
	setpgid(0, 0);		// Go into "daemon" mode
	}
    if (setupsockets(serverport)) // Create sockets to talk to
	return 1; 
    if (savefname != NULL  &&  saveFile == NULL)
	{
	saveFile = new ByteSwappedStream(savefname, ios::out | ios::app,
	  ByteSwappedStream::naturalEndedness);
	if (saveFile == NULL)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  Couldn't open '" << savefname << "': "
	      << strerr(saveErrno) << '\n';
	    return 1;
	    }
	saveFile->writeEndedness();
	char saveType = ST_RESTART;
	if (!saveFile->write(&saveType, sizeof(saveType)))
	    {
	    int saveErrno = errno;
	    InternalTimer failTime(InternalTimer::ACTUAL);
	    cerr << "correlator:  difficulty writing restart to save file: "
	      << strerr(saveErrno) << " at " << failTime.ctime() << '\n';
	    }
	}
    if (SIGNAL_CLEANUP_CAST(signal(SIGINT, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGINT, catchdeath);
    if (SIGNAL_CLEANUP_CAST(signal(SIGHUP, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGHUP, catchdeath);
    if (SIGNAL_CLEANUP_CAST(signal(SIGTERM, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGTERM, catchdeath);
    watchInput();		// Infinite loop: track observer data
    saveTable();
    if (verbose)
	cerr << "correlator:  exiting due to signal " << terminate_correlator
	  << '\n';
#ifdef USE_DMALLOC_LIBRARY
    /*
     * For memory-leak detection, it helps a lot if we clear out as
     * many hunks of allocated memory as possible.  Since the greatest
     * proportion of allocated memory comes from the File class, we'll
     * ask it to clean itself up.
     */
    File::destroyTable();
#endif // USE_DMALLOC_LIBRARY
    if (saveFile != NULL)
	delete saveFile;
    cout.flush();
    cerr.flush();
    return 0;
    }

static void usage()		// Issue a usage message
    {

    cerr << "Usage:  correlator [-a] [-c file] [-f parameter[/seed]] [-g] "
      "[-i file] "
      "[-L misslogfile] "
      "[-M manager-args] "
      "[-p port] [-r file] "
      "[-S interval[/offset]] "
      "[-s file] [-v] logfile\n";
    exit(1);
    }

void nomem(			// Issue an out-of-memory message
    const char*		where,	// Routine that failed to allocate
    int			line,	// Line number of failure
    unsigned int	size)	// Size that was wanted
    {

    cerr << "correlator:  out of memory at" << line << " in " << where;
    if (size)
	cerr << ", couldn't get " << size << " bytes";
    correlatorPanic("!");
    }

void correlatorPanic(		// Panic the correlator, cleanly
    const char*		message) // Message to issue on panic
    {
    /*
     * Close all open fds that might contain incoming observer or
     * controller connections.
     */
    for (int fd = 0;  fd < maxfd;  fd++)
	{
	if (listenFiles.hasFd(fd)  ||  controlFiles.hasFd(fd)
	  ||  readFiles.hasFd(fd))
	    close(fd);
	}
    panic(message);
    }

static void catchdeath(		// Catch signals intended to cause us to exit
    int			sig)	// Signal we caught
    {

    terminate_correlator = sig;
    (void)signal(SIGINT, SIG_IGN);
    (void)signal(SIGHUP, SIG_IGN);
    (void)signal(SIGTERM, SIG_IGN);
    }

static int setupsockets(	// Set up server sockets for observers
    int			port)	// Port to set up on
    {
    int			tcp_fd; // FD used to access TCP socket
    int			unix_fd; // FD used to access Unix socket

    if (replaymode)
	return 0;		// Sockets aren't used on replay
    tcp_fd = open_tcp_socket(port); // Set up TCP listening
    unix_fd = open_unix_socket(port); // Set up Unix listening
    if (tcp_fd < 0  &&  unix_fd < 0)
	return 1;
    if (tcp_fd >= 0)
	listenFiles.addFd(tcp_fd);
    if (unix_fd >= 0)
	listenFiles.addFd(unix_fd);
    return 0;
    }

static int open_tcp_socket(	// Open and set up a TCP listener socket
    int			port)	// Port to set up on
    {
    struct sockaddr_in	inaddr;	// Internet address of socket
    int			request; // FD of request socket

    request = socket(AF_INET, SOCK_STREAM, 0); // Make request socket
    if (request < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't make TCP socket: " << strerr(saveErrno)
	  << '\n';
	return -1;
	}
    bzero((char*) &inaddr, sizeof(inaddr));
    inaddr.sin_family = AF_INET;
    inaddr.sin_port = htons(SERVER_BASE_PORT + port);
    inaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(request, (struct sockaddr*) &inaddr, sizeof(inaddr)) < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't bind TCP socket: " << strerr(saveErrno)
	  << '\n';
	(void) close(request);
	return -1;
	}
    if (listen(request, SERVER_BACKLOG) < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't listen on TCP socket: "
	  << strerr(saveErrno) << '\n';
	(void) close(request);
	return -1;
	}
    return request;
    }

static int open_unix_socket(	// Open and set up a Unix listener socket
    int			port)	// Port to set up on
    {
    int			request; // FD of request socket
    struct sockaddr_un	unaddr;	// Address to give to Unix listener socket

    request = socket(AF_UNIX, SOCK_STREAM, 0); // Make request socket
    if (request < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't make Unix socket: "
	  << strerr(saveErrno) << '\n';
	return -1;
	}
    bzero((char*) &unaddr, sizeof(unaddr));
    unaddr.sun_family = AF_UNIX;
    int oldUmask = umask(0);
    (void) sprintf(unaddr.sun_path, "%s%d", LOCALPATH, port);
    (void) unlink(unaddr.sun_path);
    if (bind(request, (struct sockaddr*) &unaddr,
      strlen(unaddr.sun_path) + sizeof(unaddr.sun_family)))
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't bind Unix socket: "
	  << strerr(saveErrno) << '\n';
	(void) close(request);
	umask(oldUmask);
	return -1;
	}
    umask(oldUmask);
    if (listen(request, SERVER_BACKLOG) < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't listen on Unix socket: "
	  << strerr(saveErrno) << '\n';
	(void) close(request);
	return -1;
	}
    return request;
    }

static void readlog(			// Read old correlation data
    const char*		logname)	// Name of file to read
    {
    InternalTimer now(InternalTimer::ACTUAL);
    cerr << "correlator:  database read began at " << now.ctime() << '\n';
    ByteSwappedStream logfile(logname, ios::in,
      ByteSwappedStream::naturalEndedness);
    if (!logfile)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't open log file '" << logname << "': "
	  << strerr(saveErrno) << '\n';
	}
    logfile.readEndedness();
    logVersion = CORRELATOR_LOG_VERSION; // Assume a current log file
    unsigned long logMagic;
    if (logfile.read(&logMagic))
	{
	if (logMagic == CORRELATOR_DATABASE_MAGIC)
	    logfile.read(&logVersion);
	else
	    {
	    /*
	     * Old database files don't have a magic number.  Rewind
	     * and try again.
	     */
	    logVersion = 0;
	    logfile.seekg(0, ios::beg);
	    logfile.readEndedness();
	    }
	}
    if (logVersion > 2)
	{
	while (ManagerBase::readManagerParameters(logfile))
	    ;
	}
    if (logVersion <= 1)
	FileReference::setGeometricMeans(false);
					// Older version didn't have geo means
    else
	FileReference::readGeometricMeans(logfile);
    FileRelationTypeBase::readRelationTypes(logfile);
    while (File::read(logfile))
	;
    now.updateTime();
    cerr << "correlator:  database read complete at " << now.ctime() << '\n';
    }

static void replayInput()	// Replay previously-saved input
    {
    if (verbose > 0)
	{
	InternalTimer startTime(InternalTimer::ACTUAL);
	cerr << "correlator:  replay started at " << startTime.ctime()
	  << '\n';
	}
    char savetype;		// Type of save record
    replayPacket = new InputPacket(replayFile, true);
    unsigned long replayRecords = 0;
    for (  ;  replayFile->read(&savetype, sizeof savetype);  replayRecords++)
	{
	if (simulationMode)
	    {
	    InternalTimer simulatedTime(InternalTimer::VIRTUAL);
	    if (simulatedTime.isValid())
		{
		if (!nextStatisticsTime.isValid()
		  ||  simulatedTime >= nextStatisticsTime)
		    {
		    if (nextStatisticsTime.isValid())
			{
			HoardingManagerBase::showAllStatistics(cout);
			HoardingManagerBase::resetAllStatistics();
			}
		    nextStatisticsTime = simulatedTime
		      - (((struct timeval&)simulatedTime).tv_sec
			  - statisticsOffset)
			% statisticsInterval
		      + statisticsInterval;
		    }
		InternalTimer nextManagementTime(InternalTimer::VIRTUAL);
		ManagerBase::runDueManagers(simulatedTime, nextManagementTime);
		}
	    }
	else if (verbose > 1  &&  replayRecords != 0
	  &&  replayRecords % REPORT_COUNT == 0)
	    {
	    InternalTimer now(InternalTimer::ACTUAL);
	    cerr << "correlator:  " << replayRecords << " records, "
	      << replayFile->bytesRead() - sizeof savetype
	      << " bytes replayed at " << now.ctime() << '\n';
	    if (!replaymode  &&  replayRecords % MANAGE_COUNT == 0)
		ManagerBase::runAllManagers(now, NULL, false);
	    }
	switch (savetype)
	    {
	    case ST_PACKET:
		{
		replayPacket->get();
		replayPacket->consume();
		if (!processPacket(*replayPacket))
		    {
		    delete replayPacket;
		    replayPacket = NULL;
		    return;
		    }
		break;
		}
	    case ST_MANAGE:
		{
		unsigned int nameLength;
		if (!replayFile->read(&nameLength, sizeof nameLength))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for manager name length at offset"
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		char managerName[MAXPATHLEN];
		if (nameLength >= sizeof(managerName))
		    {
		    cerr << "correlator:  bad name length " << nameLength
		      << " reading fd " << replayFile->rdbuf()->fd()
		      << " for manager name at offset "
		      << replayFile->bytesRead() - sizeof nameLength;
		    correlatorPanic("");
		    }
		if (!replayFile->read(managerName, nameLength + 1))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for manager name at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		if (strlen(managerName) != nameLength)
		    {
		    cerr << "correlator:  manager name length mismatch for '"
		      << managerName << "' on fd " << replayFile->rdbuf()->fd()
		      << "' at offset "
		      << replayFile->bytesRead() - nameLength - 1
			- sizeof nameLength;
		    correlatorPanic("");
		    }
		if (replaymode  &&  !simulationMode)
		    {
		    ManagerBase* manager =
		      ManagerBase::findNamedManager(managerName);
		    if (manager == NULL)
			cerr << "correlator:  manager '" << managerName
			  << "' not found to run management, ignored,"
			  << " at offset " << replayFile->bytesRead() << '\n';
		    else
			{
			InternalTimer now(InternalTimer::ACTUAL);
			manager->manage(now, NULL, false);
			}
		    }
		break;
		}
	    case ST_STAT:
		{
		if (replaymode  &&  !simulationMode  &&  !replayedRestart)
		    {
		    cerr << "correlator:  unexpected record type ST_STAT"
			" at offset "
		      << replayFile->bytesRead() << " in replayinput\n";
		    delete replayPacket;
		    replayPacket = NULL;
		    return;
		    }
		// NEEDSWORK: this incorporates too much knowledge!
#ifdef NAME_IN_STAT_RECORD
		int nameLength;
		if (!replayFile->read(&nameLength, sizeof nameLength))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for stat name length at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		char statFileName[MAXPATHLEN];
		if ((unsigned)nameLength >= sizeof(statFileName))
		    {
		    cerr << "correlator:  bad name length " << nameLength
		      << " reading fd " << replayFile->rdbuf()->fd()
		      << " for stat name at offset "
		      << replayFile->bytesRead() - sizeof nameLength;
		    correlatorPanic("");
		    }
		if (!replayFile->read(statFileName, nameLength + 1))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for stat name at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
#endif // NAME_IN_STAT_RECORD
		char charDummy;
		if (!replayFile->read(&charDummy, sizeof charDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd() << " for state at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		if (!replayFile->read(&charDummy, sizeof charDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd() << " for type at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		unsigned long longDummy;
#ifdef LENGTH_IN_STAT_RECORD
		if (!replayFile->read(&longDummy, sizeof longDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for file size at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
#endif // LENGTH_IN_STAT_RECORD
		if (!replayFile->read(&longDummy, sizeof longDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd() << " for blocks at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		break;
		}
	    case ST_READLINK:
		{
		if (replaymode  &&  !simulationMode  &&  !replayedRestart)
		    {
		    cerr << "correlator:  unexpected record type ST_READLINK"
			" at offset "
		      << replayFile->bytesRead() - sizeof savetype
		      << " in replayinput\n";
		    delete replayPacket;
		    replayPacket = NULL;
		    return;
		    }
		// NEEDSWORK: this incorporates too much knowledge!
		int linkNameLength;
		char linkFileName[MAXPATHLEN];
#ifdef NAME_IN_STAT_RECORD
		if (!replayFile->read(&linkNameLength, sizeof linkNameLength))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for readlink name length at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		if (linkNameLength <= 0
		  ||  linkNameLength >= (int)sizeof(linkFileName))
		    {
		    cerr << "correlator:  bad name length " << linkNameLength
		      << " reading fd " << replayFile->rdbuf()->fd()
		      << " for readlink name at offset "
		      << replayFile->bytesRead() - sizeof linkNameLength;
		    correlatorPanic("");
		    }
		if (!replayFile->read(linkFileName, linkNameLength + 1))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for stat name at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
#endif // NAME_IN_STAT_RECORD
		if (!replayFile->read(&linkNameLength, sizeof linkNameLength))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for readlink target length at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		if (linkNameLength >= (int)sizeof(linkFileName))
		    {
		    cerr << "correlator:  bad name length " << linkNameLength
		      << " reading fd " << replayFile->rdbuf()->fd()
		      << " for readlink target at offset "
		      << replayFile->bytesRead() - sizeof linkNameLength
		      << '\n';
		    }
		if (linkNameLength > 0
		  &&  !replayFile->read(linkFileName, linkNameLength + 1))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for readlink target at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		break;
		}
	    case ST_RESTART:
		{
		replayedRestart = 1;
		break;
		}
	    case ST_INSTRUCTION:
		{
		char charDummy;
		while (replayFile->read(&charDummy, sizeof charDummy)
		  &&  charDummy != '\n')
		    ;
		break;
		}
	    case ST_INSTRUCTION_END:
		{
		break;
		}
	    default:
		{
		cerr << "correlator:  unexpected record type "
		  << (int) savetype << " at offset "
		  << replayFile->bytesRead() - sizeof savetype
		  << " in replayinput\n";
		delete replayPacket;
		replayPacket = NULL;
		return;
		}
	    }
	}
    if (verbose > 0)
	{
	InternalTimer now(InternalTimer::ACTUAL);
	cerr << "correlator:  replay of " << replayRecords
	  << " records, " << replayFile->bytesRead() << " bytes complete at "
	  << now.ctime() << '\n';
	}
    delete replayPacket;
    replayPacket = NULL;
    }

static void watchInput()	// Listen to observers & correlate data
    {
    bool		canManage = true;
				// True if OK to manage on this pass
    InternalTimer	nextManagementTime(InternalTimer::VIRTUAL);
				// Next time to do management
    fd_set		srfds;	// Copy of allfds, since select clobbers
    fd_set		sxfds;	// Copy of allfds for exception tests
    struct timeval	waittime; // How long to wait for information

    while (!terminate_correlator)
	{
	InternalTimer now(InternalTimer::VIRTUAL);
	if (simulationMode)
	    {
	    if (now.isValid())
		{
		if (now >= nextStatisticsTime)
		    {
		    if (nextStatisticsTime.isValid())
			{
			HoardingManagerBase::showAllStatistics(cout);
			HoardingManagerBase::resetAllStatistics();
			}
		    nextStatisticsTime = now
		      - (((struct timeval&)now).tv_sec - statisticsOffset)
			% statisticsInterval
		      + statisticsInterval;
		    }
		ManagerBase::runDueManagers(now, nextManagementTime);
		}
	    }
	else if (canManage  ||  observerReplayCount != 0)
	    ManagerBase::runDueManagers(now, nextManagementTime);
	struct timeval* wt;
	if (!nextManagementTime.isValid())
	    wt = NULL;
	else if (now >= nextManagementTime)
	    {
	    waittime.tv_sec = waittime.tv_usec = 0;
	    wt = &waittime;
	    }
	else
	    {
	    waittime = (struct timeval)(nextManagementTime - now);
	    wt = &waittime;
	    }
	bool somethingIsReady = false;
	fd_set readyFds;
	FD_ZERO(&readyFds);
	int nextFd;
	for (nextFd = 0;  nextFd < maxfd;  nextFd++)
	    {
	    if ((listenFiles.hasFd(nextFd)
		&&  listenFiles.streamFor(nextFd).dataAvailable())
	      ||  (controlFiles.hasFd(nextFd)
		&&  controlFiles.streamFor(nextFd).dataAvailable()))
		{
		FD_SET(nextFd, &readyFds);
		somethingIsReady = true;
		}
	    }
	if (somethingIsReady)
	    {
	    waittime.tv_sec = waittime.tv_usec = 0;
	    wt = &waittime;
	    }
	srfds = allfds;
	sxfds = allfds;
	InputPacket* packet = readFiles.get(srfds, sxfds, wt);
	if (packet == NULL)
	    canManage = true;	// No packets, OK to do management
	else
	    {
	    processPacket(*packet);
	    canManage = false;
	    if (observerReplayCount == 0)
		continue;	// Give packets priority over control
	    }
	for (nextFd = 0;  nextFd < maxfd;  nextFd++)
	    {
	    if (FD_ISSET(nextFd, &srfds)  ||  FD_ISSET(nextFd, &readyFds))
		{
		if (listenFiles.hasFd(nextFd))
		    newConnection(listenFiles.streamFor(nextFd));
		else if (controlFiles.hasFd(nextFd))
		    {
		    ByteSwappedStream& controlStream(
		      controlFiles.streamFor(nextFd));
		    if (!doControl(controlStream))
			(void)controlFiles.removeFd(nextFd);
		    }
		}
	    else if (FD_ISSET(nextFd, &sxfds))
		{
		if (listenFiles.hasFd(nextFd))
		    {
		    (void) listenFiles.removeFd(nextFd);
		    if (listenFiles.isEmpty())
			{
			cerr << "correlator:  all listen sockets closed\n";
			cout.flush();
			cerr.flush();
			exit(1);
			}
		    }
		else if (controlFiles.hasFd(nextFd))
		    (void) controlFiles.removeFd(nextFd);
		}
	    }
	}
    }

static void newConnection(		// Process a new connection
    ByteSwappedStream&	listenFile)	// File to accept from
    {
    struct sockaddr	addr;		// Address of newly-bound socket
    int			addrlen;	// Length of bound address
    char		conntype;	// Connection type byte
    int			nbytes;		// Number of bytes read from conn
    int			newfd;		// New FD from accept() call

    addrlen = sizeof addr;
    newfd = accept(listenFile.rdbuf()->fd(), &addr, &addrlen);
    if (newfd < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  accept failed: " << strerr(saveErrno) << '\n';
	}
    else
	{
	nbytes = read(newfd, &conntype, sizeof conntype);
	if (nbytes < 0)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  error reading connection type byte: "
	      << strerr(saveErrno) << '\n';
	    (void) close(newfd);
	    return;
	    }
	else if (nbytes != sizeof conntype)
	    {
	    cerr << "correlator:  wrong length (" << nbytes
	      << ") reading connection type\n";
	    (void) close(newfd);
	    return;
	    }
	if (conntype == CONTROL_CONNECTION)
	    {
	    controlFiles.addFd(newfd);
	    controlFiles.streamFor(newfd).readEndedness();
	    }
	else if (conntype == OBSERVER_CONNECTION)
	    {
	    readFiles.addFd(newfd, true);
	    }
	else
	    {
	    cerr << "correlator:  bad connection type " << conntype << '\n';
	    (void) close(newfd);
	    return;
	    }
	}
    }

/*
 * Handle control requests of all types.  Returns true if the request
 * was handled successfully, false if an error occurred.
 */
static bool doControl(			// Handle control requests
    ByteSwappedStream& controlStream)	// Stream to get request from
    {
    char		command;	// Control command given
    static int		nullLength = 0;	// Zero path length, for terminator

    if (!controlStream.read(&command, sizeof command))
	{
	// Socket must have gone away.
	return false;
	}
    switch (command)
	{
	case CONTROL_EXIT:		// Terminate correlator
	    {
	    saveTable();
	    if (verbose)
		cerr << "correlator:  exiting due to controller request\n";
#ifdef USE_DMALLOC_LIBRARY
	    /*
	     * For memory-leak detection, it helps a lot if we clear
	     * out as many hunks of allocated memory as possible.
	     * Since the greatest proportion of allocated memory comes
	     * from the File class, we'll ask it to clean itself up.
	     */
	    File::destroyTable();
#endif // USE_DMALLOC_LIBRARY
	    if (saveFile != NULL)
		delete saveFile;
	    cout.flush();
	    cerr.flush();
	    /*
	     * Immediately before exiting, write a "done" byte back to
	     * the control stream.  The contents of this byte are
	     * immaterial; the byte is just a way to signal to the
	     * controller that we're really ready to exit.
	     */
	    char doneByte = 0;
	    controlStream.write(&doneByte);
	    controlStream.flush();
	    exit(0);
	    // NOTREACHED
	    break;
	    }
	case CONTROL_GETPARMS:		// Get management parameters
	    {
	    int nameLength;
	    if (!controlStream.read(&nameLength, sizeof nameLength))
		{
		cerr << "correlator:  trouble getting manager name length"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		break;
		}
	    if (nameLength < 0  ||  nameLength >= MAXPATHLEN)
		{
		cerr << "correlator:  bad manager name length " << nameLength
		  << '\n';
		return false;
		}
	    else
		{
		char* managerName = new char[nameLength + 1];
		if (!controlStream.read(managerName, nameLength + 1))
		    {
		    cerr << "correlator:  trouble getting manager name"
		      " from controller at offset "
		      << controlStream.bytesRead() << '\n';
		    return false;
		    }
		else if ((int)strlen(managerName) != nameLength)
		    {
		    cerr << "correlator:  bad manager name length"
		      << strlen(managerName) << " vs. " << nameLength
		      << " at offset "
		      << controlStream.bytesRead() - nameLength - 1
			- sizeof nameLength
		      << '\n';
		    return false;
		    }
		else
		    {
		    ManagerBase::showManagerParameters(managerName,
		      controlStream);
		    controlStream << '\0';
		    controlStream.flush();
		    }
		delete[] managerName;
		managerName = NULL;
		}
	    break;
	    }
	case CONTROL_MANAGE_DUMP:	// Run management and dump results
	    {
	    int nameLength;
	    if (!controlStream.read(&nameLength, sizeof nameLength))
		{
		cerr << "correlator:  trouble getting manager name length"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		break;
		}
	    char* managerName = new char[nameLength + 1];
	    if (!controlStream.read(managerName, nameLength + 1))
		{
		cerr << "correlator:  trouble getting manager name"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		return false;
		}
	    else if ((unsigned)strlen(managerName) != (unsigned)nameLength)
		{
		cerr << "correlator:  bad manager name length"
		  << strlen(managerName) << " vs. " << nameLength
		  << " at offset "
		  << controlStream.bytesRead() - nameLength - 1
		    - sizeof nameLength
		  << '\n';
		return false;
		}
	    else
		{
		ManagerBase* manager =
		  ManagerBase::findNamedManager(managerName);
		if (manager == NULL)
		    {
		    cerr << "correlator:  request to run unknown manager '"
		      << managerName << "' at offset "
		      << controlStream.bytesRead() << '\n';
		    }
		else
		    {
		    char encodeFilenames;
		    if (!controlStream.read(&encodeFilenames,
		      sizeof encodeFilenames))
			{
			cerr << "correlator:  trouble getting MIME encoding"
			    "flag at offset "
			  << controlStream.bytesRead() - 1
			  << '\n';
			return false;
			}
		    if (saveFile != NULL)
			{
			static char saveType = ST_MANAGE;
			if (!saveFile->write(&saveType, sizeof saveType)
			  ||  !saveFile->write(&nameLength, sizeof nameLength)
			  ||  !saveFile->write(managerName, nameLength + 1)
			  ||  !saveFile->flush())
			    {
			    int saveErrno = errno;
			    long now = time((long*) NULL);
			    cerr << "correlator:  difficulty writing "
			      << "management info to save file: "
			      << strerr(saveErrno) << " at " << ctime(&now);
			    }
			}
		    InternalTimer now(InternalTimer::VIRTUAL);
		    manager->manage(now, &controlStream, encodeFilenames != 0);
		    }
		controlStream.write(&nullLength, sizeof nullLength);
		controlStream.flush();
		}
	    delete[] managerName;
	    managerName = NULL;
	    break;
	    }
	case CONTROL_MANAGE:		// Run management upon request
	    {
	    InternalTimer now(InternalTimer::ACTUAL);
	    ManagerBase::runAllManagers(now, NULL, false);
	    break;
	    }
	case CONTROL_READ_INSTRUCTIONS:
	    {
	    File::readInstructions(instrfname);
	    int nameLength;
	    if (!controlStream.read(&nameLength, sizeof nameLength))
		{
		cerr << "correlator:  trouble getting instruction file name"
		  " length from controller at offset "
		  << controlStream.bytesRead() << '\n';
		break;
		}
	    char* instructionFileName = new char[nameLength + 1];
	    if (!controlStream.read(instructionFileName, nameLength + 1))
		{
		cerr << "correlator:  trouble getting instruction file name"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		return false;
		}
	    else if ((unsigned)strlen(instructionFileName)
	      != (unsigned)nameLength)
		{
		cerr << "correlator:  bad instruction file name length"
		  << strlen(instructionFileName) << " vs. " << nameLength
		  << " at offset "
		  << controlStream.bytesRead() - nameLength - 1
		    - sizeof nameLength
		  << '\n';
		return false;
		}
	    else
		File::readInstructions(instructionFileName);
	    delete[] instructionFileName;
	    instructionFileName = NULL;
	    break;
	    }
	case CONTROL_SAVE:		// Save current tables
	    {
	    saveTable();
	    break;
	    }
	case CONTROL_SETPARMS:		// Set management parameters
	    {
	    int parameterLength;
	    if (!controlStream.read(&parameterLength, sizeof parameterLength))
		{
		cerr << "correlator:  trouble getting parameter length"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		break;
		}
	    char* parameters = new char[parameterLength + 1];
	    if (!controlStream.read(parameters, parameterLength + 1))
		cerr << "correlator:  trouble getting parameters"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
	    else if ((int)strlen(parameters) != parameterLength)
		cerr << "correlator:  bad parameter length"
		  << strlen(parameters) << " vs. " << parameterLength
		  << " at offset " << controlStream.bytesRead() << '\n';
	    else
		ManagerBase::setManagerParameters(parameters);
	    delete[] parameters;
	    parameters = NULL;
	    break;
	    }
	case CONTROL_SET_REFERENCE:	// Set file reference times
	    {
	    return setReferenceTimes(controlStream);
	    }
	case CONTROL_SET_STORAGE:	// Set file storage status
	    {
	    return setStorage(controlStream);
	    }
	case CONTROL_SET_INVESTIGATION:	// Set investigator data
	    {
	    return setInvestigation(controlStream);
	    }
	default:
	    {
	    cerr << "correlator:  unrecognized control command "
	      << (int) command << " at offset " << controlStream.bytesRead()
	      << '\n';
	    break;
	    }
	}
    return true;
    }

/*
 * Save the correlation database to a file.  If no file name is given,
 * saves it to the file given in the global 'logname'.
 */
void saveTable()			// Save correlations to file
    {
    if (strcmp(logname, "/dev/null") == 0) // Don't accidentally zap /dev/null!
	return;
    static const char*	extension = ".new"; // Extension to add while writing
    char* newName = new char[strlen(logname) + strlen(extension) + 1];
    (void)strcpy(newName, logname);
    (void)strcat(newName, extension);
    ByteSwappedStream tableSaveFile(newName, ios::out | ios::trunc,
      ByteSwappedStream::naturalEndedness);
    if (!tableSaveFile)
	{
	int saveErrno = errno;
	cerr << "correlator:  Couldn't open save file '" << newName << "': "
	  << strerr(saveErrno) << '\n';
	delete[] newName;
	return;
	}
    tableSaveFile.writeEndedness();
    if (verbose > 0)
	{
	long now = time((long*) NULL);
	cerr << "correlator:  saving to '" << logname << "' at "
	  << ctime(&now);
	}
    dumpTable(tableSaveFile);
    tableSaveFile.flush();
    fsync(tableSaveFile.rdbuf()->fd());
    tableSaveFile.close();
#ifdef FICUS_DELETE_WORKAROUND
    /*
     * Truncate the old database before renaming on top of it; this
     * will keep it from eating up disk space in the Ficus unnamed
     * directory.
     */
    int truncFd = creat(logname, 0666);
    if (truncFd >= 0)
	(void)close(truncFd);
#endif // FICUS_DELETE_WORKAROUND
    if (rename(newName, logname) < 0)
	{
	int saveErrno = errno;
	cerr << "correlator:  couldn't rename table save file '" << newName
	  << "' to '" << logname << "': " << strerr(saveErrno) << '\n';
	}
    delete[] newName;
    if (verbose > 0)
	{
	long now = time((long*) NULL);
	cerr << "correlator:  table save complete at " << ctime(&now);
	}
    }

static void dumpTable(			// Dump correlations
    ByteSwappedStream&	dumpStream)	// Stream to dump to
    {
    static const unsigned long tableMagic = CORRELATOR_DATABASE_MAGIC;
    dumpStream.write(&tableMagic);
    logVersion = CORRELATOR_LOG_VERSION;
    dumpStream.write(&logVersion);
    for (ManagerIterator manager;  manager;  ++manager)
	{
	ManagerBase& thisManager = *manager;
	thisManager.writeManagerParameterHeader(dumpStream);
	thisManager.writeParameters(dumpStream, true);
	}
    int nullLength = 0;
    dumpStream.write(&nullLength);
    FileReference::writeGeometricMeans(dumpStream);
    FileRelationTypeBase::writeRelationTypes(dumpStream);
    for (FileTableIterator iter;  iter;  ++iter)
	(*iter).write(dumpStream);
    dumpStream.flush();
    }

/*
 * Accept investigator information and install it in the file table.
 * Returns true on success, false if an error occurred.
 */
static bool setInvestigation(
    ByteSwappedStream&	controlStream)	// Stream information is arriving on
    {
    int investigatorNameLength;
    if (!controlStream.read(&investigatorNameLength,
      sizeof investigatorNameLength))
	{
	cerr << "correlator:  trouble getting investigator name length"
	  " from controller at offset " << controlStream.bytesRead()
	  << '\n';
	return false;
	}
    if (investigatorNameLength < 0  ||  investigatorNameLength > MAXPATHLEN)
	{
	cerr << "correlator:  bad investigator name length "
	  << investigatorNameLength << '\n';
	return false;
	}
    char* investigatorName = new char[investigatorNameLength + 1];
    if (!controlStream.read(investigatorName, investigatorNameLength + 1))
	{
	cerr << "correlator:  trouble getting investigator name"
	  " from controller at offset " << controlStream.bytesRead()
	  << '\n';
	delete[] investigatorName;
	return false;
	}
    else if ((int)strlen(investigatorName) != investigatorNameLength)
	{
	cerr << "correlator:  bad investigator name length"
	  << strlen(investigatorName) << " vs. " << investigatorNameLength
	  << " at offset " << controlStream.bytesRead() << '\n';
	delete[] investigatorName;
	return false;
	}
    else
	{
	InvestigatorTypeBase* investigator =
	  InvestigatorTypeBase::find(investigatorName, false);
	if (investigator == NULL)
	    {
	    cerr << "correlator:  creating new investigator '"
	      << investigatorName << "'\n";
	    investigator = InvestigatorTypeBase::find(investigatorName, true);
	    if (investigator == NULL)
		{
		delete[] investigatorName;
		return false;		// Bad investigator name, ignore it
		}
	    }
	if (!investigator->learnFromStream(controlStream))
	    return false;
	}
    delete[] investigatorName;
    return true;
    }

/*
 * Set file reference times.  Returns true on success, false if an
 * error occurred.
 */
static bool setReferenceTimes(
    ByteSwappedStream&	controlStream)	// Stream information is arriving on
    {
    int nameLength;
    if (!controlStream.read(&nameLength, sizeof nameLength))
	{
	cerr << "correlator:  trouble getting manager name length"
	  " from controller at offset " << controlStream.bytesRead()
	  << '\n';
	return false;
	}
    char* managerName = new char[nameLength + 1];
    if (!controlStream.read(managerName, nameLength + 1))
	{
	cerr << "correlator:  trouble getting manager name"
	  " from controller at offset " << controlStream.bytesRead()
	  << '\n';
	return false;
	}
    else if ((unsigned)strlen(managerName) != (unsigned)nameLength)
	{
	cerr << "correlator:  bad manager name length"
	  << strlen(managerName) << " vs. " << nameLength
	  << " at offset "
	  << controlStream.bytesRead() - nameLength - 1 - sizeof nameLength
	  << '\n';
	return false;
	}
    ManagerBase* manager = ManagerBase::findNamedManager(managerName);
    if (manager == NULL)
	{
	cerr << "correlator:  request to set references for unknown manager '"
	  << managerName << "' at offset "
	  << controlStream.bytesRead() << '\n';
	return false;
	}
    if (!manager->hasData())
	{
	cerr << "correlator:  request to set references for"
	    " non-hoarding manager '"
	  << managerName << "' at offset "
	  << controlStream.bytesRead() << '\n';
	return false;
	}
    HoardingManagerBase& hoardingManager((HoardingManagerBase&)*manager);
    File* changedFile;
    while ((changedFile = File::findFromStream(controlStream, true))
      != NULL)
	{
	long referenceChange;
	int whence;
	if (!controlStream.read(&referenceChange, sizeof referenceChange)
	  ||  !controlStream.read(&whence, sizeof whence))
	    {
	    cerr << "correlator:  trouble getting reference time change"
	      " from controller at offset " << controlStream.bytesRead()
	      << '\n';
	    return false;
	    }
	hoardingManager.setReferenceTime(*changedFile, referenceChange,
	  whence);
	}
    return true;
    }

/*
 * Set file storage statuses.  Returns true on success, false if an
 * error occurred.
 */
static bool setStorage(			// Set file storage statuses
    ByteSwappedStream&	controlStream)	// Stream information is arriving on
    {
    File* changedFile;
    while ((changedFile = File::findFromStream(controlStream, true))
      != NULL)
	{
	static bool settingReported = false;
	if (!settingReported)
	    {
	    settingReported = true;
	    cerr << "correlator: one or more storage settings received\n";
	    }
	unsigned char stored;
	if (!controlStream.read(&stored, sizeof stored))
	    {
	    cerr << "correlator:  trouble getting file storage status"
	      " from controller at offset " << controlStream.bytesRead()
	      << '\n';
	    return false;
	    }
	StorageSetting setting = (StorageSetting)stored;
	unsigned long unstoredSize = 0;
	File* targetFile = NULL;
	if (setting == SS_UNSTORE)
	    {
	    /*
	     * Files that are being set to unstored status may have
	     * the wrong size (i.e., 0) in the database, so we accept
	     * a correction from the controller.
	     */
	    if (!controlStream.read(&unstoredSize, sizeof unstoredSize))
		{
		cerr << "correlator:  trouble getting unstored file size"
		  " from controller at offset " << controlStream.bytesRead()
		  << '\n';
		return false;
		}
	    }
	else if (setting == SS_SYMLINK)
	    {
	    /*
	     * In simulation mode, it is necessary to be able to
	     * preload symlink targets.  We allow this in
	     * storage-setting mode.
	     */
	    targetFile = File::findFromStream(controlStream, true);
	    }
	changedFile->setStorage(setting, unstoredSize, targetFile);
	}
    return true;
    }

/*
 * Process an incoming observer packet.  The packet is collected,
 * assembled, logged (if verbose mode), and saved to the save file.
 * Then File::recordReference is called to record it.
 */
static bool processPacket(
    InputPacket&	packet)		// Packet received
    {
    InternalTimer::setVirtualTime(packet.packetTime());
    ReferenceType type = packet.packetType();
    if (verbose > 3)
	{
	if (type == SEER_REF_NEWPROC)
	    cout << "NEWPROC " << packet.pid() << ' ' << packet.parentPid();
	else
	    {
	    cout
	      << (type == SEER_REF_READ ? "READ"
		: type == SEER_REF_RW ? "RW"
		: type == SEER_REF_CREATE ? "CREATE"
		: type == SEER_REF_EXEC ? "EXEC"
		: type == SEER_REF_DESTROY ? "DESTROY"
		: type == SEER_REF_RENAME ? "RENAME"
		: type == SEER_REF_OPENREAD ? "OPENREAD"
		: type == SEER_REF_OPENRW ? "OPENRW"
		: type == SEER_REF_OPENCREATE ? "OPENCREATE"
		: type == SEER_REF_CLOSE ? "CLOSE"
		: type == SEER_REF_EXIT ? "EXIT"
		: type == SEER_REF_PROCEXIT ? "PROCEXIT"
		: type == SEER_REF_PEEK ? "PEEK"
		: type == SEER_REF_MAKELINK ? "MAKELINK"
		: type == SEER_REF_PEEKLINK ? "PEEKLINK"
		: type == SEER_REF_FOLLOW ? "FOLLOW"
		: type == SEER_REF_DUP ? "DUP"
		: type == SEER_REF_INTERPRET ? "INTERPRET"
		: type == SEER_REF_CONTROL ? "CONTROL"
		: "UNKNOWN")
	      << (packet.succeeded() ? " OK" : " FAIL")
	      << ' ' << (packet.path1() != NULL ? packet.path1() : "\"\"");
	    if (type == SEER_REF_RENAME)
		cout << ' ' << packet.path2();
	    }
	char usecString[9];
#ifdef TIMEVALS_ARE_INTS
	sprintf(usecString, ".%6.6ld\n", (long)packet.packetTime().tv_usec);
#else // TIMEVALS_ARE_INTS
	sprintf(usecString, ".%6.6ld\n", packet.packetTime().tv_usec);
#endif // TIMEVALS_ARE_INTS
	cout << ' ' << packet.packetTime().tv_sec << usecString;
	}
    if (saveFile != NULL)
	{
	static char savetype = ST_PACKET;
	if (!saveFile->write(&savetype, sizeof savetype)
	  ||  !packet.write(*saveFile))
	    {
	    int saveErrno = errno;
	    long now = time((long*) NULL);
	    cerr << "correlator:  difficulty writing packet to save file: "
	      << strerr(saveErrno) << " at " << ctime(&now);
	    }
	if (!saveFile->flush())
	    {
	    int saveErrno = errno;
	    long now = time((long*) NULL);
	    cerr << "correlator:  difficulty flushing packet to save file: "
	      << strerr(saveErrno) << " at " << ctime(&now);
	    }
	}
    if (type == SEER_REF_CONTROL)
	{
	/*
	 * Control packets aren't really references; they are just
	 * formatted that way for convenience.  Control packet
	 * processing is currently ad-hoc.
	 */
	if (packet.path1Length() != 1)
	    {
	    cerr
	      << "correlator:  bad control packet length (expected 1, got "
	      << packet.path1Length() << ")\n";
	    readFiles.removeSource(&packet);
	    return false;
	    }
	if (packet.path1()[0] == (char)replayMode)
	    {
	    if (!packet.isBlocking())
		cerr << "correlator:  non-blocking observer tried "
		  << "to become blocking, ignored\n";
	    observerReplayCount++;
	    }
	else if (packet.path1()[0] == (char)realTimeMode)
	    {
	    packet.dontBlock();
	    if (observerReplayCount > 0)
		{
		observerReplayCount--;
		if (simulationMode  &&  observerReplayCount == 0)
		    {
		    if (verbose)
			cerr << "correlator:  simulation complete\n";
#ifdef USE_DMALLOC_LIBRARY
		    /*
		     * For memory-leak detection, it helps a lot if we
		     * clear out as many hunks of allocated memory as
		     * possible.  Since the greatest proportion of
		     * allocated memory comes from the File class,
		     * we'll ask it to clean itself up.
		     */
		    File::destroyTable();
#endif // USE_DMALLOC_LIBRARY
		    if (saveFile != NULL)
			delete saveFile;
		    cout.flush();
		    cerr.flush();
		    exit(0);
		    // NOTREACHED
		    }
		}
	    }
	else
	    {
	    cerr << "correlator:  unrecognized control command "
	      << (unsigned int)packet.path1()[0]
	      << '\n';
	    readFiles.removeSource(&packet);
	    return false;
	    }
	return true;
	}
    else if (type == SEER_REF_PROCEXIT)
	{
	Process* process = packet.process();
	if (process != NULL)
	    {
	    File* program = process->program();
	    if (program != NULL)
		program->recordReference(process, SEER_REF_EXIT);
	    }
	packet.processExit();
	}
    else if (type == SEER_REF_NEWPROC)
	{
	// InputPacket already did all necessary work
	return true;
	}
#ifdef __linux__
    /*
     * The Linux observer might give us relative pathnames.  Since we
     * can't do anything meaningful with them, we'll ignore them
     * completely, except that renames where the old name is absolute
     * and the new relative will be turned into destroys.  Yes, this
     * is undesirable.
     */
    if (packet.path1() != NULL  &&  packet.path1()[0] != '/')
	return true;
    else if (type == SEER_REF_RENAME
      &&  packet.path2()[0] != '/')
	type = SEER_REF_DESTROY;
#endif // __linux__
    if (!packet.succeeded())
	{
	if (type == SEER_REF_RENAME)
	    type = SEER_REF_PEEK;
	const char* path = packet.path1();
	if (path != NULL)
	    File::recordMissedReference(path, packet.process(), type);
	}
    else if (type == SEER_REF_RENAME)
	{
	/*
	 * Renames get special handling.  If possible, the old file's
	 * entry is renamed to the new file's.  Then we log it as a
	 * create reference to the new file.  This works nicely
	 * because it preserves any reference information that was
	 * previously associated with the new name, if any exists, or
	 * that was associated with the old name, if the new name
	 * didn't exist.
	 *
	 * This code assumes that no manager cares about destroy
	 * references if only the old file existed.
	 */
	File::rename(packet.path1(), packet.path2(), packet.process());
	File* file2 =
	  File::findCanonical(packet.path2(), true, true, packet.process());
	file2->recordReference(packet.process(),
	  (file2->fileType() == File::FT_SYMLINK
	    ? SEER_REF_MAKELINK
	    : SEER_REF_CREATE));
	}
    else
	{
	if (packet.path1() != NULL)
	    File::recordReference(packet.path1(), packet.process(), type);
	}
    return true;
    }
