#ifndef _CORRELATOR_HH
#define _CORRELATOR_HH
/*
 * $Id: correlator.hh,v 1.18 1998/07/15 21:43:34 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: correlator.hh,v $
 * Revision 1.18  1998/07/15 21:43:34  geoff
 * Bump the save-file version.
 *
 * Revision 1.17  1998/02/19 02:10:00  geoff
 * Bump the log version to reflect adding the "incremental" flag to the
 * clustering manager.
 *
 * Revision 1.16  1997/02/12 02:52:03  geoff
 * Bump the log version to reflect the fact that oftenPercent is now a
 * double.
 *
 * Revision 1.15  1996/12/03 02:57:56  geoff
 * Bump the data base version to reflect the addition of manager
 * parameters to the database.
 *
 * Revision 1.14  1996/10/11 22:09:00  geoff
 * Bump the database version number to reflect the support for geometric
 * means.  Make the miss-log file a proper correlator global.
 *
 * Revision 1.13  1996/10/04 07:12:03  geoff
 * Add a version number for the log file (really the saved database).
 *
 * Revision 1.12  1996/09/11 22:59:44  geoff
 * Add observerReplayCount.
 *
 * Revision 1.11  1996/05/03  01:11:54  geoff
 * Update the copyright.  Add MAX_LINKS.
 *
 * Revision 1.10  1996/02/02  18:28:48  geoff
 * Make replayPacket be a global.
 *
 * Revision 1.9  1995/10/10  20:37:49  geoff
 * Get rid of the long-obsolete parameters_t structure.  Make
 * simulationMode be a global.
 *
 * Revision 1.8  1995/09/15  00:05:11  geoff
 * Globally change "cache" to "hoard".  Get rid of "nfiles", which is
 * obsolete.
 *
 * Revision 1.7  1995/05/27  00:15:51  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.6  1995/05/23  00:39:25  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.  Fix a misdefinition of
 * NULL.  Add save types for the instruction-file contents.
 *
 * Revision 1.5  1995/05/05  22:41:27  geoff
 * Add ST_READLINK.
 *
 * Revision 1.4  1995/03/30  00:44:26  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.3  1995/03/29  23:57:21  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:04:07  geoff
 * Add ST_RESTART, and some new parameter-setting and byte-swapped-stream
 * support.  Fix a longstanding bug in ISOFTEN.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#ifndef NULL
#define NULL	0
#endif // NULL

#ifdef MAIN
#define EXTERN
#define INIT(decl, val)	decl = val
#else // MAIN
#define EXTERN	extern
#define INIT(decl, val)	extern decl
#endif // MAIN

/*
 * Maximum values certain system parameters can take.
 */
const int MAX_NPREV =	100;	// Max number of previous refs to keep
const int MAX_LINKS = 100;	// Maximum symlink hops to follow

/*
 * Magic numbers and versions.
 */
const int CORRELATOR_LOG_VERSION = 6;
				// Version of save file to write

/*
 * Types of records written to a save/replay file
 */
const int ST_PACKET =		0;	// Incoming data packet
const int ST_MANAGE =		1;	// Management routine was called
const int ST_STAT =		2;	// Result of stat-ing a file
const int ST_READLINK =		3;	// Result of reading a link
const int ST_RESTART =		4;	// Correlator was restarted
const int ST_INSTRUCTION =	5;	// Instruction file contents
const int ST_INSTRUCTION_END =	6;	// End of instruction file contents

#define NUMOF(x)	(sizeof(x) / sizeof(x[0]))

INIT(int		logVersion, 0);	// Version number of log being read
INIT(class fstream*	missLogFile, NULL); // File to log misses to
INIT(int		observerReplayCount, 0);
					// Count of observers doing replays
INIT(class ByteSwappedStream* replayFile, NULL); // File to replay from
INIT(class InputPacket*	replayPacket, NULL);
					// Packet associated with replayFile
INIT(int		replaymode, 0); // NZ if replay-and-exit mode
INIT(int		replayedRestart, 0); // NZ if restart was replayed
INIT(class ByteSwappedStream* saveFile, NULL); // File to save input to
INIT(bool		simulationMode, false); // True to simulate reality
INIT(int		verbose, 0); // NZ to dump info to stdout

#endif // _CORRELATOR_HH
