#ifndef lint
static const char Rcs_Id[] =
    "$Id: deletion_mgr.cc,v 1.17 1998/10/23 23:15:07 geoff Exp $";
#endif

/*
 * $Id: deletion_mgr.cc,v 1.17 1998/10/23 23:15:07 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: deletion_mgr.cc,v $
 * Revision 1.17  1998/10/23 23:15:07  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.16  1998/05/14 01:15:28  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.15  1998/02/19 02:10:01  geoff
 * Turn recordStateChange into stateChanged, now that it's inherited from
 * the base class.
 *
 * Revision 1.14  1997/12/04 23:15:37  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.13  1996/12/03 02:57:57  geoff
 * Add support for reading and writing parameters.
 *
 * Revision 1.12  1996/09/16 22:16:44  geoff
 * Change the deletion trigger so that it won't exhibit the "three runs
 * in a row" phenomenon.  Now, the manager runs if numberMissing exceeds
 * the threshold, or if numberForgetting exceeds twice the threshold, and
 * numberForgotten doesn't enter into the equation.
 *
 * Revision 1.11  1996/05/03  01:11:55  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/04/10  01:28:28  geoff
 * Add support for the new UNSTORED file state.
 *
 * Revision 1.9  1996/03/19  18:42:50  geoff
 * Get rid of the show/read/write/clear manager-data routines now that
 * they can be inherited.  Inherit from EmptyManagerBase rather than the
 * top-level manager base.
 *
 * Revision 1.8  1996/02/02  20:09:41  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.7  1996/02/02  18:28:48  geoff
 * Inherit resetStatistics from the base class.  Fix some minor problems
 * detected by the new C++ compiler.
 *
 * Revision 1.6  1995/12/19  01:02:47  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.5  1995/12/19  00:34:11  geoff
 * Change the deletion trigger to 200 (from 1000).  Report statistics
 * after deletion, to help in loop detection.  Add process-tracking
 * support.
 *
 * Revision 1.4  1995/11/22  00:19:59  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.3  1995/11/21  21:37:48  geoff
 * Add support for doing investigations through NEW_FILE_RELATIONS.
 *
 * Revision 1.2  1995/11/10  02:58:37  geoff
 * Add zzinlinefunctions.hh.  Add NEW_FILE_RELATIONS support.  Get rid of
 * a spurious "const" declaration.
 *
 * Revision 1.1  1995/09/14  23:32:25  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "communication.hh"
#include "correlator.hh"
#include "deletion_mgr.hh"
#include "file.hh"
#include "internaltimer.hh"
#include "packet.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

/*
 * Classes used to support the deletion "manager"
 */

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			DeletionManager::DeletionManager(
			  const char* managerName,
			  const char* managerDefaults);
			DeletionManager::~DeletionManager();
bool			DeletionManager::shouldManage(const InternalTimer& now,
			    InternalTimer& nextManagementTime);
void			DeletionManager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream,
			  bool encodeFilenames);
					// Run hoard management
void			DeletionManager::setParameter(const char* description);
					// Set management parameters
void			DeletionManager::showParameters(ostream& stream) const;
					// Display management parameters
long			DeletionManager::writeParameters(
			    ByteSwappedStream& stream, bool reallyWrite)
			  const;
					// Write management params to stream
long			DeletionManager::readParameters(
			  ByteSwappedStream& stream, long parameterBytes);
					// Read management params from stream
ostream&		DeletionManager::showStatistics(ostream& stream) const;
					// Display collected quality statistics
void			DeletionManager::recordInitialState(
			  FileState::State newState);
					// Record a file's first state
void			DeletionManager::recordStateRemoval(
			  FileState::State oldState);
					// Record deletion of a file state
void			DeletionManager::stateChanged(
			  ManagerDataBase* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Record info from a file change
#endif // 0

unsigned long		DeletionManager::numberForgetting = 0;
unsigned long		DeletionManager::numberForgotten = 0;
unsigned long		DeletionManager::numberMissing = 0;

DeletionManager deletionManager(
			"Deletion",
			"Deletion::deletionTrigger=200\n"
			  "Deletion::managementInterval=3600\n");

extern InputList	readFiles;	// List of observers giving us input

extern void		correlatorPanic(const char* message);

DeletionManager::DeletionManager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : EmptyManagerBase(managerName, managerDefaults)
    {
    EmptyManagerBase::setManagerParameters(defaults);
    }

DeletionManager::~DeletionManager()
    {
    }

/*
 * Decide if this manager's management routine should run.  Regardless
 * of the return value, modifies nextManagementTime to contain the
 * next time management will be needed for this manager.
 */
bool DeletionManager::shouldManage(
    const InternalTimer&	now,	// Current time-of-day
    InternalTimer&		nextManagementTime)
					// When to run next
    {
    return ManagerBase::shouldManage(now, nextManagementTime)
      ||  numberMissing >= deletionTrigger
      ||  numberForgetting >= deletionTrigger + deletionTrigger;
    }

void DeletionManager::manage(		// Run management
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	,		// Stream to dump results to
    bool		)		// True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    if (verbose > 1)
	{
	cerr << "correlator:  deletion management began at "
	  << managementStartTime.ctime() << '\n';
	cerr << "correlator:  " << numberMissing << " missing files, "
	  << numberForgetting << " pending forgets, "
	  << numberForgotten << " pending deletions.\n";
	cerr << "correlator:  pre-deletion file table ";
	File::showHashStatistics(cerr);
	cerr << '\n';
	}
    /*
     * Make sure any UNKNOWN files have a proper status.  This is
     * necessary because some of them may actually be MISSING.
     */
    FileTableIterator fileIterator;
    for (  ;  fileIterator;  ++fileIterator)
	(*fileIterator).discoverStatusIfUnknown();
    /*
     * File deletion takes place in two phases.  In the first, the
     * file state is set to FileState::FORGETTING (this is done by the File
     * class).  At a later time, the deletion manager scans the file
     * table and ensures that all references to FileState::FORGETTING files
     * are removed.  In this same scan, all FileState::FORGETTING files are
     * set into the FileState::FORGOTTEN state in preparation for the second
     * phase (this is done by File::deleteIfObsolete).
     *
     * In the second phase, files that are in the FileState::FORGOTTEN state
     * are removed from the file table.  This is also done by
     * deleteIfObsolete.
     */
    unsigned long oldMissing = numberMissing;
    unsigned long oldForgetting = numberForgetting;
    for (FileRelationTypeIterator relationType;  relationType;  ++relationType)
	(*relationType).cleanGlobalData();
    ManagerBase::cleanAllGlobalData();	// Clean out all manager references
    readFiles.loseForgettingPointers(); // Clean out process tables
    if (replayPacket != NULL)
	replayPacket->loseForgettingPointers();
    for (fileIterator.reset();  fileIterator;  ++fileIterator)
	{
	File& file = *fileIterator;
	file.cleanRelationData();
	file.cleanManagerData();
	file.loseForgettingPointers();
	if (file.deleteIfObsolete())
	    fileIterator.removeCurrent();
	}
    if (numberForgetting != oldMissing)
	{
	cerr << "correlator:  only " << numberForgetting << " of "
	  << oldMissing;
	correlatorPanic(" missing files were changed to forgetting");
	}
    else if (numberForgotten != oldForgetting)
	{
	cerr << "correlator:  only " << numberForgotten << " of "
	  << oldForgetting;
	correlatorPanic(" forgetting files were changed to forgotten");
	}
    if (verbose > 0)
	{
	InternalTimer managementEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = managementEndTime - managementStartTime;
	cerr << "correlator:  deletion management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << managementEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  " << numberMissing << " missing files, "
	      << numberForgetting << " pending forgets, "
	      << numberForgotten << " pending deletions.\n";
	    cerr << "correlator:  post-deletion file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    }
	}
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void DeletionManager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad DeletionManager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("deletionTrigger"))
	deletionTrigger = atol(valueStart);
    else if (parameterIs("managementInterval"))
	managementInterval = atol(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized DeletionManager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void DeletionManager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "Deletion::deletionTrigger="
      << deletionTrigger << "/1/infinity/1/Trigger Level\n"
      << "Deletion::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Write manager-specific parameters to a stream.  Returns the number
 * of bytes written, or zero if trouble occurs.  If reallyWrite is
 * false, the byte count is returned, but no writing is actually done.
 */
long DeletionManager::writeParameters(
    ByteSwappedStream&	stream,		// Stream to write to
    bool		reallyWrite)	// True to actually do the writing
    const
    {
    long bytesWritten = ManagerBase::writeParameters(stream, reallyWrite);
    if (reallyWrite  &&  !stream.write(&deletionTrigger))
	return 0;
    bytesWritten += sizeof deletionTrigger;
    return bytesWritten;
    }

/*
 * Read manager-specific parameters from a stream.  The second
 * argument gives the number of bytes available for consumption.
 * Returns the number of bytes remaining.
 */
long DeletionManager::readParameters(
    ByteSwappedStream&	stream,		// Stream to read from
    long		parameterBytes)	// Number of bytes available to read
    {
    parameterBytes = ManagerBase::readParameters(stream, parameterBytes);
    parameterBytes -= sizeof deletionTrigger;
    if (parameterBytes >= 0)
	stream.read(&deletionTrigger);
    return parameterBytes;
    }

/*
 * Display collected quality statistics
 */
ostream& DeletionManager::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    return stream;
    }

/*
 * Record information from a file's initial state.  This is needed so
 * that we can properly maintain the pending-deletions count.
 */
void DeletionManager::recordInitialState(
    FileState::State	newState)	// First state of the file
    {
    switch (newState)
	{
	case FileState::UNKNOWN:
	case FileState::KNOWN:
	case FileState::UNSTORED:
	    {
	    break;
	    }
	case FileState::MISSING:
	    {
	    numberMissing++;
	    break;
	    }
	case FileState::FORGETTING:
	    {
	    numberForgetting++;
	    break;
	    }
	case FileState::FORGOTTEN:
	    {
	    numberForgotten++;
	    break;
	    }
	default:
	    cerr << "correlator:  unexpected file state " << newState;
	    correlatorPanic( " in DeletionManager::recordInitialState");
	    break;
	}
    }

/*
 * Record information from a file's final state, just before the file
 * is removed.  This is needed so that we can properly maintain the
 * pending-deletions count.
 */
void DeletionManager::recordStateRemoval(
    FileState::State	oldState)	// State file ws in
    {
    switch (oldState)
	{
	case FileState::UNKNOWN:
	case FileState::KNOWN:
	case FileState::UNSTORED:
	    break;
	case FileState::MISSING:
	    {
	    if (numberMissing > 0)
		numberMissing--;
	    else
		correlatorPanic("correlator:  attempt to reduce numberMissing"
		  " below zero in DeletionManager::recordStateRemoval");
	    break;
	    }
	case FileState::FORGETTING:
	    {
	    if (numberForgetting > 0)
		numberForgetting--;
	    else
		correlatorPanic(
		  "correlator:  attempt to reduce numberForgetting"
		  " below zero in DeletionManager::recordStateRemoval");
	    break;
	    }
	case FileState::FORGOTTEN:
	    {
	    if (numberForgotten > 0)
		numberForgotten--;
	    else
		correlatorPanic(
		  "correlator:  attempt to reduce numberForgotten"
		  " below zero in DeletionManager::recordStateRemoval");
	    break;
	    }
	default:
	    cerr << "correlator:  unexpected file state " << oldState;
	    correlatorPanic( " in DeletionManager::recordStateRemoval");
	    break;
	}
    }

/*
 * Record information from a file state change.  This is needed so
 * that we can properly maintain the pending-deletions count.
 */
void DeletionManager::stateChanged(
    ManagerDataBase*	,		// Manager-specific data for file
    FileState::State	oldState,	// Original state of the file
    FileState::State	newState)	// New state of the file
    {
    /*
     * It turns out that we can treat this as a removal and a
     * re-creation.  We do the re-creation first so that we can
     * prevent weird cases where the number of deletions goes below
     * zero.
     */
    recordInitialState(newState);
    recordStateRemoval(oldState);
    }
