#ifndef _DELETION_MGR_HH
#define _DELETION_MGR_HH
/*
 * $Id: deletion_mgr.hh,v 1.8 1998/05/14 01:15:28 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: deletion_mgr.hh,v $
 * Revision 1.8  1998/05/14 01:15:28  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.7  1998/02/19 02:10:02  geoff
 * Turn recordStateChange into stateChanged, now that it's inherited from
 * the base class.
 *
 * Revision 1.6  1996/12/03 02:57:59  geoff
 * Add support for reading and writing parameters.
 *
 * Revision 1.5  1996/05/03 01:11:56  geoff
 * Update the copyright.
 *
 * Revision 1.4  1996/03/19  18:42:51  geoff
 * Get rid of the show/read/write/clear manager-data routines now that
 * they can be inherited.  Inherit from EmptyManagerBase rather than the
 * top-level manager base.
 *
 * Revision 1.3  1996/02/02  18:28:49  geoff
 * Declare all virtual functions as such in derived classes.
 *
 * Revision 1.2  1995/11/10  02:58:40  geoff
 * Get rid of a spurious "const" declaration.
 *
 * Revision 1.1  1995/09/14  23:32:25  geoff
 * Initial revision
 *
 */

/*
 * Classes used to support a special "manager" that handles deletion of
 * files.
 */
#include "communication.hh"
#include "file.hh"
#include "managerbase.hh"

class ByteSwappedStream;
class InternalTimer;
class ostream;

class DeletionManager : public EmptyManagerBase
    {
    public:
			DeletionManager(const char* managerName,
			  const char* managerDefaults);
	virtual		~DeletionManager();
	virtual bool	shouldManage(const InternalTimer& now,
			    InternalTimer& nextManagementTime);
					// Decide if manager should run
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
					// Set management parameters
	virtual void	showParameters(ostream& stream) const;
					// Display management parameters
	virtual long	writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
	virtual long	readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
	virtual ostream& showStatistics(ostream& stream) const;
					// Display collected quality statistics
	static void	recordInitialState(FileState::State newState);
					// Record a file's first state
	static void	recordStateRemoval(FileState::State oldState);
					// Record deletion of a file state
	virtual void	stateChanged(ManagerDataBase* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Note that state has changed
    private:
	unsigned long	deletionTrigger; // Number of deletions needed to
					// ..trigger a management run
	static unsigned long
			numberForgetting; // Files in FORGETTING state now
	static unsigned long
			numberForgotten; // Files in FORGOTTEN state now
	static unsigned long
			numberMissing;	// Files in MISSING state now
    };
#endif // _DELETION_MGR_HH
