#ifndef lint
static const char Rcs_Id[] =
    "$Id: dumpcorr.cc,v 1.16 1997/12/18 01:24:41 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: dumpcorr.cc,v $
 * Revision 1.16  1997/12/18 01:24:41  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.15  1996/10/04 07:12:04  geoff
 * Add support for command-interpreter references.
 *
 * Revision 1.14  1996/09/11 22:59:45  geoff
 * Add REPORT_SELECTED_FAILURES support in many places.  Add
 * TIMEVALS_ARE_INTS support.
 *
 * Revision 1.13  1996/05/03  01:11:57  geoff
 * Update the copyright.
 *
 * Revision 1.12  1996/02/02  20:09:42  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.11  1996/02/02  18:28:51  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.10  1995/12/19  01:02:48  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.9  1995/12/19  00:34:13  geoff
 * Add support for packets with process-tracking information.  Don't use
 * fstream::get, which screws up the byte-offset tracking.
 *
 * Revision 1.8  1995/11/22  00:20:03  geoff
 * Get rid of the TIME_SORTED_INPUT and HAND_BYTE_SWAPPING ifdefs.
 *
 * Revision 1.7  1995/11/10  02:58:54  geoff
 * Add zzinlinefunctions.hh.  Use the built in swapping of
 * ByteSwappedStream.  Support SEER_REF_CONTROL.  Support
 * LENGTH_IN_STAT_RECORD.  Other minor improvements/bugfixes.
 *
 * Revision 1.6  1995/09/15  00:05:13  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.
 *
 * Revision 1.5  1995/08/07  20:02:35  geoff
 * Add SEER_REF_DUP support.
 *
 * Revision 1.4  1995/06/30  19:04:29  geoff
 * Add support for the new timestamp fields in the saved packet.  Give
 * file offsets on most error messages.  Support MAKELINK and PEEKLINK
 * references.
 *
 * Revision 1.3  1995/06/17  00:11:07  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.  Flush cerr/cout before all aborts.  Add support for
 * SEER_REF_FOLLOW.
 *
 * Revision 1.2  1995/05/23  00:39:27  geoff
 * Fix a bug caused by having a signed length buffer.  Change a bunch of
 * exits to aborts so I'll get debugging dumps.  Add support for
 * instruction-file saving.
 *
 * Revision 1.1  1995/05/05  21:47:04  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "communication.hh"
#include "byteswappedstream.hh"
#include "correlator.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <errno.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>

int			main(int argc, char* argv[]);
static void		dumpCorrelatorSaveFile(ByteSwappedStream& file);

extern "C" char*	strerr(int errnum);

int main(				// Dump an observer trace file
    int			argc,		// Argument count
    char*		argv[])		// Argument vector
    {
    int trouble = 0;
    if (argc == 1)
	{
	ByteSwappedStream stdinStream(0, ByteSwappedStream::littleEnded);
	stdinStream.readEndedness();
	dumpCorrelatorSaveFile(stdinStream);
	}
    else
	{
	for (int argno = 1;  argno < argc;  argno++)
	    {
	    ByteSwappedStream* file = new ByteSwappedStream(argv[argno],
	      ios::in, ByteSwappedStream::littleEnded);
	    if (!*file)
		{
		int saveErrno = errno;
		cerr << "dumpcorr:  couldn't open '" << argv[argno] << "': "
		  << strerr(saveErrno) << '\n';
		trouble = 1;
		}
	    else
		{
		file->readEndedness();
		dumpCorrelatorSaveFile(*file);
		}
	    delete file;
	    file = NULL;
	    }
	}
    return trouble;
    }

static void dumpCorrelatorSaveFile(	// Dump a correlator trace
    ByteSwappedStream&		file)	// File to dump from
    {
    char saveType;			// Type of save record
    while (file.read(&saveType, sizeof saveType))
	{
	cout << file.bytesRead() - sizeof saveType << ": ";
	switch (saveType)
	    {
	    case ST_PACKET:
		{
		unsigned char charType;
		unsigned char errcode;
		struct timeval packetTime;
		if (!file.read(&charType, sizeof charType)
		  ||  !file.read(&errcode, sizeof errcode)
		  ||  !file.read(&packetTime.tv_sec, sizeof packetTime.tv_sec)
		  ||  !file.read(&packetTime.tv_usec,
		    sizeof packetTime.tv_usec))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for type or time at offset "
		      << file.bytesRead();
		    panic("");
		    }
		unsigned long pid;
		if (!file.read(&pid, sizeof pid))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for pid at offset "
		      << file.bytesRead();
		    panic("");
		    }
		ReferenceType packetType = (ReferenceType)charType;
		short nameLength;
		if (packetType == SEER_REF_NEWPROC)
		    {
		    unsigned long parentPid;
		    if (!file.read(&parentPid, sizeof parentPid))
			{
			cerr << "dumpcorr:  error reading fd "
			  << file.rdbuf()->fd() << " for parent Pid at offset "
			  << file.bytesRead();
			panic("");
			}
		    cout << "PACKET NEWPROC "
		      << (errcode == 0 ? "OK " : "FAIL ") << pid
		      << ' ' << parentPid;
		    char usecString[9];
#ifdef TIMEVALS_ARE_INTS
		    sprintf(usecString, ".%6.6ld\n", (long)packetTime.tv_usec);
#else // TIMEVALS_ARE_INTS
		    sprintf(usecString, ".%6.6ld\n", packetTime.tv_usec);
#endif // TIMEVALS_ARE_INTS
		    cout << ' ' << packetTime.tv_sec << usecString;
		    break;
		    }
		if (!file.read(&nameLength, sizeof nameLength))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for length 1 at offset "
		      << file.bytesRead();
		    panic("");
		    }
		if (nameLength < 0  ||  nameLength >= MAXPATHLEN)
		    {
		    cerr << "dumpcorr:  bad length " << nameLength << " on fd "
		      << file.rdbuf()->fd()
		      << " at offset " << file.bytesRead();
		    panic("");
		    }
		char path[MAXPATHLEN];
		if (nameLength > 0)
		    {
		    if (!file.read(path, nameLength + 1))
			{
			cerr << "dumpcorr:  bad read on fd "
			  << file.rdbuf()->fd()
			  << " while reading path at offset "
			  << file.bytesRead();
			panic("");
			}
		    if ((int)strlen(path) != nameLength)
			{
			cerr << "correlator:  bad path length " << strlen(path)
			  << " for '" << path << "' on fd "
			  << file.rdbuf()->fd() << " (expected " << nameLength
			  << ") at offset"
			  << file.bytesRead() - nameLength - 1
			    - sizeof nameLength;
			panic("");
			}
		    }
		cout << "PACKET "
		  << (packetType == SEER_REF_UNKNOWN ? "UNKNOWN"
		    : packetType == SEER_REF_READ ? "READ"
		    : packetType == SEER_REF_RW ? "RW"
		    : packetType == SEER_REF_CREATE ? "CREATE"
		    : packetType == SEER_REF_EXEC ? "EXEC"
		    : packetType == SEER_REF_DESTROY ? "DESTROY"
		    : packetType == SEER_REF_RENAME ? "RENAME"
		    : packetType == SEER_REF_OPENREAD ? "OPENREAD"
		    : packetType == SEER_REF_OPENRW ? "OPENRW"
		    : packetType == SEER_REF_OPENCREATE ? "OPENCREATE"
		    : packetType == SEER_REF_CLOSE ? "CLOSE"
		    : packetType == SEER_REF_EXIT ? "EXIT"
		    : packetType == SEER_REF_PEEK ? "PEEK"
		    : packetType == SEER_REF_MAKELINK ? "MAKELINK"
		    : packetType == SEER_REF_PEEKLINK ? "PEEKLINK"
		    : packetType == SEER_REF_FOLLOW ? "FOLLOW"
		    : packetType == SEER_REF_DUP ? "DUP"
		    : packetType == SEER_REF_INTERPRET ? "INTERPRET"
		    : packetType == SEER_REF_CONTROL ? "CONTROL"
		    : packetType == SEER_REF_PROCEXIT ? "PROCEXIT"
		    : "(unknown)")
		  << (errcode == 0 ? " OK" : " FAIL")
		  << ' ' << (nameLength == 0 ? "\"\"" : path);
		if (packetType == SEER_REF_RENAME)
		    {
		    if (!file.read(&nameLength, sizeof nameLength))
			{
			cerr << "dumpcorr:  bad read on fd "
			  << file.rdbuf()->fd()
			  << " while reading length 2 at offset "
			  << file.bytesRead();
			panic("");
			}
		    if (nameLength <= 0  ||  nameLength >= (int)sizeof path)
			{
			cerr << "dumpcorr:  bad length " << nameLength
			  << " on fd " << file.rdbuf()->fd() << " at offset "
			  << file.bytesRead() - sizeof nameLength;
			panic("");
			}
		    if (!file.read(path, nameLength + 1))
			{
			cerr << "dumpcorr:  bad read on fd "
			  << file.rdbuf()->fd()
			  << " while reading path at offset "
			  << file.bytesRead() - sizeof nameLength;
			panic("");
			}
		    if ((int)strlen(path) != nameLength)
			{
			cerr << "dumpcorr:  bad path length " << strlen(path)
			  << "for '" << path << "' on fd "
			  << file.rdbuf()->fd() << "(expected " << nameLength;
			panic(")");
			}
		    cout << ' ' << path;
		    }
		cout << ' ' << pid;
		char usecString[9];
#ifdef TIMEVALS_ARE_INTS
		sprintf(usecString, ".%6.6ld\n", (long)packetTime.tv_usec);
#else // TIMEVALS_ARE_INTS
		sprintf(usecString, ".%6.6ld\n", packetTime.tv_usec);
#endif // TIMEVALS_ARE_INTS
		cout << ' ' << packetTime.tv_sec << usecString;
		break;
		}
	    case ST_MANAGE:
		{
		unsigned int nameLength;
		if (!file.read(&nameLength, sizeof nameLength))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for manager name length at offset "
		      << file.bytesRead();
		    panic("");
		    }
		char managerName[MAXPATHLEN];
		if (nameLength == 0  ||  nameLength >= sizeof(managerName))
		    {
		    cerr << "dumpcorr:  bad name length " << nameLength
		      << " reading fd " << file.rdbuf()->fd();
		    panic(" for manager name");
		    }
		if (!file.read(managerName, nameLength + 1))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for manager name at offset "
		      << file.bytesRead();
		    panic("");
		    }
		if (strlen(managerName) != nameLength)
		    {
		    cerr << "dumpcorr:  manager name length mismatch for '"
		      << managerName << "' on fd " << file.rdbuf()->fd()
		      << " at offset " << file.bytesRead();
		    panic("");
		    }
		cout << "MANAGE " << managerName << '\n';
		break;
		}
	    case ST_STAT:
		{
		// NEEDSWORK: this incorporates too much knowledge!
		cout << "STAT ";
#ifdef NAME_IN_STAT_RECORD
		unsigned int nameLength;
		if (!file.read(&nameLength, sizeof nameLength))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for stat name length at offset "
		      << file.bytesRead();
		    panic("");
		    }
		char statFileName[MAXPATHLEN];
		if (nameLength >= sizeof(statFileName))
		    {
		    cerr << "dumpcorr:  bad name length " << nameLength
		      << " reading fd " << file.rdbuf()->fd()
		      << " for stat name at offset "
		      << file.bytesRead() - sizeof nameLength;
		    panic("");
		    }
		if (!file.read(statFileName, nameLength + 1))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for stat name at offset"
		      << file.bytesRead();
		    panic("");
		    }
		cout << statFileName << ' ';
#endif // NAME_IN_STAT_RECORD
		char charState;
		if (!file.read(&charState, sizeof charState))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for state at offset "
		      << file.bytesRead();
		    panic("");
		    }
		char charType;
		if (!file.read(&charType, sizeof charType))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for type at offset "
		      << file.bytesRead();
		    panic("");
		    }
#ifdef LENGTH_IN_STAT_RECORD
		unsigned long fileSize;
		if (!file.read(&fileSize, sizeof fileSize))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for file size at offset "
		      << file.bytesRead();
		    panic("");
		    }
#endif // LENGTH_IN_STAT_RECORD
		unsigned long blocks;
		if (!file.read(&blocks, sizeof blocks))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for blocks at offset "
		      << file.bytesRead();
		    panic("");
		    }
#ifdef LENGTH_IN_STAT_RECORD
		cout << "state " << (int)charState << " type "
		  << (int)charType << " blocks " << blocks << " bytes "
		  << fileSize << '\n';
#else // LENGTH_IN_STAT_RECORD
		cout << "state " << (int)charState << " type "
		  << (int)charType << " blocks " << blocks << '\n';
#endif // LENGTH_IN_STAT_RECORD
		break;
		}
	    case ST_READLINK:
		{
		// NEEDSWORK: this incorporates too much knowledge!
		cout << "READLINK ";
		int linkNameLength;
		char linkFileName[MAXPATHLEN];
#ifdef NAME_IN_STAT_RECORD
		if (!file.read(&linkNameLength, sizeof linkNameLength))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for readlink name length at offset "
		      << file.bytesRead();
		    panic("");
		    }
		if (linkNameLength <= 0
		  ||  linkNameLength >= (int)sizeof(linkFileName))
		    {
		    cerr << "dumpcorr:  bad name length " << linkNameLength
		      << " reading fd " << file.rdbuf()->fd()
		      << " for readlink name at offset "
		      << file.bytesRead() - sizeof linkNameLength << '\n';
		    }
		if (!file.read(linkFileName, linkNameLength + 1))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd() << " for stat name at offset "
		      << file.bytesRead();
		    panic("");
		    }
		cout << linkFileName << ' ';
#endif // NAME_IN_STAT_RECORD
		if (!file.read(&linkNameLength, sizeof linkNameLength))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for readlink target length at offset "
		      << file.bytesRead();
		    panic("");
		    }
		if (linkNameLength >= (int)sizeof(linkFileName))
		    {
		    cerr << "dumpcorr:  bad name length " << linkNameLength
		      << " reading fd " << file.rdbuf()->fd()
		      << " for readlink target at offset "
		      << file.bytesRead() - sizeof linkNameLength << '\n';
		    }
		if (linkNameLength <= 0)
		    cout << "-> (readlink failure)\n";
		else if (!file.read(linkFileName, linkNameLength + 1))
		    {
		    cerr << "dumpcorr:  error reading fd "
		      << file.rdbuf()->fd()
		      << " for readlink target at offset "
		      << file.bytesRead();
		    panic("");
		    }
		else
		    cout << "-> " << linkFileName << '\n';
		break;
		}
	    case ST_RESTART:
		{
		cout << "RESTART\n";
		break;
		}
	    case ST_INSTRUCTION:
		{
		cout << "INSTRUCTION ";
		char delimiter = 0;
		while (delimiter != '\n'
		  &&  file.read(&delimiter, sizeof delimiter))
		    cout << delimiter;
		break;
		}
	    case ST_INSTRUCTION_END:
		{
		cout << "INSTRUCTION_END\n";
		break;
		}
	    default:
		{
		cout << "UNKNOWN\n";
		break;
		}
	    }
	}
    }
