#ifndef lint
static const char Rcs_Id[] =
    "$Id: ficusinfo.cc,v 1.10 1996/05/03 01:12:01 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: ficusinfo.cc,v $
 * Revision 1.10  1996/05/03 01:12:01  geoff
 * Update the copyright.
 *
 * Revision 1.9  1996/04/10  01:28:29  geoff
 * Get rid of the obsolete all-managers file storage state.
 *
 * Revision 1.8  1995/11/10  02:59:01  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.7  1995/09/15  00:05:26  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.
 *
 * Revision 1.6  1995/06/30  19:04:34  geoff
 * Generate correct error numbers when possible, and assume that
 * ficus_lstat and graftpoint_info only fail due to ENOENT.
 *
 * Revision 1.5  1995/05/23  00:39:30  geoff
 * Change a bunch of i386 ifdefs to FICUS_KNOWLEDGE ifdefs.
 *
 * Revision 1.4  1995/05/05  22:41:30  geoff
 * Add support for the SS_UFS_STORED/NOT_STORED states.
 *
 * Revision 1.3  1995/03/29  23:57:23  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:04:10  geoff
 * Include stat.h now that correlator.hh doesn't do it for us.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

/*
 * This routine retrieves Ficus-style information (attributes) about a
 * file, including replica-storage information.
 *
 * Usage:
 *
 *	int ficus_extended_stat(const char* path, int iscanonical,
 *	  storage_state_t* storage_state, struct stat* statbuf);
 *
 * The file named by "path" is examined for its attributes and storage
 * status.  If the path is already canonicalized (by "realpath" or a
 * similar routine), setting the "iscanonical" parameter nonzero will
 * improve performance.
 *
 * If the "storage_state" pointer is non-null, the state is returned
 * in that variable.  Similarly, if the "statbuf" pointer is non-null,
 * the statbuf will be filled in with the result of lstat-ing the file
 * (possible copied from the result of a "ficus_lstat").
 *
 * Return values:
 *
 *	A zero value will be returned on success.  If an error occurs,
 *	-1 will be returned and errno will be set to indicate the
 *	error.
 */

/*
 * The Ficus header file fix_sunos_headers.h contains lots of
 * incorrect declarations inherited from SunOS (who ever heard of
 * "free" returning an int, anyway), which unfortunately causes gcc to
 * barf.  So we'll pretend we already included it.  Fortunately, this
 * breaks nothing.
 */
#define fix_sunos_headers_h

#include "compileoptions.hh"
#include "cdecls.hh"
#include "file.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#ifdef FICUS_KNOWLEDGE
extern "C"
    {
#include <ficus/ficus.h>
#include <ficus/ficus_subr.h>
#include <ficus/cmd/lib/libficus/ficus_ook_subr.h>
#include <ficus/cmd/lib/libficus/fix_sunos_headers.h>
#include <ficus/cmd/lib/libficus/libficus.h>
    }
#endif // FICUS_KNOWLEDGE
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef FICUS_KNOWLEDGE
/*
 * The following union allows us to pass the right thing to
 * ficus_lstat, which writes the replica list after the struct fattr.
 * This really should be in one of the Ficus header files.
 */
typedef union
    {
    struct fattr	attrs;
    char		truesize[FFS_SMALLATTRSIZE];
    }
			ficus_lstat_t;

/*
 * The struct_fattr_fa_replist macro gets used a lot in this module,
 * and the name is just too long for readability.  This abbreviation
 * makes the code formatting much more reasonable.
 */
#define replist(x)	struct_fattr_fa_replist(x)
#endif // FICUS_KNOWLEDGE

int			ficus_extended_stat(const char* path, int iscanonical,
			  struct stat* statbuf);

int ficus_extended_stat(		// Get Ficus info about a file
    const char*		path,		// Path to get info about
    int			iscanonical,	// NZ if path is already canonical
    struct stat*	statbuf)	// Where to put struct stat
    {
    char		canpath[MAXPATHLEN]; // Canonical path, if needed
#ifdef FICUS_KNOWLEDGE
    ficus_lstat_t	fileattrs;	// Space for file attributes
    struct gptab*	gpinfo;		// Information on graft point
    struct gptab*	gpnext;		// Next element on graft-point list
    static char		hostname[MAXHOSTNAMELEN] =
			  "";		// Name of host we're running on
    int			isdropped;	// NZ if this repl has been dropped
    replid_t		local_replica;	// ID of local replica, if any
    int			local_status;	// Storage status of local replica
    char*		longhname;	// Long version of hostname
    int			ncopies;	// Number of replicas of data
    int			replno;		// Replica number being examined
    ficus_lstat_t	rootattrs;	// Space for volume root attrs
    char		rootpath[MAXPATHLEN]; // Path to volume root
    int			rootreplno;	// Index into root replica list
#endif // FICUS_KNOWLEDGE

    if (!iscanonical)
	{
	if (realpath(path, canpath) == NULL)
	    return -1;
	path = canpath;
	}
#ifdef FICUS_KNOWLEDGE
    if (get_volumeroot(path, rootpath) == NULL)
	{
	if (errno != EOPNOTSUPP)
	    return -1;
	if (statbuf != NULL)
	    return lstat(path, statbuf);
	return 0;
	}
    if (ficus_lstat(rootpath, (struct fattr*) &rootattrs) != 0)
	{
	/*
	 * Ficus_lstat doesn't set errno in all cases, so we'll force
	 * a valid one.
	 */
	if (errno != ETXTBSY)
	    errno = ENOENT;
	return -1;
	}
    if (ficus_lstat(path, (struct fattr*) &fileattrs) != 0)
	{
	/*
	 * If the file is in conflict, try harder to get attrs for it.
	 */
	if (errno != ETXTBSY)
	    return -1;
	int pathlen = strlen(path);
	if (path[pathlen - 1] == '/')
	    {
	    if (path == canpath)
		{
		canpath[pathlen - 1] = '\0';
		errno =
		  ficus_lstat_on_conflict(canpath, (struct fattr*) &rootattrs,
		    (struct fattr*) &fileattrs);
		}
	    else
		{
		char* temporaryPath = new char[pathlen];
		strncpy(temporaryPath, path, pathlen);
		temporaryPath[pathlen - 1] = '\0';
		errno = ficus_lstat_on_conflict(temporaryPath,
		    (struct fattr*) &rootattrs,
		    (struct fattr*) &fileattrs);
		delete[] temporaryPath;
		}
	    }
	if (errno != 0)
	    return -1;
	}
    if (statbuf != NULL)
	{
	statbuf->st_dev = fileattrs.attrs.fa_va.va_fsid;
	statbuf->st_ino = fileattrs.attrs.fa_va.va_nodeid;
	statbuf->st_mode = fileattrs.attrs.fa_va.va_mode;
	statbuf->st_nlink = fileattrs.attrs.fa_va.va_nlink;
	statbuf->st_uid = fileattrs.attrs.fa_va.va_uid;
	statbuf->st_gid = fileattrs.attrs.fa_va.va_gid;
	statbuf->st_rdev = fileattrs.attrs.fa_va.va_rdev;
	statbuf->st_size = fileattrs.attrs.fa_va.va_size;
	statbuf->st_atime = fileattrs.attrs.fa_va.va_atime.tv_sec;
	statbuf->st_spare1 = fileattrs.attrs.fa_va.va_atime.tv_usec;
	statbuf->st_mtime = fileattrs.attrs.fa_va.va_mtime.tv_sec;
	statbuf->st_spare2 = fileattrs.attrs.fa_va.va_mtime.tv_usec;
	statbuf->st_ctime = fileattrs.attrs.fa_va.va_ctime.tv_sec;
	statbuf->st_spare3 = fileattrs.attrs.fa_va.va_ctime.tv_usec;
	statbuf->st_blksize = fileattrs.attrs.fa_va.va_blocksize;
	statbuf->st_blocks = fileattrs.attrs.fa_va.va_blocks;
	}
    return 0;
#else // FICUS_KNOWLEDGE
    if (statbuf != NULL)
	return lstat(path, statbuf);
    return 0;
#endif // FICUS_KNOWLEDGE
    }
