#ifndef _FILEHISTORY_HH
#define _FILEHISTORY_HH
/*
 * $Id: filehistory.hh,v 1.7 1998/10/23 23:15:16 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filehistory.hh,v $
 * Revision 1.7  1998/10/23 23:15:16  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.6  1998/07/15 21:43:41  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.5  1996/06/11 06:13:17  geoff
 * Add the size() function.
 *
 * Revision 1.4  1996/05/03  01:12:11  geoff
 * Update the copyright.
 *
 * Revision 1.3  1996/02/02  18:28:57  geoff
 * Declare inline functions as such in the class definition.  Add
 * operator=.  Add the returnNulls feature to the iterator.
 *
 * Revision 1.2  1995/12/19  01:02:35  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.1  1995/12/19  00:02:56  geoff
 * Initial revision
 *
 */

/*
 * Classes used to keep track of the recent history of file references.
 */
#include "basictypes.hh"
#include "managerbase.hh"

class File;
class FileHistory;
class FileHistoryIterator;

class HistoryElement
    {
	friend		FileHistory;
	friend		FileHistoryIterator;
    public:
	inline		HistoryElement();
	inline		~HistoryElement();
	inline File*	file() const;	// Return referenced file
	inline ReferenceCount
			time() const;	// Return time of the reference
	inline void	clear();	// Clear record of reference
    private:
	File*		referencedFile;	// File that was referenced
	ReferenceCount	referenceTime;	// Time the reference occurred
    };

class FileHistory
    {
	friend		FileHistoryIterator;
    public:
			FileHistory(int initialSize);
	FileHistory&	operator=(const FileHistory& that);
			~FileHistory();
	inline int	size() const;	// Return size of history
	void		newFile(HoardingManagerBase& manager, File& file);
					// Add a new file to the table
	void		loseForgettingPointers();
					// Clean refs to forgetting files
	FileHistory&	combine(FileHistory& that);
					// Add that's references into this
    private:
			FileHistory(FileHistory& that);
	const int	tableSize;	// Size of the history table
	int		first;		// Index of start of circular buffer
	HistoryElement*	elements;	// Elements in the table
    };

class FileHistoryIterator
    {
    public:
	enum Direction
	    {
	    FORWARD,			// Iterator goes from oldest to newest
	    REVERSE			// Iterator goes from newest to oldest
	    };
	inline		FileHistoryIterator(FileHistory& table,
			  Direction plannedDirection = FORWARD,
			  bool wantNulls = false);
	inline		~FileHistoryIterator();
	inline void	reset(Direction plannedDirection = FORWARD,
			  bool wantNulls = false);
					// Restart iterator from beginning
	inline		operator bool() const;
					// Do completion testing
	inline HistoryElement&
			operator*() const; // Return current element
	inline FileHistoryIterator&
			operator++();	// Step to next element
	inline FileHistoryIterator&
			operator--();	// Step to previous element
	inline int	currentPosition() const;
					// Return index of current position
    private:
	FileHistory&	target;		// Target of the iterator
	int		index;		// Index of next element to return
	bool		live;		// True if iterator not yet expired
	bool		returnNulls;	// True to return even NULL entries
    };

#define _FILEHISTORY_ICC_NEEDED

#endif // _FILEHISTORY_HH
