#ifndef lint
static const char Rcs_Id[] =
    "$Id: filereference.cc,v 1.30 1998/10/23 23:15:17 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filereference.cc,v $
 * Revision 1.30  1998/10/23 23:15:17  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.29  1998/07/15 21:43:41  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.28  1998/07/09 21:02:02  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.27  1998/07/08 19:59:03  geoff
 * Add quickAverageDistance, and make averageDistance return the true
 * average in geometric-means mode.  This is needed by the
 * non-shared-neighbors clustering code.
 *
 * Revision 1.26  1998/06/05 18:06:45  geoff
 * Only call relationsChanged for distance changes if we aren't using
 * shared neighbors.
 *
 * Revision 1.25  1998/05/14 01:15:31  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.24  1998/02/19 02:10:07  geoff
 * When relations change, inform the owning file.
 *
 * Revision 1.23  1997/12/04 23:15:54  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.22  1997/01/02 06:48:21  geoff
 * Make geometric means be the default.
 *
 * Revision 1.21  1996/10/11 22:09:03  geoff
 * Add geometric-mean support.  Break out of the history-update loop when
 * we encounter a previous occurrence of this file.
 *
 * Revision 1.20  1996/10/04 07:12:12  geoff
 * Add support for command-interpreter references.
 *
 * Revision 1.19  1996/09/11 22:59:51  geoff
 * Use a type-specific allocator.
 *
 * Revision 1.18  1996/06/11  06:13:18  geoff
 * Add the previous reference time as an argument to newReference and
 * updateDistance, so that the latter can do aging of outdated
 * references.  Save and restore the reference flags in the database.
 * Ignore references to any file marked as alwaysHoardableType.
 * Calculate local distance based on file-table depth, not reference
 * times.  Add new NPREV correction that works with process trees.  Add
 * aging of old references.
 *
 * Revision 1.17  1996/05/03  01:12:14  geoff
 * Update the copyright.
 *
 * Revision 1.16  1996/04/10  01:28:35  geoff
 * Make RENAME be one of the prohibited reference types.
 *
 * Revision 1.15  1996/02/02  20:09:45  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.14  1996/02/02  18:28:58  geoff
 * Move some virtual functions out of the inlined code.  Get rid of some
 * excess discoverStatus() calls.  Fix some minor problems detected by
 * the new C++ compiler.  Use count() routines rather than counting
 * loops.  Don't do NPREV correction in process-tree mode, because I
 * decided I don't understand how it should work.
 *
 * Revision 1.13  1995/12/19  01:02:55  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.12  1995/12/19  00:34:22  geoff
 * Add process-tracking support.  This primarily involves making the
 * history table an argument, rather than a hard-wired global.  Add
 * FinalFileReferenceDestructor to get rid of statically-allocated stuff.
 * Get rid of some multiple calls to functions, for efficiency.
 *
 * Revision 1.11  1995/11/22  00:20:43  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.10  1995/11/21  21:38:30  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support.  Put back the temporary-variable solution to 80-bit
 * floating-point returns
 *
 * Revision 1.9  1995/11/10  02:59:10  geoff
 * Add zzinlinefunctions.hh.  Add extensive NEW_FILE_RELATIONS support.
 * Other minor fixes.
 *
 * Revision 1.8  1995/09/15  00:05:36  geoff
 * Change the RCS ID to be a const.  Various changes to support the
 * deletion manager and compressed input files.
 *
 * Revision 1.7  1995/06/30  19:04:42  geoff
 * Change show output to include file ID's.  Report file offsets in most
 * error messages.  Use underlyingFind, rather than findCanonical, when
 * reading in saved references.
 *
 * Revision 1.6  1995/06/17  00:11:19  geoff
 * Add show and clear.
 *
 * Revision 1.5  1995/05/23  00:39:50  geoff
 * Put a tab before the file name when outputting a reference.
 *
 * Revision 1.4  1995/04/20  21:57:48  geoff
 * Use findCanonical to find referenced files.
 *
 * Revision 1.3  1995/03/30  00:44:28  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.2  1995/03/29  21:04:29  geoff
 * Switch to byte-swapped streams.  Add some new-parameter-setting support.
 * Fix a few bugs.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "clustering_mgr.hh"
#include "correlator.hh"
#include "file.hh"
#include "filehistory.hh"
#include "filereference.hh"
#include "globalmanager.hh"
#include "managerbase.hh"
#include "panic.hh"
#include "process.hh"
#include "zzinlinefunctions.hh"
#include <fstream.h>
#include <iomanip.h>
#include <math.h>
#include <sys/param.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			FileReferenceType::~FileReferenceType();
FileRelationBase&	FileReferenceType::newRelation() const;
					// Create new relation instance
FileRelationBase*	FileReferenceType::readRelation(
			    ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
void			FileReferenceType::recordReference(File& file,
			    ReferenceType type, Process* process,
			    bool isMeaningless)
			  const;
					// Record a reference to a given file
void			FileReferenceType::cleanGlobalData() const;
					// Clean out type's global data
			FinalFileReferenceDestructor::
			  FinalFileReferenceDestructor();
			FinalFileReferenceDestructor::
			  ~FinalFileReferenceDestructor();
			FileReference::~FileReference();
const FileRelationTypeBase&
			FileReference::type() const;
					// Return type of the relation
ostream&		operator<<(ostream& stream,
			  const FileReference& reference);
					// Show reference in printable form
					// ..appends \n
ostream&		FileReference::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show reference in printable form,
					// ..with options.  Appends \n
void			FileReference::newReference(File* target,
			  ReferenceCount distance,
			  ReferenceCount previousReferenceTime);
					// Set file being referenced
ByteSwappedStream&	FileReference::write(ByteSwappedStream& stream) const;
					// Write a reference entry to a stream
void			FileReference::updateDistance(File& sourceFile,
			  File& targetFile, ReferenceCount localDistance,
			  ReferenceCount previousReferenceTime,
			  HoardingManagerBase& manager);
					// Update distance between two files
void			FileReference::correct(ReferenceCount lastReference,
			  HoardingManagerBase& manager);
					// Perform NPREV correction
void			FileReference::uncorrect(ReferenceCount referenceTime,
			  HoardingManagerBase& manager);
					// Undo NPREV correction if necessary
double			FileReference::averageDistance() const;
					// Return avg distance of reference
double			FileReference::quickAverageDistance() const;
					// Return avg distance of reference,
					// ..calculated quickly and only useful
					// ..for relative comparisons
void			FileReference::setGeometricMeans(bool useGeometric);
					// Set whether to use geometric means
void			FileReference::readGeometricMeans(
					  ByteSwappedStream& stream);
					// Read geometric-mean flag from stream
void			FileReference::writeGeometricMeans(
					  ByteSwappedStream& stream);
					// Write geometric-mean flag to stream
			FileReferenceIterator::~FileReferenceIterator();
#endif // 0

/*
 * Static class variables
 */
const char*		FileReferenceType::typeClass = "FileReference";
					// Master type class name
FinalFileReferenceDestructor
			FileReference::finalFileReferenceDestructor;
FileReference*		FileReference::freeReferences = NULL;
					// List of items for allocation
int			FileReference::freeCount = 0;
					// Number of items on free list
bool			FileReference::geometricMeans = true;
					// True to use geometric means
FileHistory		FileReference::globalHistory(MAX_NPREV);
					// History of previous references
ReferenceCount		FileReference::tieBreaker = 0;
const FileReferenceType	FileReference::typeInformation("FileReference");
#ifdef USE_TYPE_ALLOCATOR
TypeAllocator		FileReference::allocator(sizeof(FileReference), 0);
#endif // USE_TYPE_ALLOCATOR

extern void		correlatorPanic(const char* message);

FileReferenceType::~FileReferenceType()
    {
    }

/*
 * Create a new file-reference relation.  The relation is created with
 * "new", even though it is returned as a reference, and must someday
 * be deleted.
 */
FileRelationBase& FileReferenceType::newRelation()
    const
    {
    FileReference* result = new FileReference;
    return *result;
    }

/*
 * Read a new file reference from a stream.
 */
FileRelationBase* FileReferenceType::readRelation(
    ByteSwappedStream&	stream)		// Stream to read from
    const
    {
    FileReference* newReference =
      (FileReference*)&FileReference::typeInformation.newRelation();
    stream.read(&newReference->numReferences,
      sizeof newReference->numReferences);
    if (FileReference::geometricMeans)
	stream.read(&newReference->accumulator.totalLogDistance);
    else
	stream.read(&newReference->accumulator.totalDistance);
    stream.read(&newReference->flags, sizeof newReference->flags);
    return newReference;
    }

void FileReferenceType::recordReference( // Record a reference to a given file
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of the reference
    Process*		process,	// Process doing the reference
    bool		isMeaningless)	// True if reference is meaningless
    const
    {
    if (isMeaningless  ||  file.alwaysHoardableType())
	return;
    switch (type)
	{
	case SEER_REF_CLOSE:		// Close of open file
	case SEER_REF_EXIT:		// Process exit or re-exec
	case SEER_REF_DESTROY:		// File deletion
	    {
	    return;			// Don't record these as references
	    }
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_RW:		// Reads and/or writes of existing file
	case SEER_REF_EXEC:		// Execution
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	case SEER_REF_OPENREAD:		// Open for read
	case SEER_REF_OPENRW:		// Open for read/write
	case SEER_REF_OPENCREATE:	// Create or truncate/rewrite
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	case SEER_REF_MAKELINK:		// Create symbolic link
	case SEER_REF_PEEKLINK:		// Peek at a link (lstat or readlink)
	case SEER_REF_FOLLOW:		// Follow symbolic link
	    {
	    break;
	    }
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_DUP:		// Duplicate an existing open
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	case SEER_REF_RENAME:		// Rename a file
	default:
	    {
	    cerr << "correlator: unexpected reference type " << type;
	    correlatorPanic(" in FileReference::recordReference");
	    break;	
	    }
	}
    // @h NEEDSWORK: shouldn't have clustering manager wired in
    for (FileReferenceIterator reference(file);  reference;  ++reference)
	(*reference).uncorrect(clusteringManager.managerData(file)
	    .lastReferenceTime(),
	  clusteringManager);
    // @h NEEDSWORK: shouldn't have clustering manager wired in
    if (clusteringManager.isOften(file, GlobalManager::getOftenPercent()))
	return;
    if (file.isForgetting())
	return;
    /*
     * Update reference records.
     */
    FileHistory* history = &FileReference::globalHistory;
    if (process != NULL  &&  GlobalManager::useProcessTree())
	history = &process->fileHistory();
    for (FileHistoryIterator previous(*history, FileHistoryIterator::REVERSE);
      previous;
      --previous)
	{
	HistoryElement& element = *previous;
	File* previousFile = element.file();
	if (previousFile == &file)
	    {
	    element.clear();
	    break;
	    }
	else
	    {
	    ReferenceCount localDistance;
	    if (GlobalManager::useProcessTree())
		localDistance = process != NULL
		    &&  process->isOpenByProcess(*previousFile)
		  ? 0
		  : (ReferenceCount)(history->size()
		    - previous.currentPosition());
	    else
		{
		// @h NEEDSWORK: shouldn't have clustering manager wired in
		localDistance = previousFile->isOpen() ? 0
		  : clusteringManager.currentReferenceTime() - element.time();
		}
	    // @h NEEDSWORK: shouldn't have clustering manager wired in
	    FileReference::updateDistance(*previousFile, file, localDistance,
	      clusteringManager.managerData(file).lastReferenceTime(),
	      clusteringManager);
	    }
	}
    // @h NEEDSWORK: shouldn't have clustering manager wired in
    history->newFile(clusteringManager, file);
    }

/*
 * Get rid of any references in the 'previous' table which point at
 * files being forgotten.
 */
void FileReferenceType::cleanGlobalData()
    const
    {
    FileReference::globalHistory.loseForgettingPointers();
    }

FinalFileReferenceDestructor::FinalFileReferenceDestructor()
    {
    }

FinalFileReferenceDestructor::~FinalFileReferenceDestructor()
    {
#ifndef USE_TYPE_ALLOCATOR
    for (  ;  FileReference::freeCount > 0;  FileReference::freeCount--)
	{
	FileReference* goingReference = FileReference::freeReferences;
	FileReference::freeReferences =
	  *(FileReference**)FileReference::freeReferences;
	::operator delete(goingReference);
	}
#endif // USE_TYPE_ALLOCATOR
    }

FileReference::~FileReference()
    {
    }

/*
 * Return the type of the file-reference relation.  For file
 * references, the type is always the same.
 */
const FileRelationTypeBase& FileReference::type()
    const
    {
    return FileReference::typeInformation;
    }

/*
 * Show reference in printable form (appends \n)
 */
ostream& operator<<(
    ostream&		stream,		// Stream to print to
    const FileReference& reference)	// Reference to be printed
    {
    return reference.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Show reference in printable form, with options.  Appends \n.
 *
 * Show options honored: none
 */
ostream& FileReference::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		,		// Desired display options
    const char*		)		// Manager to show for, if SO_MANAGER
    const
    {
    stream << "RFD_REF\t    " << numReferences;
    stream << ' ' << accumulator.totalDistance
      << ' ' << averageDistance() << '\n';
    return stream;
    }

void FileReference::newReference(	// Set file being referenced
    File*		target,		// File referenced
    ReferenceCount	localDistance,	// Local distance of this reference
    ReferenceCount	previousReferenceTime)
					// Last time this file was referenced
    {
    if (!referenceIsTo(target))
	correlatorPanic("newReference called on non-target file");
    else if (flags & RF_CORRECTED)
	{
	if (GlobalManager::useProcessTree())
	    {
	    // @h NEEDSWORK: shouldn't have clustering manager wired in
	    uncorrect(previousReferenceTime, clusteringManager);
	    }
	else
	    {
	    cerr
	      << "correlator:  unexpected corrected entry when referencing '"
	      << target->encodedName(SO_MIME_ENCODE) << "'\n";
	    flags &= ~RF_CORRECTED;
	    }
	}
    numReferences++;
    if (geometricMeans)
	accumulator.totalLogDistance += log((double)localDistance + 1.0);
    else
	accumulator.totalDistance += localDistance;
    }

ByteSwappedStream& FileReference::write( // Write a reference entry to a stream
    ByteSwappedStream&	stream)		// Stream to use
    const
    {
    if (!targetRelatedFile().targetFile().isForgetting())
	{
	FileRelationBase::write(stream);
	stream.write(&numReferences, sizeof numReferences);
	if (geometricMeans)
	    stream.write(&accumulator.totalLogDistance);
	else
	    stream.write(&accumulator.totalDistance);
	stream.write(&flags, sizeof flags);
	}
    return stream;
    }

/*
 * Update the distance to a nearby file
 */
void FileReference::updateDistance(
    File&		sourceFile,	// File previously referenced
    File&		targetFile,	// File currently referenced
    ReferenceCount	localDistance,	// Local distance to that file
    ReferenceCount	previousReferenceTime,
					// Previous time source was referenced
    HoardingManagerBase& manager)	// Manager we're working on behalf of
    {
    // @h NEEDSWORK: should we throw out references if keepReferences
    // decreases?
    FileReference* reference = NULL;
    FileReference* missingReference = NULL;
    FileReference* forgettingReference = NULL;
    FileReference* forgottenReference = NULL;
    FileReference* oldReference = NULL;
    ReferenceCount oldThreshold = 0;
    int referenceCount = typeInformation.count(sourceFile);
    FileReferenceIterator referenceIterator(sourceFile);
    for (  ;  referenceIterator;  ++referenceIterator)
	{
	FileReference* testReference(&(*referenceIterator));
	File& referenceTarget =
	  testReference->targetRelatedFile().targetFile();
	if (&targetFile == &referenceTarget)
	    {
	    reference = testReference;
	    break;
	    }
	switch (referenceTarget.currentState())
	    {
	    case FileState::MISSING:
		{
		missingReference = testReference;
		break;
		}
	    case FileState::FORGETTING:
		{
		forgettingReference = testReference;
		break;
		}
	    case FileState::FORGOTTEN:
		{
		forgottenReference = testReference;
		break;
		}
	    default:
		{
		/*
		 * See if this reference is "old".  If so, allow it to
		 * be replaced.
		 */
		ReferenceCount referenceThreshold;
		if (geometricMeans)
		    {
		    /*
		     * We estimate the last-update-time of the reference
		     * as being the target file's last reference time.  We
		     * then adjust this by the average distance of the
		     * reference, multiplied by a weight and by the
		     * frequency of occurrence, and also add the aging
		     * threshold, to get a value which, if greater
		     * than the current reference time, defines this
		     * reference as an "old" one.
		     *
		     * The logic behind adding the average distance is
		     * to allow the reference to be updated before
		     * being discarded.  By multiplying the average by
		     * the total reference ocunt, we give
		     * frequently-occurring references priority over
		     * rare ones.
		     */
		    referenceThreshold =
		      manager.managerData(referenceTarget).lastReferenceTime()
			+ (unsigned long)(exp(testReference->accumulator
				.totalLogDistance)
			      * testReference->numReferences
			      * GlobalManager::ageWeight()
			    + 0.5)
			+ GlobalManager::ageThreshold();
		    }
		else
		    {
		    /*
		     * We estimate the last-update-time of the reference
		     * as being the target file's last reference time.  We
		     * then adjust this by the weighted total distance and
		     * the aging threshold to get a value which, if
		     * greater than the current reference time, defines
		     * this reference as an "old" one.
		     *
		     * The logic behind adding the weighted total distance
		     * is twofold.  First, we want to add in the average
		     * distance (slightly inflated, which is the reason
		     * for the weight) so that a reference has a chance to
		     * be updated before being discarded.  Second, by
		     * using the total distance instead of the average, we
		     * give frequently-occurring references priority over
		     * rare ones.
		     */
		    referenceThreshold =
		      manager.managerData(referenceTarget).lastReferenceTime()
			+ (unsigned long)(testReference->accumulator
				.totalDistance
			      * GlobalManager::ageWeight()
			    + 0.5)
			+ GlobalManager::ageThreshold();
		    }
		if (referenceThreshold < manager.currentReferenceTime())
		    {
		    if (oldReference == NULL
		      ||  oldThreshold > referenceThreshold)
			{
			oldReference = testReference;
			oldThreshold = referenceThreshold;
			}
		    }
		break;
		}
	    }
	}
    if (forgottenReference != NULL)
	missingReference = forgottenReference;
    else if (forgettingReference != NULL)
	missingReference = forgettingReference;
    if (reference == NULL
      &&  referenceCount < GlobalManager::referencesToKeep())
	{
	/*
	 * targetFile wasn't in the table, and we've got fewer
	 * references than the maximum, so add a new one.
	 */
	reference = new FileReference;
	sourceFile.addRelation(reference, targetFile);
	}
    else if (reference == NULL  &&  missingReference != NULL)
	{
	/*
	 * The previously-referenced file's reference table is full,
	 * but there's a reference to a file that has been deleted (or
	 * is otherwise going to disappear).  Replace the reference to
	 * the deleted file with the new one.
	 */
	sourceFile.removeRelation(missingReference);
	reference = new FileReference;
	sourceFile.addRelation(reference, targetFile);
	}
    else if (reference == NULL)
	{
	/*
	 * The previously-referenced file's reference table is full.
	 * Locate the entry with the largest distance, and replace it.
	 * Break ties among equal distances by selecting the smallest
	 * number of references (this occurs primarily when the
	 * distance is a small integer, such as 1).  Break ties among
	 * the smallest number of references semi-randomly.
	 *
	 * If the new reference has a distance larger than any
	 * currently in the table, leave "reference" NULL, which will
	 * cause the new reference to be ignored.
	 */
	referenceIterator.reset();
	reference = &(*referenceIterator);
	double bigDistance = reference->quickAverageDistance();
	ReferenceCount smallestNumReferences = reference->referenceCount();
	int numberOfTies = 0;
	for (++referenceIterator;  referenceIterator;  ++referenceIterator)
	    {
	    FileReference* testReference(&(*referenceIterator));
	    ReferenceCount currentNumReferences =
	      testReference->referenceCount();
	    double currentDistance =
	      testReference->quickAverageDistance();
	    if (currentDistance > bigDistance)
		{
		reference = testReference;
		bigDistance = currentDistance;
		smallestNumReferences = currentNumReferences;
		numberOfTies = 0;
		}
	    else if (currentDistance == bigDistance)
		{
		if (currentNumReferences < smallestNumReferences)
		    {
		    reference = testReference;
		    bigDistance = currentDistance;
		    smallestNumReferences = currentNumReferences;
		    numberOfTies = 0;
		    }
		else if (currentNumReferences == smallestNumReferences)
		    numberOfTies++;
		}
	    }
	/*
	 * @h NEEDSWORK: should I use >= here, or just >?  I figured
	 * that >= will introduce a bit of stability, so I chose that,
	 * but one could argue that instability might be better.
	 */
	if (geometricMeans
	  ? (log((double)localDistance + 1.0) >= bigDistance)
	  : (localDistance >= bigDistance))
	    reference = NULL;		// New ref bigger than any current
	else
	    {
	    /*
	     * The new reference belongs in the table.  If there are
	     * no ties, "reference" points to the reference it should
	     * replace.  Otherwise, we must choose a semi-random
	     * reference to discard.  In either case, we remove the
	     * old reference and then add the new one.
	     */
	    if (numberOfTies != 0)
		{
		/*
		 * There are ties for the proper slot.  Pick one, and
		 * assign it to "reference".
		 *
		 * Note that this code is inefficient because it
		 * searches the list twice, once to identify the
		 * reference to be removed and once to actually delete
		 * it.  This could be improved by using
		 * removeCurrent(), but I decided to go for
		 * maintainability first.
		 */
		numberOfTies = ++tieBreaker % (numberOfTies + 1);
		for (referenceIterator.reset();
		  referenceIterator;
		  ++referenceIterator)
		    {
		    reference = &(*referenceIterator);
#ifdef __i386__
		    /*
		     * On the 386, some compilers may return
		     * quickAverageDistance() in an 80-bit register,
		     * causing the equality comparison to fail unless
		     * we use a temporary.  We ought to be able to
		     * work around this by casting to double, but that
		     * doesn't work with GCC.  So we have to ifdef it.
		     */
		    double trialDistance = reference->quickAverageDistance();
		    if (reference->referenceCount() == smallestNumReferences
		      &&  trialDistance == bigDistance)
#else // __i386__
		    if (reference->referenceCount() == smallestNumReferences
		      &&  reference->quickAverageDistance() == bigDistance)
#endif // __i386__/
			{
			if (--numberOfTies < 0)
			    break;
			}
		    }
		}
	    sourceFile.removeRelation(reference);
	    reference = new FileReference;
	    sourceFile.addRelation(reference, targetFile);
	    }
	}
    if (reference == NULL  &&  oldReference != NULL)
	{
	/*
	 * The previously-reference file's reference table is full,
	 * but there's a reference that is old enough that we believe
	 * it to be obsolete.  Replace the outdated reference with the
	 * new one.
	 */
	sourceFile.removeRelation(oldReference);
	reference = new FileReference;
	sourceFile.addRelation(reference, targetFile);
	}
    if (reference != NULL)
	{
	reference->newReference(&targetFile, localDistance,
	   previousReferenceTime);
	// NEEDSWORK: this test should be in the clustering manager,
	// and relationsChanged should accept a parameter telling how
	// they changed.
	if (!ClusteringManager::usingSharedNeighbors())
	    sourceFile.relationsChanged();
	}
    }

void FileReference::correct(		// Perform NPREV correction
    ReferenceCount	lastReference,	// Last reference to file being dropped
    HoardingManagerBase& manager)	// Manager we're working on behalf of
    {
    /*
     * PERFORM MAX_NPREV CORRECTION:
     *
     * For the file that is going to be dropped from "previous", scan
     * all references to see if they are older than the file.  If so,
     * we will assume that the next reference to that file will be at
     * least MAX_NPREV+1 references away, so we add that much to the total
     * distance.  This assumption will be verified or corrected later,
     * in one of two places.  Call the file being dropped now "D", and
     * the file in its reference table (which we are correcting) "R".
     *
     * At the next reference to "D", there either has or has not been
     * an intervening reference to "R".  If there has, then the total
     * distance (incorporating the MAX_NPREV correction) is correct, but
     * we must increment the number of references.  If there has not,
     * then we must remove the correction.  The reason is the
     * possibility of a sequence like this:
     *
     *	    D, R, D, <many references>, D, R
     *
     * In this case, the proper distance from D to R is (1+1)/2.
     * Contrast this with the similar case:
     *
     *	    D, R, D, <many references>, R, D
     *
     * In this case, the best estimate of the proper distance is
     * (MAX_NPREV+1+1)/2.
     *
     * At file management time, a similar situation occurs, except
     * that there has not been a reference to D.  It is more accurate
     * to remove the correction in this case.  Consider the sequence:
     *
     *	    D, R, D, R, D, R, D, <many references>, <management>
     *
     * In this case, the best estimate of the correct distance from D
     * to R is (1+1+1)/3, rather than (MAX_NPREV+1+3)/4.
     */
    if (
      manager.managerData(targetRelatedFile().targetFile()).lastReferenceTime()
	< lastReference)
	{
	if (geometricMeans)
	    accumulator.totalLogDistance += log(MAX_NPREV + 1 + 1.0);
	else
	    accumulator.totalDistance += MAX_NPREV + 1;
	flags |= RF_CORRECTED;
	}
    }

void FileReference::uncorrect(		// Undo NPREV correction if necessary
    ReferenceCount	referenceTime,	// Containing file's reference time
    HoardingManagerBase& manager)	// Manager we're working on behalf of
    {
    /*
     * Perform MAX_NPREV recorrection.  If the referenced file has
     * received correction, but has not been referenced more recently
     * than the current file, we must remove the correction.
     * Otherwise, we must increment the reference count to reflect the
     * corrected reference.  See "PERFORM MAX_NPREV CORRECTION" above
     * for more discussion.
     */
    if ((flags & RF_CORRECTED) != 0)
	{
	flags &= ~RF_CORRECTED;
	if (referenceTime
	  > manager.managerData(targetRelatedFile().targetFile())
	    .lastReferenceTime())
	    {
	    if (geometricMeans)
		accumulator.totalLogDistance -= log(MAX_NPREV + 1 + 1.0);
	    else
		accumulator.totalDistance -= MAX_NPREV + 1;
	    }
	else
	    numReferences++;
	}
    }

double FileReference::averageDistance()	// Return avg dist of reference
    const
    {
    if (geometricMeans)
	return exp(quickAverageDistance());
    else
	return quickAverageDistance();
    }

/*
 * Return a quick version of the average distance, guaranteed to be
 * useful only for relative comparisons.  If geometric means are being
 * used, this is the log of the average distance; otherwise it is the
 * true average distance.
 */
double FileReference::quickAverageDistance()
    const
    {
    /*
     * Before calculating average distance, perform MAX_NPREV
     * recorrection.  See the comments under "PERFORM MAX_NPREV
     * CORRECTION" in the function "correlate" for more information.
     */
    if (geometricMeans)
	{
	/*
	 * We actually return the log of the geometric mean, rather
	 * than the geometric mean itself.  Since most of our callers
	 * use the value for comparison purposes, the log is adequate.
	 */
	double correctedDistance = accumulator.totalLogDistance;
	if (flags & RF_CORRECTED)
	    correctedDistance -= log(MAX_NPREV + 1 + 1.0);
	return correctedDistance / numReferences;
	}
    else
	{
	ReferenceCount correctedDistance = accumulator.totalDistance;
	if (flags & RF_CORRECTED)
	    correctedDistance -= MAX_NPREV + 1;
	return (double) correctedDistance / numReferences;
	}
    }

void FileReference::setGeometricMeans(	// Set whether to use geometric means
    bool		useGeometric)	// True to calculate geometric means
    {
    geometricMeans = useGeometric;
    }

void FileReference::readGeometricMeans(	// Read geometric-mean flag from stream
    ByteSwappedStream&	stream)		// Stream to read from
    {
    char flag;
    if (stream.read(&flag, sizeof flag))
	geometricMeans = flag;
    }

void FileReference::writeGeometricMeans( // Write geometric-mean flag to stream
    ByteSwappedStream&	stream)		// Stream to write to
    {
    char flag = geometricMeans;
    stream.write(&flag, sizeof flag);
    }

FileReferenceIterator::~FileReferenceIterator()
    {
    }
