#ifndef lint
static const char Rcs_Id[] =
    "$Id: filerelation.cc,v 1.17 1998/10/23 23:15:18 geoff Exp $";
#endif

/*
 * $Id: filerelation.cc,v 1.17 1998/10/23 23:15:18 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filerelation.cc,v $
 * Revision 1.17  1998/10/23 23:15:18  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.16  1998/07/15 21:43:43  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.15  1998/07/09 21:02:02  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.14  1998/05/14 01:15:32  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.13  1998/04/09 00:33:01  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.12  1997/12/18 01:25:08  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.11  1997/12/04 23:15:59  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.10  1996/12/03 02:58:05  geoff
 * Change the name of the routine that records missed references.  When
 * reading a saved database, reassign existing type IDs to make sure they
 * don't conflict with the new ones recorded in the database.
 *
 * Revision 1.9  1996/10/04 07:12:13  geoff
 * When reading relations, don't create an invalid relation if
 * File::findFromStream fails for some reason.
 *
 * Revision 1.8  1996/09/11 22:59:55  geoff
 * Add REPORT_SELECTED_FAILURES support.  Convert from dlists to slists.
 * Use a type-specific allocator.
 *
 * Revision 1.7  1996/05/03  01:12:17  geoff
 * Update the copyright.
 *
 * Revision 1.6  1996/02/02  20:09:46  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.5  1996/02/02  18:29:00  geoff
 * Move a couple of virtual functions into the C code.  Fix some minor
 * problems detected by the new C++ compiler.  Add a count() routine.
 *
 * Revision 1.4  1995/12/19  00:34:29  geoff
 * Add FinalRelatedFileDestructor to get rid of statically-allocated
 * stuff at exit.
 *
 * Revision 1.3  1995/11/22  00:20:57  geoff
 * Fix the RCS comment leader.  Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.2  1995/11/21  21:38:43  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support.
 *
 * Revision 1.1  1995/11/10  02:09:40  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "file.hh"
#include "filerelation.hh"
#include "investigator.hh"
#include "list.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <string.h>
#include <sys/param.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			FileRelationTypeBase::FileRelationTypeBase(
			  const char* name, const char*const className);
			FileRelationTypeBase::~FileRelationTypeBase();
FileRelationTypeBase*	FileRelationTypeBase::findByName(const char* name);
					// Locate type entry for named relation
FileRelationTypeBase*	FileRelationTypeBase::findById(int id);
					// Locate relation by ID number
const char*		FileRelationTypeBase::encodedName(ShowOptions options)
			  const;	// Return encoded name of relation type
const char*		FileRelationTypeBase::encodedClassName(
			    ShowOptions options)
			  const;	// Return encoded class name
unsigned long		FileRelationTypeBase::count(const File& file) const;
					// Return count of the number of
					// relations of this type that a given
					// ..file has
ByteSwappedStream&	FileRelationTypeBase::writeRelationTypes(
			  ByteSwappedStream& stream);
					// Write relation-type table to stream
ByteSwappedStream&	FileRelationTypeBase::readRelationTypes(
			  ByteSwappedStream& stream);
					// Read relation-type table from stream
ByteSwappedStream&	FileRelationTypeBase::write(ByteSwappedStream& stream)
			  const;	// Write type-specific info to stream
ByteSwappedStream&	FileRelationTypeBase::read(ByteSwappedStream& stream);
					// Read type-specific info from stream
void			FileRelationTypeBase::recordMissedReference(File& file,
			    ReferenceType type,
			    Process* process, bool isMeaningless)
			  const;
			FileRelationBase::~FileRelationBase();
ByteSwappedStream&	FileRelationBase::write(ByteSwappedStream& stream)
			  const;
					// Write relation to a stream
FileRelationBase*	FileRelationBase::read(ByteSwappedStream& stream);
					// Read relation from a stream
			FinalRelatedFileDestructor::
			  ~FinalRelatedFileDestructor();
			FinalRelatedFileDestructor::
			  FinalRelatedFileDestructor();
			RelatedFile::~RelatedFile();
bool			RelatedFile::hasType(const FileRelationTypeBase& type)
			  const;	// See if file has a given rel. type
FileRelationBase*	RelatedFile::chosenRelation(
			    const FileRelationTypeBase& type, bool create);
					// Return relation of chosen type
ByteSwappedStream&	RelatedFile::write(ByteSwappedStream& stream) const;
					// Write a related file's relations
ByteSwappedStream&	RelatedFile::read(ByteSwappedStream& stream,
			    File& sourceFile);
					// Read a related file's relations
ostream&		RelatedFile::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show related file in printable form
#endif // 0

/*
 * Static class variables
 */
SList			FileRelationTypeBase::allTypes(false);
unsigned char		FileRelationTypeBase::nextId = 0;
RelatedFile*		RelatedFile::freeRelatedFiles = NULL;
int			RelatedFile::freeCount = 0;
FinalRelatedFileDestructor RelatedFile::finalRelatedFileDestructor;
#ifdef USE_TYPE_ALLOCATOR
TypeAllocator		RelatedFile::allocator(sizeof(RelatedFile), 0);
#endif // USE_TYPE_ALLOCATOR

extern void		correlatorPanic(const char* message);
extern "C" char*	mimeEncode(const char* source,
			  const char* whatToEncode);
					// MIME-encode a string

/*
 * Construction for the file-relation-type base class.  The caller
 * must ensure that the className is a pointer to a non-volatile area
 * that will survive for the life of this object.
 */
FileRelationTypeBase::FileRelationTypeBase(
    const char*		name,		// Name of the relation
    const char*		className)	// Name of the relation's type class
    : typeName(strcpy(new char[strlen(name) + 1], name)),
      typeClassName(className),
      typeId(nextId++)
    {
    allTypes.pushHead(this);
    }
    
FileRelationTypeBase::~FileRelationTypeBase()
    {
    allTypes.remove(this);
    delete[] (char*)typeName;
    }

/*
 * Locate the type entry for a named relation.  Returns NULL if no
 * relation by that name can be found.
 */
FileRelationTypeBase* FileRelationTypeBase::findByName(
    const char*		name)		// Name to search for
    {
    for (SList::Iterator iterator(allTypes);  iterator;  ++iterator)
	{
	FileRelationTypeBase* entry =
	  (FileRelationTypeBase*)(allTypes[iterator]);
	if (strcmp(entry->name(), name) == 0)
	    return entry;
	}
    return NULL;
    }

/*
 * Locate the type entry for a relation, based on the relation ID.
 * Returns NULL if no relation with the given ID can be found.
 */
FileRelationTypeBase* FileRelationTypeBase::findById(
    int			id)		// ID to search for
    {
    for (SList::Iterator iterator(allTypes);  iterator;  ++iterator)
	{
	FileRelationTypeBase* entry =
	  (FileRelationTypeBase*)(allTypes[iterator]);
	if (entry->typeId == id)
	    return entry;
	}
    return NULL;
    }

/*
 * Return the MIME-encoded name of the relation type.  This cannot be
 * called multiple times in a single statement.
 */
const char* FileRelationTypeBase::encodedName(
    ShowOptions		options)	// Display options
    const
    {
    if (options & SO_MIME_ENCODE)
	{
	static const char* encoding = NULL;
	if (encoding != NULL)
	    free((void*)encoding);
	encoding = mimeEncode(name(), NULL);
	return encoding;
	}
    else
	return name();
    }

/*
 * Return the MIME-encoded class name.  This cannot be called multiple
 * times in one statement.
 */
const char* FileRelationTypeBase::encodedClassName(
    ShowOptions		options)	// Display options
    const
    {
    if (options & SO_MIME_ENCODE)
	{
	static const char* encoding = NULL;
	if (encoding != NULL)
	    free((void*)encoding);
	encoding = mimeEncode(className(), NULL);
	return encoding;
	}
    else
	return name();
    }

/*
 * Return count of the number of relations of this type that a given
 * file has.
 */
unsigned long FileRelationTypeBase::count(
    const File&		file)		// File that we will count for
      const
    {
    unsigned long numberOfRelations = 0;
    for (SelectedRelatedFileIteratorBase relatedFile((File&)file, *this);
      relatedFile;
      ++relatedFile)
	numberOfRelations++;
    return numberOfRelations;
    }

/*
 * Write a table of all known relation types to a stream.  The format
 * of the table is very simple.  First is a byte giving the number of
 * entries.  Each entry consists of a byte ID, a short name length,
 * and a null-terminated relation name.
 */
ByteSwappedStream& FileRelationTypeBase::writeRelationTypes(
    ByteSwappedStream&	stream)		// Stream to write to
    {
    unsigned char relationTypeCount = allTypes.count();
    stream.write(&relationTypeCount, sizeof relationTypeCount);
    nextId = 0;
    for (SList::Iterator iterator(allTypes);  iterator;  ++iterator)
	{
	FileRelationTypeBase* entry =
	  (FileRelationTypeBase*)(allTypes[iterator]);
	entry->typeId = nextId++;
	stream.write(&entry->typeId, sizeof entry->typeId);
	short typeNameLength = strlen(entry->name());
	stream.write(&typeNameLength);
	stream.write(entry->name(), typeNameLength + 1);
	entry->write(stream);		// Write type-specific information
	}
    return stream;
    }

/*
 * Read a table of known relation types from a stream.  See
 * writeRelationTypes for the table format.
 *
 * Each relation name read from the stream is looked up in the
 * relation-type table.  If the named type already exists, its ID
 * number is changed to match the ID read in.  (Note that this
 * invalidates the type table for any other active streams.)  If the
 * named type does not exist, an error message is issued and the entry
 * is ignored.
 */
ByteSwappedStream& FileRelationTypeBase::readRelationTypes(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    unsigned char relationTypeCount;
    if (!stream.read(&relationTypeCount, sizeof relationTypeCount))
	return stream;
    unsigned char* usedIds = new unsigned char[relationTypeCount];
    FileRelationTypeBase** newRelations =
      new FileRelationTypeBase*[relationTypeCount];
    for (int relationNo = 0;  relationNo < relationTypeCount;  relationNo++)
	{
	unsigned char relationId;
	short typeNameLength;
	if (!stream.read(&relationId, sizeof relationId)
	  ||  !stream.read(&typeNameLength, sizeof typeNameLength))
	    correlatorPanic(
	      "correlator:  trouble reading file relation table");
	char typeName[MAXPATHLEN];
	if (typeNameLength < 0  ||  typeNameLength >= (int)sizeof typeName)
	    {
	    cerr << "correlator:  bad relation type name length "
	      << typeNameLength;
	    correlatorPanic(" when reading file relation table");
	    }
	if (!stream.read(typeName, typeNameLength + 1))
	    break;
	if ((int)strlen(typeName) != typeNameLength)
	    {
	    cerr << "correlator:  bad relation type name length for '"
	      << typeName << "': expected " << typeNameLength << " got "
	      << strlen(typeName);
	    correlatorPanic("");
	    }
	/*
	 * We've read the entry.  Process it.
	 */
	FileRelationTypeBase* relationType = findByName(typeName);
	if (relationType == NULL)
	    {
	    /*
	     * The relation doesn't exist as a built-in type.  See if
	     * we can create it as an investigator type.
	     */
	    cerr << "correlator:  creating new investigator type '" << typeName
	      << "' while reading relation type table\n";
	    relationType = InvestigatorTypeBase::find(typeName, true);
	    }
	if (relationType == NULL)
	    cerr << "correlator:  couldn't find or create relation type '"
	      << typeName << "' when reading file relation type table\n";
	else
	    {
	    usedIds[relationNo] = relationId;
	    newRelations[relationNo] = relationType;
	    relationType->typeId = relationId;
	    relationType->read(stream);	// Read type-specific information
	    }
	}
    /*
     * Look through all relations, finding those that previously
     * existed, and assign them a new ID that doesn't conflict with
     * any currently being read.
     */
    nextId = 0;
    for (SList::Iterator iterator(allTypes);  iterator;  ++iterator)
	{
	FileRelationTypeBase* entry =
	  (FileRelationTypeBase*)(allTypes[iterator]);
	nextId = 0;
	int relationNo;
	for (relationNo = 0;  relationNo < relationTypeCount;  relationNo++)
	    {
	    if (newRelations[relationNo] == entry)
		break;
	    }
	if (relationNo == relationTypeCount)
	    {
	    /*
	     * This relation isn't new.  Give it an unused type.
	     */
	    for (relationNo = 0;
	      relationNo < relationTypeCount;
	      relationNo++)
		{
		if (usedIds[relationNo] == nextId)
		    {
		    nextId++;
		    relationNo = 0;	// NOTE: restarts loop!
		    }
		}
	    entry->typeId = nextId++;
	    }
	}
    return stream;
    }

/*
 * Dummy routine to record a file miss.
 */
void FileRelationTypeBase::recordMissedReference(
    File&		,		// File that was referenced
    ReferenceType	,		// Type of the reference
    Process*		,		// Process doing the reference
    bool		)		// True if meaningless reference
    const
    {
    }

/*
 * Dummy routine to write type-specific information to a stream,
 * provided for relation types that have no type-specific information.
 */
ByteSwappedStream& FileRelationTypeBase::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    return stream;
    }

/*
 * Dummy routine to read type-specific information from a stream,
 * provided for relation types that have no type-specific information.
 */
ByteSwappedStream& FileRelationTypeBase::read(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    return stream;
    }

FileRelationBase::~FileRelationBase()
    {
    }

/*
 * Write a relation to a stream.  Usually, the derived class calls
 * this before writing class-specific information.
 */
ByteSwappedStream& FileRelationBase::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    unsigned char id = type().id();
    stream.write(&id, sizeof id);
    return stream;
    }

/*
 * Read a relation from a stream.  The first byte contains the
 * relation type ID.  Following this is the path of the target file.
 * These two items are common to all relations, so they are read and
 * then passed to the type-specific relation-reading code, which
 * constructs and reads the rest of the relation.
 *
 * Returns a pointer to the new relation, or NULL if the stream has no
 * more relations.  If NULL is returned, the stream may be closed, but
 * this is not guaranteed.
 */
FileRelationBase* FileRelationBase::read(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    char id;
    if (!stream.read(&id, sizeof id))
	return NULL;
    FileRelationTypeBase* type = FileRelationTypeBase::findById(id);
    if (type == NULL)
	{
	cerr << "correlator:  couldn't find file relation type " << (int)id
	  << '\n';
	stream.close();
	return NULL;
	}
    return type->readRelation(stream);
    }

FinalRelatedFileDestructor::FinalRelatedFileDestructor()
    {
    }

FinalRelatedFileDestructor::~FinalRelatedFileDestructor()
    {
#ifndef USE_TYPE_ALLOCATOR
    for (  ;  RelatedFile::freeCount > 0;  RelatedFile::freeCount--)
	{
	RelatedFile* goingFile = RelatedFile::freeRelatedFiles;
	RelatedFile::freeRelatedFiles =
	  *(RelatedFile**)RelatedFile::freeRelatedFiles;
	::operator delete(goingFile);
	}
#endif // USE_TYPE_ALLOCATOR
    }

RelatedFile::~RelatedFile()
    {
    /*
     * Destroy any existing relations
     */
    for (RelationIterator relation(*this);  relation;  ++relation)
	{
	delete &(*relation);
	relation.removeCurrent();
	}
    }


bool RelatedFile::hasType(		// See if file has a given rel. type
    const FileRelationTypeBase&
			type)		// Type desired
    const
    {
    for (RelationIterator relation((RelatedFile&)*this);
      relation;
      ++relation)
	{
	if (&(*relation).type() == &type)
	    return true;
	}
    return false;
    }

/*
 * Return a relation (to the target file) of a chosen type.  If no
 * such relation is found, returns NULL if create is false, or creates
 * one if create is true.
 */
FileRelationBase* RelatedFile::chosenRelation(
    const FileRelationTypeBase&
			type,		// Type desired
    bool		create)		// True to create if not found
    {
    for (RelationIterator relation(*this);  relation;  ++relation)
	{
	if (&(*relation).type() == &type)
	    return &(*relation);
	}
    if (!create)
	return NULL;
    FileRelationBase* newRelation = &type.newRelation();
    relations.pushHead(newRelation);
    newRelation->target = this;
    return newRelation;
    }

/*
 * Write all of a target file's relations to a stream.
 */
ByteSwappedStream& RelatedFile::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    unsigned short relationCount = relations.count();
    stream.write(&relationCount, sizeof relationCount);
    if (relationCount == 0)
	{
	cerr << "correlator:  unexpectedly writing empty relation\n";
	return stream;
	}
    target.writePath(stream, false);
    for (RelationIterator relation((RelatedFile&)*this);
      relation;
      ++relation)
	(*relation).write(stream);
    return stream;
    }

/*
 * Read a target file and a set of relations for it, adding the target
 * and relations to the source file.
 */
ByteSwappedStream& RelatedFile::read(
    ByteSwappedStream&	stream,		// Stream to read from
    File&		sourceFile)	// File that relations come FROM
    {
    short relationCount;
    if (!stream.read(&relationCount, sizeof relationCount))
	{
	cerr << "correlator:  unexpected read failure looking for"
	  << " relation count\n";
	return stream;
	}
    if (relationCount == 0)
	{
	cerr << "correlator:  unexpectedly read empty relation\n";
	return stream;
	}
    File* targetFile = File::findFromStream(stream, true);
    if (targetFile == NULL)
	return stream;
    RelatedFile* relatedFile = new RelatedFile(*targetFile);
    sourceFile.relatedFiles.pushHead(relatedFile);
    for (  ;  relationCount > 0;  relationCount--)
	{
	FileRelationBase* relation = FileRelationBase::read(stream);
	if (relation == NULL)
	    return stream;
	relatedFile->addRelation(relation);
	}
    return stream;
    }

ostream& RelatedFile::show(		// Show related file in printable form
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    if (relations.isEmpty())
	return stream;
    int relationCount = relations.count();
    stream << "RF\t" << target.encodedName(options) << ' ' << target.id()
      << ' ' << relationCount << '\n';
    for (RelationIterator relation((RelatedFile&)*this);
      relation;
      ++relation)
	(*relation).show(stream, options, managerName);
    return stream;
    }
