#ifndef lint
static const char Rcs_Id[] =
    "$Id: filestate.cc,v 1.2 1998/10/23 23:15:21 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filestate.cc,v $
 * Revision 1.2  1998/10/23 23:15:21  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.1  1998/02/10 19:29:35  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "deletion_mgr.hh"
#include "file.hh"
#include "filestate.hh"
#include "zzinlinefunctions.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
ostream&		operator<<(ostream& stream, const FileState& state);
					// Write a state in printable fashion
			FileState::FileState(State initialState);
			FileState::~FileState();
void			FileState::setState(File& containingFile,
			  State newState);
					// Set state field as desired
#endif // 0

extern void		correlatorPanic(const char* message);

ostream& operator<<(			// Write a state in printable fashion
    ostream&		stream,		// Stream to write to
    const FileState&	state)		// State to be written
    {
    switch (state.currentState)
	{
	case FileState::UNKNOWN:
	    {
	    stream << "UNKNOWN";
	    break;
	    }
	case FileState::KNOWN:
	    {
	    stream << "KNOWN";
	    break;
	    }
	case FileState::UNSTORED:
	    {
	    stream << "UNSTORED";
	    break;
	    }
	case FileState::MISSING:
	    {
	    stream << "MISSING";
	    break;
	    }
	case FileState::FORGETTING:
	    {
	    stream << "FORGETTING";
	    break;
	    }
	case FileState::FORGOTTEN:
	    {
	    stream << "FORGOTTEN";
	    break;
	    }
	default:
	    {
	    correlatorPanic("Unknown file state in FileState::operator<<");
	    break;
	    }
	}
    return stream;
    }

FileState::FileState(
    State		initialState)	// Initial state to begin with
    : currentState(initialState)
    {
    DeletionManager::recordInitialState(initialState);
    }

FileState::~FileState()
    {
    DeletionManager::recordStateRemoval(currentState);
    }

void FileState::setState(		// Give a new value to state field
    File&		containingFile,	// File that state belongs to
    State		newState)	// New state to assign
					// Set state field as desired
    {
    if (currentState != newState)
	{
	containingFile.stateChanged(this, currentState, newState);
	currentState = newState;
	}
    }
