#ifndef _FILESTATE_HH
#define _FILESTATE_HH
/*
 * $Id: filestate.hh,v 1.3 1998/10/23 23:15:22 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1998 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filestate.hh,v $
 * Revision 1.3  1998/10/23 23:15:22  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.2  1998/07/15 21:43:45  geoff
 * Include iostream.hh; it should have always been included.
 *
 * Revision 1.1  1998/02/10 19:29:35  geoff
 * Initial revision
 *
 */

#include <iostream.h>

class File;

/*
 * Class used to keep track of file storage states.  This is needed so
 * that we can ensure control over the setting of the file's state
 * field.
 */
class FileState
    {
	friend ostream&	operator<<(ostream& stream, const FileState& state);
					// Write a state in printable fashion
    public:
	/*
	 * Possible file states
	 */
	enum State
	    {
	    UNKNOWN,			// File state is unknown
	    KNOWN,			// We know all about the file
	    UNSTORED,			// The file exists, but not locally
	    MISSING,			// File no longer exists
	    FORGETTING,			// We want to forget the file
	    FORGOTTEN			// We've forgotten the file, can delete
	    };
			FileState(State initialState);
			~FileState();
	inline State	state() const;	// Return current state field
	void		setState(File& containingFile,
			  State newState); // Set state field as desired
    private:
			FileState(const FileState& that); // Disable copy
	FileState&	operator=(const FileState& that); // Disable assignment
	State		currentState;	// Current state field
    };

#define _FILESTATE_ICC_NEEDED

#endif // _FILESTATE_HH
