#ifndef lint
static const char Rcs_Id[] =
    "$Id: globalmanager.cc,v 1.23 1998/10/23 23:15:24 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: globalmanager.cc,v $
 * Revision 1.23  1998/10/23 23:15:24  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.22  1998/07/15 21:43:46  geoff
 * Get rid of baseReference; it's now manager-specific.
 *
 * Revision 1.21  1998/05/14 01:15:33  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.20  1997/07/11 17:40:57  geoff
 * Include file.hh (don't ask my why...)
 *
 * Revision 1.19  1997/02/12 02:52:07  geoff
 * Make oftenPercent be a double.
 *
 * Revision 1.18  1997/01/02 06:48:22  geoff
 * Add resetStatistics and support for the base reference.
 *
 * Revision 1.17  1996/12/03 02:58:06  geoff
 * Add support for reading and writing parameters.
 *
 * Revision 1.16  1996/10/11 22:09:05  geoff
 * Get rid of the workaround for the floating-output bug.
 *
 * Revision 1.15  1996/10/08 19:50:05  geoff
 * Add code to work around the floating-point output bug.
 *
 * Revision 1.14  1996/06/11 06:13:19  geoff
 * Add the peek threshold, age weight factor, and aging threshold.
 *
 * Revision 1.13  1996/05/03  01:12:22  geoff
 * Update the copyright.
 *
 * Revision 1.12  1996/03/19  18:42:59  geoff
 * Get rid of the show/read/write/clear-manager-data routines now that
 * they can be inherited.  Inherit from EmptyManagerBase.
 *
 * Revision 1.11  1996/02/02  18:29:03  geoff
 * Derive resetStatistics from the base class instead of redefining it.
 * Change the default value of keepReferences to 20.  Move some virtual
 * functions into the C code.
 *
 * Revision 1.10  1995/12/19  00:34:34  geoff
 * Add GlobalManager::processTree.
 *
 * Revision 1.9  1995/11/22  00:21:17  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.8  1995/11/10  02:59:16  geoff
 * Add zzinlinefunctions.hh.  Get rid of a spurious "const".  Add
 * NEW_FILE_RELATIONS support (keepReferences).
 *
 * Revision 1.7  1995/09/15  00:05:40  geoff
 * Change the RCS ID to be a const.  Support the new (virtual) internal
 * timers.  Add statistics support.  Globally change "cache" to "hoard".
 *
 * Revision 1.6  1995/08/07  20:03:00  geoff
 * Make the cache be 50 Mb, and the management interval be one hour.
 *
 * Revision 1.5  1995/06/17  00:11:34  geoff
 * Add show options.  Add clearManagerData.  Make the default management
 * interval be 30 minutes.
 *
 * Revision 1.4  1995/05/23  00:39:57  geoff
 * Add new routines required by PRIVATE_MANAGEMENT_DATA.
 *
 * Revision 1.3  1995/05/05  22:41:53  geoff
 * Add parameter descriptions to support the new setparams code.
 *
 * Revision 1.2  1995/04/20  21:58:04  geoff
 * Add support for dumping management results to a stream, and make the
 * maximum cache size be a global parameter.
 *
 * Revision 1.1  1995/03/29  21:06:45  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "file.hh"
#include "globalmanager.hh"
#include "internaltimer.hh"
#include "zzinlinefunctions.hh"
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <strstream.h>
#include <sys/param.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			GlobalManager::GlobalManager(const char* managerName,
			  const char* managerDefaults);
			GlobalManager::~GlobalManager();
void			GlobalManager::setParameter(const char* description);
					// Change management parameters
void			GlobalManager::showParameters(ostream& stream) const;
					// Display management parameters
long			GlobalManager::writeParameters(
			    ByteSwappedStream& stream, bool reallyWrite)
			  const;
					// Write management params to stream
long			GlobalManager::readParameters(
			  ByteSwappedStream& stream, long parameterBytes);
					// Read management params from stream
ostream&		GlobalManager::showStatistics(ostream& stream) const;
					// Display collected quality statistics
void			GlobalManager::resetStatistics();
					// Reset quality statistics
bool			GlobalManager::shouldManage(const InternalTimer& now,
			  InternalTimer& nextManagementTime);
void			GlobalManager::manage(const InternalTimer& now,
			    ByteSwappedStream* dumpStream,
			    bool encodeFilenames);
					// Perform dummy management
#endif // 0

GlobalManager		globalManager(
			  "Global",
			  "Global::oftenPercent=1.0\n"
			    "Global::peekThreshold=40\n"
			    "Global::maximumHoardBlocks=102400\n"
			    "Global::keepReferences=20\n"
			    "Global::processTree=true\n"
			    "Global::ageWeightFactor=1.2\n"
			    "Global::agingThreshold=25000\n"
			  );

double			GlobalManager::ageWeightFactor = 0;
unsigned long		GlobalManager::agingThreshold = 0;
double			GlobalManager::oftenPercent = 0;
int			GlobalManager::peekThreshold = 0;
long			GlobalManager::maximumHoardBlocks = 0;
int			GlobalManager::keepReferences = 0;
bool			GlobalManager::processTree = true;

GlobalManager::GlobalManager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
	: EmptyManagerBase(managerName, managerDefaults)
    {
    EmptyManagerBase::setManagerParameters(defaults);
    }

GlobalManager::~GlobalManager ()
    {
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void GlobalManager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad GlobalManager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("ageWeightFactor"))
	ageWeightFactor = atof(valueStart);
    else if (parameterIs("agingThreshold"))
	agingThreshold = atol(valueStart);
    else if (parameterIs("oftenPercent"))
	oftenPercent = atof(valueStart);
    else if (parameterIs("peekThreshold"))
	peekThreshold = atoi(valueStart);
    else if (parameterIs("maximumHoardBlocks"))
	maximumHoardBlocks = atol(valueStart);
    else if (parameterIs("keepReferences"))
	keepReferences = atol(valueStart);
    else if (parameterIs("processTree"))
	processTree = valueStart[0] == 't';
    else
	// NEEDSWORK: should throw exception
	cerr << "correlator: unrecognized GlobalManager parameter '"
	  << description << "'\n";
#undef parameterIs
    }

/*
 * Display management parameters
 */
void GlobalManager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "Global::ageWeightFactor=" << ageWeightFactor
      << "/0.0/infinity/1/Weighting Factor for Aging References\n";
    stream << "Global::agingThreshold=" << agingThreshold
      << "/0/infinity/1/Threshold for Aging References\n";
    stream << "Global::oftenPercent=" << oftenPercent
      << "/0/100/1.0/Often-Referenced File Threshold\n";
    stream << "Global::peekThreshold=" << peekThreshold
      << "/0/100/1/Threshold Above Which Peeks Are Meaningless\n";
    stream << "Global::maximumHoardBlocks=" << maximumHoardBlocks
      << "/0/infinity/2048/Hoard Size (Mb)\n";
    stream << "Global::keepReferences="
      << keepReferences
      << "/1/infinity/1/No. of Related Files to Keep\n";
    stream << "Global::processTree="
      << (processTree ? "true" : "false")
      << "/false/true/1/Use Process Trees in Calculating Semantic Distance\n";
    }

/*
 * Write manager-specific parameters to a stream.  Returns the number
 * of bytes written, or zero if trouble occurs.  If reallyWrite is
 * false, the byte count is returned, but no writing is actually done.
 */
long GlobalManager::writeParameters(
    ByteSwappedStream&	stream,		// Stream to write to
    bool		reallyWrite)	// True to actually do the writing
    const
    {
    long bytesWritten = ManagerBase::writeParameters(stream, reallyWrite);
    if (reallyWrite  &&  !stream.write(&ageWeightFactor))
	return 0;
    bytesWritten += sizeof ageWeightFactor;
    if (reallyWrite  &&  !stream.write(&agingThreshold))
	return 0;
    bytesWritten += sizeof agingThreshold;
    if (reallyWrite  &&  !stream.write(&oftenPercent))
	return 0;
    bytesWritten += sizeof oftenPercent;
    if (reallyWrite  &&  !stream.write(&peekThreshold))
	return 0;
    bytesWritten += sizeof peekThreshold;
    if (reallyWrite  &&  !stream.write(&maximumHoardBlocks))
	return 0;
    bytesWritten += sizeof maximumHoardBlocks;
    if (reallyWrite  &&  !stream.write(&keepReferences))
	return 0;
    bytesWritten += sizeof keepReferences;
    if (reallyWrite  &&  !stream.write(&processTree))
	return 0;
    bytesWritten += sizeof processTree;
    return bytesWritten;
    }

/*
 * Read manager-specific parameters from a stream.  The second
 * argument gives the number of bytes available for consumption.
 * Returns the number of bytes remaining.
 */
long GlobalManager::readParameters(
    ByteSwappedStream&	stream,		// Stream to read from
    long		parameterBytes)	// Number of bytes available to read
    {
    parameterBytes = ManagerBase::readParameters(stream, parameterBytes);
    parameterBytes -= sizeof ageWeightFactor;
    if (parameterBytes >= 0)
	stream.read(&ageWeightFactor);
    parameterBytes -= sizeof agingThreshold;
    if (parameterBytes >= 0)
	stream.read(&agingThreshold);
    if (logVersion > 3)
	{
	parameterBytes -= sizeof oftenPercent;
	if (parameterBytes >= 0)
	    stream.read(&oftenPercent);
	}
    else
	{
	int intOftenPercent;
	parameterBytes -= sizeof intOftenPercent;
	if (parameterBytes >= 0)
	    {
	    stream.read(&intOftenPercent);
	    oftenPercent = intOftenPercent;
	    }
	}
    parameterBytes -= sizeof peekThreshold;
    if (parameterBytes >= 0)
	stream.read(&peekThreshold);
    parameterBytes -= sizeof maximumHoardBlocks;
    if (parameterBytes >= 0)
	stream.read(&maximumHoardBlocks);
    parameterBytes -= sizeof keepReferences;
    if (parameterBytes >= 0)
	stream.read(&keepReferences);
    parameterBytes -= sizeof processTree;
    if (parameterBytes >= 0)
	stream.read(&processTree);
    return parameterBytes;
    }

/*
 * Display collected quality statistics
 */
ostream& GlobalManager::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    return stream;
    }

void GlobalManager::resetStatistics() // Reset quality statistics
    {
    }

/*
 * Decide if this manager's management routine should run.  Regardless
 * of the return value, modifies nextManagementTime to contain the
 * next time management will be needed for this manager.
 */
bool GlobalManager::shouldManage(
    const InternalTimer&	,	// Current time-of-day
    InternalTimer&		nextManagementTime)
					// When to run next
    {
    nextManagementTime.makeInvalid();
    return false;			// This manager never needs to run
    }

void GlobalManager::manage(		// Dummy management routine
    const InternalTimer&	,	// Current time-of-day
    ByteSwappedStream*		,	// Stream to dump results to
    bool			)	// Whether to encode file names
    {
    }
