#ifndef lint
static const char Rcs_Id[] =
    "$Id: growmalloc.cc,v 1.4 1996/05/03 01:12:26 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: growmalloc.cc,v $
 * Revision 1.4  1996/05/03 01:12:26  geoff
 * Update the copyright.
 *
 * Revision 1.3  1995/11/10  02:59:21  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.2  1995/09/15  00:05:45  geoff
 * Change the RCS ID to be a const.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "zzinlinefunctions.hh"
#include <stdlib.h>
#include <unistd.h>

void *			growmalloc(void * curblock, unsigned int * curelems,
			  unsigned int nelems, unsigned int elemsize,
			  unsigned int expincr);

#define DOUBLESIZE	128
#define GROWPERCENT(x)	((x) + (x) / 10)
#define MINSIZE		16

/*
 * Handle automatically-expanding allocation of a block of memory.  On
 * entry, curblock is either null or a pointer to a
 * previously-allocated memory block.  *Curelems contains the number
 * of elements of size elemsize that are currently stored in curblock.
 * (If *curelems is zero, curblock is assumed null, and conversely, if
 * curblock is null, *curelems is assumed to be zero.)  The routine
 * returns a pointer to an area of memory large enough to hold at
 * least nelems elements of size elemsize; it may be larger than
 * necessary at the discretion of the routine.  The new size of the
 * block is stored into *curelems.  Any information previously stored
 * in the block is preserved, but any newly-allocated memory is not
 * cleared.  If it is necessary to expand the block, and expincr is
 * nonzero, the block is expanded by expincr elements.  Otherwise, the
 * block is expanded by a value dependent on the current size, as
 * follows:
 *
 * 1.	Regardless of any following rules, the block is expanded by at
 *	least (nelems - *curelems) elements.
 * 2.	If the current size is zero, the growmalloc attempts to allocate
 *	enough elements to fill MINSIZE bytes.
 * 3.	If the current size is less than DOUBLESIZE bytes, the new
 *	size is twice the oldsize.
 * 4.	Otherwise, the new size is GROWPERCENT larger than the old,
 *	rounded up.
 *
 * As is usual for such routines, NULL is returned on allocation
 * failures.  However, NULL is also returned if there is no current
 * block and growmalloc is called with nelems == 0, so don't write
 * code that will fail on that case.
 */
void* growmalloc(
    void*		curblock,	// Currently-allocated block, if any
    unsigned int*	curelems,	// Number of elements in current blk
    unsigned int	nelems,		// Number of elements needed
    unsigned int	elemsize,	// Size of a single element
    unsigned int	expincr)	// Amount to expand by
    {
    unsigned int	size;		// Size now or to be allocated

    size = *curelems;
    if (size == 0)
	curblock = NULL;
    else if (curblock == NULL)
	*curelems = size = 0;
    if (size >= nelems)
	return curblock;
    /*
     * We must allocate or reallocate.  Pick a size.
     */
    if (expincr != 0)
	size = (size + expincr) * elemsize;
    else if (size == 0)
	size = MINSIZE - MINSIZE % elemsize;
    else
	{
	size *= elemsize;
	if (size < DOUBLESIZE)
	    size *= 2;
	else
	    {
	    size = GROWPERCENT(size) + elemsize - 1;
	    size -= size % elemsize;
	    }
	}
    nelems *= elemsize;
    if (size < nelems)
	size = nelems;
    if (curblock == NULL)
	curblock = malloc(size);
    else
	curblock = realloc(curblock, size);
    if (curblock == NULL)
	*curelems = 0;
    else
	*curelems = size / elemsize;
    return curblock;
    }
