#ifndef lint
static const char Rcs_Id[] =
    "$Id: internaltimer.cc,v 1.7 1998/10/23 23:15:28 geoff Exp $";
#endif

/*
 * $Id: internaltimer.cc,v 1.7 1998/10/23 23:15:28 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: internaltimer.cc,v $
 * Revision 1.7  1998/10/23 23:15:28  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.6  1997/12/04 23:16:02  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.5  1996/09/11 22:59:59  geoff
 * Add TIMEVALS_ARE_INTS support.
 *
 * Revision 1.4  1996/05/03  01:12:27  geoff
 * Update the copyright.
 *
 * Revision 1.3  1996/02/02  20:09:47  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.2  1995/11/10  02:59:22  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.1  1995/09/14  23:32:25  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cdecls.hh"
#include "internaltimer.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <stdio.h>
#include <string.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			InternalTimer::InternalTimer(TimerType desiredType);
const InternalTimer&	InternalTimer::operator=(const InternalTimer& source);
const char*		InternalTimer::ctime() const;
					// Return timer as string
const char*		InternalTimer::cmicrotime() const;
					// Return timer as string
const char*		InternalTimer::numericTime(Conversion conversion)
			  const;
					// Return timer as numeric string
const char*		InternalTimer::numericMicrotime(Conversion conversion)
			  const;
					// Return timer w/ microseconds
void			InternalTimer::updateTime();
					// Update timer's recorded time
void			InternalTimer::timerMode(TimerType desiredMode);
					// Set global mode to real/virtual
#endif // 0

InternalTimer::TimerType InternalTimer::mode = InternalTimer::VIRTUAL;
					// Current mode for virtual timers
struct timeval		InternalTimer::simulatedTime = {0, 0};
					// Current simulated time

extern void		correlatorPanic(const char* message);

InternalTimer::InternalTimer(
    TimerType		desiredType)	// Type of timer to create
    : type(desiredType)
    {
    if (type == VIRTUAL  &&  mode == VIRTUAL)
	currentTime = simulatedTime;
    else
	(void)gettimeofday(&currentTime, (struct timezone *)NULL);
    markChanged();
    }

const InternalTimer& InternalTimer::operator=(
    const InternalTimer& source)	// Timer to assign from
    {
    if (this == &source)
	return *this;			// Assignment to self!
    type = source.type;
    currentTime = source.currentTime;
    markChanged();
    if (source.seconds[0] != '\0')
	(void)strcpy(seconds, source.seconds);
    if (source.microseconds[0] != '\0')
	(void)strcpy(microseconds, source.microseconds);
    if (source.numericMicroseconds[0] != '\0')
	(void)strcpy(numericMicroseconds, source.numericMicroseconds);
    return *this;
    }

const char* InternalTimer::ctime()	// Return timer as string
    const
    {
    if (seconds[0] == '\0')
	{
#ifdef TIMEVALS_ARE_INTS
	(void)strcpy(seconds, ::ctime((long*)&currentTime.tv_sec));
#else // TIMEVALS_ARE_INTS
	(void)strcpy(seconds, ::ctime(&currentTime.tv_sec));
#endif // TIMEVALS_ARE_INTS
	seconds[sizeof "Wed Jun 30 21:49:08 1993" - 1] = '\0';
	}
    return seconds;
    }

const char* InternalTimer::cmicrotime()	// Return timer as string
    const
    {
    if (microseconds[0] == '\0')
	{
	(void)ctime();			// Fill in seconds part
	(void)strncpy(microseconds, seconds, sizeof "Wed Jun 30 21:49:08" - 1);
	(void)sprintf(&microseconds[sizeof "Wed Jun 30 21:49:08" - 1],
	  ".%6.6lu", (unsigned long)currentTime.tv_usec);
	(void)strcat(microseconds, &seconds[sizeof "Wed Jun 30 21:49:08" - 1]);
	}
    return microseconds;
    }

/*
 * Return the seconds part of the timer as a numeric string.
 */
const char* InternalTimer::numericTime(
    Conversion		conversion)	// Chosen method of converting
    const
    {
    switch (conversion)
	{
	case S:
	case MS:
	case HMS:
	case DHMS:
	    break;			// No decision needs to be made here
	case S_MS:
	    {
	    if (currentTime.tv_sec >= 60)
		conversion = MS;
	    else
		conversion = S;
	    break;
	    }
	case MS_HMS:
	    {
	    if (currentTime.tv_sec >= 60 * 60)
		conversion = HMS;
	    else
		conversion = MS;
	    break;
	    }
	case S_MS_HMS:
	    {
	    if (currentTime.tv_sec >= 60 * 60)
		conversion = HMS;
	    else if (currentTime.tv_sec >= 60)
		conversion = MS;
	    else
		conversion = S;
	    break;
	    }
	case HMS_DHMS:
	    {
	    if (currentTime.tv_sec >= 24 * 60 * 60)
		conversion = DHMS;
	    else
		conversion = HMS;
	    break;
	    }
	case MS_HMS_DHMS:
	    {
	    if (currentTime.tv_sec >= 24 * 60 * 60)
		conversion = DHMS;
	    else if (currentTime.tv_sec >= 60 * 60)
		conversion = HMS;
	    else
		conversion = MS;
	    break;
	    }
	case S_MS_HMS_DHMS:
	    {
	    if (currentTime.tv_sec >= 24 * 60 * 60)
		conversion = DHMS;
	    else if (currentTime.tv_sec >= 60 * 60)
		conversion = HMS;
	    else if (currentTime.tv_sec >= 60)
		conversion = MS;
	    else
		conversion = S;
	    break;
	    }
	default:
	    cerr << "correlator:  illegal conversion type " << conversion;
	    correlatorPanic(" in InternalTimer::numericTime");
	    break;
	}
    switch (conversion)
	{
	case S:
	    {
	    (void)sprintf(numericMicroseconds, "%lu",
	      (unsigned long)currentTime.tv_sec);
	    break;
	    }
	case MS:
	    {
	    (void)sprintf(numericMicroseconds, "%lu:%2.2lu",
	      (unsigned long)currentTime.tv_sec / 60,
	      (unsigned long) currentTime.tv_sec % 60);
	    break;
	    }
	case HMS:
	    {
	    (void)sprintf(numericMicroseconds, "%lu:%2.2lu:%2.2lu",
	      (unsigned long)currentTime.tv_sec / (60 * 60),
	      ((unsigned long)currentTime.tv_sec / 60) % 60,
	      (unsigned long) currentTime.tv_sec % 60);
	    break;
	    }
	case DHMS:
	    {
	    (void)sprintf(numericMicroseconds, "%lud %lu:%2.2lu:%2.2lu",
	      (unsigned long)currentTime.tv_sec / (60 * 60 * 24),
	      ((unsigned long)currentTime.tv_sec / (60 * 60)) % 24,
	      ((unsigned long)currentTime.tv_sec / 60) % 60,
	      (unsigned long) currentTime.tv_sec % 60);
	    break;
	    }
	default:
	    cerr << "correlator:  unexpected conversion type " << conversion
	      << " in InternalTimer::numericTime\n ";
	    break;
	}
    return numericMicroseconds;
    }

/*
 * Return the seconds and microseconds part of the timer as a numeric
 * string, with a decimal point separating the two parts.
 */
const char* InternalTimer::numericMicrotime(
    Conversion		conversion)	// Chosen method of converting
    const
    {
    (void)numericTime(conversion);
    (void)sprintf(&numericMicroseconds[strlen(numericMicroseconds)], ".%6.6lu",
      (unsigned long)currentTime.tv_usec);
    return numericMicroseconds;
    }

void InternalTimer::updateTime()	// Update timer's recorded time
    {
    if (type == VIRTUAL  &&  mode == VIRTUAL)
	currentTime = simulatedTime;
    else
	(void)gettimeofday(&currentTime, (struct timezone*)NULL);
    markChanged();
    }

void InternalTimer::timerMode(		// Set global mode to real/virtual
    TimerType		desiredMode)	// Mode to run in
    {
    if (mode == ACTUAL  &&  desiredMode == VIRTUAL)
	correlatorPanic(
	  "correlator: attempt to return from actual to virtual time");
    mode = desiredMode;
    }
