#ifndef lint
static const char Rcs_Id[] =
    "$Id: investigator.cc,v 1.21 1998/10/23 23:15:32 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This file contains classes used to support external investigators.
 */

/*
 * $Log: investigator.cc,v $
 * Revision 1.21  1998/10/23 23:15:32  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.20  1998/08/13 18:38:42  geoff
 * Integrate changes from the bug-fix branch:
 *     Add the meaningless argument to the reference-recording routines.
 *
 *     Many changes to make the last-reference time and similar things be
 *     manager-specific.
 *
 * Revision 1.19  1998/06/30 20:20:45  geoff
 * Preliminary changes to support separate tracking of overlapped and
 * nonoverlapped cluster members.  This code is only partially debugged
 * at this point.
 *
 * Revision 1.18  1998/05/14 01:15:34  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.17  1998/05/02 00:29:08  geoff
 * Put the element iterator in the cluster namespace, instead of using a
 * naming convention.
 *
 * Revision 1.16  1998/02/19 02:10:10  geoff
 * When relations change, inform the owning file.
 *
 * Revision 1.15  1997/12/04 23:16:03  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.14  1997/01/02 06:48:23  geoff
 * Don't require investigatedClusters to be empty on destruction; this
 * cures a core dump at the expense of leaving dmalloc with a bit of
 * dirty memory.
 *
 * Revision 1.13  1996/09/11 23:00:01  geoff
 * Convert from dlists to slists.  Make minor changes in the interface to
 * the controller, to allow more flexible sending of information.
 *
 * Revision 1.12  1996/06/11  06:13:22  geoff
 * Major changes to support cluster investigators.  Add learnFromStream
 * routines, all necessary cluster-investigator stuff, new virtual
 * routines required by cluster-investigator support.
 *
 * Revision 1.11  1996/05/03  01:12:31  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/02/02  20:09:47  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.9  1996/02/02  18:29:07  geoff
 * Move some virtual functions into the C code.  Fix some minor problems
 * detected by the new C++ compiler.
 *
 * Revision 1.8  1995/12/19  01:02:57  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.7  1995/12/19  00:34:39  geoff
 * Add process-tracking support.
 *
 * Revision 1.6  1995/11/22  00:21:29  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.5  1995/11/21  21:38:56  geoff
 * Rewrite the investigator support to make use of NEW_FILE_RELATIONS.
 *
 * Revision 1.4  1995/11/10  02:59:41  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.3  1995/10/10  20:38:02  geoff
 * Get rid of the obsolete TRACK_MEMORY_LEAKS stuff.
 *
 * Revision 1.2  1995/09/15  00:05:47  geoff
 * Change the RCS ID to be a const.  Add abortive TRACK_MEMORY_LEAKS
 * support.  Add proper bracketing to all switch statements.
 *
 * Revision 1.1  1995/08/07  19:14:49  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "file.hh"
#include "filerelation.hh"
#include "hashtable.hh"
#include "investigator.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <string.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			InvestigatorTypeBase::~InvestigatorTypeBase();
void			InvestigatorTypeBase::setParameter(
			  const char* description);
					// Set a changeable parameter
InvestigatorTypeBase*	InvestigatorTypeBase::find(const char* nameSought,
			  bool create);
					// Find a named investigator
ByteSwappedStream&	InvestigatorTypeBase::write(ByteSwappedStream& stream)
			  const;
					// Write type-specific info to stream
ByteSwappedStream&	InvestigatorTypeBase::read(ByteSwappedStream& stream);
					// Read type-specific info from stream
void			InvestigatorTypeBase::recordReference(File& file,
			    ReferenceType type,
			    Process* process, bool isMeaningless)
			  const;
					// Record a reference 
void			InvestigatorTypeBase::cleanGlobalData() const;
					// Clean out type's global data
			ClusterInvestigatorType::~ClusterInvestigatorType();
InvestigatorTypeBase*	ClusterInvestigatorType::newInvestigatorType(
			  const char* name);
					// Create a new investigator type
FileRelationBase&	ClusterInvestigatorType::newRelation()
			  const;
					// Create new relation instance
FileRelationBase*	ClusterInvestigatorType::readRelation(
			    ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
bool			ClusterInvestigatorType::learnFromStream(
			  ByteSwappedStream& controlStream);
					// Add investigated data provided
					// ..externally
ByteSwappedStream&	ClusterInvestigatorType::write(
			    ByteSwappedStream& stream)
			  const;
					// Write type-specific info to stream
ByteSwappedStream&	ClusterInvestigatorType::read(
			  ByteSwappedStream& stream);
					// Read type-specific info from stream
void			ClusterInvestigatorType::cleanGlobalData() const;
					// Clean out type's global data
			InvestigatedCluster::~InvestigatedCluster();
void			InvestigatedCluster::computeWeight();
					// Calculate hoarding weight
const char*		InvestigatedCluster::typeName() const;
					// Return cluster's type name
const char*		InvestigatedCluster::encodedTypeName(
			  ShowOptions options) const;
					// Return cluster's type name
const char*		InvestigatedCluster::name() const;
					// Return cluster's name
const char*		InvestigatedCluster::encodedName(ShowOptions options)
			  const;
					// Return cluster's name
			RelationInvestigatorType::~RelationInvestigatorType();
InvestigatorTypeBase*	RelationInvestigatorType::newInvestigatorType(
			  const char* name);
					// Create a new investigator type
FileRelationBase&	RelationInvestigatorType::newRelation()
			  const;
					// Create new relation instance
FileRelationBase*	RelationInvestigatorType::readRelation(
			    ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
bool			RelationInvestigatorType::learnFromStream(
			  ByteSwappedStream& controlStream);
					// Add investigated data provided
					// ..externally
			RelationInvestigation::RelationInvestigation(
			  const FileRelationTypeBase& chosenType);
			RelationInvestigation::~RelationInvestigation();
const FileRelationTypeBase&
			RelationInvestigation::type() const;
					// Return type of the relation
ostream&		operator<<(ostream& stream,
			  const RelationInvestigation& relation);
					// Show relation in printable form
					// ..appends \n
ostream&		RelationInvestigation::show(ostream& stream,
			    ShowOptions options,
			    const char* managerName)
			  const;	// Show relation in printable form,
					// ..with options.  Appends \n
ByteSwappedStream&	RelationInvestigation::write(ByteSwappedStream& stream)
			  const;	// Write a relation entry to a stream
#endif // 0

/*
 * Static class variables
 */
const char*		ClusterInvestigatorType::typeClass = "Cluster";
					// Master type class name for clusters
SList			ClusterInvestigatorType::investigatedClusters(false);
					// Clusters provided by investigators
const char*		RelationInvestigatorType::typeClass = "Relation";
					// Master type class name for relations

extern void		correlatorPanic(const char* message);
extern "C" char*	mimeEncode(const char* source,
			  const char* whatToEncode);
					// MIME-encode a string
extern void		nomem(const char* where, int line, unsigned int size);
extern char*		strdup(const char* string, int size);

InvestigatorTypeBase::~InvestigatorTypeBase()
    {
    }

/*
 * Set a single investigator parameter.  The description string is of
 * the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void InvestigatorTypeBase::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad Investigator parameter format"
	    " in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("weight"))
	overallWeight = atof(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized Investigator parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Find a named investigator, optionally creating it if it doesn't exist.
 *
 * Investigator names are of the form:
 *
 *	<type>:<name>
 *
 * where "type" is the type of the investigator, selected from a short
 * list of valid types, and "name" is an arbitrary name to be given to
 * the investigator.
 *
 * Current valid types are: Cluster, Relation
 *
 * The investigator table is searched for the given investigator.  If
 * none is found, and "create" is false, the routine returns NULL.
 * Otherwise, it creates a new investigator of the given type and
 * returns that.
 *
 * If an invalid type is specified, the routine issues an error
 * message and returns a NULL pointer, even if "create" is true.
 *
 * Return values:
 *
 *	NULL	if an error occurs, or if the investigator is not
 *		found and "create" is false
 */
InvestigatorTypeBase* InvestigatorTypeBase::find(
    const char*		nameSought,	// Name of investigator to locate
    bool		create)		// True to create if nonexistent
    {
    static struct			// Table of legal investigator types
	{
	const char*	typeName;	// Name of the type
	InvestigatorTypeMaker* maker;	// Routine to make the type
	}
			typeTable[] =
	{
	{"Cluster:", ClusterInvestigatorType::newInvestigatorType},
	{"Relation:", RelationInvestigatorType::newInvestigatorType},
	{NULL, NULL}
	};
    for (int typeNum = 0;  typeTable[typeNum].typeName != NULL;  typeNum++)
	{
	int typeLength = strlen(typeTable[typeNum].typeName);
	if (strncmp(nameSought, typeTable[typeNum].typeName, typeLength) == 0)
	    {
	    InvestigatorTypeBase* investigator =
	      (InvestigatorTypeBase*)
		FileRelationTypeBase::findByName(nameSought);
	    if (investigator == NULL  &&  create)
		investigator = (*typeTable[typeNum].maker)(nameSought);
	    return investigator;
	    }
	}
    cerr << "correlator: bad type in investigator name '" << nameSought
      << "'\n";
    return NULL;
    }

/*
 * Write type-specific information to a stream.
 */
ByteSwappedStream& InvestigatorTypeBase::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    stream.write(&overallWeight, sizeof overallWeight);
    return stream;
    }

/*
 * Read type-specific information from a stream.
 */
ByteSwappedStream& InvestigatorTypeBase::read(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    stream.read(&overallWeight, sizeof overallWeight);
    return stream;
    }

/*
 * Record a reference to a given file.  For investigated relations,
 * this is normally a no-op.
 */
void InvestigatorTypeBase::recordReference(
    File&		,		// File that was referenced
    ReferenceType	,		// Type of the reference
    Process*		,		// Process doing the reference
    bool		)		// True if reference is meaningless
    const
    {
    }

/*
 * Clean global data associated with a relation.  For investigated
 * relations, this is normally a no-op.
 */
void InvestigatorTypeBase::cleanGlobalData()
    const
    {
    }

ClusterInvestigatorType::~ClusterInvestigatorType()
    {
    for (InvestigatedClusterIterator iterator;  iterator;  ++iterator)
	{
	InvestigatedCluster* cluster = &(*iterator);
	if (&cluster->type() == this)
	    {
	    iterator.removeCurrent();
	    delete cluster;
	    }
	}
    }

/*
 * Create a new cluster investigator type
 */
InvestigatorTypeBase* ClusterInvestigatorType::newInvestigatorType(
    const char*		name)		// Name to give to the new type
    {
    return new ClusterInvestigatorType(name);
    }

/*
 * Create a new cluster relation instance
 */
FileRelationBase& ClusterInvestigatorType::newRelation()
					// File that's target of the relation
    const
    {
    correlatorPanic("ClusterInvestigatorType::newRelation called");
    return *(FileRelationBase*)NULL;
    }

/*
 * Read a new cluster relation from a stream
 */
FileRelationBase* ClusterInvestigatorType::readRelation(
    ByteSwappedStream&	)		// Stream containing the relation
    const
    {
    correlatorPanic("ClusterInvestigatorType::readRelation called");
    return NULL;
    }

/*
 * Add investigated data provided on a stream from an external source.
 *
 * We allow only one cluster of each type.  There is only one reason
 * for this: it allows us to delete and replace clusters.  If there
 * were a way to identify a cluster (other than as an investigated
 * type), we could allow multiple clusters of a given type.  This
 * would probably be the preferable implementation.
 *
 * Returns true if all went well, false if trouble arose with the
 * input stream.
 */
bool ClusterInvestigatorType::learnFromStream(
    ByteSwappedStream&	stream)		// Stream to read information from
    {
    int clusterNameLength;
    if (!stream.read(&clusterNameLength, sizeof clusterNameLength))
	return false;
    if (clusterNameLength == 0)
	return true;
    else if (clusterNameLength < 0  ||  clusterNameLength > MAXPATHLEN)
	{
	cerr << "correlator:  bad cluster name length "
	  << clusterNameLength << '\n';
	return false;
	}
    char* clusterName = new char[clusterNameLength + 1];
    if (!stream.read(clusterName, clusterNameLength + 1))
	{
	cerr << "correlator:  trouble getting cluster name"
	    " from controller at offset "
	  << stream.bytesRead() << '\n';
	delete[] clusterName;
	return false;
	}
    else if ((int)strlen(clusterName) != clusterNameLength)
	{
	cerr << "correlator:  bad cluster name length"
	  << strlen(clusterName) << " vs. " << clusterNameLength
	  << " at offset " << stream.bytesRead() << '\n';
	delete[] clusterName;
	return false;
	}
    for (InvestigatedClusterIterator iterator;  iterator;  ++iterator)
	{
	InvestigatedCluster* cluster = &(*iterator);
	if (&cluster->type() == this
	  &&  strcmp(clusterName, cluster->name()) == 0)
	    {
	    iterator.removeCurrent();
	    delete cluster;
	    }
	}
    InvestigatedCluster* newCluster =
      new InvestigatedCluster(*this, clusterName);
    File* clusterFile;
    // NEEDSWORK: do we really want the symlink, or do we want its target?
    while ((clusterFile = File::findFromStream(stream, false)) != NULL)
	newCluster->addNonoverlappedFile(clusterFile);
    if (newCluster->isEmpty())
	delete newCluster;
    else
	investigatedClusters.pushHead(newCluster);
    return true;
    }

/*
 * Write type-specific information to a stream.  The information
 * written includes all investigated clusters of the current type.
 */
ByteSwappedStream& ClusterInvestigatorType::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    InvestigatorTypeBase::write(stream);
    static unsigned int nullLength = 0;
    for (InvestigatedClusterIterator clusterIterator;
      clusterIterator;
      ++clusterIterator)
	{
	InvestigatedCluster& cluster(*clusterIterator);
	if (&cluster.type() == this)
	    {
	    unsigned int clusterNameLength(strlen(cluster.name()));
	    stream.write(&clusterNameLength);
	    stream.write(cluster.name(), clusterNameLength + 1);
	    for (Cluster::AllElementIterator fileIterator(cluster);
	      fileIterator;
	      ++fileIterator)
		{
		File& file(*fileIterator);
		file.writePath(stream, true);
		}
	    File::writePathTerminator(stream);
	    }
	}
    stream.write(&nullLength);
    return stream;
    }

/*
 * Read type-specific information from a stream.  The information read
 * includes all investigated clusters of the current type.
 */
ByteSwappedStream& ClusterInvestigatorType::read(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    InvestigatorTypeBase::read(stream);
    unsigned int clusterNameLength;
    while (stream.read(&clusterNameLength)  &&  clusterNameLength != 0)
	{
	char clusterName[MAXPATHLEN];
	if (clusterNameLength >= sizeof clusterName)
	    {
	    cerr << "correlator: bad cluster name length "
	      << clusterNameLength << " at offset " << stream.bytesRead()
	      << '\n';
	    stream.close();
	    return stream;
	    }
	stream.read(clusterName, clusterNameLength + 1);
	if (strlen(clusterName) != clusterNameLength)
	    {
	    cerr << "correlator: cluster name length mismatch at offset "
	      << stream.bytesRead() << ": expected " << clusterNameLength
	      << " got " << strlen(clusterName) << '\n';
	    stream.close();
	    return stream;
	    }
	for (InvestigatedClusterIterator iterator;  iterator;  ++iterator)
	    {
	    InvestigatedCluster* cluster = &(*iterator);
	    if (&cluster->type() == this
	      &&  strcmp(clusterName, cluster->name()) == 0)
		{
		iterator.removeCurrent();
		delete cluster;
		}
	    }
	InvestigatedCluster* newCluster =
	  new InvestigatedCluster(*this, clusterName);
	File* clusterFile;
	// NEEDSWORK: do we really want the symlink, or do we want its target?
	while ((clusterFile = File::findFromStream(stream, false)) != NULL)
	    newCluster->addNonoverlappedFile(clusterFile);
	if (newCluster->isEmpty())
	    delete newCluster;
	else
	    investigatedClusters.pushHead(newCluster);
	}
    return stream;
    }

/*
 * Clean any forgetting/forgotten files from global data kept by the
 * cluster investigator.
 */
void ClusterInvestigatorType::cleanGlobalData()
    const
    {
    for (InvestigatedClusterIterator iterator;  iterator;  ++iterator)
	{
	InvestigatedCluster& cluster(*iterator);
	for (Cluster::AllElementIterator fileIterator(cluster);
	  fileIterator;
	  ++fileIterator)
	    {
	    File& file(*fileIterator);
	    if (file.isForgetting())
		fileIterator.removeCurrent(); // Remove forgetting files
	    }
	}
    }

InvestigatedCluster::~InvestigatedCluster()
    {
    /*
     * For convenience, deleting an investigated cluster empties it.
     */
    nonoverlappedMembers.popAll();
    overlappedMembers.popAll();
    delete (char*)clusterName;
    clusterName = NULL;
    }

void InvestigatedCluster::computeWeight() // Calculate hoarding weight
    {
    double weight = clusterType.weight();
    if (weight <= 0)
	hoardWeight = INT_MAX;
    else
        hoardWeight =
          (long)((clusteringManager.currentReferenceTime() - lastReference)
	    / weight);
    }

/*
 * Return the type name of an investigated cluster
 */
const char* InvestigatedCluster::typeName()
    const
    {
    return type().name();
    }

/*
 * Return the encoded type name of an investigated cluster
 */
const char* InvestigatedCluster::encodedTypeName(
    ShowOptions		options)	// Encoding options
    const
    {
    return type().encodedName(options);
    }

/*
 * Return the name of an investigated cluster
 */
const char* InvestigatedCluster::name()
    const
    {
    return clusterName;
    }

/*
 * Return the name of an investigated cluster.  Returns a pointer to a
 * static buffer, so be careful!
 */
const char* InvestigatedCluster::encodedName(
    ShowOptions		options)	// Encoding options
    const
    {
    if (options & SO_MIME_ENCODE)
	{
	static const char* encoding = NULL;
	if (encoding != NULL)
	    free((void*)encoding);
	encoding = mimeEncode(clusterName, NULL);
	return encoding;
	}
    else
	return clusterName;
    }

RelationInvestigatorType::~RelationInvestigatorType()
    {
    }

/*
 * Create a new relation investigator type
 */
InvestigatorTypeBase* RelationInvestigatorType::newInvestigatorType(
    const char*		name)		// Name to give to the new type
    {
    return new RelationInvestigatorType(name);
    }

/*
 * Create a new investigated relation instance
 */
FileRelationBase& RelationInvestigatorType::newRelation()
					// File that's target of the relation
    const
    {
    RelationInvestigation* result = new RelationInvestigation(*this);
    return *result;
    }

/*
 * Read a new investigated relation from a stream
 */
FileRelationBase* RelationInvestigatorType::readRelation(
    ByteSwappedStream&	stream)		// Stream containing the relation
					// File that's target of the relation
    const
    {
    RelationInvestigation* result = (RelationInvestigation*)&(newRelation());
    stream.read(&result->relationWeight, sizeof result->relationWeight);
    return result;
    }

/*
 * Add investigated data to a given file's entry.  The investigated
 * data is provided on the given stream by an external process.
 *
 * Returns true if all went well, false if trouble arose with the
 * input stream.
 */
bool RelationInvestigatorType::learnFromStream(
    ByteSwappedStream&	controlStream)	// Stream to read information from
    {
    // NEEDSWORK: do we really want the symlink, or do we want its target?
    File* investigatedFile(File::findFromStream(controlStream, false));
    if (investigatedFile == NULL)
	return false;
    File* targetFile;
    while ((targetFile = File::findFromStream(controlStream, false))
      != NULL)
	{
	double weight;
	if (!controlStream.read(&weight, sizeof weight))
	    {
	    cerr << "correlator: trouble reading weight for relationship"
	      << " between '" << investigatedFile->encodedName(SO_MIME_ENCODE)
	      << "' and '" << targetFile->name() << "'.\n";
	    return false;
	    }
	if (weight <= 0.0)
	    {
	    /*
	     * A non-positive weight means that the relationship
	     * should be removed.
	     */
	    FileRelationBase* relation =
	      investigatedFile->chosenRelation(*targetFile, *this, false);
	    investigatedFile->removeRelation(relation);
	    investigatedFile->relationsChanged();
	    }
	else if (!investigatedFile->isForgotten()
	  &&  !targetFile->isForgetting())
	    {
	    RelationInvestigation* relation =
	      (RelationInvestigation*)investigatedFile->chosenRelation(
		*targetFile, *this, true);
	    relation->setWeight(weight);
	    investigatedFile->relationsChanged();
	    targetFile->relationsChanged();
	    }
	}
    return true;
    }

RelationInvestigation::~RelationInvestigation()
    {
    }

/*
 * Return the type of an investigated cluster relation
 */
const FileRelationTypeBase& RelationInvestigation::type()
    const
    {
    return relationType;
    }

/*
 * Show relation in printable form (appends \n).
 */
ostream& operator<<(
    ostream&		stream,		// Stream to print on
    const RelationInvestigation&
			relation)	// Relation to be printed
    {
    return relation.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Show relation in printable form, with options.  Appends \n.
 *
 * Show options honored: none
 */
ostream& RelationInvestigation::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		)		// Manager to show for, if SO_MANAGER
    const
    {
    stream << "RFD_REL\t    "
      <<  relationType.encodedName(options)
      << ' ' << relationWeight
      << '\n';
    return stream;
    }

/*
 * Write an investigated relation to a stream.
 */
ByteSwappedStream& RelationInvestigation::write(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    if (targetRelatedFile().targetFile().isForgetting())
	return stream;
    FileRelationBase::write(stream);
    stream.write(&relationWeight, sizeof relationWeight);
    return stream;
    }
