#ifndef lint
static const char Rcs_Id[] =
    "$Id: lrumanager.cc,v 1.28 1998/10/23 23:15:39 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: lrumanager.cc,v $
 * Revision 1.28  1998/10/23 23:15:39  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.27  1998/07/15 21:43:49  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.26  1998/07/09 21:02:03  geoff
 * Keep our own internal reference counter (this is implemented partially
 * incorrectly).  Add our own reference-recording routines.
 *
 * Revision 1.25  1998/05/14 01:15:35  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.24  1998/03/05 03:18:15  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.23  1997/12/04 23:16:07  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.22  1997/05/16 21:26:22  geoff
 * Remove the debugging code used to find the LRU/clustering inconsistencies.
 *
 * Revision 1.21  1997/05/16 17:10:38  geoff
 * Use includeInSimulation to decide what files to put in the hoard.  Add
 * temporary debugging code to catch discrepancies between the LRU and
 * clustering managers.  Fix problems in loop breaking that could cause a
 * file to be double-counted in the working set and hoard.  Add a clear()
 * routine.  This is the version used to run the simulations in the
 * dissertation.
 *
 * Revision 1.20  1997/04/18 19:59:46  geoff
 * Sort open files as if they have a current reference time.
 *
 * Revision 1.19  1997/01/02 06:48:23  geoff
 * Add cleanGlobalData and the requiredSize routine.
 *
 * Revision 1.18  1996/09/11 23:00:06  geoff
 * Turn off time-based management.
 *
 * Revision 1.17  1996/05/03  01:12:35  geoff
 * Update the copyright.  Make the management interval be 100 days.
 * Always hoard symlinks and directories.
 *
 * Revision 1.16  1996/04/10  01:28:36  geoff
 * Add support for manager-specific storage statuses.  Add the new
 * ManagerData::manager() routine.
 *
 * Revision 1.15  1996/03/19  18:43:01  geoff
 * Get rid of the show/read/write/clear-manager-data and record-reference
 * routines now that they can be inherited.  Inherit from
 * HoardingManagerBase.  Add support for the DERIVED_CLASS_BUG ifdef.
 *
 * Revision 1.14  1996/02/02  20:09:49  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.13  1996/02/02  18:29:09  geoff
 * Derive resetStatistics from the base class instead of redefining it.
 * Derive data accessors from the base class.  Fix some minor problems
 * detected by the new C++ compiler.
 *
 * Revision 1.12  1995/12/19  01:02:59  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.11  1995/12/19  00:34:42  geoff
 * Fix a parameter title.  Get rid of some multiple function calls, for
 * efficiency.  Add process-tracking support.
 *
 * Revision 1.10  1995/11/22  00:21:40  geoff
 * Get rid of the TIME_SORTED_INPUT ifdef.
 *
 * Revision 1.9  1995/11/10  02:59:46  geoff
 * Add zzinlinefunctions.hh.  Get rid of a spurious "const".  Add
 * SEER_REF_CONTROL support.
 *
 * Revision 1.8  1995/10/11  19:05:02  geoff
 * When managing, give the files sequential output ID numbers.
 *
 * Revision 1.7  1995/10/10  20:38:06  geoff
 * Get rid of the COMPARISON_SUPPORT and SIMULATION_SUPPORT ifdefs.
 *
 * Revision 1.6  1995/09/15  00:05:53  geoff
 * Change the RCS ID to be a const.  Add internal-timer and statistics
 * support.  Globally change "cache" to "hoard".  Add abortive
 * TRACK_MEMORY_LEAKS support.
 *
 * Revision 1.5  1995/06/30  19:04:44  geoff
 * Change to the new storage method of first unstoring everything, then
 * setting desired files to stored.  Use countStorage to estimate space
 * needed by a file.  Ask for full file information when showing files.
 *
 * Revision 1.4  1995/06/17  00:11:40  geoff
 * Add show options and clearManagerData.  Add a couple of const
 * keywords.  Flush cerr/count before all aborts.
 *
 * Revision 1.3  1995/05/27  00:16:02  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.2  1995/05/23  00:40:02  geoff
 * Add some new routines required by PRIVATE_MANAGEMENT_DATA.
 *
 * Revision 1.1  1995/05/05  21:47:04  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "correlator.hh"
#include "file.hh"
#include "globalmanager.hh"
#include "internaltimer.hh"
#include "lrumanager.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			LRU_Manager::LRU_Manager(const char* managerName,
			  const char* managerDefaults);
			LRU_Manager::~LRU_Manager();
void			LRU_Manager::setParameter(const char* description);
void			LRU_Manager::showParameters(ostream& stream) const;
					// Display management parameters
void			LRU_Manager::cleanGlobalData();
					// Clean out global file references
ostream&		LRU_Manager::showStatistics(ostream& stream) const;
					// Display collected quality statistics
ostream&		LRU_Manager::requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const;
					// Display hoard size needed for
					// ..no misses
void			LRU_Manager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream,
			  bool encodeFilenames);
					// Run hoard management
LRU_ManagerData&	LRU_Manager::managerData(File& file)
			  const;
					// Provide data access
const LRU_ManagerData&	LRU_Manager::managerData(const File& file)
			  const;
					// Provide data access
int			LRU_Manager::compareLRU(File** file1, File** file2);
					// Compare for qsort of LRU
			LRU_ManagerData::~LRU_ManagerData();
HoardingManagerBase&	LRU_ManagerData::manager() const;
					// Return pointer to our manager
void			LRU_ManagerData::recordMissedReference(File& file,
			  ReferenceType referenceType, bool isMeaningless);
					// Record that a file miss happened
void			LRU_ManagerData::recordReference(File& file,
			  ReferenceType type, bool isMeaningless);
					// Keep track of a file reference
void			LRU_ManagerData::clear(const File& file);
					// Clear management data
ostream&		LRU_ManagerData::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
void			LRU_ManagerData::cleanReferences();
					// Get rid of obsolete references
LRU_ManagerData&	LRU_ManagerData::managerData(File& file)
			  const;
					// Provide recursive data access
const LRU_ManagerData&	LRU_ManagerData::managerData(const File& file)
			  const;
					// Provide recursive data access
#endif // 0
 
LRU_Manager		lruManager(
			  "LRU",
			  "LRU::managementInterval=0\n");

extern void		correlatorPanic(const char* message);

LRU_Manager::LRU_Manager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : HoardingManagerBase(managerName, managerDefaults),
      lruFileTable(NULL),
      lruFileTableSize(0),
      lruActiveCount(0)
    {
    HoardingManagerBase::setManagerParameters(defaults);
    }

LRU_Manager::~LRU_Manager()
    {
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void LRU_Manager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad LRU_Manager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("managementInterval"))
	managementInterval = atoi(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized LRU_Manager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void LRU_Manager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "LRU::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Clean out pointers to files being forgotten.
 */
void LRU_Manager::cleanGlobalData()
    {
    if (lruFileTable == NULL)
	return;
    for (long fileNumber = lruFileTableSize;  --fileNumber >= 0;  )
	{
	File* file = lruFileTable[fileNumber];
	if (file != NULL  &&  file->isForgetting())
	    lruFileTable[fileNumber] = NULL;
	}
    }

/*
 * Display collected quality statistics
 */
ostream& LRU_Manager::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    HoardingManagerBase::showStatistics(stream);
    return stream;
    }

/*
 * Calculate and display the hoard size needed to achieve zero misses.
 * For LRU managers, this is simple: we just scan the last previous
 * file table backwards until we encounter a file referenced more
 * recently than GlobalManager::baseReference; we then add up the size
 * of that file and all newer files.
 */
ostream& LRU_Manager::requiredSize(
    ostream&		stream,		// Stream to display result on
    ostream*		duplicateStream) // Stream to use for duplicate results
    const
    {
    if (lruFileTable == NULL)
	return stream;
    for (FileTableIterator fileIterator;  fileIterator;  ++fileIterator)
	(*fileIterator).unmark();
    long fileNumber = lruFileTableSize;
    unsigned long totalSize = 0;	// Size of all files
    unsigned long totalCount = 0;	// Count of all files
    /*
     * The following variables count up various versions of the most
     * recent period's working set (with the period being determined
     * by the global manager's base reference).  Variables ending in
     * "size" count blocks, while those ending in "count" count files.
     * The counts computed here differ from the counts computed by the
     * clustering manager in that these counts only consider files
     * that were known in the previous period (i.e., that are in the
     * LRU table), while the clustering manager considers all files
     * known now; thus, the counts here ignore files created or
     * discovered during the current period.
     *
     * Variables prefixed with "total" count essentially all files.
     * Those prefixed with nothing count files that *could* be hoarded
     * if they were listed as controlled.  The "controlled"
     * prefix marks files that are actually controlled.
     *
     * We only calculate this for the base LRU manager, since there's
     * no point in doing it multiple times.  Yes, the test is ugly.
     */
    unsigned long controlledWorkingSetSize = 0;
					// Blocks in in this period's 
    unsigned long totalWorkingSetSize = 0;
    unsigned long workingSetSize = 0;
    unsigned long controlledWorkingSetCount = 0;
    unsigned long totalWorkingSetCount = 0;
    unsigned long workingSetCount = 0;
    unsigned long globalBase = getBaseReferenceTime();
    if (this == &lruManager)
	{
	while (--fileNumber >= (long)lruActiveCount)
	    {
	    File* file = lruFileTable[fileNumber];
	    if (file != NULL
	      &&  managerData(*file).lastReferenceTime() >= globalBase
	      &&  file->includeInSimulation())
		{
		unsigned long fileSize = file->size();
		totalWorkingSetCount++;
		totalWorkingSetSize += fileSize;
		if (file->couldHoard())
		    {
		    workingSetCount++;
		    workingSetSize += fileSize;
		    if (file->canHoard())
			{
			controlledWorkingSetSize += fileSize;
			controlledWorkingSetCount++;
			}
		    }
		}
	    }
	fileNumber++;
	}
    while (--fileNumber >= 0)
	{
	File* file = lruFileTable[fileNumber];
	if (file == NULL  ||  !file->includeInSimulation())
	    continue;
	if (file->couldHoard()
	  &&  managerData(*file).lastReferenceTime() >= globalBase)
	    {
#ifdef LRU_DUMP_TRIGGER_FILES
	    if (this == &lruManager)
		stream << "DEBUG: trigger file '"
		  << file->encodedName(SO_MIME_ENCODE) << "' "
		  << file->size() << ' '
		  << managerData(*file).lastReferenceTime() << '\n';
#endif // LRU_DUMP_TRIGGER_FILES
	    fileNumber++;		// Re-examine this file
	    break;
	    }
	if (this == &lruManager
	  &&  managerData(*file).lastReferenceTime() >= globalBase)
	    {
	    unsigned long fileSize = file->size();
	    totalWorkingSetCount++;
	    totalWorkingSetSize += fileSize;
	    if (file->couldHoard())
		{
		workingSetCount++;
		workingSetSize += fileSize;
		if (file->canHoard())
		    {
		    controlledWorkingSetSize += fileSize;
		    controlledWorkingSetCount++;
		    }
		}
	    }
	}
    unsigned long controlledSize = 0;	// Size of controlled files
    unsigned long controlledCount = 0;	// Count of controlled files
    while (--fileNumber >= 0)
	{
	File* file = lruFileTable[fileNumber];
	if (file == NULL  ||  !file->includeInSimulation())
	    continue;
	if (managerData(*file).lastReferenceTime() != 0)
	    {
	    if (file->isControlled())
		{
		fileNumber++;		// Re-examine this file
		break;
		}
	    else if (file->couldHoard())
		{
		file->mark();
#ifdef LRU_DUMP_EXTRA_HOARD
		if (this == &lruManager
		  &&  managerData(*file).lastReferenceTime() < globalBase)
		    stream << "DEBUG: old file '"
		      << file->encodedName(SO_MIME_ENCODE) << "' "
		      << file->size() << ' '
		      << managerData(*file).lastReferenceTime()
		      << '\n';
#endif // LRU_DUMP_EXTRA_HOARD
		totalSize += file->size();
		totalCount++;
		}
	    }
	if (this == &lruManager
	  &&  managerData(*file).lastReferenceTime() >= globalBase)
	    {
	    unsigned long fileSize = file->size();
	    totalWorkingSetCount++;
	    totalWorkingSetSize += fileSize;
	    if (file->couldHoard())
		{
		workingSetCount++;
		workingSetSize += fileSize;
		if (file->canHoard())
		    {
		    controlledWorkingSetSize += fileSize;
		    controlledWorkingSetCount++;
		    }
		}
	    }
	}
    while (--fileNumber >= 0)
	{
	File* file = lruFileTable[fileNumber];
	if (file == NULL ||  !file->includeInSimulation())
	    continue;
	if (this == &lruManager
	  &&  managerData(*file).lastReferenceTime() >= globalBase)
	    {
	    unsigned long fileSize = file->size();
	    totalWorkingSetCount++;
	    totalWorkingSetSize += fileSize;
	    if (file->couldHoard())
		{
		workingSetCount++;
		workingSetSize += fileSize;
		if (file->canHoard())
		    {
		    controlledWorkingSetSize += fileSize;
		    controlledWorkingSetCount++;
		    }
		}
	    }
	if (managerData(*file).lastReferenceTime() != 0)
	    {
	    if (file->isControlled())
		{
		file->mark();
#ifdef LRU_DUMP_EXTRA_HOARD
		if (this == &lruManager
		  &&  managerData(*file).lastReferenceTime() < globalBase)
		    stream << "DEBUG: old file '"
		      << file->encodedName(SO_MIME_ENCODE) << "' "
		      << file->size() << ' '
		      << managerData(*file).lastReferenceTime()
		      << '\n';
#endif // LRU_DUMP_EXTRA_HOARD
		unsigned long blocks = file->hoardableSize();
		controlledSize += blocks;
		controlledCount++;
		totalSize += blocks;
		totalCount++;
		}
	    else if (file->couldHoard())
		{
		file->mark();
#ifdef LRU_DUMP_EXTRA_HOARD
		if (this == &lruManager
		  &&  managerData(*file).lastReferenceTime() < globalBase)
		    stream << "DEBUG: old file '"
		      << file->encodedName(SO_MIME_ENCODE) << "' "
		      << file->size() << ' '
		      << managerData(*file).lastReferenceTime()
		      << '\n';
#endif // LRU_DUMP_EXTRA_HOARD
		totalSize += file->size();
		totalCount++;
		}
	    }
	}
    if (this == &lruManager)
	{
	stream << managerName() << " working set size is "
	  << controlledWorkingSetSize << " controlled, " << workingSetSize
	  << " controllable, " << totalWorkingSetSize << " total\n";
	stream << managerName() << " working set count is "
	  << controlledWorkingSetCount << " controlled, " << workingSetCount
	  << " controllable, " << totalWorkingSetCount << " total\n";
	}
    for (FileTableIterator fileIterator;  fileIterator;  ++fileIterator)
	{
	File& file = *fileIterator;
	if (!file.isMarked()  &&  file.alwaysHoardableType()
	  &&  file.includeInSimulation())
	    {
	    if (file.isControlled())
		{
		unsigned long blocks = file.hoardableSize();
		controlledSize += blocks;
		controlledCount++;
		totalSize += blocks;
		totalCount++;
		}
	    else if (file.couldHoard())
		{
		totalSize += file.size();
		totalCount++;
		}
	    }
	}
    stream << managerName() << " hoard size needed is " << controlledSize
      << " controlled, " << totalSize << " total\n";
    stream << managerName() << " hoard count needed is " << controlledCount
      << " controlled, " << totalCount << " total\n";
    if (duplicateStream != NULL)
	{
	*duplicateStream
	  << managerName() << " hoard size needed is " << controlledSize
	  << " controlled, " << totalSize << " total\n";
	*duplicateStream
	  << managerName() << " hoard count needed is " << controlledCount
	  << " controlled, " << totalCount << " total\n";
	}
    return stream;
    }

void LRU_Manager::manage(		// Manage the file hoard
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump results to
    bool		encodeFilenames) // True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    if (verbose > 1)
	cerr << "correlator:  LRU management began at "
	  << managementStartTime.ctime() << '\n';
    /*
     * LRU management is very simple.  We simply sort the files into
     * order by access, and then scan through them until the hoard is
     * full.
     */
    // NEEDSWORK:  should handle FORGOTTEN files, need other method
    delete[] lruFileTable;
    lruFileTableSize = File::numberOfFiles();
    lruFileTable = new File*[lruFileTableSize];
    unsigned long LRU_Index = 0;
    for (FileTableIterator iter;  iter;  ++iter)
	{
	if (LRU_Index >= lruFileTableSize)
	    correlatorPanic("correlator:  bad file count in LRU manager");
	File& file(*iter);
	if (file.alwaysHoardableType())
	    managerData(file).store(file);
	else
	    {
	    managerData(file).unstore(file);
	    if (!file.willForget()  &&  file.canHoard())
		{
		file.discoverStatusIfUnknown();
		lruFileTable[LRU_Index++] = &file;
		}
	    }
	}
    lruFileTableSize = LRU_Index;		// Actual # files hoarded
    qsort((void*) lruFileTable, lruFileTableSize, sizeof lruFileTable[0],
      (int (*)(const void*, const void*)) LRU_Manager::compareLRU);
    /*
     * Scan the files, fitting in as many as possible.  Note that a
     * lower-priority file might be inserted if it will fit where the
     * higher-priority one will not.
     */
    lruActiveCount = 0;
    unsigned long totalBlocks = 0;	// Total blocks in the hoard
    for (LRU_Index = 0;  LRU_Index < lruFileTableSize;  LRU_Index++)
	{
	File* file = lruFileTable[LRU_Index];
	if (lruActiveCount == 0
	  &&  managerData(*file).lastReferenceTime() == 0)
	    lruActiveCount = LRU_Index;
	if (!file->canHoard()  ||  file->willForget()  ||  file->isDirectory())
	    continue;
	unsigned long requiredSpace = file->countStorage(*this);
	if (totalBlocks + requiredSpace <= GlobalManager::hoardSize())
	    {
	    managerData(*file).store(*file);
	    totalBlocks += requiredSpace;
	    }
	else
	    {
	    if (verbose > 2  &&  totalBlocks < GlobalManager::hoardSize()
	      &&  requiredSpace >= GlobalManager::hoardSize() / 10)
		{
		cerr << "correlator:  skipping file " << LRU_Index
		  << " ('" << file->encodedName(SO_MIME_ENCODE)
		  << "') because it won't fit, size "
		  << requiredSpace << " blocks\n";
		}
	    }
	}
    if (lruActiveCount == 0  &&  lruFileTableSize > 0
      &&  managerData(*lruFileTable[lruFileTableSize - 1]).lastReferenceTime()
	!= 0)
	lruActiveCount = lruFileTableSize;
    /*
     * Management is now complete.
     */
    if (verbose > 0)
	{
	InternalTimer managementEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = managementEndTime - managementStartTime;
	cerr << "correlator:  LRU hoard contains " << totalBlocks
	  << " blocks\n";
	cerr << "correlator:  LRU management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << managementEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    }
	}
    if (dumpStream != NULL)
	{
	File::resetIdNumbering();
	ShowOptions options = SO_FULL_FILE_INFORMATION | SO_MANAGER;
	if (encodeFilenames)
	    options |= SO_MIME_ENCODE;
#ifdef LRU_SHOW_ONLY_HOARDABLE
	for (LRU_Index = 0;  LRU_Index < lruFileTableSize;  LRU_Index++)
	    {
	    lruFileTable[LRU_Index]->setId();
	    lruFileTable[LRU_Index]->show(*dumpStream, options, name);
	    }
#else // LRU_SHOW_ONLY_HOARDABLE
	for (FileTableIterator iter;  iter;  iter++)
	    {
	    File& file(*iter);
	    file.setId();
	    file.show(*dumpStream, options, name);
	    }
#endif // LRU_SHOW_ONLY_HOARDABLE
	}
    }

/*
 * Return a reference to the LRU manager's file-private data, so that
 * manager routines can access it.
 */
LRU_ManagerData& LRU_Manager::managerData(
    File&		file)		// File we're providing access for
    const
    {
    return file.lruManagerData;
    }

const LRU_ManagerData& LRU_Manager::managerData(
    const File&		file)		// File we're providing access for
    const
    {
    return file.lruManagerData;
    }

int LRU_Manager::compareLRU(		// Compare for qsort of LRU
    File**		file1,		// Files to compare
    File**		file2)		// ...
    {
    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (file1 == file2)
	return 0;
    if ((*file1)->isOpen()  &&  !(*file2)->isOpen())
	return -1;
    else if ((*file2)->isOpen()  &&  !(*file1)->isOpen())
	return 1;
    return lruManager.managerData(**file1).lastReferenceTime()
	  < lruManager.managerData(**file2).lastReferenceTime()
	? 1
      : lruManager.managerData(**file1).lastReferenceTime()
	  > lruManager.managerData(**file2).lastReferenceTime()
	? -1
      : 0;
    }

LRU_ManagerData::~LRU_ManagerData()
    {
    }

/*
 * Return pointer to our manager
 */
HoardingManagerBase& LRU_ManagerData::manager()
    const
    {
    return lruManager;
    }

/*
 * Record a failed reference to a file.
 */
void LRU_ManagerData::recordMissedReference(
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of the reference
    bool		isMeaningless)	// True if reference is meaningless
    {
    HoardingManagerDataBase::recordMissedReference(file, type, isMeaningless);
    lastReference = manager().nextReferenceTime();
    }

/*
 * Record a reference to a file.
 */
void LRU_ManagerData::recordReference(
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of the reference
    bool		isMeaningless)	// True if reference is meaningless
    {
    HoardingManagerDataBase::recordReference(file, type, isMeaningless);
    lastReference = manager().nextReferenceTime();
    }

/*
 * Clear the manager-specific data related to a file.  For the LRU
 * manager, this means searching the LRU file table for references to
 * the file, and removing them.
 */
void LRU_ManagerData::clear(
    const File&		file)		// File being cleared
    {
    if (!simulationMode  ||  lruManager.lruFileTable == NULL)
	return;
    for (long fileNumber = lruManager.lruFileTableSize;  --fileNumber >= 0;  )
	{
	if (lruManager.lruFileTable[fileNumber] == &file)
	    lruManager.lruFileTable[fileNumber] = NULL;
	}
    }


/*
 * Show manager data in printable form, with options
 */
ostream& LRU_ManagerData::show(
    ostream&		stream,		// Stream to write to
    ShowOptions		options,	// Options controlling how to show
    const char*		managerName)	// Manager to show for
    const
    {
    stream << "LH\t   " << manager().managerName() << ": ";
    HoardingManagerDataBase::show(stream, options, managerName);
    stream << '\n';
    return stream;
    }

void LRU_ManagerData::cleanReferences()	// Get rid of obsolete references
    {
    }

LRU_ManagerData& LRU_ManagerData::managerData(
    File&		file)		// File we're providing access for
    const
    {
    return file.lruManagerData;
    }

const LRU_ManagerData& LRU_ManagerData::managerData(
    const File&		file)		// File we're providing access for
    const
    {
    return file.lruManagerData;
    }
