#ifndef lint
static const char Rcs_Id[] =
    "$Id: managerbase.cc,v 1.33 1998/10/23 23:15:43 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: managerbase.cc,v $
 * Revision 1.33  1998/10/23 23:15:43  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.32  1998/07/15 21:43:52  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.  Also add HoardingManagerIterator and move certain
 * other things to the hoarding manager base class.
 *
 * Revision 1.31  1998/07/09 21:02:04  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.30  1998/05/14 01:15:36  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.29  1998/02/19 02:10:11  geoff
 * Add stateChanged and relationsChanged as null functions that
 * subclasses can override.
 *
 * Revision 1.28  1997/12/18 01:25:11  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.27  1997/12/04 23:16:10  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.26  1997/07/11 17:55:02  geoff
 * Report misses only on plain files
 *
 * Revision 1.25  1997/05/16 17:10:39  geoff
 * Fix the clear() routine to take an argument.  This is the version used
 * to run the simulations in the dissertation.
 *
 * Revision 1.24  1997/01/02 06:48:24  geoff
 * Add support for the requiredSize routines.
 *
 * Revision 1.23  1996/12/03 02:58:08  geoff
 * Add support for reading and writing parameters.  Change the name of
 * the routines that record miss statistics and missed references.
 *
 * Revision 1.22  1996/10/29 18:42:30  geoff
 * Flush the miss-log file after recording misses.
 *
 * Revision 1.21  1996/10/11 22:09:06  geoff
 * Log into the miss-log file, as well as to the caller's stream, in
 * showAllStatistics.
 *
 * Revision 1.20  1996/10/04 07:12:14  geoff
 * Add support for command-interpreter references.  Add support for the
 * new linear LRU manager.
 *
 * Revision 1.19  1996/09/11 23:00:08  geoff
 * Add REPORT_SELECTED_FAILURES support.  Allow a zero management
 * interval to mean "no time-based management".  In non-simulation mode,
 * don't manage right away.
 *
 * Revision 1.18  1996/05/03  01:12:39  geoff
 * Update the copyright.  Add the save manager.  Add clearInstructions.
 * Stick in a couple of strategic const declarations.
 *
 * Revision 1.17  1996/04/10  06:09:34  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.16  1996/04/10  01:28:38  geoff
 * Add support for manager-specific storage statuses.
 *
 * Revision 1.15  1996/03/19  18:43:03  geoff
 * Add support for the new manager class inheritance structure.
 *
 * Revision 1.14  1996/02/02  20:09:50  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.13  1996/02/02  18:29:12  geoff
 * Add new data accessors to the base class so that they can be derived.
 * Add the new ManagerDataBase class.  Support the new modified LRU
 * managers.  In simulation mode, only run managers once per statistics
 * interval.  Support reading manager-specific instructions from
 * alternate instruction files.  Add inheritManagerData.  Move some
 * virtual functions into the C code.
 *
 * Revision 1.12  1995/12/19  01:03:00  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.11  1995/12/19  00:34:44  geoff
 * Get rid of some multiple function calls.  Add process-tracking
 * support.
 *
 * Revision 1.10  1995/11/22  00:21:59  geoff
 * Get rid of the TIME_SORTED_INPUT ifdef.
 *
 * Revision 1.9  1995/11/10  02:59:51  geoff
 * Add zzinlinefunctions.hh.  Add SEER_REF_CONTROL support.
 *
 * Revision 1.8  1995/10/10  20:38:10  geoff
 * Get rid of the COMPARISON_SUPPORT and SIMULATION_SUPPORT ifdefs.
 *
 * Revision 1.7  1995/09/15  00:06:04  geoff
 * Change the RCS ID to be a const.  Add statistics support.  Add
 * InternalTimer support.
 *
 * Revision 1.6  1995/06/17  00:11:45  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.
 *
 * Revision 1.5  1995/05/23  00:40:08  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.4  1995/05/05  22:41:56  geoff
 * Allow a null parameter pointer in setManagerParameters.
 *
 * Revision 1.3  1995/04/20  21:58:13  geoff
 * Add support for dumping management results to a stream.  Add the
 * LRU manager to the manager list.  When running managers, put the
 * manager name in the save file.
 *
 * Revision 1.2  1995/03/29  21:04:40  geoff
 * Many changes to get working new parameter-setting support.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "clustering_mgr.hh"
#include "communication.hh"
#include "correlator.hh"
#include "deletion_mgr.hh"
#include "file.hh"
#include "filestate.hh"
#include "globalmanager.hh"
#include "internaltimer.hh"
#include "lrumanager.hh"
#include "managerbase.hh"
#include "modlrumanager.hh"
#include "panic.hh"
#include "savemanager.hh"
#include "zzinlinefunctions.hh"
#include <errno.h>
#include <iostream.h>
#include <string.h>
#include <strstream.h>

class DeletionManager;
class GlobalManager;
class LRU_Manager;
class ClusteringManager;

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			ManagerBase::ManagerBase(const char* managerName,
			  const char* managerDefaults,
			  bool dataStoringManager);
			ManagerBase::~ManagerBase();
bool			ManagerBase::shouldManage(const InternalTimer& now,
			    InternalTimer& nextManagementTime);
					// Decide if manager should run
void			ManagerBase::readInstructions(const char* fileName)
			  const;
					// Read manager-specific instructions
void			ManagerBase::clearInstructions(File& file) const;
					// Get rid of manager-specific instrs
long			ManagerBase::writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
long			ManagerBase::readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
ByteSwappedStream&	ManagerBase::writeManagerParameterHeader(
			    ByteSwappedStream& stream)
			const;
					// Write header for manager parameters
bool			ManagerBase::readManagerParameters(
			  ByteSwappedStream& stream);
					// Read params for named manager
ManagerBase*		ManagerBase::findNamedManager(const char* name);
					// Locate a named manager
void			ManagerBase::runAllManagers(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Execute all managers
void			ManagerBase::runDueManagers(const InternalTimer& now,
			  InternalTimer& nextManagementTime);
					// Execute managers that are due to run
void			ManagerBase::cleanAllGlobalData();
					// Clean out manager file references
void			ManagerBase::cleanGlobalData();
					// Clean out global file references
void			ManagerBase::setManagerParameters(
			  const char* parameters);
					// Set parameters for named manager
void			ManagerBase::showManagerParameters(
			  const char* managerName, ostream& stream);
					// Show parameters for named manager(s)
ostream&		HoardingManagerBase::showAllStatistics(ostream& stream);
ostream&		HoardingManagerBase::showStatistics(ostream& stream)
			  const;	// Display collected quality statistics
void			HoardingManagerBase::resetAllStatistics();
					// Reset statistics for all managers
void			HoardingManagerBase::resetStatistics();
					// Reset quality statistics to zero
ostream&		HoardingManagerBase::requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const;
					// Display hoard size needed for
					// ..no misses
void			HoardingManagerBase::interpretStatus(File& file) const;
					// Apply manager-specific
					// ..interpretation to file status
void			HoardingManagerBase::recordMissStatistics(File& file,
			  ReferenceType referenceType,
			  bool isMeaningless);
					// Record that a file miss happened
void			ManagerBase::stateChanged(ManagerDataBase* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Note that state has changed
			ManagerDataBase::~ManagerDataBase();
void			HoardingManagerDataBase::relationsChanged();
					// Note that relations have changed
ostream&		operator<<(ostream& stream,
			  const ManagerDataBase& data);
					// Write printable version
			EmptyManagerBase::~EmptyManagerBase();
ManagerDataBase&	EmptyManagerBase::managerData(File& file)
			  const;
					// Panic on data-access attempts
const ManagerDataBase&	EmptyManagerBase::managerData(const File& file)
			  const;
					// Panic on data-access attempts
			HoardingManagerBase::HoardingManagerBase(
			  const char* managerName,
			  const char* managerDefaults);
			HoardingManagerBase::~HoardingManagerBase();
long			HoardingManagerBase::writeParameters(
			    ByteSwappedStream& stream, bool reallyWrite)
			  const;
					// Write management params to stream
long			HoardingManagerBase::readParameters(
			  ByteSwappedStream& stream, long parameterBytes);
					// Read management params from stream
void			HoardingManagerBase::setReferenceTime(File& file,
			  long change, int whence);
			HoardingManagerDataBase::~HoardingManagerDataBase();
void			HoardingManagerDataBase::interpretStatus(File& file);
					// Apply manager-specific
					// ..interpretation to file status
void			HoardingManagerDataBase::recordMissedReference(
			  File& file, ReferenceType type, bool isMeaningless);
					// Record that a file miss happened
void			HoardingManagerDataBase::recordReference(File& file,
			  ReferenceType type, bool isMeaningless);
					// Keep track of a file reference
ostream&		HoardingManagerDataBase::show(ostream& stream,
			  ShowOptions options, const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
ByteSwappedStream&	HoardingManagerDataBase::read(
			  ByteSwappedStream& stream, File& file);
					// Read binary version from stream
ByteSwappedStream&	HoardingManagerDataBase::write(
			  ByteSwappedStream& stream,
			  const File& file)
			    const;
					// Write binary version to stream
void			HoardingManagerDataBase::clear(const File& file);
					// Clear management data
void			HoardingManagerDataBase::inherit(const File& file,
			  const File& parent);
					// Inherit manager data from parent
void			HoardingManagerDataBase::cleanReferences();
					// Clean out any dangling references
void			HoardingManagerDataBase::setReferenceTime(long change,
			  int whence);
					// Set (fake) file reference time
#endif // 0

extern BoundedLRU_Manager boundedLRU_Manager;
extern ClusteringManager clusteringManager;
extern DeletionManager	deletionManager;
extern GlobalManager	globalManager;
extern LRU_Manager	lruManager;
extern LinearLRU_Manager linearLRU_Manager;
extern SaveManager	saveManager;
extern WeightedLRU_Manager weightedLRU_Manager;

ManagerBase*		ManagerBase::managerTable[] =
    {
    &globalManager,
    &deletionManager,
    &saveManager,
    &clusteringManager,
    &lruManager,
    &boundedLRU_Manager,
    &linearLRU_Manager,
    &weightedLRU_Manager,
    NULL
    };

extern void		correlatorPanic(const char* message);
extern "C" char*	strerr(int errnum);

ManagerBase::ManagerBase(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults, // Default parameter settings
    bool		dataStoringManager) // True if manager stores file data
    : defaults(managerDefaults),
      name(managerName),
      managementInterval(600),
      lastManagementTime(InternalTimer::VIRTUAL),
      managerHasData(dataStoringManager)
    {
    }

ManagerBase::~ManagerBase()
    {
    }

/*
 * Default shouldManage routine, for time-based managers.  Decides if
 * this manager's management routine should run, and returns true if
 * so.  Regardless of the return value, modifies nextManagementTime to
 * contain the next time management will be needed for this manager.
 */
bool ManagerBase::shouldManage(
    const InternalTimer&	now,	// Current time-of-day
    InternalTimer&		nextManagementTime)
					// When to run next
    {
    /*
     * In simulation mode, we run management only after statistics
     * have been collected.  This is accomplished by only running it
     * if lastManagementTime has been invalidated by resetStatistics.
     */
    if (simulationMode)
	return !lastManagementTime.isValid();
    /*
     * If the management interval is zero, never run time-based
     * management.
     */
    if (managementInterval == 0)
	return false;
    /*
     * In non-simulation mode, lastManagementTime is invalid only when
     * we first start up.  In this case, there's no point in running
     * management right away.
     */
    if (!lastManagementTime.isValid())
	{
	lastManagementTime = now;
	nextManagementTime = now + managementInterval;
	return false;
	}
    nextManagementTime = lastManagementTime + managementInterval;
    if (now >= nextManagementTime)
	{
	nextManagementTime += managementInterval;
	return true;
	}
    else
	return false;
    }

/*
 * Read manager-specific instructions from a file.  This is a virtual
 * routine.  The default action is to simply issue an error message
 * and ignore the file.
 */
void ManagerBase::readInstructions(
    const char*		fileName)	// File containing instructions
    const
    {
    cerr << "correlator:  manager " << managerName()
      << " doesn't accept instructions, file '" << fileName << "' ignored\n";
    }

/*
 * Get rid of manager-specific instructions from a file.  This is a
 * virtual routine.  The default action is null.
 */
void ManagerBase::clearInstructions(
    File&		)		// File to fix
    const
    {
    }

/*
 * Write manager-specific parameters to a stream.  Returns the number
 * of bytes written, or zero if trouble occurs.  If reallyWrite is
 * false, the byte count is returned, but no writing is actually done.
 */
long ManagerBase::writeParameters(
    ByteSwappedStream&	stream,		// Stream to write to
    bool		reallyWrite)	// True to actually do the writing
    const
    {
    long bytesWritten = 0;
    if (reallyWrite  &&  !stream.write(&managementInterval))
	return 0;
    bytesWritten += sizeof managementInterval;
    return bytesWritten;
    }

/*
 * Read manager-specific parameters from a stream.  The second
 * argument gives the number of bytes available for consumption.
 * Returns the number of bytes remaining.
 */
long ManagerBase::readParameters(
    ByteSwappedStream&	stream,		// Stream to read from
    long		parameterBytes)	// Number of bytes available to read
    {
    parameterBytes -= sizeof managementInterval;
    if (parameterBytes >= 0)
	stream.read(&managementInterval);
    if (logVersion <= 5)
	{
	/*
	 * Older save files recorded firstMissReference, missCount,
	 * and possibly referenceBase here.  None of these are
	 * critically important, so we'll just discard them.
	 */
	ReferenceCount dummy;
	parameterBytes -= sizeof dummy;
	if (parameterBytes >= 0)
	    stream.read(&dummy);
	parameterBytes -= sizeof dummy;
	if (parameterBytes >= 0)
	    stream.read(&dummy);
	if (this == &globalManager)
	    {
	    parameterBytes -= sizeof dummy;
	    if (parameterBytes >= 0)
		stream.read(&dummy);
	    }
	}
    return parameterBytes;
    }

/*
 * Write a header for manager parameters.
 */
ByteSwappedStream& ManagerBase::writeManagerParameterHeader(
    ByteSwappedStream&	stream)		// Stream to write to
    const
    {
    int nameLength = (int)strlen(name);
    stream.write(&nameLength);
    stream.write(name, strlen(name) + 1);
    long parameterBytes = this->writeParameters(stream, false);
    return stream.write(&parameterBytes);
    }

/*
 * Read parameters for a named manager.  Reads the name, finds the
 * matching manager, and calls its parameter-read routine.  Returns
 * true if the read was successful and more manager parameters follow.
 */
bool ManagerBase::readManagerParameters(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    int length;
    if (!stream.read(&length, sizeof length))
	return false;
    char managerName[MAXPATHLEN];
    if (length == 0)
	return false;			// Zero-length path is a terminator
    else if (length < 0  ||  length > (int)sizeof managerName - 1)
	{
	cerr << "correlator:  bad path length " << length
	  << " in readManagerParameters\n";
	return false;
	}
    if (!stream.read(managerName, length + 1))
	{
	cerr << "correlator:  trouble reading name in readManagerParameters\n";
	return false;
	}
    if ((int)strlen(managerName) != length)
	{
	cerr << "correlator:  bad name length "
	  << strlen(managerName) << " for '" << managerName << "', expected "
	  << length << " in readManagerParameters\n";
	return false;
	}
    long parameterBytes;
    if (!stream.read(&parameterBytes))
	return false;
    ManagerBase* manager = findNamedManager(managerName);
    if (manager != NULL)
	{
	parameterBytes = manager->readParameters(stream, parameterBytes);
	if (parameterBytes < 0)
	    return false;
	else if (parameterBytes > 0)
	    stream.rdbuf()->seekoff(parameterBytes, ios::cur, ios::in);
	}
    else
	{
	cerr << "correlator:  unrecognized manager name '" << managerName
	  << "' in readManagerParameters\n";
	stream.rdbuf()->seekoff(parameterBytes, ios::cur, ios::in);
	}
    return true;
    }

/*
 * Return a pointer to a chosen manager, given its name.  We have to
 * be a bit careful, because this routine can be called during
 * startup, when some things aren't completely initialized.
 */
ManagerBase* ManagerBase::findNamedManager(
    const char*		name)		// Name of desired manager
    {
    for (ManagerIterator iter;  iter;  ++iter)
	{
	ManagerBase& manager(*iter);
	if (manager.name != NULL  &&  strcmp(manager.name, name) == 0)
	    return &manager;
	}
    return NULL;
    }

/*
 * Execute all known managers in sequence.  This is used for
 * comparative management.
 */
void ManagerBase::runAllManagers(
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump to after management
    bool		encodeFilenames) // True to MIME-encode file names
    {
    for (ManagerIterator iter;  iter;  ++iter)
	{
	ManagerBase& manager(*iter);
	if (saveFile != NULL)
	    {
	    static char savetype = ST_MANAGE;
	    unsigned int nameLength = strlen(manager.name);
	    if (!saveFile->write(&savetype, sizeof savetype)
	      ||  !saveFile->write(&nameLength, sizeof nameLength)
	      ||  !saveFile->write(manager.name, nameLength + 1)
	      ||  !saveFile->flush())
		{
		int saveErrno = errno;
		InternalTimer realNow(InternalTimer::ACTUAL);
		cerr << "correlator:  difficulty writing management info "
		  << "to save file: " << strerr(saveErrno)
		  << " at " << realNow.ctime() << '\n';
		}
	    }
	manager.manage(now, dumpStream, encodeFilenames);
	}
    }

/*
 * Execute all managers that claim they are due to run.  As a side
 * effect, this routine sets *nextManagementTime to contain the next
 * time that this routine expects to be needed.  This can be used to
 * calculate timeout values for select calls.  However, the caller
 * should still call this routine even if the timeout is not reached,
 * because some managers may decide to run based on criteria other
 * than time.
 */
void ManagerBase::runDueManagers(
    const InternalTimer& now,		// Current time-of-day
    InternalTimer&	nextManagementTime) // Next time to call this routine
    {
    nextManagementTime.makeInvalid();
    for (ManagerIterator iter;  iter;  ++iter)
	{
	ManagerBase& manager(*iter);
	InternalTimer thisManagerNextDue(InternalTimer::VIRTUAL);
	if (manager.shouldManage(now, thisManagerNextDue))
	    {
	    if (saveFile != NULL)
		{
		static char savetype = ST_MANAGE;
		unsigned int nameLength = strlen(manager.name);
		if (!saveFile->write(&savetype, sizeof savetype)
		  ||  !saveFile->write(&nameLength, sizeof nameLength)
		  ||  !saveFile->write(manager.name, nameLength + 1)
		  ||  !saveFile->flush())
		    {
		    int saveErrno = errno;
		    InternalTimer realNow(InternalTimer::ACTUAL);
		    cerr << "correlator:  difficulty writing management info "
		      << "to save file: " << strerr(saveErrno)
		      << " at " << realNow.ctime() << '\n';
		    }
		}
	    manager.manage(now, NULL, false);
	    }
	if (thisManagerNextDue.isValid()
	  &&  (!nextManagementTime.isValid()
	    ||  thisManagerNextDue < nextManagementTime))
	    nextManagementTime = thisManagerNextDue;
	}
    }

/*
 * Ask each manager to clean out any global pointers to obsolete files
 * that it might maintain.
 */
void ManagerBase::cleanAllGlobalData()
    {
    for (ManagerIterator iter;  iter;  ++iter)
	(*iter).cleanGlobalData();
    }

/*
 * Default global-data cleaning routine for managers that have none.
 */
void ManagerBase::cleanGlobalData()
    {
    }

/*
 * Set parameters for named managers.
 *
 * Format of input parameter string:
 *
 *	{managerName::parameterName=value[garbage]\n}*
 *
 * where "managerName" is the name of a particular manager,
 * "parameterName" is a parameter withing that manager, and "value" is
 * a value for the parameter.  A parameter value may be terminated by
 * a newline, a null byte, or any character invalid for that value
 * (this allows comments, etc., to be placed after the values).
 * However, the search for the next parameter value will only commence
 * after a newline.
 *
 * This is the same format as that generated by the showParameters
 * routines.
 */
void ManagerBase::setManagerParameters(
    const char*		parameters)	// Name and parameters to set
    {
    if (parameters == NULL)
	return;
    while (parameters[0] != '\0')
	{
	const char* colon = strchr(parameters, ':');
	if (colon == NULL  ||  colon[1] != ':')
	    {
	    cerr << "correlator:  bad manager parameter string '" << parameters
	      <<  "'\n";
	    return;
	    }
	char* nameBuffer = new char[colon - parameters + 1];
	strncpy(nameBuffer, parameters, colon - parameters);
	nameBuffer[colon - parameters] = '\0';
	ManagerBase* manager = ManagerBase::findNamedManager(nameBuffer);
	if (manager == NULL)
	    cerr << "correlator:  setManagerParameters couldn't find manager '"
	      << nameBuffer << "'\n";
	else
	    manager->setParameter(colon + 2);
	delete[] nameBuffer;
	nameBuffer = NULL;
	parameters = strchr(parameters, '\n');
	if (parameters == NULL)
	    break;
	parameters++;
	}
    }

/*
 * Show the parameters for a named manager.  If the name is '*', parameters
 * for all managers are shown.
 */
void ManagerBase::showManagerParameters(
    const char*		managerName,	// Manager to show parameters for
    ostream&		stream)		// Stream to display them on
    {
    if (strcmp(managerName, "*") == 0)
	{
	for (ManagerIterator iter;  iter;  ++iter)
	    (*iter).showParameters(stream);
	}
    else
	{
	ManagerBase* manager = ManagerBase::findNamedManager(managerName);
	if (manager == NULL)
	    cerr
	      << "correlator:  showManagerParameters couldn't find manager '"
	      << managerName << "'\n";
	else
	    manager->showParameters(stream);
	}
    }

/*
 * Display statistics collected by all managers, presumably in the
 * last simulation interval.
 */
ostream& HoardingManagerBase::showAllStatistics(
    ostream&		stream)		// Stream to display to
    {
    InternalTimer now(InternalTimer::VIRTUAL);
    // @h NEEDSWORK: shouldn't depend on clustering manager
    /*
     * I don't think the casts below should be necessary, but if I
     * don't do them, I get link-time errors.  Odd.
     */
    stream << "Manager statistics at clustering manager reference "
      << clusteringManager.currentReferenceTime() << " ("
      << clusteringManager.currentReferenceTime()
	- clusteringManager.referenceBase
      << " this interval), time " << now.numericTime(InternalTimer::S) << " = "
      << now.ctime() << '\n';
    for (HoardingManagerIterator manager;  manager;  ++manager)
	{
	(*manager).showStatistics(stream);
	(*manager).requiredSize(stream, missLogFile);
	if (missLogFile != NULL)
	    (*manager).showStatistics(*missLogFile);
	}
    stream.flush();
    if (missLogFile != NULL)
	missLogFile->flush();
    return stream;
    }

/*
 * Display collected quality statistics
 */
ostream& HoardingManagerBase::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << managerName() << "Manager " << missCount << " miss"
      << (missCount == 1 ? "" : "es")
      << ", first miss at "
      << (firstMissReference != 0 ? firstMissReference - referenceBase : 0)
      << " of " << currentReferenceTime() - getBaseReferenceTime() << '\n';
    return stream;
    }

/*
 * Reset statistics collected by all managers.
 */
void HoardingManagerBase::resetAllStatistics()
    {
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).resetStatistics();
    }

/*
 * Reset quality statistics to zero
 */
void HoardingManagerBase::resetStatistics()
    {
    firstMissReference = 0;
    missCount = 0;
    referenceBase = currentReferenceTime();
    if (simulationMode)
	lastManagementTime.makeInvalid(); // Force management to run ASAP
    }

/*
 * Apply a manager-specific interpretation to file status information.
 */
void HoardingManagerBase::interpretStatus(
    File&		file)		// File with new status
    const
    {
    managerData(file).interpretStatus(file);
    }

/*
 * Record that a file miss happened, for statistical purposes.
 */
void HoardingManagerBase::recordMissStatistics(
    File&		file,		// File that missed
    ReferenceType	referenceType,	// Type of reference that missed
    bool		)		// True if reference was meaningless
    {
    ReferenceCount previousReferenceTime =
      managerData(file).lastReferenceTime();
					// Time file was last referenced
    missCount++;
    if (firstMissReference == 0)
	firstMissReference = currentReferenceTime();
    if (missLogFile != NULL)
	{
	InternalTimer timestamp(InternalTimer::VIRTUAL);
	*missLogFile << managerName()
	  << ' ' << file.encodedName(SO_MIME_ENCODE)
	  << ' ' << (referenceType == SEER_REF_READ ? "READ"
	    : referenceType == SEER_REF_RW ? "RW"
	    : referenceType == SEER_REF_CREATE ? "CREATE"
	    : referenceType == SEER_REF_EXEC ? "EXEC"
	    : referenceType == SEER_REF_DESTROY ? "DESTROY"
	    : referenceType == SEER_REF_RENAME ? "RENAME"
	    : referenceType == SEER_REF_OPENREAD ? "OPENREAD"
	    : referenceType == SEER_REF_OPENRW ? "OPENRW"
	    : referenceType == SEER_REF_OPENCREATE ? "OPENCREATE"
	    : referenceType == SEER_REF_CLOSE ? "CLOSE"
	    : referenceType == SEER_REF_EXIT ? "EXIT"
	    : referenceType == SEER_REF_PEEK ? "PEEK"
	    : referenceType == SEER_REF_MAKELINK ? "MAKELINK"
	    : referenceType == SEER_REF_PEEKLINK ? "PEEKLINK"
	    : referenceType == SEER_REF_FOLLOW ? "FOLLOW"
	    : referenceType == SEER_REF_DUP ? "DUP"
	    : referenceType == SEER_REF_INTERPRET ? "INTERPRET"
	    : referenceType == SEER_REF_CONTROL ? "CONTROL"
	    : referenceType == SEER_REF_NEWPROC ? "NEWPROC"
	    : referenceType == SEER_REF_PROCEXIT ? "PROCEXIT"
	    : "UNKNOWN")
	  << ' ' << previousReferenceTime
	  << ' ' << currentReferenceTime()
	  << ' ' << timestamp.numericMicrotime(InternalTimer::S)
	  << ' ' << timestamp.cmicrotime()
	  << '\n';
	missLogFile->flush();
	}
    }

/*
 * Take note of changed storage state.  By default, managers don't
 * care, so this is a no-op unless overridden.
 */
void ManagerBase::stateChanged(
    ManagerDataBase*	,		// Manager-specific data
    FileState::State	,		// Original state of the file
    FileState::State	)		// New state of the file
    {
    }

ManagerDataBase::~ManagerDataBase()
    {
    }

/*
 * Take note of changed relations.  By default, managers don't care,
 * so this is a no-op unless overridden.
 */
void HoardingManagerDataBase::relationsChanged()
    {
    }

/*
 * Provide an output operator that can be used to conveniently write
 * manager data to a stream.  The show options are chosen to show
 * everything.
 */
ostream& operator<<(
    ostream&		stream,		// Stream to write to
    const ManagerDataBase&
			data)		// Data to be written
    {
    return data.show(stream, SO_EVERYTHING, NULL);
    }

EmptyManagerBase::~EmptyManagerBase()
    {
    }

/*
 * Panic on data-access attempts
 */
ManagerDataBase& EmptyManagerBase::managerData(
    File&		)		// File to provide access for
    const
    {
    correlatorPanic("managerData called unexpectedly");
    return (ManagerDataBase&)*this;	// Dummy return to silence compiler
    }

/*
 * Panic on data-access attempts
 */
const ManagerDataBase& EmptyManagerBase::managerData(
    const File&		)		// File to provide access for
    const
    {
    correlatorPanic("managerData called unexpectedly");
    return (ManagerDataBase&)*this;	// Dummy return to silence compiler
    }

HoardingManagerBase::HoardingManagerBase(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : ManagerBase(managerName, managerDefaults, true),
      currentReference(0),
      firstMissReference(0),
      missCount(0),
      referenceBase(0)
    {
    }

HoardingManagerBase::~HoardingManagerBase()
    {
    }

/*
 * Write manager-specific parameters to a stream.  Returns the number
 * of bytes written, or zero if trouble occurs.  If reallyWrite is
 * false, the byte count is returned, but no writing is actually done.
 */
long HoardingManagerBase::writeParameters(
    ByteSwappedStream&	stream,		// Stream to write to
    bool		reallyWrite)	// True to actually do the writing
    const
    {
    long bytesWritten = ManagerBase::writeParameters(stream, reallyWrite);
    if (reallyWrite  &&  !stream.write(&currentReference))
	return 0;
    bytesWritten += sizeof currentReference;
    if (reallyWrite  &&  !stream.write(&firstMissReference))
	return 0;
    bytesWritten += sizeof firstMissReference;
    if (reallyWrite  &&  !stream.write(&missCount))
	return 0;
    bytesWritten += sizeof missCount;
    if (reallyWrite  &&  !stream.write(&referenceBase))
	return 0;
    bytesWritten += sizeof referenceBase;
    return bytesWritten;
    }

/*
 * Read manager-specific parameters from a stream.  The second
 * argument gives the number of bytes available for consumption.
 * Returns the number of bytes remaining.
 */
long HoardingManagerBase::readParameters(
    ByteSwappedStream&	stream,		// Stream to read from
    long		parameterBytes)	// Number of bytes available to read
    {
    parameterBytes = ManagerBase::readParameters(stream, parameterBytes);
    if (logVersion > 5)
	{
	parameterBytes -= sizeof currentReference;
	if (parameterBytes >= 0)
	    stream.read(&currentReference);
	parameterBytes -= sizeof firstMissReference;
	if (parameterBytes >= 0)
	    stream.read(&firstMissReference);
	parameterBytes -= sizeof missCount;
	if (parameterBytes >= 0)
	    stream.read(&missCount);
	parameterBytes -= sizeof referenceBase;
	if (parameterBytes >= 0)
	    stream.read(&referenceBase);
	}
    return parameterBytes;
    }

/*
 * Set the manager's reference time (to allow initialization and fakery)
 */
void HoardingManagerBase::setReferenceTime(
    File&		file,		// File to change for
    long		change,		// Change to be applied
    int			whence)		// Where change is from (vis. lseek)
    {
    managerData(file).setReferenceTime(change, whence);
    }

HoardingManagerDataBase::~HoardingManagerDataBase()
    {
    }

void HoardingManagerDataBase::interpretStatus(
    File&		file)		// File that has new status
    {
    if (file.isControlled())
	return;				// Nothing needed for controlled files
    if (file.currentState() == FileState::KNOWN)
	ourStorageState = STORED;
    else
	ourStorageState = NOT_STORED;
    }

/*
 * Record that a missed file reference happened.
 */
void HoardingManagerDataBase::recordMissedReference(
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of the reference
    bool		isMeaningless)	// True if reference is meaningless
    {
    /*
     * We explicitly qualify here so that we won't wind up in the
     * derived class's virtual routine, which would surprise the
     * derived class's recordMissedReference code.
     */
    HoardingManagerDataBase::recordReference(file, type, isMeaningless);
    }

/*
 * Record a file reference.
 */
void HoardingManagerDataBase::recordReference(
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of the reference
    bool		isMeaningless)	// True if reference is meaningless
    {
    switch (type)
	{
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_RW:		// Reads and/or writes of existing file
	case SEER_REF_EXEC:		// Execution
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	case SEER_REF_OPENRW:		// Open for read/write
	case SEER_REF_PEEKLINK:		// Peek at a symbolic link's contents
	case SEER_REF_FOLLOW:		// Follow a symbolic link
	case SEER_REF_DUP:		// Duplicate an existing file reference
	    {
	    if (ourStorageState == NOT_STORED  &&  file.canHoard()
	      &&  !file.willForget()  &&  file.fileType == File::FT_FILE)
		{
		((HoardingManagerBase&)manager()).recordMissStatistics(file,
		  type, isMeaningless);
		ourStorageState = NOT_STORED_BUT_MISSED;
		}
	    break;
	    }
	case SEER_REF_OPENREAD:		// Open for read
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	    {
	    if (ourStorageState == NOT_STORED  &&  file.canHoard()
	      &&  !file.willForget()  &&  file.fileType == File::FT_FILE)
		{
		((HoardingManagerBase&)manager()).recordMissStatistics(file,
		  type, isMeaningless);
		ourStorageState = NOT_STORED_BUT_MISSED;
		}
	    break;
	    }
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	case SEER_REF_OPENCREATE:	// Create or truncate/rewrite
	case SEER_REF_MAKELINK:		// Make a symbolic link
	    {
	    ourStorageState = STORED;
	    break;
	    }
	case SEER_REF_CLOSE:		// Close of open file
	case SEER_REF_DESTROY:		// Destruction (delete)
	case SEER_REF_EXIT:		// Process exit or re-exec
	    {
	    break;
	    }
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_RENAME:		// Change of file name
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	default:
	    {
	    cerr << "HoardingManagerDataBase::recordReference: "
		"unexpected reference type "
	      <<  type;
	    correlatorPanic("");
	    break;
	    }
	}
    }

/*
 * Show manager data in printable form, with options
 */
ostream& HoardingManagerDataBase::show(
    ostream&		stream,		// Stream to write to
    ShowOptions		,		// Options controlling how to show
    const char*		)		// Manager to show for
    const
    {
    switch (ourStorageState)
	{
	case STORED:
	    {
	    stream << "STORED";
	    break;
	    }
	case NOT_STORED:
	    {
	    stream << "NOT_STORED";
	    break;
	    }
	case NOT_STORED_BUT_MISSED:
	    {
	    stream << "NOT_STORED_BUT_MISSED";
	    break;
	    }
	}
    stream << ' ' << lastReference;
    return stream;
    }

/*
 * Read binary version of manager data from stream
 */
ByteSwappedStream& HoardingManagerDataBase::read(
    ByteSwappedStream&	stream,		// Stream to read from
    File&		)		// File we're a member of
    {
    if (logVersion > 5)
	stream.read(&lastReference, sizeof lastReference);
    return stream;
    }

/*
 * Write binary version of manager data to stream
 */
ByteSwappedStream& HoardingManagerDataBase::write(
    ByteSwappedStream&	stream,		// Stream to write to
    const File&		)		// File we're a member of
    const
    {
    stream.write(&lastReference, sizeof lastReference);
    return stream;
    }

/*
 * Clear management data
 */
void HoardingManagerDataBase::clear(	// Clear management data
    const File&		)		// File we're clearing for
    {
    lastReference = 0;
    ourStorageState = NOT_STORED;
    }

void HoardingManagerDataBase::inherit(	// Inherit manager data from parent
    const File&		,		// File inheriting information
    const File&		)		// Parent to inherit from
    {
    // The default is to inherit nothing
    }

/*
 * Clean out any dangling references.  By default, hoarding managers
 * don't keep any references that might dangle.
 */
void HoardingManagerDataBase::cleanReferences()
    {
    }
/*
 * Set the manager's reference time (to allow initialization and fakery)
 */
void HoardingManagerDataBase::setReferenceTime(
    long		change,		// Change to be applied
    int			whence)		// Where change is from (vis. lseek)
    {
    switch (whence)
	{
	case 0:				// Change is measured from zero
	    {
	    lastReference = (ReferenceCount)change;
	    break;
	    }
	case 1:				// Change is measured from current
					// ..last-reference time
	    {
	    lastReference += change;
	    break;
	    }
	case 2:				// Change is measured from end
	    {
	    lastReference =
	      ((HoardingManagerBase&)manager()).currentReferenceTime()
		+ change;
	    break;
	    }
	default:
	    {
	    cerr << "correlator: bad whence " << whence
	      << " ignored in HoardingManagerDataBase::setReferenceTime\n";
	    break;
	    }
	}
    }
