#ifndef lint
static const char Rcs_Id[] =
    "$Id: modlrumanager.cc,v 1.19 1998/10/23 23:15:46 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: modlrumanager.cc,v $
 * Revision 1.19  1998/10/23 23:15:46  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.18  1998/07/15 21:43:54  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.17  1998/07/09 21:02:04  geoff
 * Add support for an internal (LRU-specific) reference counter.
 *
 * Revision 1.16  1998/05/14 01:15:38  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.15  1998/03/05 03:18:16  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.14  1997/12/04 23:16:13  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.13  1997/05/16 17:10:40  geoff
 * Issue error messages when manager instruction files can't be opened.
 * This is the version used to run the simulations in the dissertation.
 *
 * Revision 1.12  1997/04/18 19:59:47  geoff
 * Sort open files as if they have a current reference time.
 *
 * Revision 1.11  1997/02/12 02:52:08  geoff
 * Fix a bunch of bugs in the linear-LRU manager: the default line
 * parameters were wrong (intercept of 1, slope of 0); the bounding was
 * inverted; the intercept was applied to the last reference time instead
 * of the age.  Fix some similar problems in the weighted LRU manager.
 *
 * Revision 1.10  1997/01/02 06:48:26  geoff
 * Add support for the requiredSize routines.
 *
 * Revision 1.9  1996/10/11 22:09:07  geoff
 * Get rid of the workaround for the floating-output bug.
 *
 * Revision 1.8  1996/10/08 19:50:06  geoff
 * Add code to work around the floating-point output bug.
 *
 * Revision 1.7  1996/10/04 07:12:15  geoff
 * Add the new linear LRU manager.
 *
 * Revision 1.6  1996/09/11 23:00:10  geoff
 * Turn off time-based management.  After reading instructions, remember
 * to delete the final fields structure.
 *
 * Revision 1.5  1996/05/03  01:12:42  geoff
 * Update the copyright.  Add clearInstructions for both managers.
 * Support always-hoardable file types.  Change the management intervals
 * to 100 days.  Accept tilde forms in instruction files, and apply
 * instructions to all descendants.
 *
 * Revision 1.4  1996/04/10  01:28:40  geoff
 * Add the new ManagerData::manager() routine.  Add support for
 * manager-specific storage statuses.
 *
 * Revision 1.3  1996/03/19  18:43:07  geoff
 * Get rid of the show/read/write/clear-manager-data and record-reference
 * routines now that they can be inherited.  Inherit from
 * HoardingManagerBase.  Add support for the DERIVED_CLASS_BUG ifdef.
 * Use the fields package to read instruction files.
 *
 * Revision 1.2  1996/02/02  20:09:51  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.1  1996/02/02  17:16:42  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "correlator.hh"
#include "fields.hh"
#include "file.hh"
#include "globalmanager.hh"
#include "modlrumanager.hh"
#include "panic.hh"
#include "showoptions.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			BoundedLRU_Manager::BoundedLRU_Manager(
			  const char* managerName,
			  const char* managerDefaults);
			BoundedLRU_Manager::~BoundedLRU_Manager();
void			BoundedLRU_Manager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
void			BoundedLRU_Manager::setParameter(
			  const char* description);
					// Change a management parameter
void			BoundedLRU_Manager::showParameters(ostream& stream)
			  const;
					// Display management parameters
void			BoundedLRU_Manager::readInstructions(
			    const char* fileName)
			  const;
					// Read manager-specific instructions
void			BoundedLRU_Manager::clearInstructions(File& file)
			  const;
					// Get rid of manager-specific instrs
BoundedLRU_ManagerData&	BoundedLRU_Manager::managerData(File& file)
			  const;
					// Provide data access
const BoundedLRU_ManagerData&
			BoundedLRU_Manager::managerData(
			    const File& file)
			  const;
					// Provide data access
int			BoundedLRU_Manager::compareLRU(File** file1,
			  File** file2);
					// Compare for qsort of bounded LRU
			BoundedLRU_ManagerData::~BoundedLRU_ManagerData();
HoardingManagerBase&	BoundedLRU_ManagerData::manager() const;
					// Return pointer to our manager
ByteSwappedStream&	BoundedLRU_ManagerData::read(ByteSwappedStream& stream,
			  File& file);
					// Read binary version from stream
ByteSwappedStream&	BoundedLRU_ManagerData::write(
			    ByteSwappedStream& stream, const File& file)
			  const;
					// Write binary version to stream
void			BoundedLRU_ManagerData::inherit(const File& file,
			  const File& parent);
					// Inherit information from parent
ostream&		operator<<(class ostream& stream,
			  const BoundedLRU_ManagerData& data);
					// Write printable version
ostream&		BoundedLRU_ManagerData::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
BoundedLRU_ManagerData&	BoundedLRU_ManagerData::managerData(
			    File& file)
			  const;
					// Provide recursive data access
const BoundedLRU_ManagerData&
			BoundedLRU_ManagerData::managerData(
			    const File& file)
			  const;
					// Provide recursive data access
			LinearLRU_Manager::LinearLRU_Manager(
			  const char* managerName,
			  const char* managerDefaults);
			LinearLRU_Manager::~LinearLRU_Manager();
void			LinearLRU_Manager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
void			LinearLRU_Manager::setParameter(
			  const char* description);
					// Change a management parameter
void			LinearLRU_Manager::showParameters(ostream& stream)
			  const;
					// Display management parameters
void			LinearLRU_Manager::readInstructions(
			    const char* fileName)
			  const;
					// Read manager-specific instructions
void			LinearLRU_Manager::clearInstructions(File& file)
			  const;
					// Get rid of manager-specific instrs
LinearLRU_ManagerData&	LinearLRU_Manager::managerData(File& file)
			  const;
					// Provide data access
const LinearLRU_ManagerData&
			LinearLRU_Manager::managerData(
			    const File& file)
			  const;
					// Provide data access
int			LinearLRU_Manager::compareLRU(File** file1,
			  File** file2);
					// Compare for qsort of linear LRU
			LinearLRU_ManagerData::~LinearLRU_ManagerData();
HoardingManagerBase&	LinearLRU_ManagerData::manager() const;
					// Return pointer to our manager
ByteSwappedStream&	LinearLRU_ManagerData::read(ByteSwappedStream& stream,
			  File& file);
					// Read binary version from stream
ByteSwappedStream&	LinearLRU_ManagerData::write(
			    ByteSwappedStream& stream, const File& file)
			  const;
					// Write binary version to stream
void			LinearLRU_ManagerData::inherit(const File& file,
			  const File& parent);
					// Inherit information from parent
ostream&		operator<<(class ostream& stream,
			  const LinearLRU_ManagerData& data);
					// Write printable version
ostream&		LinearLRU_ManagerData::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
LinearLRU_ManagerData&	LinearLRU_ManagerData::managerData(
			    File& file)
			  const;
					// Provide recursive data access
const LinearLRU_ManagerData&
			LinearLRU_ManagerData::managerData(
			    const File& file)
			  const;
					// Provide recursive data access
			WeightedLRU_Manager::WeightedLRU_Manager(
			  const char* managerName,
			  const char* managerDefaults);
			WeightedLRU_Manager::~WeightedLRU_Manager();
void			WeightedLRU_Manager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
void			WeightedLRU_Manager::setParameter(
			  const char* description);
					// Change a management parameter
void			WeightedLRU_Manager::showParameters(ostream& stream)
			  const;
					// Display management parameters
void			WeightedLRU_Manager::readInstructions(
			    const char* fileName)
			  const;
					// Read manager-specific instructions
void			WeightedLRU_Manager::clearInstructions(File& file)
			  const;
					// Get rid of manager-specific instrs
WeightedLRU_ManagerData& WeightedLRU_Manager::managerData(File& file)
			  const;
					// Provide data access
const WeightedLRU_ManagerData&
			WeightedLRU_Manager::managerData(const File& file)
			  const;
					// Provide data access
int			WeightedLRU_Manager::compareLRU(File** file1,
			  File** file2);
					// Compare for qsort of weighted LRU
			WeightedLRU_ManagerData::~WeightedLRU_ManagerData();
HoardingManagerBase&	WeightedLRU_ManagerData::manager() const;
					// Return pointer to our manager
ByteSwappedStream&	WeightedLRU_ManagerData::read(
			  ByteSwappedStream& stream, File& file);
					// Read binary version from stream
ByteSwappedStream&	WeightedLRU_ManagerData::write(
			    ByteSwappedStream& stream, const File& file)
			  const;
					// Write binary version to stream
void			WeightedLRU_ManagerData::inherit(const File& file,
			  const File& parent);
					// Inherit information from parent
ostream&		operator<<(class ostream& stream,
			  const WeightedLRU_ManagerData& data);
					// Write printable version
ostream&		WeightedLRU_ManagerData::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
WeightedLRU_ManagerData&
			WeightedLRU_ManagerData::managerData(File& file)
			  const;
					// Provide recursive data access
const WeightedLRU_ManagerData&
			WeightedLRU_ManagerData::managerData(const File& file)
			  const;
					// Provide recursive data access
#endif // 0
 
BoundedLRU_Manager	boundedLRU_Manager(
			  "BoundedLRU",
			  "BoundedLRU::managementInterval=0\n");
LinearLRU_Manager	linearLRU_Manager(
			  "LinearLRU",
			  "LinearLRU::managementInterval=0\n");
WeightedLRU_Manager	weightedLRU_Manager(
			  "WeightedLRU",
			  "WeightedLRU::managementInterval=0\n");

extern void		correlatorPanic(const char* message);
extern "C" char*	strerr(int errnum);

BoundedLRU_Manager::BoundedLRU_Manager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : LRU_Manager(managerName, managerDefaults)
    {
    ManagerBase::setManagerParameters(defaults);
    }

BoundedLRU_Manager::~BoundedLRU_Manager()
    {
    }

void BoundedLRU_Manager::manage(	// Manage the file hoard
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump results to
    bool		encodeFilenames) // True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    if (verbose > 1)
	cerr << "correlator:  Bounded LRU management began at "
	  << managementStartTime.ctime() << '\n';
    /*
     * Bounded LRU management is very simple.  We simply sort the
     * files into order by bounded access time, and then scan through
     * them until the hoard is full.  (The bounded access time is
     * simply the larger of the access time and the difference between
     * the current time and the file's bound.)
     */
    // NEEDSWORK:  should handle FORGOTTEN files, need other method
    delete[] lruFileTable;
    lruFileTableSize = File::numberOfFiles();
    lruFileTable = new File*[lruFileTableSize];
    unsigned long BoundedLRU_Index = 0;
    for (FileTableIterator iter;  iter;  ++iter)
	{
	if (BoundedLRU_Index >= lruFileTableSize)
	    correlatorPanic(
	      "correlator:  bad file count in Bounded LRU manager");
	File& file(*iter);
	if (file.alwaysHoardableType())
	    managerData(file).store(file);
	else
	    {
	    managerData(file).unstore(file);
	    if (simulationMode
	      ||  (file.canHoard()  &&  !file.willForget()
		&&  !file.isDirectory()))
		lruFileTable[BoundedLRU_Index++] = &file;
	    }
	}
    lruFileTableSize = BoundedLRU_Index; // Actual # files hoarded
    qsort((void*) lruFileTable, lruFileTableSize, sizeof lruFileTable[0],
      (int (*)(const void*, const void*)) BoundedLRU_Manager::compareLRU);
    /*
     * Scan the files, fitting in as many as possible.  Note that a
     * lower-priority file might be inserted if it will fit where the
     * higher-priority one will not.
     */
    unsigned long totalBlocks = 0;	// Total blocks in the hoard
    for (BoundedLRU_Index = 0;
      BoundedLRU_Index < lruFileTableSize;
      BoundedLRU_Index++)
	{
	File* file = lruFileTable[BoundedLRU_Index];
	if (!file->canHoard()  ||  file->willForget()  ||  file->isDirectory())
	    continue;
	unsigned long requiredSpace = file->countStorage(*this);
	if (totalBlocks + requiredSpace <= GlobalManager::hoardSize())
	    {
	    managerData(*file).store(*file);
	    totalBlocks += requiredSpace;
	    }
	else
	    {
	    if (verbose > 2  &&  totalBlocks < GlobalManager::hoardSize()
	      &&  requiredSpace >= GlobalManager::hoardSize() / 10)
		{
		cerr << "correlator:  skipping file " << BoundedLRU_Index
		  << " ('" << file->encodedName(SO_MIME_ENCODE)
		  << "') because it won't fit, size "
		  << requiredSpace << " blocks\n";
		}
	    }
	}
    /*
     * Management is now complete.
     */
    if (verbose > 0)
	{
	InternalTimer managementEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = managementEndTime - managementStartTime;
	cerr << "correlator:  Bounded LRU hoard contains " << totalBlocks
	  << " blocks\n";
	cerr << "correlator:  Bounded LRU management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << managementEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    }
	}
    if (dumpStream != NULL)
	{
	File::resetIdNumbering();
	ShowOptions options = SO_FULL_FILE_INFORMATION | SO_MANAGER;
	if (encodeFilenames)
	    options |= SO_MIME_ENCODE;
#ifdef LRU_SHOW_ONLY_HOARDABLE
	for (BoundedLRU_Index = 0;
	  BoundedLRU_Index < lruFileTableSize;
	  BoundedLRU_Index++)
	    {
	    lruFileTable[BoundedLRU_Index]->setId();
	    lruFileTable[BoundedLRU_Index]->show(*dumpStream, options, name);
	    }
#else // LRU_SHOW_ONLY_HOARDABLE
	for (FileTableIterator iter;  iter;  iter++)
	    {
	    File& file(*iter);
	    file.setId();
	    file.show(*dumpStream, options, name);
	    }
#endif // LRU_SHOW_ONLY_HOARDABLE
	}
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void BoundedLRU_Manager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad BoundedLRU_Manager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("managementInterval"))
	managementInterval = atoi(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized BoundedLRU_Manager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void BoundedLRU_Manager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "BoundedLRU::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Read manager-specific instructions from an auxiliary file.
 *
 * Instructions for this manager are of the form "filename bound",
 * where filename is a full pathname of a file (symlinks are OK), and
 * bound is an integer giving the priority to assign to the file
 * (lower numbers make the file more likely to be cached).
 */
void BoundedLRU_Manager::readInstructions(
    const char*		fileName)	// Name of file to read from
    const
    {
    ByteSwappedStream instructionFile(fileName, ios::in,
      ByteSwappedStream::bigEnded);
    if (!instructionFile)
	{
	int saveErrno = errno;
	cerr << "Couldn't open bounded LRU instruction file '"
	  << fileName << "': " << strerr(saveErrno) << '\n';
	return;
	}
    ParsedFields* fields = NULL;
    for (  ;
      (fields = new ParsedFields(instructionFile))  &&  *fields;
      delete fields)
	{
	if (fields->count() == 0  ||  (*fields)[0][0] == '#')
	    continue;
	else if (fields->count() != 2)
	    cerr << "correlator:  Bounded LRU instruction must contain"
		" a bound in '"
	      << *fields << "'\n";
	const char* expandedPath = File::tildeExpand((*fields)[0]);
	if (expandedPath == NULL)
	    cerr << "correlator:  Bounded LRU manager can't expand ~user  "
		"notation in '"
	      << *fields << "'\n";
	else if (expandedPath[0] != '/')
	    cerr << "correlator:  Bounded LRU manager instruction path name "
		"must be absolute in '"
	      << *fields << "'\n";
	else
	    {
	    File* file = File::findCanonical(expandedPath, false, true, NULL);
	    ReferenceCount bound = atol((*fields)[1]);
	    managerData(*file).setBound(bound);
	    for (FileDescendantsIterator child(*file);  child;  ++child)
		managerData(*child).setBound(bound);
	    }
	}
    delete fields;
    }

/*
 * Get rid of manager-specific instructions for a file.
 */
void BoundedLRU_Manager::clearInstructions(
    File&		file)		// File to fix
    const
    {
    managerData(file).setBound(INT_MAX);
    }

/*
 * Return a reference to the bounded LRU manager's file-private data,
 * so that manager routines can access it.
 */
BoundedLRU_ManagerData& BoundedLRU_Manager::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.boundedLRU_ManagerData;
    }

const BoundedLRU_ManagerData& BoundedLRU_Manager::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.boundedLRU_ManagerData;
    }

int BoundedLRU_Manager::compareLRU(	// Compare for qsort of bounded LRU
    class File**	file1,		// Files to compare
    class File**	file2)		// ...
    {
    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (file1 == file2)
	return 0;
    ReferenceCount now = boundedLRU_Manager.currentReferenceTime();
    ReferenceCount referenceTime1 =
      boundedLRU_Manager.managerData(**file1).lastReferenceTime();
    if ((*file1)->isOpen())
	referenceTime1 = now;
    ReferenceCount bound = boundedLRU_Manager.managerData(**file1).bound;
    if (now < bound)
	bound = now;
    if (referenceTime1 < now - bound)
	referenceTime1 = now - bound;
    ReferenceCount referenceTime2 =
      boundedLRU_Manager.managerData(**file2).lastReferenceTime();
    if ((*file2)->isOpen())
	referenceTime2 = now;
    bound = boundedLRU_Manager.managerData(**file2).bound;
    if (now < bound)
	bound = now;
    if (referenceTime2 < now - bound)
	referenceTime2 = now - bound;
    return referenceTime1 < referenceTime2 ? 1
      : referenceTime1 > referenceTime2 ? -1
      : 0;
    }

BoundedLRU_ManagerData::~BoundedLRU_ManagerData()
    {
    }

/*
 * Return pointer to our manager
 */
HoardingManagerBase& BoundedLRU_ManagerData::manager()
    const
    {
    return boundedLRU_Manager;
    }

/*
 * Read binary version of manager data from stream
 */
ByteSwappedStream& BoundedLRU_ManagerData::read(
    ByteSwappedStream&	stream,		// Stream to read data from
    File&		file)		// File we are a member of
    {
    HoardingManagerDataBase::read(stream, file);
    stream.read(&bound, sizeof bound);
    return stream;
    }

/* 
 * Write binary version of manager data to stream
 */
ByteSwappedStream& BoundedLRU_ManagerData::write(
    ByteSwappedStream&	stream,		// Stream to write data to
    const File&		file)		// File we are a member of
    const
    {
    HoardingManagerDataBase::write(stream, file);
    stream.write(&bound, sizeof bound);
    return stream;
    }

void BoundedLRU_ManagerData::inherit(	// Inherit information from parent
    const File&		,		// File inheriting information
    const File&		parent)		// Parent to inherit from
    {
    bound = managerData(parent).bound;
    }

/*
 * Write a printable version of the bounded LRU manager data to a stream
 */
ostream& operator<<(
    ostream&		stream,		// Stream to write to
    const BoundedLRU_ManagerData& data)	// Data to write
    {
    return data.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Write a printable version of the bounded LRU manager data to a
 * stream, with options.
 *
 * Show options honored: none
 */
ostream& BoundedLRU_ManagerData::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    stream << "BLH\t    BoundedLRU_Manager: ";
    HoardingManagerDataBase::show(stream, options, managerName);
    stream << ' ' << bound << '\n';
    return stream;
    }
 
/*
 * Return a reference to the bounded LRU manager's file-private data,
 * so that manager routines can access it recursively.
 */
BoundedLRU_ManagerData& BoundedLRU_ManagerData::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.boundedLRU_ManagerData;
    }

const BoundedLRU_ManagerData& BoundedLRU_ManagerData::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.boundedLRU_ManagerData;
    }

LinearLRU_Manager::LinearLRU_Manager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : LRU_Manager(managerName, managerDefaults)
    {
    ManagerBase::setManagerParameters(defaults);
    }

LinearLRU_Manager::~LinearLRU_Manager()
    {
    }

void LinearLRU_Manager::manage(		// Manage the file hoard
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump results to
    bool		encodeFilenames) // True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    if (verbose > 1)
	cerr << "correlator:  Linear LRU management began at "
	  << managementStartTime.ctime() << '\n';
    /*
     * Linear LRU management is very simple.  We simply sort the files
     * into order by transformed access time, and then scan through
     * them until the hoard is full.
     */
    // NEEDSWORK:  should handle FORGOTTEN files, need other method
    lruFileTableSize = File::numberOfFiles();
    delete[] lruFileTable;
    lruFileTable = new File*[lruFileTableSize];
    unsigned long LinearLRU_Index = 0;
    for (FileTableIterator iter;  iter;  ++iter)
	{
	if (LinearLRU_Index >= lruFileTableSize)
	    correlatorPanic(
	      "correlator:  bad file count in Linear LRU manager");
	File& file(*iter);
	if (file.alwaysHoardableType())
	    managerData(file).store(file);
	else
	    {
	    managerData(file).unstore(file);
	    if (simulationMode
	      ||  (file.canHoard()  &&  !file.willForget()
		&&  !file.isDirectory()))
		lruFileTable[LinearLRU_Index++] = &file;
	    }
	}
    lruFileTableSize = LinearLRU_Index;	// Actual # files hoarded
    qsort((void*) lruFileTable, lruFileTableSize, sizeof lruFileTable[0],
      (int (*)(const void*, const void*)) LinearLRU_Manager::compareLRU);
    /*
     * Scan the files, fitting in as many as possible.  Note that a
     * lower-priority file might be inserted if it will fit where the
     * higher-priority one will not.
     */
    unsigned long totalBlocks = 0;	// Total blocks in the hoard
    for (LinearLRU_Index = 0;
      LinearLRU_Index < lruFileTableSize;
      LinearLRU_Index++)
	{
	File* file = lruFileTable[LinearLRU_Index];
	if (!file->canHoard()  ||  file->willForget()  ||  file->isDirectory())
	    continue;
	unsigned long requiredSpace = file->countStorage(*this);
	if (totalBlocks + requiredSpace <= GlobalManager::hoardSize())
	    {
	    managerData(*file).store(*file);
	    totalBlocks += requiredSpace;
	    }
	else
	    {
	    if (verbose > 2  &&  totalBlocks < GlobalManager::hoardSize()
	      &&  requiredSpace >= GlobalManager::hoardSize() / 10)
		{
		cerr << "correlator:  skipping file " << LinearLRU_Index
		  << " ('" << file->encodedName(SO_MIME_ENCODE)
		  << "') because it won't fit, size "
		  << requiredSpace << " blocks\n";
		}
	    }
	}
    /*
     * Management is now complete.
     */
    if (verbose > 0)
	{
	InternalTimer managementEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = managementEndTime - managementStartTime;
	cerr << "correlator:  Linear LRU hoard contains " << totalBlocks
	  << " blocks\n";
	cerr << "correlator:  Linear LRU management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << managementEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    }
	}
    if (dumpStream != NULL)
	{
	File::resetIdNumbering();
	ShowOptions options = SO_FULL_FILE_INFORMATION | SO_MANAGER;
	if (encodeFilenames)
	    options |= SO_MIME_ENCODE;
#ifdef LRU_SHOW_ONLY_HOARDABLE
	for (LinearLRU_Index = 0;
	  LinearLRU_Index < lruFileTableSize;
	  LinearLRU_Index++)
	    {
	    lruFileTable[LinearLRU_Index]->setId();
	    lruFileTable[LinearLRU_Index]->show(*dumpStream, options, name);
	    }
#else // LRU_SHOW_ONLY_HOARDABLE
	for (FileTableIterator iter;  iter;  iter++)
	    {
	    File& file(*iter);
	    file.setId();
	    file.show(*dumpStream, options, name);
	    }
#endif // LRU_SHOW_ONLY_HOARDABLE
	}
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void LinearLRU_Manager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad LinearLRU_Manager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("managementInterval"))
	managementInterval = atoi(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized LinearLRU_Manager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void LinearLRU_Manager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "LinearLRU::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Read manager-specific instructions from an auxiliary file.
 *
 * Instructions for this manager are of the form "filename intercept
 * slope bound", where filename is a full pathname of a file (symlinks
 * are OK).  In the hoarding computation, the LRU age of the file is
 * multiplied by the slope, and then the intercept is added.  A slope
 * and intercept of 1.0 and 0.0, respectively, mean no transformation.
 * For both slope and intercept, smaller values make the file more
 * likely to remain hoarded.  Slopes should not be negative, but
 * negative intercepts are meaningful.
 *
 * If a command does not contain all three values, the slope is
 * assumed to be 1.0 and the bound is assumed to be DBL_MAX (i.e.,
 * no bound).
 */
void LinearLRU_Manager::readInstructions(
    const char*		fileName)	// Name of file to read from
    const
    {
    ByteSwappedStream instructionFile(fileName, ios::in,
      ByteSwappedStream::bigEnded);
    if (!instructionFile)
	{
	int saveErrno = errno;
	cerr << "Couldn't open linear LRU instruction file '"
	  << fileName << "': " << strerr(saveErrno) << '\n';
	return;
	}
    ParsedFields* fields = NULL;
    for (  ;
      (fields = new ParsedFields(instructionFile))  &&  *fields;
      delete fields)
	{
	int numFields = fields->count();
	if (numFields == 0  ||  (*fields)[0][0] == '#')
	    continue;
	else if (numFields < 2  ||  numFields > 4)
	    cerr << "correlator:  Linear LRU instruction must contain"
		" intercept, optional slope, and optional bound in '"
	      << *fields << "'\n";
	const char* expandedPath = File::tildeExpand((*fields)[0]);
	if (expandedPath == NULL)
	    cerr << "correlator:  Linear LRU manager can't expand ~user  "
		"notation in '"
	      << *fields << "'\n";
	else if (expandedPath[0] != '/')
	    cerr << "correlator:  Linear LRU manager instruction path name "
		"must be absolute in '"
	      << *fields << "'\n";
	else
	    {
	    File* file = File::findCanonical(expandedPath, false, true, NULL);
	    double intercept;
	    double slope = 1.0;
	    double bound = DBL_MAX;
	    intercept = atof((*fields)[1]);
	    if (numFields > 2)
		slope = atof((*fields)[2]);
	    if (numFields > 3)
		bound = atof((*fields)[3]);
	    managerData(*file).setLineParameters(slope, intercept, bound);
	    for (FileDescendantsIterator child(*file);  child;  ++child)
		managerData(*child).setLineParameters(slope, intercept, bound);
	    }
	}
    delete fields;
    }

/*
 * Get rid of manager-specific instructions for a file.
 */
void LinearLRU_Manager::clearInstructions(
    File&		file)		// File to fix
    const
    {
    managerData(file).setLineParameters(0.0, 1.0, DBL_MAX);
    }

/*
 * Return a reference to the linear LRU manager's file-private data,
 * so that manager routines can access it.
 */
LinearLRU_ManagerData& LinearLRU_Manager::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.linearLRU_ManagerData;
    }

const LinearLRU_ManagerData& LinearLRU_Manager::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.linearLRU_ManagerData;
    }

int LinearLRU_Manager::compareLRU(	// Compare for qsort of linear LRU
    class File**	file1,		// Files to compare
    class File**	file2)		// ...
    {
    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (file1 == file2)
	return 0;
    ReferenceCount now = linearLRU_Manager.currentReferenceTime();
    ReferenceCount referenceTime1 =
      linearLRU_Manager.managerData(**file1).lastReferenceTime();
    if ((*file1)->isOpen())
	referenceTime1 = now;
    ReferenceCount referenceTime2 =
      linearLRU_Manager.managerData(**file2).lastReferenceTime();
    if ((*file2)->isOpen())
	referenceTime2 = now;
    double age1 =
      (linearLRU_Manager.currentReferenceTime() - referenceTime1)
	* linearLRU_Manager.managerData(**file1).slope
	+ linearLRU_Manager.managerData(**file1).intercept;
    double bound = linearLRU_Manager.managerData(**file1).bound;
    if (age1 > bound)
	age1 = bound;
    double age2 =
      (linearLRU_Manager.currentReferenceTime() - referenceTime2)
	* linearLRU_Manager.managerData(**file2).slope
	+ linearLRU_Manager.managerData(**file2).intercept;
    bound = linearLRU_Manager.managerData(**file2).bound;
    if (age2 > bound)
	age2 = bound;
    return age1 > age2 ? 1
      : age1 < age2 ? -1
      : 0;
    }

LinearLRU_ManagerData::~LinearLRU_ManagerData()
    {
    }

/*
 * Return pointer to our manager
 */
HoardingManagerBase& LinearLRU_ManagerData::manager()
    const
    {
    return linearLRU_Manager;
    }

/*
 * Read binary version of manager data from stream
 */
ByteSwappedStream& LinearLRU_ManagerData::read(
    ByteSwappedStream&	stream,		// Stream to read data from
    File&		file)		// File we are a member of
    {
    if (logVersion > 0)
	{
	HoardingManagerDataBase::read(stream, file);
	stream.read(&intercept);
	stream.read(&slope);
	stream.read(&bound);
	}
    return stream;
    }

/* 
 * Write binary version of manager data to stream
 */
ByteSwappedStream& LinearLRU_ManagerData::write(
    ByteSwappedStream&	stream,		// Stream to write data to
    const File&		file)		// File we are a member of
    const
    {
    HoardingManagerDataBase::write(stream, file);
    stream.write(&intercept);
    stream.write(&slope);
    stream.write(&bound);
    return stream;
    }

void LinearLRU_ManagerData::inherit(	// Inherit information from parent
    const File&		,		// File inheriting information
    const File&		parent)		// Parent to inherit from
    {
    intercept = managerData(parent).intercept;
    slope = managerData(parent).slope;
    bound = managerData(parent).bound;
    }

/*
 * Write a printable version of the linear LRU manager data to a stream
 */
ostream& operator<<(
    ostream&		stream,		// Stream to write to
    const LinearLRU_ManagerData& data)	// Data to write
    {
    return data.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Write a printable version of the linear LRU manager data to a
 * stream, with options.
 *
 * Show options honored: none
 */
ostream& LinearLRU_ManagerData::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    stream << "LLH\t    LinearLRU_Manager: ";
    HoardingManagerDataBase::show(stream, options, managerName);
    stream << ' ' << intercept << ' ' << slope << ' ' << bound << '\n';
    return stream;
    }
 
/*
 * Return a reference to the linear LRU manager's file-private data,
 * so that manager routines can access it recursively.
 */
LinearLRU_ManagerData& LinearLRU_ManagerData::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.linearLRU_ManagerData;
    }

const LinearLRU_ManagerData& LinearLRU_ManagerData::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.linearLRU_ManagerData;
    }

WeightedLRU_Manager::WeightedLRU_Manager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : LRU_Manager(managerName, managerDefaults)
    {
    ManagerBase::setManagerParameters(defaults);
    }

WeightedLRU_Manager::~WeightedLRU_Manager ()
    {
    }

void WeightedLRU_Manager::manage(	// Manage the file hoard
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump results to
    bool		encodeFilenames) // True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    if (verbose > 1)
	cerr << "correlator:  Weighted LRU management began at "
	  << managementStartTime.ctime() << '\n';
    /*
     * Weighted LRU management is very simple.  We simply sort the
     * files into order by weighted access time, and then scan through
     * them until the hoard is full.  (The weighted access time is
     * simply the product of the access time and the weight.)
     */
    // NEEDSWORK:  should handle FORGOTTEN files, need other method
    lruFileTableSize = File::numberOfFiles();
    delete[] lruFileTable;
    lruFileTable = new File*[lruFileTableSize];
    unsigned long WeightedLRU_Index = 0;
    for (FileTableIterator iter;  iter;  ++iter)
	{
	if (WeightedLRU_Index >= lruFileTableSize)
	    correlatorPanic(
	      "correlator:  bad file count in Weighted LRU manager");
	File& file(*iter);
	if (file.alwaysHoardableType())
	    managerData(file).store(file);
	else
	    {
	    managerData(file).unstore(file);
	    if (simulationMode
	      ||  (file.canHoard()  &&  !file.willForget()
		&&  !file.isDirectory()))
		lruFileTable[WeightedLRU_Index++] = &file;
	    }
	}
    lruFileTableSize = WeightedLRU_Index; // Actual # files hoarded
    qsort((void*) lruFileTable, lruFileTableSize, sizeof lruFileTable[0],
      (int (*)(const void*, const void*)) WeightedLRU_Manager::compareLRU);
    /*
     * Scan the files, fitting in as many as possible.  Note that a
     * lower-priority file might be inserted if it will fit where the
     * higher-priority one will not.
     */
    unsigned long totalBlocks = 0;	// Total blocks in the hoard
    for (WeightedLRU_Index = 0;
      WeightedLRU_Index < lruFileTableSize;
      WeightedLRU_Index++)
	{
	File* file = lruFileTable[WeightedLRU_Index];
	if (!file->canHoard()  ||  file->willForget()  ||  file->isDirectory())
	    continue;
	unsigned long requiredSpace = file->countStorage(*this);
	if (totalBlocks + requiredSpace <= GlobalManager::hoardSize())
	    {
	    managerData(*file).store(*file);
	    totalBlocks += requiredSpace;
	    }
	else
	    {
	    if (verbose > 2  &&  totalBlocks < GlobalManager::hoardSize()
	      &&  requiredSpace >= GlobalManager::hoardSize() / 10)
		{
		cerr << "correlator:  skipping file " << WeightedLRU_Index
		  << " ('" << file->encodedName(SO_MIME_ENCODE)
		  << "') because it won't fit, size "
		  << requiredSpace << " blocks\n";
		}
	    }
	}
    /*
     * Management is now complete.
     */
    if (verbose > 0)
	{
	InternalTimer managementEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = managementEndTime - managementStartTime;
	cerr << "correlator:  Weighted LRU hoard contains " << totalBlocks
	  << " blocks\n";
	cerr << "correlator:  Weighted LRU management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << managementEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    }
	}
    if (dumpStream != NULL)
	{
	File::resetIdNumbering();
	ShowOptions options = SO_FULL_FILE_INFORMATION | SO_MANAGER;
	if (encodeFilenames)
	    options |= SO_MIME_ENCODE;
#ifdef LRU_SHOW_ONLY_HOARDABLE
	for (WeightedLRU_Index = 0;
	  WeightedLRU_Index < lruFileTableSize;
	  WeightedLRU_Index++)
	    {
	    lruFileTable[WeightedLRU_Index]->setId();
	    lruFileTable[WeightedLRU_Index]->show(*dumpStream, options, name);
	    }
#else // LRU_SHOW_ONLY_HOARDABLE
	for (FileTableIterator iter;  iter;  iter++)
	    {
	    File& file(*iter);
	    file.setId();
	    file.show(*dumpStream, options, name);
	    }
#endif // LRU_SHOW_ONLY_HOARDABLE
	}
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void WeightedLRU_Manager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad WeightedLRU_Manager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("managementInterval"))
	managementInterval = atoi(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized WeightedLRU_Manager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void WeightedLRU_Manager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "WeightedLRU::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Read manager-specific instructions from an auxiliary file.
 *
 * Instructions for this manager are of the form "filename weight",
 * where filename is a full pathname of a file (symlinks are OK), and
 * weight is a floating-point value giving the weight to assign to the
 * file (lower numbers make the file more likely to be cached, with
 * 1.0 being neutral).
 */
void WeightedLRU_Manager::readInstructions(
    const char*		fileName)	// Name of file to read from
    const
    {
    ByteSwappedStream instructionFile(fileName, ios::in,
      ByteSwappedStream::bigEnded);
    if (!instructionFile)
	{
	int saveErrno = errno;
	cerr << "Couldn't open weighted LRU instruction file '"
	  << fileName << "': " << strerr(saveErrno) << '\n';
	return;
	}
    ParsedFields* fields = NULL;
    for (  ;
      (fields = new ParsedFields(instructionFile))  &&  *fields;
      delete fields)
	{
	if (fields->count() == 0  ||  (*fields)[0][0] == '#')
	    continue;
	else if (fields->count() != 2)
	    cerr << "correlator:  Weighted LRU instruction must contain"
		" a weight in '"
	      << *fields << "'\n";
	const char* expandedPath = File::tildeExpand((*fields)[0]);
	if (expandedPath == NULL)
	    cerr << "correlator:  Weighted LRU manager can't expand ~user  "
		"notation in '"
	      << *fields << "'\n";
	else if (expandedPath[0] != '/')
	    cerr << "correlator:  Weighted LRU manager instruction path name "
		"must be absolute in '"
	      << *fields << "'\n";
	else
	    {
	    File* file = File::findCanonical(expandedPath, false, true, NULL);
	    double weight = atof((*fields)[1]);
	    managerData(*file).setWeight(weight);
	    for (FileDescendantsIterator child(*file);  child;  ++child)
		managerData(*child).setWeight(weight);
	    }
	}
    delete fields;
    }

/*
 * Get rid of manager-specific instructions for a file.
 */
void WeightedLRU_Manager::clearInstructions(
    File&		file)		// File to fix
    const
    {
    managerData(file).setWeight(1.0);
    }

/*
 * Return a reference to the weighted LRU manager's file-private data,
 * so that manager routines can access it.
 */
WeightedLRU_ManagerData& WeightedLRU_Manager::managerData(
    File&		file)
    const
    {
    return file.weightedLRU_ManagerData;
    }

const WeightedLRU_ManagerData& WeightedLRU_Manager::managerData(
    const File&		file)
    const
    {
    return file.weightedLRU_ManagerData;
    }

int WeightedLRU_Manager::compareLRU(	// Compare for qsort of weighted LRU
    class File**	file1,		// Files to compare
    class File**	file2)		// ...
    {
    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (file1 == file2)
	return 0;
    ReferenceCount now = weightedLRU_Manager.currentReferenceTime();
    ReferenceCount referenceTime1 =
      weightedLRU_Manager.managerData(**file1).lastReferenceTime();
    if ((*file1)->isOpen())
	referenceTime1 = now;
    ReferenceCount referenceTime2 =
      weightedLRU_Manager.managerData(**file2).lastReferenceTime();
    if ((*file2)->isOpen())
	referenceTime2 = now;
    double age1 =
      (weightedLRU_Manager.currentReferenceTime() - referenceTime1)
	* weightedLRU_Manager.managerData(**file1).weight;
    double age2 =
      (weightedLRU_Manager.currentReferenceTime() - referenceTime2)
	* weightedLRU_Manager.managerData(**file2).weight;
    return age1 > age2 ? 1
      : age1 < age2 ? -1
      : 0;
    }

WeightedLRU_ManagerData::~WeightedLRU_ManagerData()
    {
    }

/*
 * Return pointer to our manager
 */
HoardingManagerBase& WeightedLRU_ManagerData::manager()
    const
    {
    return weightedLRU_Manager;
    }

/*
 * Read binary version of manager data from stream
 */
ByteSwappedStream& WeightedLRU_ManagerData::read(
    ByteSwappedStream&	stream,		// Stream to read data from
    File&		file)		// File we are a member of
    {
    HoardingManagerDataBase::read(stream, file);
    stream.read(&weight, sizeof weight);
    return stream;
    }

/* 
 * Write binary version of manager data to stream
 */
ByteSwappedStream& WeightedLRU_ManagerData::write(
    ByteSwappedStream&	stream,		// Stream to write data to
    const File&		file)		// File we are a member of
    const
    {
    HoardingManagerDataBase::write(stream, file);
    stream.write(&weight, sizeof weight);
    return stream;
    }

void WeightedLRU_ManagerData::inherit(	// Inherit information from parent
    const File&		,		// File inheriting information
    const File&		parent)		// Parent to inherit from
    {
    weight = managerData(parent).weight;
    }

/*
 * Write a printable version of the weighted LRU manager data to a stream
 */
ostream& operator<<(
    ostream&		stream,		// Stream to write to
    const WeightedLRU_ManagerData& data)	// Data to write
    {
    return data.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Write a printable version of the weighted LRU manager data to a
 * stream, with options.
 *
 * Show options honored: none
 */
ostream& WeightedLRU_ManagerData::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    stream << "WLH\t    WeightedLRU_Manager: ";
    HoardingManagerDataBase::show(stream, options, managerName);
    stream << ' ' << weight << '\n';
    return stream;
    }

/*
 * Return a reference to the weighted LRU manager's file-private data,
 * so that manager routines can access it recursively.
 */
WeightedLRU_ManagerData& WeightedLRU_ManagerData::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.weightedLRU_ManagerData;
    }

const WeightedLRU_ManagerData& WeightedLRU_ManagerData::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.weightedLRU_ManagerData;
    }
