#ifndef lint
static const char Rcs_Id[] =
    "$Id: observer.cc,v 1.35 2000/01/03 21:56:50 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: observer.cc,v $
 * Revision 1.35  2000/01/03 21:56:50  geoff
 * Add a typecast to the output of an enum to silence a warning.
 *
 * Revision 1.34  1998/06/02 23:59:46  geoff
 * Add the -n (avoid filesystem calls) switch.  This is incompletely
 * implemented, because exec checks still use the filesystem.
 *
 * Revision 1.33  1998/04/09 00:33:02  geoff
 * Use the new notation for hash and list iterators.
 *
 * Revision 1.32  1998/01/31 05:38:53  geoff
 * Add some limited usages of simple strings, mostly as a way to test
 * that class.
 *
 * Revision 1.31  1997/12/18 01:25:15  geoff
 * Get rid of some obsolete development ifdefs.  Move some ifdefs to
 * compiloptions.hh.  Make some function arguments const char* instead of
 * char*.  Make tmp_mnt be a define (is it even needed as a built-in any
 * more?)
 *
 * Revision 1.30  1997/12/04 23:16:17  geoff
 * Add observerPanic, which closes the /dev/seer fd before panicking, and
 * use it to do all aborts.  Arrange to pause on panics.
 *
 * Revision 1.29  1997/07/14 21:43:36  geoff
 * Fix a place where a null file descriptor could be closed.  Take out
 * the inquiry of dtablesize, and replace it with dynamic management of
 * the descriptor table, all because it turns out that some systems use
 * INT_MAX as the maximum descriptor table size (I'd like to see somebody
 * try to allocate *that* many descriptors!
 *
 * Revision 1.28  1997/07/11 17:40:58  geoff
 * Support limited replay of Web traces.
 *
 * Revision 1.27  1997/03/13 23:00:16  geoff
 * Add the 'q' flag.
 *
 * Revision 1.26  1996/10/11 22:09:09  geoff
 * Add code to automatically check the save file to be sure that it's all
 * valid, and to truncate any garbage data found.
 *
 * Revision 1.25  1996/10/04 07:12:20  geoff
 * Add support for command-interpreter references, detecting script execs
 * and extracting the #! information.  Don't ever send process-exit
 * records for process 1 (this can happen if process 1 gets into the
 * table for some reason and then we see a restart record while replaying
 * a file).
 *
 * Revision 1.24  1996/09/23 23:41:27  geoff
 * Don't reset the byte-swap value before every save file.  This keeps us
 * from being able to replay dissimilar save files from a single
 * observer, which is a minor loss, but fixes a bug introduced by the
 * automatic save-file switching code.  Also put the buffer version in
 * the timestamp record, so that it can be used as an alternate header,
 * which is also necessitated by the automatic-switching code.
 *
 * Revision 1.23  1996/09/13 18:01:28  geoff
 * Change the save-file-switching code to base its decision on file size,
 * rather than date.  As part of this, just use decimal numbering rather
 * than a time-based scheme.  Remember to pass the save file name to the
 * new process when re-executing.
 *
 * Revision 1.22  1996/09/11 23:00:12  geoff
 * Add REPORT_SELECTED_FAILURES support.  Add code to automatically
 * switch save files once a month and gzip the old one.  Don't try to
 * open the seer device until just before entering the observation loop.
 * Add TIMEVALS_ARE_INTS support.
 *
 * Revision 1.21  1996/08/02  21:31:36  geoff
 * After catching USR1, reset the handler
 *
 * Revision 1.20  1996/06/11  07:05:01  geoff
 * If restart_self fails, clear do_reexec so we don't loop forever.
 *
 * Revision 1.19  1996/06/11  06:13:24  geoff
 * Change when absolutePath is called in addfileref, to avoid core dumps
 * due to having a static clobbered.
 *
 * Revision 1.18  1996/05/03  01:12:48  geoff
 * Update the copyright.  Change the declaration and invocation of
 * showent to support the (here unused) new parameter.
 *
 * Revision 1.17  1996/03/19  18:43:10  geoff
 * Set the call time properly in restart records.  Use canonicalize_pn to
 * clean up CWDs and roots, so that ".." references will be properly
 * removed.  Add utime/utimes support.  Only get rid of Ficus replica
 * ID's, etc., if FICUS_KNOWLEDGE is enabled.  Add code to prevent
 * too-long path names.
 *
 * Revision 1.16  1996/02/02  20:09:52  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.15  1996/02/02  18:29:14  geoff
 * Undefine and redefine NULL to get around a stupid g++ header file on
 * Linux.  Fix some minor problems detected by the new C++ compiler.
 * Ensure that every NEWPROC generated has a matching PROCEXIT (this is
 * done by creating a dummy process entry).  Fix some serious bugs in the
 * way CWD tracking worked on Linux (it determined the process' new
 * directory, and then issued a chdir relative to that rather than
 * relative to the old one).
 *
 * Revision 1.14  1995/12/19  01:03:06  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.13  1995/12/19  00:34:45  geoff
 * Add process-tracking support, including compatibity code for old save
 * files.  When a process does a long string of relative chdirs, make
 * sure that the CWD path doesn't become excessively long.  Fix a bug
 * that crashed the observer when the CWD overflowed.
 *
 * Revision 1.12  1995/11/22  00:22:06  geoff
 * Get rid of the TIME_SORTED_INPUT ifdef.
 *
 * Revision 1.11  1995/11/10  02:59:55  geoff
 * Add zzinlinefunctions.hh.  Add TIME_SORTED_INPUT support.  Add Linux
 * code to find out a process' CWD from /proc.
 *
 * Revision 1.10  1995/10/10  20:38:13  geoff
 * Add further Linux support, in particular modifying the chdir and
 * chroot tracking so that it actually works correctly.  Add support for
 * automating byte-swapping of the Seer save file.  Range-check all fd's
 * supplied by the kernel, just to be sure we don't overwrite the fd
 * array.  Add a few new Linux-specific system calls.
 *
 * Revision 1.9  1995/09/15  00:06:10  geoff
 * Change the RCS ID to be a const.  Add death catching and smooth
 * cleanup.  Add proper bracketing to all switch statements.  Add support
 * for multiple and compressed replay files.  Clean up memory on exit to
 * help catch leaks.  Do a better job of handling open flags.  Re-swap
 * arguments before writing to the save file, so that the save file is
 * canonical.  Add lots of Linux support.
 *
 * Revision 1.8  1995/08/07  20:03:03  geoff
 * Add TRACK_CWD support.  Add support for tracking fcntl calls,
 * including close-on-exec support.  Remove the truncreplay option.  Use
 * seertell to find out how many bytes have been replayed.
 *
 * Revision 1.7  1995/06/30  19:04:46  geoff
 * Add support for getting timestamps from the kernel and passing them on
 * to the correlator.  Put a data-structure version in the restart
 * message.  Add SUPPRESS_FFSH and STRIP_FICUS_QUALIFIERS.  Add symlink
 * support.
 *
 * Revision 1.6  1995/06/17  00:11:51  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.  Flush cerr/cout before all aborts.  Use symbolic values
 * for all system call numbers.
 *
 * Revision 1.5  1995/05/23  00:40:12  geoff
 * Change an exit to an abort, clean up a trivial formatting error, and
 * add a test condition to handle non-null server names that are still
 * local.
 *
 * Revision 1.4  1995/05/05  22:41:58  geoff
 * Add TRIM_DOUBLE_SLASH.  Add byte-swapping support of the replay file
 * for 386 boxes (but only on replay, so it's a major kludge).
 *
 * Revision 1.3  1995/04/20  21:58:19  geoff
 * Only canonicalize paths that are going to remote correlators;
 * otherwise let the correlator do it.  Fix a bug in re-execution
 * argument checking.  Use the new enums in calls to connectServer.
 * Remove AFTER_EXEC.  Add code to detect disappearing fd's.
 * Fix a minor bug in the re-execution message.
 *
 * Revision 1.2  1995/03/29  21:04:48  geoff
 * Change replay-reporting times downward.  Add the new argument to
 * connectServer.  Remember to flush cerr in a couple of important
 * places.  Add debugging code to try to catch whoever is swallowing
 * fd's.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "fields.hh"
#include "hashtable.hh"
#include "observer.hh"
#include "panic.hh"
#include "simplestring.hh"
#include "zzinlinefunctions.hh"
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <fstream.h>
extern "C"
    {
#include <signal.h>
    }
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
extern "C"
    {
#include <seer/seer.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>
    }

/*
 * Some stupid include files define NULL as a pointer to void*, which
 * is incorrect and causes compilation errors.  This is needed to work
 * around the problem.
 */
#undef NULL
#define NULL 0

#define SEERDEV		"/dev/seer"	// Device to open to get seer data

/*
 * Structure used to keep track of open files
 */
typedef struct
    {
    const char*		path;		// Pathname of the file
    int			refcount;	// Number of references
    }
			openfile_t;

/*
 * Structure used to keep track of active processes
 */
typedef struct
    {
    pid_t		pid;		// Unix process ID
    uid_t		uid;
    char*		lastexec;	// Last path exec was tried on
    openfile_t*		procname;	// Pathname of running process
#ifdef TRACK_CWD
    openfile_t*		cwd;		// Path of current working directory,
					// ..kept relative to currentRoot,
					// ..rather than to / 
    openfile_t*		currentRoot;	// Process' current root directory
					// ..non-NULL only if  not "/"
#endif // TRACK_CWD
    openfile_t**	dtable;		// File-descriptor table
    int			dtableSize;	// Size of file-descriptor table
    unsigned char*	closeOnExecFlags; // Per-descriptor close-on-exec flags
					// ..one per byte
    }
			proc_t;

#define DTABLE_INCREMENT 20		// Increment for expanding dtable

/*
 * 
 * Structure used to keep track of connections to servers
 */
typedef struct
    {
    uid_t		uid;		// UID to send to this server
    char*		servername;	// Name of server to connect to
    FILE*		file;		// File to send on
    long		retrytime;	// Time to retry connection
    int			blocked;	// NZ if server output is blocked
    }
			server_t;

int			main(int argc, char* argv[]);
static void		usage(void);
void			nomem(const char* where, int line, unsigned int size);
void			observerPanic(const char* message);
static int		openseer(void);
static void		catchusr1(int sig);
static void		catchdeath(int sig);
static void		cleanSaveFile(SimpleString& saveName);
static void		replaydata(SimpleString& replayfname);
static void		webPlay(void);	// Replay web data
static void		procexit(proc_t* procent, struct timeval* timestamp);
static void		observedata(void);
static void		servertrouble(server_t* server);
static void		closeserver(server_t* server);
static void		restart_self(void);
static int		process_entry(seerbuf_t* seerp);
static proc_t*		findproc(pid_t pid, uid_t uid, bool create);
static bool		expand_dtable(proc_t* procent, int desiredFd);
static char*		findarg(seerbuf_t* seerp, int argno);
static int		recordNewProcess(uid_t uid, pid_t pid, pid_t parentPid,
			  struct timeval* timestamp);
static int		recordReference(proc_t* procent, const char* path1,
			  const char* path2, bool chrootApplies,
			  ReferenceType type,
			  unsigned int errcode, struct timeval* timestamp);
static int		sendPacket(ReferenceType type, uid_t uid,
			  char* packetBuffer, int packetLength,
			  bool& fakeExit);
static int		packetize_path(const char* path, char* buf,
			  proc_t* procent,
			  bool canonicalize, bool chrootApplies,
			  struct timeval* timestamp);
static void		sendControl(ConnectionControl command,
			  struct timeval* timestamp);
					// Send a dummy control packet
static unsigned int	pidhash(proc_t* procent);
static int		pidcmp(proc_t* a, proc_t* b);
static unsigned int	filehash(openfile_t* fileent);
static int		filecmp(openfile_t* a, openfile_t* b);
static void		canonicalize_pn(const char* path, char* result,
			  proc_t* procent, bool chrootApplies,
			  struct timeval* timestamp);
static const char*	absolutePath(char* target, const char* source,
			  proc_t* procent, bool mustCopy,
			  struct timeval* timestamp);
static int		addfileref(proc_t* process, const char* path,
			  bool chrootApplies, openfile_t** tableent,
			  struct timeval* timestamp);
static int		delfileref(proc_t* procent, openfile_t** tableent,
			  int sendit, struct timeval* timestamp);
static bool		remove_ref(proc_t* procent, openfile_t* refent,
			  struct timeval* timestamp);
#ifdef TRACK_CWD
static void		setCWD(proc_t* process, struct timeval* timestamp);
					// Set a process' CWD
#endif // TRACK_CWD

/*
 * Time to wait between buffer flushes to servers, in seconds
 */
#define FLUSHTIME	5

/*
 * Time to wait before retrying server connections
 */
#define RETRYTIME	30

/*
 * Time to wait between recording timestamps
 */
#define STAMPTIME	600

/*
 * Time to wait between replay reports
 */
#define REPORTTIME	600

/*
 * Number of replay records to go through before checking report time
 */
#define REPORTCOUNT	5000

/*
 * Maximum size of a save file, before switching to a new one.  It's a
 * bit under a power of two because we don't check the size after ever
 * record we write.
 */
#define MAX_SAVE_SIZE	(10*1000*1000)

/*
 * Command to use for compressing save files.  A trailing space is required.
 */
#define GZIPIT		"gzip -9 "

/*
 * "Magic number" that indicates script interpreters
 */
#define SCRIPT_MAGIC	"#!"

/*
 * Check for both EAGAIN and EWOULDBLOCK, because some supposedly POSIX
 * systems are broken and return EWOULDBLOCK when they should return EAGAIN
 *
 * (stolen from the X server)
 */
#if defined(EAGAIN) && defined(EWOULDBLOCK)
#define ETEST(err) (err == EAGAIN || err == EWOULDBLOCK)
#else
#ifdef EAGAIN
#define ETEST(err) (err == EAGAIN)
#else
#define ETEST(err) (err == EWOULDBLOCK)
#endif
#endif

/*
 * NEEDSWORK. SunOS 4.1.1 declares signal and SIG_IGN in a fashion
 * that causes warning messages from g++.  The following casts
 * suppress the warnings.
 */
#ifdef __sparc__
#define SIGNAL_CLEANUP_CAST(x)	((void*)(x))
#else // __sparc__
#define SIGNAL_CLEANUP_CAST(x)	x
#endif // __sparc__

static int		avoidFilesystem = 0; // NZ to avoid accessing the
					// ..filesystem as part of
					// ..handling trace records
					// ..(only active during
					// ..replay mode)
static int		do_reexec = 0;	// NZ to re-execute ourselves
HashTable		filetable(1000,
			  (unsigned int (*)(const void*)) filehash,
			  (int (*)(const void*, const void*)) filecmp);
char			ourWD[MAXPATHLEN]; // Our original working directory
static HashTable	proctable(500,	// Process tracking table
			  (unsigned int (*)(const void*)) pidhash,
			  (int (*)(const void*, const void*)) pidcmp);
SimpleString		progname;	// Name we were run under
static int		needflush = 0;	// NZ if buffers need flushing
static int		nservers = 0;	// Number of servers in table
static FILE*		replayfile = NULL; // File to replay input from
FILE*			savefile = NULL; // File to save logs to
WritableSimpleString	savefname;	// Name of file to save to
int			savelen = 0;	// Length of savefname
unsigned long		savesize = 0;	// Size of save file
static int		seerfd;		// FD to use to access /dev/seer
static server_t*	servers = NULL;	// Servers to communicate with
static int		terminate_observer = 0; // NZ to exit observer cleanly
int			verbose = 0;	// NZ to dump info to stdout

extern seerbuf_t*	readent(FILE* replayfile);
extern int		seerclose(FILE* replayFile);
extern FILE*		seeropen(const char* fileName);
extern unsigned long	seertell(FILE* replayFile);
extern void		seertruncate(const char* fileName);
extern int		seerversion();
extern void		setseerswap(bool needToSwap);
extern void		showent(seerbuf_t* seerp, const char* procNames[],
			  int maxNames);
#ifndef SEER_BLOCKING
extern void		setnonblock(int fd);
#endif // SEER_BLOCKING
extern char*		strdup(const char* string, int size);
extern "C" char*	strerr(int errnum);

int main(			// Observe kernel trace data
    int			argc,	// Argument count
    char*		argv[])	// Argument vector
    {
    char*		colonp;	// Pointer to colon in UID/server arg
    seerbuf_t		dummyent; // Dummy entry for marker
    int			fd;	// Fd current server is connected on
    int			numReplayFiles = 0; // Number of files to replay
    int			quitAfterReplay = 0; // NZ to exit after replay done
    int			reexec = 0; // NZ if reexec happened
    SimpleString*	replayFileNames = NULL;
				// Names of files to replay from
    int			savefd;	// FD save file is open on
    int			serverno; // Current server number
    bool		webTrace = false; // True to do web trace from stdin

    progname = argv[0];
    pauseOnPanics();
    if (getcwd(ourWD, sizeof ourWD) == NULL)
	(void)strcpy(ourWD, "/tmp");
    /*
     * We assume that every argument pair specifies a UID/server
     * combination.  This causes overkill in memory allocation, but
     * the wastage isn't enough to worry about.
     */
    servers = (server_t*) malloc((argc / 2) * sizeof(server_t));
    if (servers == NULL)
	nomem("main", __LINE__, (argc / 2) * sizeof(server_t));

    for (  ;  argc > 1  &&  argv[1][0] == '-';  argc--, argv++)
	{
	switch (argv[1][1])
	    {
	    case 'n':			// Avoid using the filesystem
		{
		avoidFilesystem = 1;
		break;
		}
	    case 'q':			// Exit after replay finishes
		{
		quitAfterReplay = 1;
		break;
		}
	    case 'R':			// Re-execution call
		{
		reexec = 1;
		/*
		 * Reexec arguments:
		 *
		 * seerfd savefd
		 */
		argc -= 2;
		if (argc != 2)
		    usage();
		++argv;
		seerfd = atoi(argv[1]);
		++argv;
		savefd = atoi(argv[1]);
		++argv;
		if (savefd >= 0)
		    {
		    savefname = argv[1];
		    savefile = fdopen(savefd, "a");
		    savesize = lseek(savefd, 0L, 1);
		    }
		break;
		}
	    case 'r':			// Specify replay file
		{
		if (--argc <= 1)
		    usage();
		++argv;
		if (replayFileNames == NULL)
		    {
		    /*
		     * We'll assume that all following arguments are
		     * -r options.  This may cause us to over-allocate
		     * the replayFileNames array, but the space wasted
		     * is negligible.
		     */
		    replayFileNames = new SimpleString[argc / 2];
		    if (replayFileNames == NULL)
			nomem("main", __LINE__,
			  sizeof(SimpleString) * argc / 2);
		    }
		replayFileNames[numReplayFiles++] = argv[1];
		break;
		}
	    case 's':			// Specify save file
		{
		if (--argc <= 1)
		    usage();
		++argv;
		savefname = argv[1];
		break;
		}
	    case 'u':			// Record UID to specified server
		{
		if (--argc <= 1)
		    usage();
		++argv;
		colonp = argv[1];
		servers[nservers].file = NULL;
		if (*colonp == '!')	// FD given on reexec
		    {
		    servers[nservers].file = fdopen(atoi(argv[1]), "w");
		    colonp = strchr(argv[1], ':');
		    if (colonp == NULL)
			usage();
		    ++colonp;
		    }
		servers[nservers].uid = atoi(colonp);
		colonp = strchr(colonp, ':');
		if (colonp == NULL)
		    servers[nservers].servername = NULL;
		else
		    servers[nservers].servername = colonp + 1;
		servers[nservers].retrytime = 0;
		servers[nservers].blocked = 0;
		nservers++;
		break;
		}
	    case 'v':			// Run in verbose mode
		{
		if (argv[1][2] == '\0')
		    verbose++;
		else
		    verbose = atoi(&argv[1][2]);
		break;
		}
	    case 'w':			// Replay Web trace from stdin
		{
		webTrace = true;
		break;
		}
	    default:
		{
		break;
		}
	    }
	}
    if (argc != 1)
	usage();
    (void) signal(SIGUSR1, catchusr1);
    (void) signal(SIGPIPE, SIG_IGN);
    for (serverno = nservers;  --serverno >= 0;  )
	{
	if (servers[serverno].file == NULL)
	    {
#ifdef SEER_BLOCKING
	    fd = connectServer(servers[serverno].servername,
	      OBSERVER_CONNECTION, blockingConnection, reportFailures);
#else // SEER_BLOCKING
	    fd = connectServer(servers[serverno].servername,
	      OBSERVER_CONNECTION,
	      numReplayFiles ? blockingConnection : nonBlockingConnection,
	      reportFailures);
#endif // SEER_BLOCKING
	    if (fd < 0)
		{
		servers[serverno].file = NULL;
		servers[serverno].retrytime = time((long*) NULL) + RETRYTIME;
		}
	    else
		{
		servers[serverno].file = fdopen(fd, "w");
		servers[serverno].retrytime = 0;
		servers[serverno].blocked = 0;
		}
	    }
	}
    for (serverno = nservers;  --serverno >= 0;  )
	{
	if (servers[serverno].file != NULL)
	    break;
	}
    if (serverno < 0)
	cerr << "observer:  no servers at startup!\n";
    if (webTrace)
	{
	struct timeval controlTime = {0, 0};
	sendControl(replayMode, &controlTime);
#if 0
	// WEBWORK: need to enable modes
	sendControl(sizeMode, &controlTime);
	sendControl(webMode, &controlTime);
#endif
	time_t now;
	(void) time(&now);
	cerr << "observer: web replay begun at " << ctime(&now);
	webPlay();
	for (serverno = nservers;  --serverno >= 0;  )
	    {
	    if (servers[serverno].file != NULL)
		(void) fflush(servers[serverno].file);
	    }
	(void) time(&now);
	cerr << "observer: web replay complete, exiting at " << ctime(&now);
	return 0;
	}
    else if (numReplayFiles != 0)
	{
	struct timeval controlTime = {0, 0};
	sendControl(replayMode, &controlTime);
#ifdef __i386__
	setseerswap(true);
#else // __i386__
	setseerswap(false);
#endif // __i386__
	for (int replayIndex = 0;
	  replayIndex < numReplayFiles;
	  replayIndex++)
	    replaydata(replayFileNames[replayIndex]);
	delete[] replayFileNames;
	for (serverno = nservers;  --serverno >= 0;  )
	    {
	    if (servers[serverno].file != NULL)
		{
		(void) fflush(servers[serverno].file);
#ifndef SEER_BLOCKING
		setnonblock(fileno(servers[serverno].file));
#endif // SEER_BLOCKING
		}
	    }
	}
    if (SIGNAL_CLEANUP_CAST(signal(SIGINT, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGINT, catchdeath);
    if (SIGNAL_CLEANUP_CAST(signal(SIGHUP, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGHUP, catchdeath);
    if (SIGNAL_CLEANUP_CAST(signal(SIGTERM, SIG_IGN))
      != SIGNAL_CLEANUP_CAST(SIG_IGN))
	(void)signal(SIGTERM, catchdeath);
    if (savefname != NULL)
	{
	cleanSaveFile(savefname);
	savelen = savefname.strlen();
	if (!reexec)
	    savefile = fopen(savefname, "a");
	if (savefile == NULL)
	    {
	    int saveErrno = errno;
	    cerr << "observer:  Couldn't open save file '" << savefname
	      << "': " << strerr(saveErrno) << '\n';
	    }
	}
    setseerswap(false);
    if (savefile != NULL)
	{
	/*
	 * Write a dummy buffer entry to indicate restart.  Note that
	 * we assume that sizeof(time_t) is the same as sizeof(int).
	 *
	 * The time given in this record isn't necessarily the same as
	 * that in the first packet we'll write to the save file.  In
	 * fact, since the kernel probably has some old entries
	 * buffered up, the first packet will probably have a time
	 * earlier than this timestamp.
	 */
	dummyent.callcode = reexec;
	dummyent.errcode = SEERBUF_VERSION;
	dummyent.flags = SEER_RESTART;
	dummyent.argsize = 1;
	dummyent.uid = 0;
	dummyent.pid = 0;
	dummyent.retval = 0;
	(void)gettimeofday(&dummyent.calltime, (struct timezone*)NULL);
	dummyent.args[0] = dummyent.calltime.tv_sec;
	(void) fwrite(&dummyent, sizeof dummyent, 1, savefile);
	}
    struct timeval controlTime;
    (void)gettimeofday(&controlTime, (struct timezone*)NULL);
    sendControl(realTimeMode, &controlTime);
    if (quitAfterReplay)
	{
	time_t now;
	(void) time(&now);
	cerr << "observer: replay complete, exiting at " << ctime(&now);
	return 0;
	}
    avoidFilesystem = 0;
    if (!reexec)
	{
	if (openseer())
	    return 1;
	}
    observedata();
    if (savefile != NULL)
	(void)fclose(savefile);
    /*
     * Clean out the process table so that the malloc-debug code will
     * be able to detect other non-freed areas.
     */
    struct timeval exit_time;
    (void)gettimeofday(&exit_time, (struct timezone*)NULL);
    for (HashTable::Iterator iter(proctable);  iter;  ++iter)
	{
	procexit ((proc_t*)proctable[iter], &exit_time);
	iter.removeCurrent();
	}
    return 0;
    }

static void usage()		// Issue a usage message
    {

    cerr << "Usage:  observer [-n] [-q] [-s file] [-r file...] [-u uid:name:port]"
      " [-v] [-R args]\n";
    exit(1);
    }

void nomem(			// Issue an out-of-memory message
    const char*		where,	// Routine that failed to allocate
    int			line,	// Line number of failure
    unsigned int	size)	// Size that was wanted
    {

    cerr << "observer:  out of memory at " << line << " in " << where;
    if (size)
	cerr << ", couldn't get " << size << " bytes";
    observerPanic("!");
    }

void observerPanic(		// Panic the observer, cleanly
    const char*		message) // Message to issue on panic
    {
    /*
     * Close the Seer file descriptor so that the kernel doesn't hang
     * when the panic pauses.  Then use the standard panic routine.
     */
    close(seerfd);
    panic(message);
    }

static int openseer()		// Open the seer device
    {

    seerfd = open(SEERDEV, 0);
    if (seerfd < 0)
	{
	int saveErrno = errno;
	cerr << "observer: couldn't open " << SEERDEV << ": "
	  << strerr(saveErrno) << '\n';
	return 1;
	}
    return 0;
    }

static void catchusr1(		// Catch USR1 signals
    int			)	// Signal we caught
    {

    do_reexec = 1;
    (void) signal(SIGUSR1, catchusr1);
    }

static void catchdeath(		// Catch signals intended to cause us to exit
    int			)	// Signal we caught
    {

    terminate_observer = 1;
    }

static void cleanSaveFile(		// Make sure save file has valid data
    SimpleString&	saveName)	// Name of save file to clean
    {
    FILE* saveFile = seeropen(saveName);
    if (saveFile != NULL)
	{
	while (readent(saveFile))
	    ;
	(void)seerclose(saveFile);
	}
    seertruncate(saveName);
    }

static void replaydata(			// Replay previously-saved trace data
    SimpleString&	replayfname)	// Name of file to replay from
    {
    static unsigned long totalBytes = 0; // Total bytes replayed
    replayfile = seeropen(replayfname);
    if (replayfile == NULL)
	{
	int saveErrno = errno;
	cerr << "observer:  Couldn't open replay file '" << replayfname
	  << "': " << strerr(saveErrno) << '\n';
	exit(1);
	}
    time_t now;
    (void) time(&now);
    cerr << "observer:  replay of " << replayfname << " started at "
      << ctime(&now);
    cerr.flush();
    seerbuf_t* seerp = NULL;
    time_t nextReport = now + REPORTTIME;
    for (long recordCount = 0;
      (seerp = readent (replayfile)) != NULL;
      recordCount++)
	{
	if (recordCount % REPORTCOUNT == 0)
	    {
	    (void) time(&now);
	    if (now >= nextReport)
		{
		nextReport = now + REPORTTIME;
		cerr << "observer:  " << recordCount << " records, "
		  << seertell(replayfile) << " bytes replayed at "
		  << ctime(&now);
		}
	    }
	/*
	 * On every restart of the observer, assume that all
	 * existing processes have exited.  Don't do this for
	 * reexecs, however.
	 */
	if (seerp->flags & SEER_RESTART)
	    {
	    if (seerp->callcode == 0)
		{
		/*
		 * Restart that wasn't a reexec.  Walk the process
		 * table to close all files.
		 */
		for (HashTable::Iterator iter(proctable);  iter;  ++iter)
		    {
		    procexit ((proc_t*) proctable[iter], &seerp->calltime);
		    iter.removeCurrent();
		    }
		}
	    }
	else if ((seerp->flags & SEER_TIMESTAMP) == 0)
	    (void) process_entry(seerp);
	}
    (void) time(&now);
    cerr << "observer:  replay of " << seertell(replayfile) - totalBytes
      << " bytes from " << replayfname << " (" << seertell(replayfile)
      << " total) complete at " << ctime(&now);
    cerr.flush();
    totalBytes = seertell(replayfile);
    (void)seerclose(replayfile);
    replayfile = NULL;
    }

static void webPlay()			// Replay web data from stdin
    {
    ParsedFields* fields = NULL;
    while (1)
	{
	delete fields;
	fields = new ParsedFields(cin, " \t", ParsedFields::RUNS, 6);
	if (!*fields)
	    {
	    delete fields;
	    break;
	    }
	if (fields->count() < 6  ||  (*fields)[0][0] == '#')
	    continue;
	ReferenceType type = SEER_REF_READ;
	if (strcmp((*fields)[0], "PUT") == 0)
	    type = SEER_REF_RW;
	struct timeval timeStamp = {atoi((*fields)[1]), 0};
	long IP_Addr = atoi((*fields)[2]);
#if 0
	// WEBWORK: need to communicate size.
	long size = atoi((*fields[3]));
	long pathLength = atoi((*fields[4]));
#endif // 0
	proc_t* procent = findproc(IP_Addr, 0, true);
	(void)recordReference(procent, (*fields)[5], NULL, false, type, 0,
	  &timeStamp);
	// WEBWORK: need to prune process table based on time
	}
    }

static void procexit(			// Handle process exit
    proc_t*		procent,	// Process that's exiting
    struct timeval*	timestamp)	// Timestamp to assign to the exit
    {
    int			fdno;		// Process fd being handled

    for (fdno = procent->dtableSize;  --fdno >= 0;  )
	(void) delfileref(procent, &procent->dtable[fdno], 1, timestamp);
    if (procent->procname == NULL)
	(void) recordReference(procent, NULL, NULL, false,
	  SEER_REF_EXIT, 0, timestamp);
    else
	{
	(void) recordReference(procent, procent->procname->path,
	  NULL, false, SEER_REF_EXIT, 0, timestamp);
	(void) delfileref(procent, &procent->procname, 0, timestamp);
	}
#ifdef TRACK_CWD
    (void) delfileref(procent, &procent->cwd, 0, timestamp);
    (void) delfileref(procent, &procent->currentRoot, 0, timestamp);
#endif // TRACK_CWD
    if (procent->pid != 1)
	(void) recordReference(procent, NULL, NULL, false, SEER_REF_PROCEXIT,
	  0, timestamp);
    if (procent->lastexec != NULL)
	free((void*) procent->lastexec);
    free((void*)procent->dtable);
    free((void*)procent->closeOnExecFlags);
    free((void*)procent);
    }

static void observedata()	// Observe trace data from kernel
    {
    char		buf[SEERBUFSIZE]; // Bfr to hold data from kernel
    seerbuf_t		dummyent; // Dummy entry for timestamp
    int			entsize; // Size of current entry
    int			fd;	// FD of reopened server
    int			flushno = 0; // Server number to flush next
    int			maxsaveno; // Maximum saveno filename can support
    int			nbytes;	// No. of bytes read from seer device
    struct timeval	nextfltime; // Next time to flush buffers
    struct timeval	nextstamptime; // Next time to issue timestamp
    int			nfds;	// Number of fd's that select completed on
    struct timeval	now;	// Current time
    int			saveno;	// Save file counter
    char*		saveptr; // Pointer into save file name
    fd_set		selfds; // FD list to select on
    seerbuf_t*		seerp;	// Pointer to next seer structure
    int			serverno; // Server being scanned for retry
    struct timeval	waittime; // How long to wait for information
    struct timeval*	wt;	// Pointer to waittime, or null

    (void) gettimeofday(&nextfltime, (struct timezone*) NULL);
    nextstamptime = nextfltime;
    nextfltime.tv_sec += FLUSHTIME;
    nextstamptime.tv_sec += STAMPTIME;
    while (!terminate_observer)
	{
	if (do_reexec)
	    {
	    restart_self();
	    do_reexec = 0;	// Only get here if failure; prevent loops
	    }
	(void) gettimeofday(&now, (struct timezone*) NULL);
	if (now.tv_sec >= nextstamptime.tv_sec
	  &&  savefile != NULL)
	    {
	    /*
	     * Write a periodic timestamp to the save file.  We assume
	     * sizeof(time_t) equals sizeof(int).  Note that the
	     * timestamp might be later than the time of the next
	     * packet we write, if the kernel has any data buffered up
	     * for us.
	     */
	    dummyent.callcode = 0;
	    dummyent.errcode = SEERBUF_VERSION;
	    dummyent.flags = SEER_TIMESTAMP;
	    dummyent.argsize = 1;
	    dummyent.uid = 0;
	    dummyent.pid = 0;
	    dummyent.retval = 0;
	    dummyent.calltime = now;
	    dummyent.args[0] = now.tv_sec;
	    nextstamptime.tv_sec = now.tv_sec + STAMPTIME;
	    (void) fwrite(&dummyent, sizeof dummyent, 1, savefile);
	    /*
	     * See whether the save file is too big.  If so, and the
	     * save file ends with one or more digits, close the save
	     * file, compress it, and open a new one (not quite in
	     * that order).
	     */
	    saveptr = &savefname[savelen - 1];
	    if (ftell(savefile) >= MAX_SAVE_SIZE  &&  isdigit(*saveptr))
		{
		maxsaveno = 1;
		while (saveptr > &savefname[0]  &&  isdigit(*saveptr)
		  &&  maxsaveno < 2000000000)
		    {
		    saveptr--;
		    maxsaveno *= 10;
		    }
		if (isdigit(*saveptr))
		    maxsaveno *= 10;
		else
		    saveptr++;
		saveno = atoi(saveptr) + 1;
		if (saveno < maxsaveno)
		    {
		    (void)fclose(savefile);
		    SimpleString zipcmd(GZIPIT, savelen + sizeof GZIPIT + 1);
		    zipcmd += savefname;
		    zipcmd += "&";
		    system(zipcmd);
		    (void)sprintf(saveptr, "%*.*d",
		      savelen - (saveptr - &savefname[0]),
		      savelen - (saveptr - &savefname[0]),
		      saveno);
		    savefile = fopen(savefname, "w");
		    if (savefile == NULL)
			{
			int saveErrno = errno;
			cerr << "observer:  Couldn't open save file '"
			  << savefname << "': " << strerr(saveErrno) << '\n';
			}
		    else
			(void)fwrite(&dummyent, sizeof dummyent, 1, savefile);
		    }
		}
	    }
	if (now.tv_sec > nextfltime.tv_sec
	  ||  (now.tv_sec == nextfltime.tv_sec
	    &&  now.tv_usec >= nextfltime.tv_usec))
	    {
	    nextfltime = now;
	    nextfltime.tv_sec += FLUSHTIME;
	    if (servers[flushno].file != NULL)
		{
		errno = 0;
		if (fflush(servers[flushno].file) < 0)
		    servertrouble(&servers[flushno]);
		else
		    {
		    if (servers[flushno].blocked)
			{
#ifdef TIMEVALS_ARE_INTS
			cerr << "observer:  server on fd "
			  << fileno(servers[flushno].file)
			  << " unblocked at " << ctime((long*)&now.tv_sec);
#else // TIMEVALS_ARE_INTS
			cerr << "observer:  server on fd "
			  << fileno(servers[flushno].file)
			  << " unblocked at " << ctime(&now.tv_sec);
#endif // TIMEVALS_ARE_INTS
			servers[flushno].blocked = 0;
			}
		    }
		}
	    if (++flushno >= nservers)
		{
		flushno = 0;
		needflush = 0;
		}
	    }
	FD_ZERO(&selfds);
	FD_SET(seerfd, &selfds);
	wt = NULL;
	waittime.tv_usec = 0;
	for (serverno = 0;  serverno < nservers;  serverno++)
	    {
	    if (servers[serverno].file == NULL)
		{
		if (now.tv_sec >= servers[serverno].retrytime)
		    {
#ifdef SEER_BLOCKING
		    fd = connectServer(servers[serverno].servername,
		      OBSERVER_CONNECTION, blockingConnection, silentFailures);
#else // SEER_BLOCKING
		    fd = connectServer(servers[serverno].servername,
		      OBSERVER_CONNECTION, nonBlockingConnection,
		      silentFailures);
#endif // SEER_BLOCKING
		    if (fd < 0)
			{
			servers[serverno].file = NULL;
			servers[serverno].retrytime = now.tv_sec + RETRYTIME;
			}
		    else
			{
			servers[serverno].file = fdopen(fd, "w");
			servers[serverno].retrytime = 0;
			servers[serverno].blocked = 0;
#ifdef TIMEVALS_ARE_INTS
			cerr << "observer:  reopened connection to "
			  << (servers[serverno].servername == NULL
			    ? "local server"
			    : servers[serverno].servername)
			  << " at " << ctime((long*)&now.tv_sec);
#else // TIMEVALS_ARE_INTS
			cerr << "observer:  reopened connection to "
			  << (servers[serverno].servername == NULL
			    ? "local server"
			    : servers[serverno].servername)
			  << " at " << ctime(&now.tv_sec);
#endif // TIMEVALS_ARE_INTS
			sendControl(realTimeMode, &now);
			}
		    }
		else if (wt == NULL
		  ||  waittime.tv_sec
		    > servers[serverno].retrytime - now.tv_sec)
		    {
		    waittime.tv_sec = servers[serverno].retrytime - now.tv_sec;
		    wt = &waittime;
		    }
		}
	    }
	if (needflush)
	    {
	    if (wt == NULL
	      ||  waittime.tv_sec > nextfltime.tv_sec - now.tv_sec)
		{
		waittime.tv_sec = nextfltime.tv_sec - now.tv_sec;
		waittime.tv_usec = nextfltime.tv_usec - now.tv_usec;
		if (waittime.tv_usec < 0)
		    {
		    waittime.tv_usec += 1000000;
		    waittime.tv_sec--;
		    }
		wt = &waittime;
		}
	    }
	nfds =
	  select(seerfd + 1, &selfds, (fd_set*) NULL, (fd_set*) NULL, wt);
	if (nfds <= 0)
	    nbytes = 0;
	else
	    nbytes = read(seerfd, buf, sizeof buf);
	seerp = (seerbuf_t*) buf;
	if (verbose > 1  &&  nbytes > 0)
	    {
	    cerr << "observer:  got " << nbytes << " bytes from /dev/seer\n";
	    cerr.flush();
	    }
	while (nbytes > 0)
	    {
	    entsize = sizeof(seerbuf_t)
	      + (seerp->argsize - 1) * sizeof seerp->args[0];
	    if (verbose > 2)
		{
		showent(seerp, NULL, 0);
		cerr.flush();
		}
	    if (process_entry(seerp)  &&  savefile != NULL)
		(void) fwrite((char*) seerp, entsize, 1, savefile);
	    seerp = (seerbuf_t*) ((char*) seerp + entsize);
	    nbytes -= entsize;
	    }
	if (savefile != NULL)
	    (void) fflush(savefile);
	}
    (void) gettimeofday(&now, (struct timezone*)NULL);
#ifdef TIMEVALS_ARE_INTS
    cerr << "observer:  exiting at " << ctime((long*)&now.tv_sec);
#else // TIMEVALS_ARE_INTS
    cerr << "observer:  exiting at " << ctime(&now.tv_sec);
#endif // TIMEVALS_ARE_INTS
    }

static void servertrouble(		// Handle difficulties with a server
    server_t*		server)		// Server that we wered talking to
    {
    long		now;		// Current time

    /*
     * This routine must be called after an fwrite or fflush failure.
     */
    if (ETEST(errno)
#ifdef SUNSYSV // check for another brain-damaged OS bug
      || errno == 0
#endif
#ifdef EMSGSIZE // check for another brain-damaged OS bug
      || errno == EMSGSIZE
#endif
      )
	{
	if (!server->blocked)
	    {
	    (void) time(&now);
	    cerr << "observer:  server on fd " << fileno(server->file)
	      << " blocked at " << ctime(&now);
	    server->blocked = 1;
	    }
	}
    else
	{
	int saveErrno = errno;
	cerr << "observer:  packet write failed at offset "
	  << ftell(server->file) << ": " << strerr(saveErrno) << '\n';
	closeserver(server);
	}
    }

static void closeserver(		// Close connection to a server
    server_t*		server)		// Server that we lost
    {
    int saveErrno = errno;
    server->retrytime = time((long*) NULL);
    cerr << "observer:  lost connection to "
      << (server->servername == NULL ? "local server" : server->servername)
      << ": " << strerr(saveErrno) << " at " << ctime(&server->retrytime);
    (void) fclose(server->file);
    server->file = NULL;
    server->retrytime += RETRYTIME;
    }

static void restart_self()		// Restart the observer
    {
    int			arglen;		// Length of argument to build
    int			argno;		// Argument being built
    char*		argp;		// Pointer into arg being built
    char		csavefd[12];	// Character form of save file fd
    char		cseerfd[12];	// Character form of seer fd
    const char**	execargs;	// Argument vector for the re-exec
    int			serverno;	// Server number being unparsed

    (void) sprintf(cseerfd, "%d", seerfd);
    if (savefile != NULL)
	(void) sprintf(csavefd, "%d", fileno(savefile));
    else
	(void) strcpy(csavefd, "-1");
    execargs =
      (const char**) malloc((nservers * 2 + 6) * sizeof(const char**));
    if(execargs == NULL)
	nomem("restart_self", __LINE__,
	  (nservers * 2 + 5) * sizeof(const char**));
    cerr << "observer:  restart '" << progname;
    execargs[0] = progname;
    for (argno = 1, serverno = nservers;  --serverno >= 0;  )
	{
	execargs[argno++] = "-u";
	cerr << (argno == 1 ? "" : " ") << "-u";
	if (servers[serverno].file == NULL)
	    arglen = 1;
	else
	    arglen = 11 + 1 + 1 + 1;	// 32-bit int plus ! plus :
	arglen += 11;			// 32-bit int
	if (servers[serverno].servername != NULL)
	    arglen += strlen(servers[serverno].servername) + 1;
	argp = (char*) malloc(arglen);
	if (argp == NULL)
	    nomem("restart_self", __LINE__, arglen);
	execargs[argno] = argp;
	if (servers[serverno].file != NULL)
	    {
	    (void) sprintf(argp, "!%d:", fileno(servers[serverno].file));
	    argp += strlen(argp);
	    (void) fflush(servers[serverno].file);
	    }
	if (servers[serverno].servername != NULL)
	    (void) sprintf(argp, "%d:%s", servers[serverno].uid,
	      servers[serverno].servername);
	else
	    (void) sprintf(argp, "%d", servers[serverno].uid);
	cerr << ' ' << execargs[argno];
	argno++;
	}
    cerr << (argno == 1 ? "" : " ") << "-R " << cseerfd << ' '
      << csavefd << "'\n";
    execargs[argno++] = "-R";
    execargs[argno++] = cseerfd;
    execargs[argno++] = csavefd;
    execargs[argno++] = savefname == NULL ? "" : (const char*)savefname;
    execargs[argno] = NULL;
    (void)chdir(ourWD);
    execvp(execargs[0], (char *const *)execargs);
    /*
     * If we fail, we just return and continue executing in the
     * current guise
     */
    int saveErrno = errno;
    cerr << "observer:  re-exec failed: " << strerr(saveErrno) << '\n';
    }

static int process_entry(		// Process a trace entry
					// Returns NZ to save it in trace
    seerbuf_t*		seerp)		// Trace buffer entry to process
    {
    int			fdno;		// FD number being processed
    openfile_t		dummyent;	// Dummy entry for hash find
    openfile_t*		fileent;	// Entry for file being handled
    char*		newpath;	// Path for new name in rename
    const char*		path;		// Path referenced in current entry
    proc_t*		parentent;	// Parent's proc table entry
    proc_t*		procent;	// Table entry for current proc
    openfile_t*		replent;	// File being replaced in rename
    int			saveit = 0;	// NZ if we should trace this entry

    procent = findproc(seerp->pid, seerp->uid, false);
    switch (seerp->callcode)
	{
	case SEER_SYS_exit:
	    {
	    if (procent != NULL)
		{
		saveit = 1;
		(void) proctable.remove((void*) procent);
		procexit(procent, &seerp->calltime);
		procent = NULL;
		}
	    break;
	    }
	case SEER_SYS_fork:
	case SEER_SYS_vfork:
	    {
	    if (seerp->errcode == 0  &&  seerp->retval == 0)
		{
		saveit = 1;
		(void)recordNewProcess(seerp->uid, seerp->pid, seerp->args[0],
		  &seerp->calltime);
		if (procent == NULL)
		    {
		    /*
		     * We need to create a process entry so that the
		     * corresponding exit will be sent to the
		     * correlators.  There is no other reason to
		     * create it.
		     */
		    procent = findproc(seerp->pid, seerp->uid, true);
		    }
		else
		    {
#ifdef TRACK_CWD
		    (void) delfileref(procent, &procent->cwd, 0,
		      &seerp->calltime);
		    (void) delfileref(procent, &procent->currentRoot, 0,
		      &seerp->calltime);
#endif // TRACK_CWD
		    for (fdno = procent->dtableSize;  --fdno >= 0;  )
			{
			(void) delfileref(procent, &procent->dtable[fdno], 1,
			  &seerp->calltime);
			procent->closeOnExecFlags[fdno] = 0;
			}
		    }
		parentent = findproc(seerp->args[0], seerp->uid, false);
		if (parentent != NULL)
		    {
		    if (parentent->procname != NULL)
			(void) addfileref(parentent,
			  parentent->procname->path, true, &procent->procname,
			  &seerp->calltime);
#ifdef TRACK_CWD
		    if (parentent->cwd != NULL)
			{
			procent->cwd = parentent->cwd;
			procent->cwd->refcount++;
			}
		    if (parentent->currentRoot != NULL)
			{
			procent->currentRoot = parentent->currentRoot;
			procent->currentRoot->refcount++;
			}
#endif // TRACK_CWD
		    for (fdno = parentent->dtableSize;  --fdno >= 0;  )
			{
			if (parentent->dtable[fdno] != NULL)
			    {
			    (void)expand_dtable(procent, fdno);
			    procent->dtable[fdno] = parentent->dtable[fdno];
			    procent->dtable[fdno]->refcount++;
			    saveit |= recordReference(procent,
			      procent->dtable[fdno]->path, NULL, false,
			      SEER_REF_DUP, 0, &seerp->calltime);
			    }
			}
		    }
		}
	    break;
	    }
	case SEER_SYS_open:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		/*
		 * Figure out what type of open this is.
		 */
		int openFlags = seerp->args[seerp->argsize - 1];
		ReferenceType referenceType = SEER_REF_OPENREAD;
		if ((openFlags & SEER_ACCMODE) != SEER_O_RDONLY)
		    referenceType = SEER_REF_OPENRW;
		// NEEDSWORK: this is too strong an assumption
		if (openFlags & (SEER_O_TRUNC | SEER_O_CREAT))
		    referenceType = SEER_REF_OPENCREATE;
		path = findarg(seerp, 0);
		if (seerp->errcode != 0
		  ||  !expand_dtable(procent, seerp->retval))
		    {
		    if (seerp->errcode == 0)
			cerr << "observer:  bad open return " << seerp->retval
			  << " from process " << seerp->pid << '\n';
		    switch (referenceType)
			{
			case SEER_REF_OPENREAD:
			    {
			    referenceType = SEER_REF_READ;
			    break;
			    }
			case SEER_REF_OPENRW:
			    {
			    referenceType = SEER_REF_OPENRW;
			    break;
			    }
			case SEER_REF_OPENCREATE:
			    {
			    referenceType = SEER_REF_CREATE;
			    break;
			    }
			default:
			    {
			    cerr << "observer: unexpected reference type "
			      << (unsigned long)referenceType;
			    observerPanic("");
			    break;
			    }
			}
		    }
		else
		    {
		    saveit |= addfileref(procent, path, true,
		      &procent->dtable[seerp->retval], &seerp->calltime);
		    procent->closeOnExecFlags[seerp->retval] = 0;
		    }
		saveit |= recordReference(procent, path, NULL, true,
		  referenceType, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_close:
	    {
	    fdno = seerp->args[0];
	    if (seerp->errcode == 0 &&  procent != NULL)
		{
		if (fdno >= 0  &&  fdno < procent->dtableSize)
		    {
		    if (procent->dtable[fdno] != NULL)
			{
			saveit |=
			  delfileref(procent, &procent->dtable[fdno], 1,
			    &seerp->calltime);
			procent->closeOnExecFlags[fdno] = 0;
			}
		    }
		else
		    cerr << "observer:  close accepted bad fd " << fdno
		      << " from process " << seerp->pid << '\n';
		}
	    break;
	    }
	case SEER_SYS_creat:
	    {
	    if (seerp->errcode == 0)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		path = findarg(seerp, 0);
		ReferenceType referenceType = SEER_REF_OPENCREATE;
		if (!expand_dtable(procent, seerp->retval))
		    {
		    cerr << "observer:  bad open return " << seerp->retval
		      << " from process " << seerp->pid << '\n';
		    referenceType = SEER_REF_CREATE;
		    }
		else
		    {
		    addfileref(procent, path, true,
		      &procent->dtable[seerp->retval], &seerp->calltime);
		    procent->closeOnExecFlags[seerp->retval] = 0;
		    }
		saveit |= recordReference(procent, path, NULL, true,
		  referenceType, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_link:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 1),
		  NULL, true, SEER_REF_CREATE, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_unlink:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_DESTROY, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_execv:
	case SEER_SYS_execve:
	    {
	    if (seerp->uid != 0)
		saveit = 1;
	    if (seerp->flags & SEER_BEFORE)
		{
		/*
		 * Because the pathname of an exec is lost when it
		 * succeeds, we must save the path before the exec,
		 * and then report it later.
		 */
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		else
		    {
		    if (procent->lastexec != NULL)
			free((void*) procent->lastexec);
		    }
		path = findarg(seerp, 0);
		procent->lastexec = strdup(path, 0);
		}
	    else
		{
		if (seerp->errcode != 0  &&  seerp->errcode != ENOENT)
		    saveit = 0;
		if (procent == NULL)
		    cerr << "observer:  exec completed for unknown process "
		      << seerp->pid << '\n';
		else if (procent->lastexec != NULL)
		    {
		    char fullPath[MAXPATHLEN]; // Space for expanded path
		    const char* execPath = absolutePath(fullPath,
		      procent->lastexec, procent, false, &seerp->calltime);
		    if (seerp->errcode == 0)
			{
			for (fdno = procent->dtableSize;  --fdno >= 0;  )
			    {
			    if (procent->closeOnExecFlags[fdno])
				{
				saveit |= delfileref(procent,
				  &procent->dtable[fdno], 1, &seerp->calltime);
				procent->closeOnExecFlags[fdno] = 0;
				}
			    }
			if (procent->procname != NULL)
			    {
			    (void) delfileref(procent,
			      &procent->procname, 0, &seerp->calltime);
			    }
			/*
			 * When an exec happens, we need to find out
			 * whether an interpreter was executed using
			 * the #! mechanism.  For historical reasons,
			 * this is done here, rather than in the
			 * kernel, which might be a better place.
			 */
#if 0
// Test disabled because it's not a serious simulation-hang problem,
// and it gives better simulations to disable it
			int execFile =
			  avoidFilesystem ? -1 : open(execPath, 0);
#else // 0
			int execFile = open(execPath, 0);
#endif // 0
			if (execFile >= 0)
			    {
			    char interpreterBuffer[MAXPATHLEN
			      + sizeof SCRIPT_MAGIC + 1];
			    int bytesRead = read(execFile, interpreterBuffer,
			      sizeof interpreterBuffer - 1);
			    if (bytesRead >= (int)sizeof SCRIPT_MAGIC)
				{
				interpreterBuffer[bytesRead] = '\0';
				if (strncmp(interpreterBuffer, SCRIPT_MAGIC,
				    sizeof SCRIPT_MAGIC - 1)
				  == 0)
				    {
				    char* whiteSpace =
				      &interpreterBuffer[sizeof SCRIPT_MAGIC];
				    while (*whiteSpace != '\0'
				      &&  !isspace(*whiteSpace))
					whiteSpace++;
				    *whiteSpace = '\0';
				    saveit |= recordReference(procent,
				      &interpreterBuffer[sizeof SCRIPT_MAGIC
					- 1],
				      NULL, true, SEER_REF_INTERPRET,
				      seerp->errcode, &seerp->calltime);
				    }
				}
			    (void)close(execFile);
			    }
			}
		    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
			{
			saveit |= recordReference(procent,
			  execPath, NULL, true, SEER_REF_EXEC,
			  seerp->errcode, &seerp->calltime);
			if (seerp->errcode == 0)
			    (void) addfileref(procent, execPath, true,
			      &procent->procname, &seerp->calltime);
			}
		    free((void*) procent->lastexec);
		    procent->lastexec = NULL;
		    }
		}
	    break;
	    }
	case SEER_SYS_chdir:
	    {
	    if (seerp->errcode == ENOENT)
		{
		saveit |= recordReference(procent, findarg(seerp, 0), NULL,
		  false, SEER_REF_READ, seerp->errcode, &seerp->calltime);
		}
	    else if (seerp->errcode == 0)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
#ifdef TRACK_CWD
		bool hadCWD = procent->cwd != NULL;
#endif // TRACK_CWD
		char newCWD[MAXPATHLEN];
		path = absolutePath(newCWD, findarg(seerp, 0), procent, false,
		  &seerp->calltime);
		saveit |= recordReference(procent, path, NULL, false,
		  SEER_REF_READ, seerp->errcode, &seerp->calltime);
#ifdef TRACK_CWD
		/*
		 * If absolutePath had to generate a fake chdir for
		 * us, we don't want to save this packet, because it
		 * is relative to the old working directory and setCWD
		 * works relative to the new one (this is because of
		 * the relative timing of the trace and when the
		 * correlator runs).
		 */
		if (!hadCWD  &&  procent->cwd != NULL)
		    saveit = 0;
		/*
		 * We need to make sure that current-directory paths
		 * don't get too long when a process does a series of
		 * chdir(foo)/chdir(..) calls.  We achieve this by
		 * using realpath to remove the ..'s.  This also has
		 * the advantage of removing any symlinks from the
		 * CWD, so that subsequent relative path references
		 * don't generate extra FOLLOW references.
		 */
		char cleanCWD[MAXPATHLEN];
		canonicalize_pn(path, cleanCWD, NULL, false, NULL);
		(void)addfileref(procent, cleanCWD, false, &procent->cwd,
		  &seerp->calltime);
#endif // TRACK_CWD
		}
	    break;
	    }
	case SEER_SYS_mknod:
	case SEER_SYS_mkfifo:
	    {
	    if (seerp->errcode == 0)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_CREATE, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_oldstat:
	case SEER_SYS_access:
	case SEER_SYS_stat:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_PEEK, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_lstat:
	case SEER_SYS_oldlstat:
	case SEER_SYS_readlink:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_PEEKLINK, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_uselib:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0), NULL,
		  true, SEER_REF_READ, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_dup:
	    {
	    if (procent != NULL  &&  seerp->errcode == 0)
		{
		fdno = seerp->retval;
		if (!expand_dtable(procent, fdno))
		    cerr << "observer:  bad dup return " << seerp->retval
		      << " from process " << seerp->pid << '\n';
		else
		    {
		    saveit |=
		      delfileref(procent, &procent->dtable[fdno], 1,
			&seerp->calltime);
		    procent->dtable[fdno] =
		      procent->dtable[seerp->args[0]];
		    if (procent->dtable[fdno] != NULL)
			{
			procent->dtable[fdno]->refcount++;
			saveit |= recordReference(procent,
			  procent->dtable[fdno]->path, NULL, false,
			  SEER_REF_DUP, seerp->errcode, &seerp->calltime);
			}
		    procent->closeOnExecFlags[fdno] = 0;
		    }
		}
	    break;
	    }
	case SEER_SYS_symlink:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_MAKELINK, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_chroot:
	    {
	    if (seerp->errcode == ENOENT)
		{
		saveit |= recordReference(procent, findarg(seerp, 0), NULL,
		  false, SEER_REF_READ, seerp->errcode, &seerp->calltime);
		}
	    else if (seerp->errcode == 0)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		char newRoot[MAXPATHLEN];
		path = absolutePath(newRoot, findarg(seerp, 0), procent, false,
		  &seerp->calltime);
		saveit |= recordReference(procent, path, NULL, false,
		  SEER_REF_READ, seerp->errcode, &seerp->calltime);
#ifdef TRACK_CWD
		/*
		 * Clean out '..' and symbolic links.  See chdir for
		 * more detailed commentary.
		 */
		char cleanRoot[MAXPATHLEN];
		canonicalize_pn(path, cleanRoot, NULL, false, NULL);
		(void)addfileref(procent, cleanRoot, false,
		  &procent->currentRoot, &seerp->calltime);
#endif // TRACK_CWD
		}
	    break;
	    }
	case SEER_SYS_dup2:
	    {
	    if (procent != NULL  &&  seerp->errcode == 0)
		{
		fdno = seerp->args[1];
		if (fdno >= 0  &&  fdno < procent->dtableSize)
		    saveit |= delfileref(procent, &procent->dtable[fdno], 1,
		      &seerp->calltime);
		if (fdno != seerp->retval)
		    {
		    cerr << "observer:  bad dup2 return value for process "
		      << seerp->pid << ": expected " << fdno << ", got "
		      << seerp->retval << '\n';
		    fdno = seerp->retval;
		    }
		if (!expand_dtable(procent, fdno))
		    cerr << "observer:  bad dup2 return " << fdno
		      << " from process " << seerp->pid << '\n';
		else if (!expand_dtable(procent, seerp->args[0]))
		    cerr << "observer:  bad dup2 argument " << seerp->args[0]
		      << " from process " << seerp->pid << '\n';
		else
		    {
		    procent->dtable[fdno] = procent->dtable[seerp->args[0]];
		    if (procent->dtable[fdno] != NULL)
			{
			procent->dtable[fdno]->refcount++;
			saveit = recordReference(procent,
			  procent->dtable[fdno]->path, NULL, false,
			  SEER_REF_DUP, seerp->errcode, &seerp->calltime);
			}
		    procent->closeOnExecFlags[fdno] = 0;
		    }
		}
	    break;
	    }
	case SEER_SYS_fcntl:
	    {
	    if (seerp->errcode == 0)
		{
		switch (seerp->args[1])
		    {
		    case F_DUPFD:	/* Duplicate file descriptor */
			if (procent != NULL)
			    {
			    fdno = seerp->retval;
			    if (!expand_dtable(procent, fdno))
				cerr << "observer:  bad F_DUPFD return "
				  << fdno << " from process " << seerp->pid
				  << '\n';
			    else if (!expand_dtable(procent, seerp->args[0]))
				cerr << "observer:  bad F_DUPFD argument "
				  << seerp->args[0] << "from process "
				  << seerp->pid << '\n';
			    else
				{
				saveit |= delfileref(procent,
				  &procent->dtable[fdno], 1, &seerp->calltime);
				procent->dtable[fdno] =
				  procent->dtable[seerp->args[0]];
				if (procent->dtable[fdno] != NULL)
				    {
				    procent->dtable[fdno]->refcount++;
				    saveit = recordReference(procent,
				      procent->dtable[fdno]->path, NULL,
				      false, SEER_REF_DUP, seerp->errcode,
				      &seerp->calltime);
				    }
				procent->closeOnExecFlags[fdno] = 0;
				}
			    }
			break;
		    case F_GETFD:	/* Get close-on-exec flag */
			if (procent == NULL)
			    procent = findproc(seerp->pid, seerp->uid, true);
			if (!expand_dtable(procent, seerp->args[0]))
			    cerr << "observer:  F_GETFD accepted bad fd "
			      <<  seerp->args[0] << '\n';
			else
			    procent->closeOnExecFlags[seerp->args[0]] =
			      seerp->retval & 1;
			break;
		    case F_SETFD:	/* Set close-on-exec flag */
			if (procent == NULL)
			    procent = findproc(seerp->pid, seerp->uid, true);
			if (!expand_dtable(procent, seerp->args[0]))
			    cerr << "observer:  F_GETFD accepted bad fd "
			      <<  seerp->args[0] << '\n';
			else
			    procent->closeOnExecFlags[seerp->args[0]] =
			      seerp->args[2] & 1;
			break;
		    default:
			// It's a call that we don't care about
			break;
		    }
		}
	    break;
	    }
	case SEER_SYS_rename:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		path = findarg(seerp, 0);
		newpath = findarg(seerp, 1);
		saveit |= recordReference(procent, path, newpath,
		  true, SEER_REF_RENAME, seerp->errcode, &seerp->calltime);
		dummyent.path = path;
		fileent = (openfile_t*) filetable.find((void*) &dummyent,
		  HashTable::FAIL);
		if (fileent != NULL)
		    {
		    (void) filetable.remove((void*) fileent);
		    free ((void*) fileent->path);
		    fileent->path = strdup(newpath, 0);
		    replent = (openfile_t*) filetable.find((void*) fileent,
		      HashTable::INSERT);
		    if (replent != fileent)
			{
			saveit = 1;
			(void) recordReference(procent, newpath, NULL, true,
			  SEER_REF_DESTROY, seerp->errcode, &seerp->calltime);
			for (HashTable::Iterator iter(proctable);
			  iter;
			  ++iter)
			    {
			    if (remove_ref((proc_t*) proctable[iter], replent,
			      &seerp->calltime))
				break;
			    }
			(void) filetable.find((void*) fileent,
			  HashTable::INSERT);
			}
		    }
		}
	    break;
	    }
	case SEER_SYS_truncate:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_RW, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_mkdir:
	    {
	    if (seerp->errcode == 0)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_CREATE, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_rmdir:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_DESTROY, seerp->errcode,
		  &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_utime:
	case SEER_SYS_utimes:
	    {
	    if (seerp->errcode == 0  ||  seerp->errcode == ENOENT)
		{
		if (procent == NULL)
		    procent = findproc(seerp->pid, seerp->uid, true);
		/*
		 * We record utime as a PEEK because it doesn't change
		 * any of the stat information that we care about, but
		 * it is used in a manner very similar to stat-ing the
		 * file.  Other alternatives might include recording
		 * it as a READ or a CREATE, or creating a new
		 * reference type to handle it.
		 */
		saveit |= recordReference(procent, findarg(seerp, 0),
		  NULL, true, SEER_REF_READ, seerp->errcode, &seerp->calltime);
		}
	    break;
	    }
	case SEER_SYS_fchdir:
	    {
	    if (seerp->errcode == 0)
		{
		if (seerp->args[0] < 0
		  ||  (procent != NULL
		    &&  !expand_dtable(procent, seerp->args[0])))
		    cerr << "observer:  fchdir accepted bad fd "
		      << seerp->args[0] << '\n';
		if (seerp->uid != 0)
		    saveit = 1;
		openfile_t* newCWD =
		  procent == NULL ? NULL : procent->dtable[seerp->args[0]];
		if (newCWD != NULL)
		    saveit |= recordReference(procent, newCWD->path, NULL,
		      false, SEER_REF_READ, seerp->errcode, &seerp->calltime);
#ifdef TRACK_CWD
		if (procent != NULL)
		    {
		    (void)delfileref(procent, &procent->cwd, 0,
		      &seerp->calltime);
		    procent->cwd = newCWD;
		    if (newCWD != NULL)
			newCWD->refcount++;
		    }
#endif // TRACK_CWD
		}
	    break;
	    }
	case SEER_SYS_fchroot:
	    {
	    if (seerp->errcode == 0)
		{
		if (seerp->args[0] < 0
		  ||  (procent != NULL
		    &&  !expand_dtable(procent, seerp->args[0])))
		    cerr << "observer:  fchroot accepted bad fd "
		      << seerp->args[0] << '\n';
		if (seerp->uid != 0)
		    saveit = 1;
		openfile_t* newRoot =
		  procent == NULL ? NULL : procent->dtable[seerp->args[0]];
		if (newRoot != NULL)
		    saveit |= recordReference(procent, newRoot->path, NULL,
		      false, SEER_REF_READ, seerp->errcode, &seerp->calltime);
#ifdef TRACK_CWD
		if (procent != NULL)
		    {
		    (void)delfileref(procent, &procent->currentRoot, 0,
		      &seerp->calltime);
		    procent->currentRoot = newRoot;
		    if (newRoot != NULL)
			newRoot->refcount++;
		    }
#endif // TRACK_CWD
		}
	    break;
	    }
	default:
	    {
	    break;
	    }
	}
    if (procent != NULL)
	procent->uid = seerp->uid;
    return saveit;
    }

static proc_t* findproc(		// Find process entry for a PID
    pid_t		pid,		// PID to locate
    uid_t		uid,		// UID to assign, if new process
    bool		create)		// NZ to create entry if nonexistent
    {
    int			fdno;		// File descriptor being zeroed
    static proc_t*	newent = NULL;	// New process-table entry
    proc_t*		procent;	// Table entry for current proc

    if (newent == NULL)
	{
	newent = (proc_t*) malloc(sizeof(proc_t));
	if (newent == NULL)
	    nomem ("findproc", __LINE__, sizeof(proc_t));
	}
    newent->pid = pid;
    procent = (proc_t*) proctable.find((void*) newent,
      create ? HashTable::INSERT : HashTable::FAIL);
    if (procent == NULL)
	return NULL;
    if (procent == newent)		// Previously-unknown process?
	{
	procent->lastexec = NULL;
	procent->procname = NULL;
#ifdef TRACK_CWD
	procent->cwd = NULL;
	procent->currentRoot = NULL;
#endif // TRACK_CWD
	procent->dtableSize = DTABLE_INCREMENT;
	procent->dtable =
	  (openfile_t**) malloc(sizeof(openfile_t*) * DTABLE_INCREMENT);
	if (procent->dtable == NULL)
	    nomem("findproc", __LINE__,
	      sizeof(openfile_t*) * DTABLE_INCREMENT);
	procent->closeOnExecFlags =
	  (unsigned char*) malloc(sizeof(unsigned char) * DTABLE_INCREMENT);
	if (procent->closeOnExecFlags == NULL)
	    nomem("findproc", __LINE__,
	      sizeof(unsigned char) * DTABLE_INCREMENT);
	for (fdno = DTABLE_INCREMENT;  --fdno >= 0;  )
	    {
	    procent->dtable[fdno] = NULL;
	    procent->closeOnExecFlags[fdno] = 0;
	    }
	newent = NULL;
	}
    procent->uid = uid;
    return procent;
    }

/*
 * Expand the file-descriptor table for a process so that it is large
 * enough to contain desired_fd.  Returns true if all went well, false
 * if there was some sort of problem, such as a negative desired_fd.
 */
static bool expand_dtable(
    proc_t*		procent,	// Process to be expanded
    int			desiredFd)	// Fd that we'd like to store
    {
    if (desiredFd < 0)
	return false;
    if (desiredFd >= procent->dtableSize)
	{
	int newSize = procent->dtableSize + DTABLE_INCREMENT;
	if (desiredFd >= newSize)
	    newSize = desiredFd + 1;
	procent->dtable = (openfile_t**) realloc(procent->dtable,
	  sizeof(openfile_t*) * newSize);
	if (verbose > 0)
	    {
	    cerr << "observer:  expanding dtable for process " << procent->pid
	      << " to " << newSize << '\n';
	    cerr.flush();
	    }
	procent->closeOnExecFlags = (unsigned char*) realloc(
	  procent->closeOnExecFlags, sizeof(unsigned char) * newSize);
	if (procent->dtable == NULL)
	    nomem("expand_dtable", __LINE__, sizeof(openfile_t*) * newSize);
	else if (procent->closeOnExecFlags == NULL)
	    nomem("expand_dtable", __LINE__, sizeof(unsigned char) * newSize);
	for (int fdno = newSize;  --fdno >= procent->dtableSize;  )
	    {
	    procent->dtable[fdno] = NULL;
	    procent->closeOnExecFlags[fdno] = 0;
	    }
	procent->dtableSize = newSize;
	}
    return true;
    }

static char* findarg(			// Find an argument after pathname
    seerbuf_t*		seerp,		// Trace buffer to search in
    int			argno)		// Number of the argument to locate
    {
    char*		argp;		// Pointer to next argument

    argp = (char*) &seerp->args[0];
    while (--argno >= 0)
	argp += ((strlen(argp) + sizeof seerp->args[0] - 1 + 1)
	    / sizeof seerp->args[0]) * sizeof seerp->args[0];
    return argp;
    }

static int recordNewProcess(		// Record creation of new process
    uid_t		uid,		// UID of new process
    pid_t		pid,		// PID of new process
    pid_t		parentPid,	// PID of parent process
    struct timeval*	timestamp)	// Timestamp to assign to reference
    {
    char		packetbuf[MAX_PACKET_LENGTH];
					// Buffer for output packet
    packetbuf[0] = SEER_REF_NEWPROC;
    // NEEDSWORK: shouldn't hand-pack and hand-swap
    packetbuf[1] = 0;			// Error code: always success
    packetbuf[2] = timestamp->tv_sec;
    packetbuf[3] = timestamp->tv_sec >> 8;
    packetbuf[4] = timestamp->tv_sec >> 16;
    packetbuf[5] = timestamp->tv_sec >> 24;
    packetbuf[6] = timestamp->tv_usec;
    packetbuf[7] = timestamp->tv_usec >> 8;
    packetbuf[8] = timestamp->tv_usec >> 16;
    packetbuf[9] = timestamp->tv_usec >> 24;
    packetbuf[10] = pid & 0xff;
    packetbuf[11] = pid >> 8;
    packetbuf[12] = pid >> 16;
    packetbuf[13] = pid >> 24;
    packetbuf[14] = parentPid & 0xff;
    packetbuf[15] = parentPid >> 8;
    packetbuf[16] = parentPid >> 16;
    packetbuf[17] = parentPid >> 24;
    bool fakeExit;
    return sendPacket(SEER_REF_NEWPROC, uid, packetbuf, 18, fakeExit);
    }

static int recordReference(		// Record a successful file ref
    proc_t*		procent,	// Referencing process
    const char*		path1,		// First path that was referenced
    const char*		path2,		// Second path that was referenced
    bool		chrootApplies,	// True if must prepend currentRoot
    ReferenceType	type,		// What type of reference it was
    unsigned int	errcode,	// Error code from call (0 = success)
    struct timeval*	timestamp)	// Timestamp to assign to reference
    {
    char		packetbuf[MAX_PACKET_LENGTH];
					// Buffer for output packet
    int			packetlen;	// Length of packetbuf

    packetbuf[0] = type;
    // NEEDSWORK: shouldn't hand-pack and hand-swap
    packetbuf[1] = errcode;
    packetbuf[2] = timestamp->tv_sec;
    packetbuf[3] = timestamp->tv_sec >> 8;
    packetbuf[4] = timestamp->tv_sec >> 16;
    packetbuf[5] = timestamp->tv_sec >> 24;
    packetbuf[6] = timestamp->tv_usec;
    packetbuf[7] = timestamp->tv_usec >> 8;
    packetbuf[8] = timestamp->tv_usec >> 16;
    packetbuf[9] = timestamp->tv_usec >> 24;
    packetbuf[10] = procent->pid & 0xff;
    packetbuf[11] = procent->pid >> 8;
    packetbuf[12] = procent->pid >> 16;
    packetbuf[13] = procent->pid >> 24;
    packetlen = 14;
    bool needToCanonicalize = false;
    for (int serverno = nservers;  --serverno >= 0;  )
	{
	if (servers[serverno].servername != NULL
	  &&  servers[serverno].servername[0] != ':')
	    {
	    needToCanonicalize = true;
	    break;
	    }
	}
#ifdef SUPPRESS_FFSH
    if (path1 != NULL)
	{
	for (const char* cp = path1;  *cp != '\0';  cp++)
	    {
	    if (cp[0] == '/'  &&  cp[1] == ':')
		return 0;		// Assume it's a ffsh
	    }
	}
    if (path2 != NULL)
	{
	for (const char* cp = path2;  *cp != '\0';  cp++)
	    {
	    if (cp[0] == '/'  &&  cp[1] == ':')
		return 0;		// Assume it's a ffsh
	    }
	}
#endif // SUPPRESS_FFSH
    if (path1 == NULL)
	{
	packetbuf[packetlen] = 0;
	packetbuf[packetlen + 1] = 0;
	packetlen += 2;
	}
    else
	{
	if (path1[0] == '\0')
	    return 0;			// Suppress null-path references
#ifdef TRIM_DOUBLE_SLASH
	while (path1[0] == '/'  &&  path1[1] == '/')
	    path1++;
#endif // TRIM_DOUBLE_SLASH
	packetlen += packetize_path(path1, &packetbuf[packetlen], procent,
	  needToCanonicalize, chrootApplies, timestamp);
	}
    if (path2 != NULL)
	{
	if (path2[0] == '\0')
	    return 0;			// Suppress null-path references
#ifdef TRIM_DOUBLE_SLASH
	while (path2[0] == '/'  &&  path2[1] == '/')
	    path2++;
#endif // TRIM_DOUBLE_SLASH
	packetlen += packetize_path(path2, &packetbuf[packetlen],
	  procent, needToCanonicalize, chrootApplies, timestamp);
	}
    bool fakeExit;
    int sentit =
      sendPacket(type, procent->uid, packetbuf, packetlen, fakeExit);
    /*
     * Because older save files do not properly record exits, we need
     * to fake an exit in certain cases.  Sendpacket will tell us when
     * we need to do this.
     */
    if (fakeExit)
	sentit |= recordReference(procent, NULL, NULL, false,
	  SEER_REF_PROCEXIT, 0, timestamp);
    return sentit;
    }

static int sendPacket(			// Transmit a packet to active servers
    ReferenceType	type,		// Type of reference being sent
    uid_t		uid,		// UID packet belongs to
    char*		packetBuffer,	// Buffer with packet in it
    int			packetLength,	// Length of the packet buffer
    bool&		fakeExit)	// OUT: tell whether to fake proc exit
    {
    bool		allServers;	// True if packet goes to all servers
    long		now;		// Current time
    int			sentit = 0;	// NZ if trace went to a server
    switch (type)
	{
	case SEER_REF_DESTROY:
	case SEER_REF_RENAME:
	case SEER_REF_CONTROL:
	    {
	    allServers = true;
	    fakeExit = seerversion() <= 2;
	    break;
	    }
	case SEER_REF_NEWPROC:
	case SEER_REF_PROCEXIT:
	    {
	    allServers = true;
	    fakeExit = false;
	    break;
	    }
	default:
	    {
	    allServers = false;
	    fakeExit = false;
	    break;
	    }
	}
    for (int serverNumber = nservers;  --serverNumber >= 0;  )
	{
	/*
	 * Destroys, renames, controls, and new processes go to all
	 * servers; other operations go just to owner
	 */
	if (allServers)
	    {
	    /*
	     * In older save files, we need to generate a fake exit if
	     * the real exit isn't going to go the the server.  As it
	     * happens, all of those save files contain data for a
	     * single UID, so the following code is correct even
	     * though it would be wrong in the general case of
	     * multiple servers handling multiple UIDs.
	     */
	    if (servers[serverNumber].uid == uid)
		fakeExit = false;
	    }
	else if (servers[serverNumber].uid != uid)
	    continue;
	sentit = 1;			// This should be sent to a server
	if (servers[serverNumber].file == NULL)
	    continue;
	errno = 0;
	if (fwrite(packetBuffer, sizeof packetBuffer[0], packetLength,
	    servers[serverNumber].file)
	  == 0)
	    servertrouble(&servers[serverNumber]);
	else
	    {
	    if (servers[serverNumber].blocked)
		{
		(void) time(&now);
		cerr << "observer:  server on fd "
		  << fileno(servers[serverNumber].file) << " unblocked at "
		  << ctime(&now);
		servers[serverNumber].blocked = 0;
		}
	    }
	needflush = 1;
	}
    return sentit;
    }

static int packetize_path(		// Store a path in a packet buffer
    const char*		path,		// Path to be stored
    char*		buf,		// Where to store it
    proc_t*		procent,	// Relevant process-table entry
    bool		canonicalize,	// True to canonicalize the path
    bool		chrootApplies,	// True if need to prepend currentRoot
    struct timeval*	timestamp)	// Timestamp for any refs generated
    {
#ifdef STRIP_FICUS_QUALIFIERS
    char*		colon;		// Pointer to colon in path
#endif // STRIP_FICUS_QUALIFIERS
    int			pathlen;	// Length of the (canonical) path

    if (canonicalize)
	canonicalize_pn(path, &buf[2], procent, chrootApplies, timestamp);
    else
	{
#ifdef TRACK_CWD
	if (chrootApplies)
	    absolutePath(&buf[2], path, procent, true, timestamp);
	else
	    {
	    strncpy(&buf[2], path, MAXPATHLEN);
	    buf[2 + MAXPATHLEN - 1] = '\0';
	    }
#else // TRACK_CWD
	strncpy(&buf[2], path, MAXPATHLEN);
	buf[2 + MAXPATHLEN - 1] = '\0';
#endif // TRACK_CWD
	}
#ifdef STRIP_TMPMNT
    if (strncmp(&buf[2], TMPMNT_PATH, sizeof TMPMNT_PATH - 1, 9) == 0)
	(void) strcpy(&buf[2], &buf[2 + 8]);
					// Strip /tmp_mnt prefix
#endif // STRIP_TMPMNT
#ifdef STRIP_FICUS_QUALIFIERS
    while ((colon = strchr(&buf[2], ':')) != NULL)
	{
	char* slash = strchr(colon, '/');
	if (slash == NULL)
	    *colon = '\0';
	else
	    {
	    // Trim from colon to following slash
	    while ((*colon++ = *slash++) != '\0')
		;
	    }
	}
#endif // STRIP_FICUS_QUALIFIERS
    pathlen = strlen(&buf[2]);
#ifndef STRIP_FICUS_QUALIFIERS
#ifdef STRIP_COLONGP
    while (strcmp(&buf[2 + pathlen + 1] - sizeof ":gp:gp", ":gp:gp") == 0)
	{
	pathlen -= 3;
	buf[2 + pathlen] = '\0';
	}
#endif // STRIP_COLONGP
#endif // STRIP_FICUS_QUALIFIERS
    buf[0] = pathlen & 0xff;		// Convert length
    buf[1] = (pathlen >> 8) & 0xff;
    return 2 + pathlen + 1;		// 2 for length, 1 for null byte
    }

static void sendControl(		// Send a dummy control packet
    ConnectionControl	command,	// Control command to send
    struct timeval*	timestamp)	// Timestamp to assign to packet
    {
    char		packetbuf[MAX_PACKET_LENGTH];
					// Buffer for output packet
    packetbuf[0] = (char)SEER_REF_CONTROL;
    // NEEDSWORK: shouldn't hand-pack and hand-swap
    packetbuf[1] = 0;			// Control packets never fail
    packetbuf[2] = timestamp->tv_sec & 0xff;
    packetbuf[3] = (timestamp->tv_sec >> 8) & 0xff;
    packetbuf[4] = (timestamp->tv_sec >> 16) & 0xff;
    packetbuf[5] = (timestamp->tv_sec >> 24) & 0xff;
    packetbuf[6] = timestamp->tv_usec & 0xff;
    packetbuf[7] = (timestamp->tv_usec >> 8) & 0xff;
    packetbuf[8] = (timestamp->tv_usec >> 16) & 0xff;
    packetbuf[9] = (timestamp->tv_usec >> 24) & 0xff;
    packetbuf[10] = 1 & 0xff;		// Control packets always appear to
    packetbuf[11] = 1 >> 8;		// ..come from process 1
    packetbuf[12] = 1 >> 16;
    packetbuf[13] = 1 >> 24;
    packetbuf[14] = 1 & 0xff;		// Length bytes
    packetbuf[15] = 1 >> 8;		// ..(string length)
    packetbuf[16] = (char)command;
    packetbuf[17] = '\0';
    bool fakeExit;
    sendPacket(SEER_REF_CONTROL, 0, packetbuf, 18, fakeExit);
    }

static unsigned int pidhash(		// Hash a process ID
    proc_t*		procent)	// Table entry to hash
    {

    return procent->pid * (procent->pid + 3);
    }

static int pidcmp(			// Compare process ID's
    proc_t*		a,		// ID's to compare
    proc_t*		b)		// ...
    {

    return a->pid - b->pid;
    }

static unsigned int filehash(		// Hash a filename
    openfile_t*	fileent)	// File table entry
    {

    return hashString(fileent->path);
    }

static int filecmp(			// Compare two file table entries
    openfile_t*		a,		// Entries to compare
    openfile_t*		b)		// ...
    {

    return strcmp(a->path, b->path);
    }

#ifdef TRACK_CWD
static void canonicalize_pn(		// Get rid of garbage in path
    const char*		path,		// Path to canonicalize
    char*		result,		// Where to put result
    proc_t*		procent,	// Entry for referencing process
    bool		chrootApplies,	// True if need to prepend currentRoot
    struct timeval*	timestamp)	// Timestamp for any generated refs
#else // TRACK_CWD
static void canonicalize_pn(		// Get rid of garbage in path
    const char*		path,		// Path to canonicalize
    char*		result,		// Where to put result
    proc_t*		,		// Entry for referencing process
    bool		,		// True if need to prepend currentRoot
    struct timeval*	)		// Timestamp for any generated refs
#endif // TRACK_CWD
    {
    int			atdir = 1;	// NZ if we're at a dir boundary
    char*		presult;	// Pointer into result

#ifdef TRACK_CWD
    char fullPath[MAXPATHLEN];		// Space for expanded path
    if (chrootApplies)
	path = absolutePath(fullPath, path, procent, false, timestamp);
#endif // TRACK_CWD
#ifdef USE_REALPATH
    if (!avoidFilesystem  &&  realpath(path, result) != NULL)
	{
#ifdef FICUS_KNOWLEDGE
	/*
	 * Get rid of Ficus replica ID's and other Ficus information.
	 */
	char* colonp = result;
	while ((colonp = strchr(colonp, ':')) != NULL)
	    {
	    char* slashp = strchr(colonp, '/');
	    if (slashp == NULL)
		*colonp = '\0';
	    else
		(void) strcpy(colonp, slashp);
	    }
#endif // FICUS_KNOWLEDGE
	return;
	}
#endif // USE_REALPATH
    presult = result;
    while (1)
	{
	switch (*path)
	    {
	    case '\0':
		{
		if (presult > result + 1  &&  presult[-1] == '/')
		    presult--;
		*presult = '\0';
		return;
		}
#ifdef FICUS_KNOWLEDGE
	    case ':':
		{
		while (*++path != '/'  &&  *path != '\0')
		    ;
		break;
		}
#endif // FICUS_KNOWLEDGE
	    case '/':
		{
		*presult++ = '/';
		while (*++path == '/')
		    ;
		atdir = 1;
		break;
		}
	    case '.':
		{
		if (atdir)
		    {
		    if (path[1] == '/')
			path += 2;	// Skip over "./"
		    else if (path[1] == '\0')
			{
			// Delete trailing "/."
			if (presult > result)
			    presult--;
			*presult = '\0';
			return;
			}
		    else if (path[1] == '.'
		      &&  (path[2] == '/'  ||  path[2] == '\0'))
			{
			// Delete ".." reference
			// NEEDSWORK:  fails miserably on symlinks
			if (presult > result)
			    presult--;
			while (presult > result  &&  presult[-1] != '/')
			    presult--;
			if (presult == result)
			    *presult++ = '/';
			if (path[2] == '/')
			    path += 3;
			else
			    path += 2;
			}
		    else
			{
			atdir = 0;
			*presult++ = *path++;
			}
		    }
		else
		    *presult++ = *path++;
		break;
		}
	    default:
		{
		atdir = 0;
		*presult++ = *path++;
		break;
		}
	    }
	}
    }

/*
 * Convert a relative path to an absolute one, not necessarily
 * canonical.  Returns a pointer to the converted path, which may be
 * the source path if the source is already absolute and mustCopy is
 * false.  Guarantees that the target will be null-terminated and no
 * longer than MAXPATHLEN (including the null byte).
 *
 * Returns: pointer to one of source or target, whichever is absolute.
 */
#ifdef TRACK_CWD
static const char* absolutePath(
    char*		target,		// Where to put result
    const char*		source,		// Path to make absolute
    proc_t*		procent,	// Process making the reference
    bool		mustCopy,	// True if only target is OK return val
    struct timeval*	timestamp)	// Timestamp to assign generated refs
#else // TRACK_CWD
static const char* absolutePath(
    char*		target,		// Where to put result
    const char*		source,		// Path to make absolute
    proc_t*		,		// Process making the reference
    bool		mustCopy,	// True if only target is OK return val
    struct timeval*	)		// Timestamp to assign generated refs
#endif // TRACK_CWD
    {
#ifdef TRACK_CWD
    if (source[0] == '/'  &&  !mustCopy  &&  procent->currentRoot == NULL)
	return source;			// We don't have to do anything
    int targetLength = 0;
    target[MAXPATHLEN - 1] = '\0';
    if (procent->currentRoot != NULL)
	{
	/*
	 * NEEDSWORK: if the expanded path (including symlink
	 * contents) contains enough ".." references to pass the
	 * process' root, this should be turned into just the root.
	 * But this requires expanding symlinks by hand, which the
	 * observer doesn't currently do.  So we'll just assume that
	 * the path doesn't have enough up-level references to pass
	 * the root.  In some sense this is all moot anyway, since
	 * ordinary users don't normally use chroot.
	 */
	strncpy(target, procent->currentRoot->path, MAXPATHLEN - 2);
	targetLength = strlen(procent->currentRoot->path);
	if (targetLength > MAXPATHLEN - 2)
	    targetLength = MAXPATHLEN - 2;
	target[targetLength++] = '/';
	target[targetLength] = '\0';;
	}
    if (source[0] != '/')
	{
	if (procent->cwd == NULL)
	    setCWD(procent, timestamp);
	if (procent->cwd != NULL  &&  targetLength < MAXPATHLEN - 2)
	    {
	    strncpy(&target[targetLength], procent->cwd->path,
	      MAXPATHLEN - targetLength - 2);
	    targetLength += strlen(&target[targetLength]);
	    target[targetLength++] = '/';
	    target[targetLength] = '\0';;
	    }
	}
    if (targetLength < MAXPATHLEN - 1)
	strncpy(&target[targetLength], source, MAXPATHLEN - targetLength - 1);
    return target;
#else // TRACK_CWD
    if (mustCopy)
	{
	target[MAXPATHLEN - 1] = '\0';
	return strncpy(target, source, MAXPATHLEN - 1);
	}
    else
	return source;
#endif // TRACK_CWD
    }

static int addfileref(			// Add a reference to a file
    proc_t*		process,	// Process doing the reference
    const char*		path,		// Pathname of file to add
#ifdef TRACK_CWD
    bool		chrootApplies,	// True if path is non-absolute
#else // TRACK_CWD
    bool		,		// True if path is non-absolute
#endif // TRACK_CWD
    openfile_t**	tableent,	// Table entry to modify
    struct timeval*	timestamp)	// Timestamp to assign to removal
    {
    openfile_t*		foundent;	// Entry we found on lookup
#ifdef TRACK_CWD
    char		fullPath[MAXPATHLEN]; // Space to hold full-length path
#endif // TRACK_CWD
    static openfile_t*	newent = NULL;	// Entry we are creating, if needed
    int			saveit;		// NZ to save entry in log

    saveit = delfileref(process, tableent, 1, timestamp);
					// Get rid of any old reference
#ifdef TRACK_CWD
    /*
     * We must do this first because it may recursively call
     * addfileref, which might change the value of newent.
     */
    if (chrootApplies)
	path = absolutePath(fullPath, path, process, false, timestamp);
#endif // TRACK_CWD
    if (newent == NULL)
	{
	newent = (openfile_t*) malloc(sizeof(openfile_t));
	if (newent == NULL)
	    nomem("addfileref", __LINE__, sizeof(openfile_t));
	}
    newent->path = path;
    foundent =
      (openfile_t*) filetable.find((void*) newent, HashTable::INSERT);
    if (foundent == newent)
	{
	/*
	 * We just created a new entry.
	 */
	newent = NULL;
	foundent->path = strdup(path, 0);
	foundent->refcount = 1;
	}
    else
	foundent->refcount++;
    *tableent = foundent;
    return saveit;
    }

static int delfileref(			// Remove a reference to a file
    proc_t*		procent,	// Process that held reference
    openfile_t**	tableent,	// Table entry for the file
    int			sendit,		// NZ to send removal to clients
    struct timeval*	timestamp)	// Timestamp to assign to removal
    {
    int			saveit;		// NZ to save entry in log

    if (*tableent == NULL)
	return 0;
    saveit = 0;
    if (sendit)
	saveit = recordReference(procent, (*tableent)->path, NULL, false,
	  SEER_REF_CLOSE, 0, timestamp);
    if (--(*tableent)->refcount <= 0)
	{
	if (filetable.remove((void*) *tableent))
	    observerPanic("observer:  failed to remove known file entry!");
	free((void*) (*tableent)->path);
	free((void*) *tableent);
	}
    *tableent = NULL;
    return saveit;
    }

static bool remove_ref(			// Remove a file ref from a proc
    proc_t*		procent,	// Process to remove from
    openfile_t*		refent,		// File to be removed
    struct timeval*	timestamp)	// Time stamp to assign to removal
    {
    int			fdno;		// File descriptor being examined

    for (fdno = procent->dtableSize;  --fdno >= 0;  )
	{
	if (procent->dtable[fdno] == refent)
	    {
	    procent->closeOnExecFlags[fdno] = 0;
	    if (refent->refcount <= 1)
		{
		(void) delfileref(procent, &procent->dtable[fdno], 0,
		  timestamp);
		return true;		// Don't need to walk further
		}
	    else
		(void) delfileref(procent, &procent->dtable[fdno], 0,
		  timestamp);
	    }
	}
    if (procent->procname == refent)
	{
	if (refent->refcount <= 1)
	    {
	    (void) delfileref(procent, &procent->procname, 0, timestamp);
	    return true;		// Don't need to walk further
	    }
	else
	    (void) delfileref(procent, &procent->procname, 0, timestamp);
	}
    return false;
    }

#ifdef TRACK_CWD
// NEEDSWORK: there should be a setRoot as well as a setCWD.
static void setCWD(			// Set a process' CWD
    proc_t*		process,	// Process to modify
    struct timeval*	timestamp)	// Time to assign to any references
    {
    if (process->cwd != NULL  ||  replayfile != NULL)
	return;
#ifdef __linux__
    /*
     * Try to find out the process' CWD from /proc.  This only works
     * if the process is still alive.  As a side effect, the
     * observer's CWD is left in /tmp after all is done.
     */
    char procCWD[sizeof "/proc/1234567890/cwd"];
    (void)sprintf(procCWD, "/proc/%d/cwd", process->pid);
    if (chdir(procCWD) < 0)
	return;
    char CWD_Name[MAXPATHLEN];
    if (getcwd(CWD_Name, sizeof CWD_Name) == NULL)
	{
	(void)chdir(ourWD);
	return;
	}
    (void)chdir(ourWD);
    (void)addfileref(process, CWD_Name, false, &process->cwd, timestamp);
    /*
     * When we are replaying from a save file, the existing process
     * will never be correct, and so we need to make sure that setCWD
     * isn't needed.  We do this by writing a fake absolute-path chdir
     * record to the save file.  This will work because the packet
     * that had the relative path that triggered a setCWD call isn't
     * written to the save file until all other processing is
     * complete.
     */
    if (savefile != NULL)
	{
	seerbuf_t fakeBuffer;
	fakeBuffer.callcode = SEER_SYS_chdir;
	fakeBuffer.errcode = 0;
	fakeBuffer.flags = SEER_GENERATED;
	fakeBuffer.argsize =
	  (strlen(CWD_Name) + 1 + sizeof fakeBuffer.args[0] - 1)
	    / sizeof fakeBuffer.args[0];
	fakeBuffer.uid = process->uid;
	fakeBuffer.pid = process->pid;
	fakeBuffer.retval = 0;
	fakeBuffer.calltime = *timestamp;
	(void)fwrite((char*)&fakeBuffer,
	  sizeof fakeBuffer - sizeof fakeBuffer.args[0], 1, savefile);
	(void)fwrite(CWD_Name, sizeof fakeBuffer.args[0], fakeBuffer.argsize,
	  savefile);
	(void)fflush(savefile);
	}
#endif // __linux__
    }
#endif // TRACK_CWD
