#ifndef lint
static const char Rcs_Id[] =
    "$Id: packet.cc,v 1.13 1999/02/04 21:05:13 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: packet.cc,v $
 * Revision 1.13  1999/02/04 21:05:13  geoff
 * Add Inputlist::hasFd.
 *
 * Revision 1.12  1998/10/23 23:15:49  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.11  1998/04/09 00:33:02  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.10  1997/12/18 01:25:24  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.9  1997/12/04 23:16:27  geoff
 * Add InputList::removeAll to get rid of all packet sources.  Use
 * correlatorPanic to do aborts.
 *
 * Revision 1.8  1996/09/11 23:00:13  geoff
 * Add REPORT_SELECTED_FAILURES support.  Convert from dlists to slists.
 *
 * Revision 1.7  1996/05/03  01:12:51  geoff
 * Update the copyright.
 *
 * Revision 1.6  1996/02/02  20:09:52  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.5  1996/02/02  18:29:15  geoff
 * Make the waitTime argument to InputList::get be a pointer so it can
 * optionally be null.  Fix some minor problems detected by the new C++
 * compiler.
 *
 * Revision 1.4  1995/12/19  01:03:09  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.3  1995/12/19  00:34:48  geoff
 * Add process-tracking support.  Get rid of HANDLE_OLD_SAVE_FILES.  Add
 * an InputList destructor that gets rid of the old packets.
 *
 * Revision 1.2  1995/11/22  00:22:10  geoff
 * Get rid of the TIME_SORTED_INPUT and HAND_BYTE_SWAPPING ifdefs.
 *
 * Revision 1.1  1995/11/10  02:09:40  geoff
 * Initial revision
 *
 */

/*
 * Classes and routines to support the reading of packets from the
 * observer and delivering them in timestamp order.
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "correlator.hh"
#include "hashtable.hh"
#include "list.hh"
#include "packet.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			InputPacket::InputPacket(ByteSwappedStream* stream,
			  bool blockForInput);
			InputPacket::InputPacket(int fd, bool blockForInput);
			InputPacket::~InputPacket();
bool			InputPacket::get(); // Collect a packet from the stream
bool			InputPacket::write(ByteSwappedStream& stream);
					// Write a packet to a stream
			InputList::~InputList();
void	    		InputList::addFd(int fd, bool blockForInput);
					// Add a new packet source
bool			InputList::hasFd(int fd) const;
					// See if fd is in list's fd_set
void			InputList::removeSource(InputPacket* oldSource);
					// Remove a packet source
void			InputList::removeAll();
					// Remove all packet sources
void			InputList::dontBlock(InputPacket* subjectPacket);
					// Mark a stream as no-longer-blocking
InputPacket*		InputList::get(fd_set& controlFds, fd_set& badFds,
			  struct timeval* waitTime);
					// Return next input packet,
					// ..or NULL if none to be
					// ..processed
void	    		InputList::parseFds(const char* descriptor);
					// Parse ASCII into an input list
char*			InputList::saveFds() const;
					// Turn input list into ASCII string
void			InputList::loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
#endif // 0

extern int		maxfd;		// Maximum known FD + 1

extern void		correlatorPanic(const char* message);
extern void		nomem(const char* where, int line, unsigned int size);
extern "C" char*	strerr(int errnum);

InputPacket::InputPacket(
    ByteSwappedStream*	stream,		// Stream that will provide input
    bool		blockForInput)	// True to block until input is ready
    : sourceStream(stream),
      blockingStream(blockForInput),
      deleteStreamOnDestruction(false),
      valid(false),
      callSucceeded(false),
      type(SEER_REF_UNKNOWN),
      currentProcess(NULL),
      length1(0),
      pathName1(NULL),
      length2(0),
      pathName2(NULL)
    {
    timestamp.tv_sec = timestamp.tv_usec = 0;
    /*
     * We leave the buffer uninitialized.  Actually, we could leave
     * everything uninitialized if we chose, but it's probably not
     * inefficient enough to be worth the sloppiness.
     */
    }

InputPacket::InputPacket(
    int			fd,		// Fd that will provide input
    bool		blockForInput)	// True to block until input is ready
    : blockingStream(blockForInput),
      deleteStreamOnDestruction(true),
      valid(false),
      callSucceeded(false),
      type(SEER_REF_UNKNOWN),
      currentProcess(NULL),
      length1(0),
      pathName1(NULL),
      length2(0),
      pathName2(NULL)
    {
    sourceStream = new ByteSwappedStream(fd, ByteSwappedStream::littleEnded);
    timestamp.tv_sec = timestamp.tv_usec = 0;
    /*
     * We leave the buffer uninitialized.  Actually, we could leave
     * everything uninitialized if we chose, but it's probably not
     * inefficient enough to be worth the sloppiness.
     */
    }

InputPacket::~InputPacket()
    {
    if (deleteStreamOnDestruction)
	delete sourceStream;
    sourceStream = NULL;
    }

/*
 * Read an incoming observer packet from the source stream.
 *
 * Returns true if the packet was read without error, false if some
 * sort of error (formatting or stream) occurs.
 */
bool InputPacket::get()
    {
    if (valid)
	return true;
    /*
     * Every packet begins with a one-byte type, a one-byte
     * success/failure code (NZ is failure), a timestamp, and a (long)
     * process ID.
     */
    char byteType;
    if (!sourceStream->read(&byteType, sizeof byteType))
	return false;
    type = (ReferenceType)byteType;
    char byteFailure;
    if (!sourceStream->read(&byteFailure, sizeof byteFailure))
	return false;
    callSucceeded = byteFailure == 0;
    if (!sourceStream->read(&timestamp.tv_sec, sizeof timestamp.tv_sec)
      ||  !sourceStream->read(&timestamp.tv_usec, sizeof timestamp.tv_usec))
	{
	cerr << "correlator:  trouble reading timestamp on fd "
	  << sourceStream->rdbuf()->fd() << " at offset "
	  << sourceStream->bytesRead() << '\n';
	return false;
	}
    /*
     * On some systems pid_t is a long, while on others it is a short.
     * In the packet, though, it is always a long.
     */
    unsigned long processId;
    if (!sourceStream->read(&processId, sizeof processId))
	{
	cerr << "correlator:  trouble reading process ID on fd "
	  << sourceStream->rdbuf()->fd() << " at offset "
	  << sourceStream->bytesRead() << '\n';
	return false;
	}
    if (type == SEER_REF_NEWPROC)
	{
	unsigned long parentProcessId;
	if (!sourceStream->read(&parentProcessId, sizeof parentProcessId))
	    {
	    cerr << "correlator:  trouble reading process ID on fd "
	      << sourceStream->rdbuf()->fd() << " at offset "
	      << sourceStream->bytesRead() << '\n';
	    return false;
	    }
	currentProcess =
	  processes.findProcess(processId, parentProcessId, true);
	length1 = length2 = 0;
	pathName1 = pathName2 = buffer;
	buffer[0] = '\0';
	valid = true;
	return true;
	}
    currentProcess = processes.findProcess(processId, 0, true);
    unsigned short shortLength;
    if (!sourceStream->read(&shortLength, sizeof shortLength))
	{
	cerr << "correlator:  trouble reading length 1 on fd "
	  << sourceStream->rdbuf()->fd() << " at offset "
	  << sourceStream->bytesRead() << '\n';
	return false;
	}
    length1 = shortLength;
    pathName1 = &buffer[0];
    if (length1 < 0  ||  length1 >= MAXPATHLEN)
	{
	cerr << "correlator:  bad length 1 " << length1 << " on fd "
	  << sourceStream->rdbuf()->fd() << " at offset "
	  << sourceStream->bytesRead() - (pathName1 - buffer) << '\n';
	return false;
	}
    if (length1 == 0)
	pathName1 = NULL;
    else
	{
	if (!sourceStream->read((char*)pathName1, length1 + 1))
	    {
	    cerr << "correlator:  bad read on fd "
	      << sourceStream->rdbuf()->fd()
	      << " while reading path 1 at offset "
	      << sourceStream->bytesRead() << '\n';
	    return false;
	    }
	if ((int)strlen(pathName1) != length1)
	    {
	    cerr << "correlator:  bad pathName1 length "
	      << strlen(pathName1) << " for '" << pathName1 << "' on fd "
	      << sourceStream->rdbuf()->fd() << " (expected " << length1
	      << ") at offset "
	      << sourceStream->bytesRead() - (pathName1 - buffer) - length1 - 1
	      << '\n';
	    return false;
	    }
	}
    if (type != SEER_REF_RENAME)
	{
	pathName2 = NULL;
	length2 = 0;
	}
    else
	{
	/*
	 * Point pathName2 at the buffer location just past pathName1.
	 * This is where the length bytes will be read; we will later
	 * bump the pointer past the length.
	 */
	pathName2 = pathName1 + length1 + 1;
	if (!sourceStream->read(&shortLength, sizeof shortLength))
	    {
	    cerr << "correlator:  trouble reading length 2 on fd "
	      << sourceStream->rdbuf()->fd() << " at offset "
	      << sourceStream->bytesRead() << '\n';
	    return false;
	    }
	length2 = shortLength;
	if (length2 <= 0  ||  length2 >= MAXPATHLEN)
	    {
	    cerr << "correlator:  bad length 2 " << length2
	      << " on fd " << sourceStream->rdbuf()->fd() << " at offset "
	      << sourceStream->bytesRead() - (pathName2 - buffer) << '\n';
	    return false;
	    }
	if (!sourceStream->read((char*)pathName2, length2 + 1))
	    {
	    cerr << "correlator:  bad read on fd "
	      << sourceStream->rdbuf()->fd()
	      << " while reading path 2 at offset "
	      << sourceStream->bytesRead() << '\n';
	    return false;
	    }
	if ((int)strlen(pathName2) != length2)
	    {
	    cerr << "correlator:  bad pathName2 length " << strlen(pathName2)
	      << " for '" << pathName2 << "' on fd "
	      << sourceStream->rdbuf()->fd()
	      << " (expected " << length2 << ") at offset "
	      << sourceStream->bytesRead() - (pathName2 - buffer) - length2 - 1
	      << '\n';
	    return false;
	    }
	}
    valid = true;
    return true;
    }

/*
 * Write a packet to a stream in little-endian order.
 *
 * Returns true if the stream is OK, false if an I/O error occurred.
 */
bool InputPacket::write(
    ByteSwappedStream&	stream)		// Stream to write to
    {
    /*
     * We will write out the packet in pieces, and the stream's
     * byte-swapping code will take care of the details.
     */
    char byteType = (char)type;
    stream.write(&byteType, sizeof byteType);
    char errcode = callSucceeded ? 0 : ENOENT;
    stream.write(&errcode, sizeof errcode);
    stream.write(&timestamp.tv_sec, sizeof timestamp.tv_sec);
    stream.write(&timestamp.tv_usec, sizeof timestamp.tv_usec);
    unsigned long processId(pid());
    stream.write(&processId, sizeof processId);
    if (type == SEER_REF_NEWPROC)
	{
	unsigned long parentProcessId(parentPid());
	return stream.write(&parentProcessId, sizeof parentProcessId);
	}
    short shortLength = length1;
    stream.write(&shortLength, sizeof shortLength);
    if (length1 != 0)
	stream.write(pathName1, length1 + 1);
    if (pathName2 != NULL)
	{
	shortLength = length2;
	stream.write(&shortLength, sizeof shortLength);
	stream.write(pathName2, length2 + 1);
	}
    return stream;
    }

InputList::~InputList()
    {
    while (!packets.isEmpty())
	{
	InputPacket* packet = (InputPacket*)packets.popHead();
	delete packet;
	}
    }

void InputList::addFd(			// Add a new packet source
    int			fd,		// Fd for new source
    bool		blockForInput)	// True if stream will block
    {
    if (fd >= maxfd)
	maxfd = fd + 1;
    InputPacket* newPacket = new InputPacket(fd, blockForInput);
    packets.pushHead(newPacket);
    }

bool InputList::hasFd(			// See if fd is in list's fd_set
    int			fd)		// Fd to search for
    const
    {
    for (SList::Iterator element((SList&)packets);  element;  ++element)
	{
	InputPacket* packet = (InputPacket*)(packets[element]);
	if (packet->sourceStream->rdbuf()->fd() == fd)
	    return true;
	}
    return false;
    }

void InputList::removeSource(		// Remove a packet source
    InputPacket*	oldSource)	// Packet for source to be removed
    {
    packets.remove(oldSource);
    delete oldSource;
    }

void InputList::removeAll()		// Remove all packet sources
    {
    while (!packets.isEmpty())
	{
	InputPacket* source = (InputPacket*)packets.popHead();
	delete source;
	}
    }

/*
 * Mark a stream as no-longer-blocking.  If a NULL pointer is given,
 * mark all strings as non-blocking.
 */
void InputList::dontBlock(		// Mark a stream as no-longer-blocking
    InputPacket*	subjectPacket)	// Packet for stream that doesn't block
    {
    for (SList::Iterator element(packets);  element;  ++element)
	{
	InputPacket* packet = (InputPacket*)(packets[element]);
	if (subjectPacket == NULL  ||  subjectPacket == packet)
	    {
	    packet->blockingStream = false;
	    break;
	    }
	}
    }

/*
 * Return next input packet, or NULL if there are none to be processed.
 *
 * Processing has three phases.
 *
 * In phase one, all blocking streams are examined to see if they have
 * a pending packet saved.  If not, a packet is read (using a blocking
 * read).  Note that this may suspend the program indefinitely, which
 * among other things means that packets coming in on control
 * connections will be ignored.  (This routine does not handle control
 * connections, but by blocking the entire program, can prevent
 * control information from being processed.)
 *
 * In phase two, all nonblocking streams are examined to see if they
 * have a pending packet saved.  If not, and if the stream has input
 * available, a packet is read from that stream.  If no stream,
 * blocking or nonblocking, has a pending packet, the select system
 * call is used to suspend the program until one becomes available or
 * the wait time expires.
 *
 * In phase three, all streams are examined to find the stream whose
 * packet has the earliest time.  This packet is marked invalid and
 * returned.  If no such packet can be found, the routine returns
 * NULL.
 *
 * At any return, controlFds has been modified to reflect the control
 * FDs that are known to be ready, and badFds has been set to reflect
 * the control FDs that have exception conditions pending.  Both of
 * these fd_sets may also have other (garbage) bits set.
 */
InputPacket* InputList::get(
    fd_set&		controlFds,	// INOUT: Control FD's to check
    fd_set&		badFds,		// OUT: Control FD's with trouble
    struct timeval*	waitTime)	// Time to wait for a packet
    {
    /*
     * Phase one: make sure all blocking streams have a valid packet.
     * As a side effect, build a list of non-blocking FD's, for use in
     * a select.
     */
    fd_set allFds = controlFds;
    fd_set nonblockingFds = controlFds;
    SList::Iterator element(packets);
    for (  ;  element;  ++element)
	{
	InputPacket* packet = (InputPacket*)(packets[element]);
	int fd = packet->sourceStream->rdbuf()->fd();
	FD_SET(fd, &allFds);
	if (fd >= maxfd)
	    maxfd = fd + 1;
	if (packet->blockingStream)
	    {
	    if (!packet->valid)
		{
		if (!packet->get())
		    {
		    element.removeCurrent();
		    FD_CLR(fd, &allFds);
		    delete packet;
		    }
		}
	    }
	else
	    FD_SET(fd, &nonblockingFds);
	}
    /*
     * Phase two: get packets from any ready nonblocking streams.
     */
    struct timeval zeroTime = {0, 0};
    badFds = nonblockingFds;
    (void)select(maxfd, &nonblockingFds, (fd_set*)NULL, &badFds,
      &zeroTime);
    controlFds = nonblockingFds;
    bool somebodyHasInput = false;
    for (element.reset();  element;  ++element)
	{
	InputPacket* packet = (InputPacket*)(packets[element]);
	int fd = packet->sourceStream->rdbuf()->fd();
	if (!packet->blockingStream  &&  !packet->valid
	  &&  (packet->sourceStream->dataAvailable()
	    ||  FD_ISSET(fd, &nonblockingFds)))
	    {
	    if (packet->get())
		somebodyHasInput = true;
	    else
		{
		element.removeCurrent();
		FD_CLR(fd, &allFds);
		delete packet;
		}
	    }
	else if (FD_ISSET(fd, &badFds))
	    {
	    element.removeCurrent();
	    FD_CLR(fd, &allFds);
	    delete packet;
	    }
	else
	    somebodyHasInput |= packet->valid;
	}
    if (!somebodyHasInput)
	{
	badFds = allFds;
	controlFds = allFds;
	int nFds =
	  select(maxfd, &controlFds, (fd_set*)NULL, &badFds, waitTime);
	if (nFds < 0)
	    {
	    int saveErrno = errno;
	    cerr << "correlator:  select failed: " << strerr(saveErrno)
	      << '\n';
	    if (errno != EINTR)
		correlatorPanic("");
	    FD_ZERO(&controlFds);
	    FD_ZERO(&badFds);
	    }
	else
	    {
	    for (element.reset();  element;  ++element)
		{
		InputPacket* packet = (InputPacket*)(packets[element]);
		int fd = packet->sourceStream->rdbuf()->fd();
		if (FD_ISSET(fd, &controlFds))
		    {
		    if (!packet->get())
			{
			element.removeCurrent();
			FD_CLR(fd, &allFds);
			delete packet;
			}
		    }
		else if (FD_ISSET(fd, &badFds))
		    {
		    element.removeCurrent();
		    FD_CLR(fd, &allFds);
		    delete packet;
		    }
		}
	    }
	}
    /*
     * Phase three: find the earliest packet and return it.
     */
    InputPacket* earliest = NULL;
    for (element.reset();  element;  ++element)
	{
	InputPacket* packet = (InputPacket*)(packets[element]);
	if (packet->valid
	  &&  (earliest == NULL
	    ||  packet->timestamp.tv_sec < earliest->timestamp.tv_sec
	    ||  (packet->timestamp.tv_sec == earliest->timestamp.tv_sec
	      &&  packet->timestamp.tv_usec < earliest->timestamp.tv_usec)))
	    earliest = packet;
	}
    if (earliest != NULL)
	earliest->consume();
    return earliest;
    }

/*
 * Parse a comma-separated list of fd's into an input.  We do almost
 * no error checking.  All of the new streams are placed into
 * non-blocking mode.
 */
void InputList::parseFds(
    const char*		descriptor)	// String describing the FD's
    {
    for (const char* cp = descriptor;  *cp != '\0';  cp++)
	{
	addFd(atoi(cp), false);
	cp = strchr(cp, ',');
	if (cp == NULL)
	    break;
	}
    }

/*
 * Convert an InputList into a ASCII list of FD's.  Returns a pointer
 * to a new[]'ed storage area containing the descriptive string.
 *
 * This routine assumes that every fd fits into three digits or less.
 */
char* InputList::saveFds()
    const
    {
    char* returnString = new char[packets.count() * 4];
    if (returnString == NULL)
	nomem("saveFds", __LINE__, packets.count() * 4);
    char* nextSpot = returnString;
    *nextSpot = '\0';
    for (SList::Iterator iter((SList&)packets);  iter;  ++iter)
	{
	InputPacket* packet = (InputPacket*)packets[iter];
	if (nextSpot != returnString)
	    *nextSpot++ = ',';
	(void) sprintf(nextSpot, "%d", packet->sourceStream->rdbuf()->fd());
	nextSpot += strlen(nextSpot);
	}
    return returnString;
    }

/*
 * Eliminate pointers to files being forgotten
 */
void InputList::loseForgettingPointers()
    {
    for (SList::Iterator iter((SList&)packets);  iter;  ++iter)
	((InputPacket*)packets[iter])->loseForgettingPointers();
    }
