#ifndef lint
static const char Rcs_Id[] =
    "$Id: process.cc,v 1.10 1998/10/23 23:15:50 geoff Exp $";
#endif

/*
 * $Id: process.cc,v 1.10 1998/10/23 23:15:50 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: process.cc,v $
 * Revision 1.10  1998/10/23 23:15:50  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.9  1998/05/14 01:15:39  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.8  1998/04/09 00:33:03  geoff
 * Use the new notation for hash and list iterators.
 *
 * Revision 1.7  1997/12/04 23:16:31  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.6  1996/09/11 23:00:17  geoff
 * Convert from dlists to slists.
 *
 * Revision 1.5  1996/05/03  01:12:56  geoff
 * Update the copyright.  Support and maintain the open-directory count.
 *
 * Revision 1.4  1996/02/02  20:09:54  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.3  1996/02/02  18:29:17  geoff
 * Fix some minor problems detected by the new C++ compiler.  Add the
 * CHECK_PROCESS_EXITS ifdef.
 *
 * Revision 1.2  1995/12/19  01:03:11  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.1  1995/12/19  00:05:36  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "hashtable.hh"
#include "panic.hh"
#include "process.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
void			Process::loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
ostream&		operator<<(ostream& stream, const Process& process);
			FinalProcessTableDestructor::
			  FinalProcessTableDestructor();
			FinalProcessTableDestructor::
			  ~FinalProcessTableDestructor();
			ProcessTable::ProcessTable();
			ProcessTable::~ProcessTable();
Process*		ProcessTable::findProcess(unsigned long processId,
			  unsigned long parentProcessId, bool create);
					// Find or create a process entry

			ProcessTable::ProcessTable();
			ProcessTable::~ProcessTable();
void			ProcessTable::processExit(Process* process);
					// Record the exit of a given process
void			ProcessTable::loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
unsigned int		ProcessTable::processHash(Process* process);
					// Hash process-table entries
int			ProcessTable::processCompare(Process* process1,
			  Process* process2);
					// Compare process-table entries, for
					// ..hash-table routines
#endif // 0

FinalProcessTableDestructor
			ProcessTable::finalProcessTableDestructor;
Process*		ProcessTable::newEntry = NULL;
					// Scratch for table lookups/inserts/

/*
 * Eliminate pointers to files being forgotten
 */
void Process::loseForgettingPointers()
    {
    history.loseForgettingPointers();
    if (runningProgram != NULL  &&  runningProgram->isForgetting())
	runningProgram = NULL;
    for (SList::Iterator file(openFiles);  file;  ++file)
	{
	File* thisFile = (File*)openFiles[file];
	if (thisFile->isForgetting())
	    {
	    file.removeCurrent();
	    if (thisFile->isDirectory()  &&  openDirectoryCount > 0)
		openDirectoryCount--;
	    }
	}
    }

/*
 * Write a process-table entry in printable form
 */
ostream& operator<<(
    ostream&		stream,		// Stream to write to
    const Process&	process)	// Entry to display
    {
    stream << "PID " << process.processId << " parent ";
    if (process.parentProcess == NULL)
	stream << "unknown";
    else
	stream << process.parentProcess->processId;
    if (process.runningProgram != NULL)
	stream << ' ' << process.runningProgram->encodedName(SO_MIME_ENCODE);
    return stream;
    }

FinalProcessTableDestructor::FinalProcessTableDestructor()
    {
    }

FinalProcessTableDestructor::~FinalProcessTableDestructor()
    {
    if (ProcessTable::newEntry != NULL)
	{
#ifdef CHECK_PROCESS_EXITS
	/*
	 * Before we delete newEntry, we must set exitCalled, or the
	 * process destructor will bomb the correlator.
	 */
	ProcessTable::newEntry->exitCalled = true;
#endif // CHECK_PROCESS_EXITS
	delete ProcessTable::newEntry;
	ProcessTable::newEntry = NULL;
	}
    }

ProcessTable::ProcessTable()
    : HashTable(30, (unsigned int(*)(const void*))processHash,
	(int(*)(const void*, const void*))processCompare)
    {
    }

ProcessTable::~ProcessTable()
    {
    for (HashTable::Iterator process(*this);  process;  ++process)
	{
	Process* goingProcess = (Process*)((*this)[process]);
#ifdef CHECK_PROCESS_EXITS
	/*
	 * Before we delete the process, we must set exitCalled, or
	 * the process destructor will bomb the correlator.
	 *
	 * NEEDSWORK: the history tables for all processes should be
	 * collapsed, and eventually inserted into the global history
	 * table./
	 */
	goingProcess->exitCalled = true;
#endif // CHECK_PROCESS_EXITS
	delete goingProcess;
	}
    }

/*
 * Find or create a process-table entry.  The parent process ID is
 * assumed valid only if it is nonzero.
 */
Process* ProcessTable::findProcess(	// Record spawning of a new process
    unsigned long	processId,	// PID of new process
    unsigned long	parentProcessId, // PID of process' parent
    bool		create)		// True to create if not found
    {
    if (newEntry == NULL)
	newEntry = new Process(processId);
    else
	newEntry->processId = processId;
    Process* foundEntry = (Process*)find((void*)newEntry,
      create ? HashTable::INSERT : HashTable::FAIL);
    if (foundEntry == NULL)
	return NULL;
    else if (foundEntry == newEntry)
	{
	/* 
	 * The entry is new.  If a parent process ID is not specified,
	 * assume the process is a child of process 1.
	 */
	newEntry = NULL;
	if (parentProcessId == 0)
	    parentProcessId = 1;
	/*
	 * Report whenever the process-table size grows past a
	 * 100-process boundary.
	 */
	if (verbose)
	    {
	    static unsigned long lastsize = 0;
	    if (size() % 100 == 0  &&  size() > lastsize)
		{
		lastsize = size();
		cerr << "correlator:  process table (unknown input fd) ";
		showStatistics(cerr);
		cerr << '\n';
		if (verbose > 1)
		    {
		    for (HashTable::Iterator entry(*this);  entry;  ++entry)
			cerr << *(Process*)((*this)[entry]) << '\n';
		    }
		cerr.flush();
		}
	    }
	}
    /*
     * If necessary, set the parent process pointer.  Note that the
     * recursive calls may have unexpected effects on newEntry, so we
     * can't depend on its value after this point.
     */
    if (parentProcessId != 0
      &&  (foundEntry->parent() == NULL
	||  foundEntry->parent()->pid() != parentProcessId))
	{
	if (processId == 1)
	    foundEntry->parentProcess = foundEntry;
	else
	    {
	    foundEntry->parentProcess = findProcess(parentProcessId, 0, false);
	    if (foundEntry->parentProcess != NULL)
		{
		foundEntry->runningProgram =
		  foundEntry->parentProcess->runningProgram;
		foundEntry->history = foundEntry->parentProcess->history;
		foundEntry->flags |=
		  foundEntry->parentProcess->flags & Process::PF_INHERIT_LIST;
		}
	    }
	}
    return foundEntry;
    }

void ProcessTable::processExit(		// Record the exit of a given process
    Process*		process)	// Process that is going away
    {
    if (process->processId == 1)
	correlatorPanic("correlator:  unexpected exit for process 1");
    /*
     * Combine our history table with the parent's.
     */
    if (process->parentProcess != NULL  &&  process->parentProcess != process)
	process->parentProcess->history.combine(process->history);
    /*
     * Scan the hash table, looking for processes that are children of
     * this one.  Reparent them to be children of our parent, which is
     * what happens in the Unix kernel.
     */
    for (HashTable::Iterator iterator(*this);  iterator;  ++iterator)
	{
	Process* otherProcess = (Process*)((*this)[iterator]);
	if (otherProcess == process)
	    iterator.removeCurrent();	// Remove process from hash table
	else if (otherProcess->parentProcess == process)
	    otherProcess->parentProcess = process->parentProcess;
	}
#ifdef CHECK_PROCESS_EXITS
    process->exitCalled = true;
#endif // CHECK_PROCESS_EXITS
    delete process;
    }

/*
 * Eliminate pointers to files being forgotten
 */
void ProcessTable::loseForgettingPointers()
    {
    for (HashTable::Iterator iterator(*this);  iterator;  ++iterator)
	((Process*)((*this)[iterator]))->loseForgettingPointers();
    }

unsigned int ProcessTable::processHash(	// Hash process-table entries
    Process*		process)
    {
    return process->processId * (process->processId + 3);
    }

/*
 * Compare process-table entries, for use by the hash-table routines
 */
int ProcessTable::processCompare(
    Process*		process1,	// Processes to compare
    Process*		process2)	// ...
    {
    return process1->processId - process2->processId;
    }
