#ifndef lint
static const char Rcs_Id[] =
    "$Id: readent.cc,v 1.18 1997/12/18 01:25:28 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: readent.cc,v $
 * Revision 1.18  1997/12/18 01:25:28  geoff
 * Make some function arguments const char* instead of char*.
 *
 * Revision 1.17  1997/03/13 23:00:16  geoff
 * Fix a byte-swapping bug in dealing with restart records.
 *
 * Revision 1.16  1996/10/11 22:09:11  geoff
 * Add open/close/truncate support.  Never panic the caller; just return
 * NULL on failures instead.  Improve some error messages.
 *
 * Revision 1.15  1996/09/23 23:41:28  geoff
 * Add code to accept the timestamp entry as a file version header.  This
 * is very kludgy code because I didn't realize it would be necessary
 * until some save files already existed.
 *
 * Revision 1.14  1996/06/11 06:13:27  geoff
 * Make the "unrecognized buffer version" message slightly more
 * informative.
 *
 * Revision 1.13  1996/05/03  01:12:59  geoff
 * Update the copyright.
 *
 * Revision 1.12  1996/03/19  18:43:11  geoff
 * Add code to work around older observers that didn't set the call time
 * correctly in RESTART records.
 *
 * Revision 1.11  1996/02/02  20:09:56  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.10  1996/02/02  18:29:19  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.9  1995/12/19  00:34:52  geoff
 * Add seerversion() to report the current buffer version.  Support a new
 * buffer version.  Make the bad-version error message report the file
 * offset.
 *
 * Revision 1.8  1995/11/10  02:59:58  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.7  1995/10/10  20:38:15  geoff
 * Add byte-swapping support, including code that tries to detect what
 * sort of swapping is needed.
 *
 * Revision 1.6  1995/09/15  00:06:14  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.
 *
 * Revision 1.5  1995/08/07  20:03:07  geoff
 * Add byte offsets to error messages.  Fix some bugs in the
 * compatibility processing of time stamps.
 *
 * Revision 1.4  1995/06/30  19:04:48  geoff
 * Add support for the new data-structure version, and support for
 * reading save files in the old format.  Add seertell().
 *
 * Revision 1.3  1995/06/17  00:11:54  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.
 *
 * Revision 1.2  1995/05/05  22:42:01  geoff
 * Add byte-swapping support of the replay file for 386 boxes (but only
 * on replay, so it's a major kludge).
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cdecls.hh"
#include "communication.hh"
#include "zzinlinefunctions.hh"
#include <errno.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <seer/seer.h>
#include <sys/param.h>
#include <sys/stat.h>

FILE*			seeropen(const char* fileName);
int			seerclose(FILE* replayFile);
void			seertruncate(const char* fileName);
seerbuf_t*		readent(FILE* replayfile);
void			setseerswap(bool needToSwap);
unsigned long		seertell(FILE* replayFile);
int			seerversion();

extern char*		progname;	// Name we were run under

static char		buf[SEERBUFSIZE]; // Bfr to hold incoming data
static long		bufbytes = 0;	// No. of bytes currently in buffer
static unsigned long	lastValidOffset; // Offset of last valid record
#ifdef __i386__
static bool		needSwap = true; // True to byte-swap input
#else // __i386__
static bool		needSwap = false; // True to byte-swap input
#endif // __i386__
static bool		replayIsPipe;	// True if replaying from pipe
static int		seer_buffer_version = 0; // Buffer version expected now
static seerbuf_t*	seerp = (seerbuf_t*) buf; // Ptr to next seer structure
static unsigned long	totalBytesRead = 0; // Bytes read so far, for seertell

#define swapshort(x) (((x) & 0xff) << 8) | (((x) >> 8) & 0xff)
#define swaplong(x) (((x) & 0xff) << 24) | (((x) & 0xff00) << 8) \
  | (((x) >> 8) & 0xff00) | (((x) >> 24) & 0xff)

extern void		nomem(const char* where, int line, unsigned int size);
extern "C" char*	strerr(int errnum);

/*
 * Open an observer save file to be read with readent().  The file
 * should be closed with seerclose.  Automatically handles compressed
 * files.
 *
 * Only one replay file can be open at a time.
 */
FILE* seeropen(
    const char*		fileName)	// Name of file to be opened
    {
    lastValidOffset = 0;
    /*
     * A NULL filename means use stdin.
     */
    if (fileName == NULL)
	{
	replayIsPipe = false;
	return stdin;
	}
    FILE* replayFile;
    char* dotPointer = strrchr(fileName, '.');
    if (dotPointer != NULL  &&  strcmp(dotPointer, ".gz") == 0)
	{
	replayIsPipe = true;
	char* command = (char*)malloc(strlen(fileName) + sizeof "gzcat ");
	if (command == NULL)
	    nomem("replaydata", __LINE__,
	      strlen(fileName) + sizeof "gzcat ");
	strcpy(command, "gzcat ");
	strcat(command, fileName);
	replayFile = popen(command, "r");
	free((void*)command);
	}
    else
	{
	replayIsPipe = false;
	replayFile = fopen(fileName, "r");
	}
    return replayFile;
    }

int seerclose(				// Close file opened with seeropen
    FILE*		replayFile)	// File to be closed
    {
    if (replayIsPipe)
	return pclose(replayFile);
    else
	return fclose(replayFile);
    }

/*
 * Truncate a replay file to the last valid offset.  Truncation should
 * only be performed after the file has been closed, and before the
 * next one has been opened.  Truncation cannot be performed on
 * compressed files.
 */
void seertruncate(
    const char*		fileName)	// Name of file to truncate
    {
    char* dotPointer = strrchr(fileName, '.');
    if (dotPointer != NULL  &&  strcmp(dotPointer, ".gz") == 0)
	cerr << progname << ": cannot truncate compressed file '" << fileName
	  << "'\n";
    else
	{
	struct stat fileStatus;
	if (stat(fileName, &fileStatus) < 0)
	    return;
	else if ((unsigned)fileStatus.st_size > lastValidOffset)
	    {
	    cerr << progname << ":  truncating '" << fileName << "' to "
	      << lastValidOffset << " bytes\n";
	    if (truncate(fileName, lastValidOffset) < 0)
		{
		int saveErrno = errno;
		cerr << progname << ":  trouble truncating '" << fileName
		  << "': " << strerr(saveErrno) << '\n';
		}
	    }
	}
    }

/*
 * Read a seerbuf_t entry from a file, and return a pointer to it.
 * The pointer returned is *not* malloc'ed, but rather is a pointer
 * into an internal buffer.  This routine may be used on only one file
 * at a time.
 *
 * This routine keeps track of the entry format version number (from
 * the SEER_RESTART record) and converts older-style entries into the
 * latest style.  The caller is also allowed to change the version by
 * setting the appropriate global.
 */
seerbuf_t* readent(			// Read a trace entry from a save file
    FILE*		replayfile)	// File to read from
    {
    /*
     * Current-format buffer to hold a copy of old-format entries.
     * The union is just to make sure it is padded to an adequate
     * size.  We depend on the fact that this is initialized to zero.
     */
    static union			// Space for alternate-format return
	{
	seerbuf_t	buf;		// Buffer to hold copy of old-format
	char		padding[MAXPATHLEN * 2
			  + sizeof(seerbuf_t) - sizeof seerp->args[0]
			  + 2 * sizeof seerp->args[0]];
	}
			alternate_return;
    if (bufbytes < (int)(sizeof(seerbuf_t) - sizeof seerp->args[0]))
	{
	if (bufbytes != 0)
	    (void) memcpy(&buf[0], (char*) seerp, bufbytes);
	seerp = (seerbuf_t*) &buf[0];
	long nbytes = fread(&buf[bufbytes], sizeof buf[0],
	  sizeof buf - bufbytes, replayfile);
	if (nbytes == 0)
	    {
	    if (feof(replayfile))
		{
		if (bufbytes != 0)
		    cerr << progname << ": early EOF in readent\n";
		}
	    else
		{
		int saveErrno = errno;
		cerr << progname << ":  I/O error in readent: "
		  << strerr(saveErrno) << '\n';
		}
	    bufbytes = 0;
	    return NULL;
	    }
	bufbytes += nbytes;
	}
    /*
     * Detect whether we need to do byte swapping.  We will do this
     * only if it looks like the current entry is a RESTART entry.  We
     * assume that all flags are below 0x80.  There is a historical
     * save file in big-endian format that does not begin with a
     * RESTART entry, so by default we swap on the 386 and not on the
     * Sparc.  Also, when reading from /dev/seer itself, we will never
     * see a RESTART entry, so the caller is responsible for setting
     * needSwap to false in this case.
     */
    if ((seerp->flags & (SEER_RESTART << 8)) != 0)
	needSwap = true;
    else if ((seerp->flags & SEER_RESTART) != 0)
	needSwap = false;
    /*
     * Due to an oversight when I wrote the buffer-switching code,
     * there are some save files that start with SEER_TIMESTAMP
     * records instead of SEER_RESTART records.  To make matters even
     * more fun, they don't include a version number, setting it to
     * zero instead.  As it happens, all of these are buffer version
     * 3.  The following code handles this.
     */
    if (totalBytesRead == 0  &&  (seerp->flags & (SEER_TIMESTAMP << 8)) != 0)
	{
	needSwap = true;
	if (seerp->errcode == 0)
	    seer_buffer_version = 3;
	else
	    seer_buffer_version = seerp->errcode;
	}
    else if (totalBytesRead == 0  &&  (seerp->flags & SEER_TIMESTAMP) != 0)
	{
	needSwap = false;
	if (seerp->errcode == 0)
	    seer_buffer_version = 3;
	else
	    seer_buffer_version = seerp->errcode;
	}
    if (needSwap)
	seerp->flags = swapshort(seerp->flags);
    if (seerp->flags & SEER_RESTART)
	seer_buffer_version = seerp->errcode;
    seerbuf_t* entryp;
    long entsize;
    switch (seer_buffer_version)
	{
	case 0:
	    {
	    seerbuf_v0_t* version0p = (seerbuf_v0_t*)seerp;
	    entryp = &alternate_return.buf;
	    // Convert old format into new.
	    entryp->callcode = version0p->callcode;
	    entryp->errcode = version0p->errcode;
	    entryp->flags = version0p->flags;
	    entryp->argsize = version0p->argsize;
	    entryp->uid = version0p->uid;
	    entryp->pid = version0p->pid;
	    entryp->retval = version0p->retval;
	    entsize = sizeof(seerbuf_v0_t);
	    break;
	    }
	case 1:
	case SEERBUF_VERSION:
	    {
	    entryp = seerp;
	    entsize = sizeof(seerbuf_t);
	    break;
	    }
	default:
	    {
	    cerr << "readent: unrecognized buffer version "
	      << seer_buffer_version << " at offset " << seertell(replayfile)
	      << '\n';
	    return NULL;
	    break;
	    }
	}
    if (needSwap)
	{
	entryp->argsize = swapshort(entryp->argsize);
	entryp->uid = swapshort(entryp->uid);
	entryp->pid = swaplong(entryp->pid);
	entryp->retval = swaplong(entryp->retval);
	}
    entsize += (entryp->argsize - 1) * sizeof entryp->args[0];
    if (bufbytes < entsize)
	{
	// Need to read more to complete current entry
	(void) memcpy(&buf[0], (char*) seerp, bufbytes);
	seerp = (seerbuf_t*) &buf[0];
	long nbytes = fread(&buf[bufbytes], sizeof buf[0],
	  sizeof buf - bufbytes, replayfile);
	if (nbytes == 0)
	    {
	    if (feof(replayfile))
		{
		if (bufbytes != 0)
		    cerr << progname << ":  early EOF in readent at offset "
		      << seertell(replayfile) << '\n';
		}
	    else
		{
		int saveErrno = errno;
		cerr << progname << ":  I/O error in readent at offset "
		  << seertell(replayfile) << ": " << strerr(saveErrno) << '\n';
		}
	    bufbytes = 0;
	    return NULL;
	    }
	bufbytes += nbytes;
	switch (seer_buffer_version)
	    {
	    case 0:
		{
		break;
		}
	    case 1:
	    case SEERBUF_VERSION:
		{
		entryp = seerp;
		break;
		}
	    }
	}
    /*
     * Do some validity checks on the entry.  These are:
     *
     * 1. Total entry size less than SEERBUFSIZE.
     * 2. Total argument size less than 2 * MAXPATHLEN, with
     *	  slop left over for null bytes.
     */
    if (entsize < 0
      ||  entsize > SEERBUFSIZE
      ||  entryp->argsize * sizeof entryp->args[0]
	> 2 * MAXPATHLEN + 2 * sizeof(entryp->args[0]))
	{
	cerr << progname << ":  validity check failed in readent at offset "
	  << seertell(replayfile) << '\n';
	cerr << progname << ":  entsize = " << entsize << ", argsize = "
	  << entryp->argsize << ", pid = " << entryp->pid << '\n';
	return NULL;
	}
    /*
     * Copy arguments into alternate return, if necessary.
     */
    switch (seer_buffer_version)
	{
	case 0:
	    {
	    if (entryp->argsize != 0)
		(void) memcpy((char*)&entryp->args[0],
		  (char*)&((seerbuf_v0_t*)seerp)->args[0],
		  sizeof entryp->args[0] * entryp->argsize);
	    if (entryp->flags & SEER_TIMESTAMP)
		{
		if (needSwap)
		    entryp->calltime.tv_sec = swaplong(entryp->args[0]);
		else
		    entryp->calltime.tv_sec = entryp->args[0];
		}
	    break;
	    }
	case 1:
	case SEERBUF_VERSION:
	    {
	    if (needSwap)
		{
		entryp->calltime.tv_sec = swaplong(entryp->calltime.tv_sec);
		entryp->calltime.tv_usec = swaplong(entryp->calltime.tv_usec);
		}
	    /*
	     * Some older observers didn't set the calltime in RESTART records.
	     */
	     if (entryp->flags & SEER_RESTART)
		{
		if (entryp->calltime.tv_sec != entryp->args[0])
		    {
		    entryp->calltime.tv_sec = entryp->args[0];
		    entryp->calltime.tv_usec = 0;
		    }
		}
	    break;
	    }
	}
    seerp = (seerbuf_t*) ((char*) seerp + entsize);
    bufbytes -= entsize;
    totalBytesRead += entsize;
    lastValidOffset += entsize;
    /*
     * Byte-swap any calls that have non-string arguments.
     */
    if (needSwap)
	{
	if (entryp->flags & (SEER_RESTART | SEER_TIMESTAMP))
	    entryp->args[0] = swaplong(entryp->args[0]);
	switch (entryp->callcode)
	    {
	    case SEER_SYS_open:
		{
		entryp->args[entryp->argsize - 1] =
		  swaplong(entryp->args[entryp->argsize - 1]);
		break;
		}
	    case SEER_SYS_exit:
	    case SEER_SYS_fork:
	    case SEER_SYS_vfork:
	    case SEER_SYS_close:
	    case SEER_SYS_dup:
	    case SEER_SYS_fchdir:
	    case SEER_SYS_fchroot:
		{
		entryp->args[0] = swaplong(entryp->args[0]);
		break;
		}
	    case SEER_SYS_dup2:
		{
		entryp->args[0] = swaplong(entryp->args[0]);
		entryp->args[1] = swaplong(entryp->args[1]);
		break;
		}
	    case SEER_SYS_fcntl:
		{
		entryp->args[0] = swaplong(entryp->args[0]);
		entryp->args[1] = swaplong(entryp->args[1]);
		entryp->args[2] = swaplong(entryp->args[2]);
		break;
		}
	    }
	}
    return entryp;
    }

/*
 * Set the needSwap flag to reflect whether the caller thinks swapping
 * is necessary.  This may be overridden by readent.
 */
void setseerswap(
    bool		needToSwap)	// True if caller wants swapping
    {
    needSwap = needToSwap;
    }

/*
 * Give the current offset in the input file.  The file must be the
 * same one that readent() is currently reading from.
 *
 * Note that in the current implementation, the offset given is
 * calculated across all input files.
 */
unsigned long seertell(
    FILE*		)		// File we're reading from
    {
    return totalBytesRead;
    }

/*
 * Provide the current buffer version number
 */
int seerversion()
    {
    return seer_buffer_version;
    }
