#ifndef lint
static const char Rcs_Id[] =
    "$Id: savemanager.cc,v 1.6 1998/10/23 23:15:54 geoff Exp $";
#endif

/*
 * $Id: savemanager.cc,v 1.6 1998/10/23 23:15:54 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: savemanager.cc,v $
 * Revision 1.6  1998/10/23 23:15:54  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.5  1998/07/15 21:43:54  geoff
 * Reflect the name change for saveTable in correlator.cc.
 *
 * Revision 1.4  1998/05/14 01:15:39  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.3  1997/12/04 23:16:34  geoff
 * Don't include panic.hh; it's not needed any more.
 *
 * Revision 1.2  1996/09/11 23:00:18  geoff
 * Don't do saves in simulation mode.
 *
 * Revision 1.1  1996/05/03  01:13:47  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "correlator.hh"
#include "byteswappedstream.hh"
#include "savemanager.hh"
#include "internaltimer.hh"
#include "zzinlinefunctions.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

/*
 * Classes used to support the save "manager"
 */

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			SaveManager::SaveManager(const char* managerName,
			  const char* managerDefaults);
			SaveManager::~SaveManager();
void			SaveManager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
void			SaveManager::setParameter(const char* description);
					// Set management parameters
void			SaveManager::showParameters(ostream& stream) const;
					// Display management parameters
ostream&		SaveManager::showStatistics(ostream& stream) const;
					// Display collected quality statistics
#endif // 0

SaveManager saveManager("Save",
			  "Save::managementInterval=3600\n");

extern void		saveTable();	// Write correlator database to file

SaveManager::SaveManager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : EmptyManagerBase(managerName, managerDefaults)
    {
    EmptyManagerBase::setManagerParameters(defaults);
    }

SaveManager::~SaveManager()
    {
    }

void SaveManager::manage(		// Run management
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	,		// Stream to dump results to
    bool		)		// True to encode filenames in dump
    {
    lastManagementTime = now;
    if (simulationMode)
	return;
    saveTable();
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void SaveManager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad SaveManager parameter format in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("managementInterval"))
	managementInterval = atol(valueStart);
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized SaveManager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters
 */
void SaveManager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "Save::managementInterval=" << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n";
    }

/*
 * Display collected quality statistics
 */
ostream& SaveManager::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    return stream;
    }
