#ifndef lint
static const char Rcs_Id[] =
    "$Id: showent.cc,v 1.16 1997/12/18 01:25:30 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: showent.cc,v $
 * Revision 1.16  1997/12/18 01:25:30  geoff
 * Make some function arguments const char* instead of char*.
 *
 * Revision 1.15  1997/07/14 21:43:38  geoff
 * Add support for the mkfifo call.
 *
 * Revision 1.14  1996/10/11 22:09:12  geoff
 * Fix a nasty little bug that could cause the same memory to be freed twice.
 *
 * Revision 1.13  1996/10/04 07:12:23  geoff
 * Show the raw time in restart records as well as the converted time.
 *
 * Revision 1.12  1996/09/11 23:00:20  geoff
 * Add TIMEVALS_ARE_INTS support.
 *
 * Revision 1.11  1996/05/03  01:13:02  geoff
 * Update the copyright.  Add support for a table of process names.
 *
 * Revision 1.10  1996/03/19  18:43:12  geoff
 * Add utime/utimes support.  Get rid of some places that had doubled
 * semicolons.
 *
 * Revision 1.9  1996/02/02  20:09:56  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.8  1995/11/10  03:00:01  geoff
 * Add zzinlinefunctions.hh.  Add SEER_GENERATED support.
 *
 * Revision 1.7  1995/10/10  20:38:17  geoff
 * Get rid of the byte-swapping support, since it's now wired into
 * readent.  Add oldlstat and uselib.
 *
 * Revision 1.6  1995/09/15  00:06:20  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.  Put out open flags in symbolic form.
 *
 * Revision 1.5  1995/08/07  20:03:09  geoff
 * Add the printInHex macro to allow a more maintainable way of printing
 * integers in hexadecimal.  Add fcntl support.  Put out the decimal
 * value of the timestamp in TIMESTAMP records.  Put out the error code
 * as an unsigned integer, not a character.
 *
 * Revision 1.4  1995/06/30  19:04:51  geoff
 * Add code to display the timestamp on entries.
 *
 * Revision 1.3  1995/06/17  00:11:59  geoff
 * Use symbolic values for all system call numbers.
 *
 * Revision 1.2  1995/05/05  22:42:06  geoff
 * Add byte-swapping support of the replay file for 386 boxes (but only
 * on replay, so it's a major kludge).
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "cdecls.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <fcntl.h>
#include <iostream.h>
#include <seer/seer.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

void			showent(seerbuf_t* seerp, const char* procNames[],
			  int maxNames);
static int		show_pn(seerbuf_t* seerp, int pathno);

/*
 * The following macro recasts an integer so that it will print in
 * hex.  This is truly ugly, but it works.
 */
#define printInHex(x)	((void*)(x))

extern char*		strdup(const char* string, int size);

void showent(			// Display a trace entry on cout
    seerbuf_t*		seerp,	// Trace entry to be dumped
    const char*		procNames[], // Table of process names
    int			maxNames) // Size of process-name table
    {
    /*
     * Handle special buffer entries
     */
    if (seerp->flags & SEER_RESTART)
	{
	cout << "RESTART (" << (seerp->callcode ? "reexec" : "scratch")
	  << ") at " << seerp->args[0] << ' ' << ctime((long*)&seerp->args[0]);
	for (int i = 0;  i < maxNames;  i++)
	    {
	    if (procNames[i] != NULL)
		{
		free((void*)procNames[i]);
		procNames[i] = NULL;
		}
	    }
	return;
	}
    else if (seerp->flags & SEER_TIMESTAMP)
	{
	cout << "TIMESTAMP " << seerp->args[0] << ' '
	  << ctime((long*)&seerp->args[0]);
	return;
	}
    /*
     * Give special handling to fork and vfork so that we can give the
     * name of the process doing the fork.
     */
    if (seerp->callcode == SEER_SYS_fork
      ||  seerp->callcode == SEER_SYS_vfork)
	{
	if (seerp->pid < maxNames)
	    {
	    if (procNames[seerp->pid] != NULL)
		free((void*)procNames[seerp->pid]);
	    procNames[seerp->pid] = NULL;
	    if (seerp->args[0] < maxNames
	      &&  procNames[seerp->args[0]] != NULL)
		procNames[seerp->pid] = strdup(procNames[seerp->args[0]], 0);
	    }
	}
    /*
     * Handle standard buffer entries
     */
    char microbuf[8];
#ifdef TIMEVALS_ARE_INTS
    sprintf (microbuf, ".%6.6ld", (long)seerp->calltime.tv_usec);
#else // TIMEVALS_ARE_INTS
    sprintf (microbuf, ".%6.6ld", seerp->calltime.tv_usec);
#endif // TIMEVALS_ARE_INTS
    const char* procName =
      seerp->pid < maxNames ? procNames[seerp->pid] : (const char*)NULL;
    if (procName == NULL)
	procName = "??";
    cout << "UID " << seerp->uid << " PID " << seerp->pid << ' '
      << procName << ' '
      << ((seerp->flags & SEER_BEFORE) ? "B" : "")
      << ((seerp->flags & SEER_AFTER) ? "A" : "")
      << ((seerp->flags & SEER_SPECIAL) ? "S" : "")
      << ((seerp->flags & SEER_GENERATED) ? "G" : "")
      << ' '
      << seerp->calltime.tv_sec << microbuf
      << ' ';
    switch (seerp->callcode)
	{
	case SEER_SYS_exit:
	    {
	    cout << "exit(" << seerp->args[0] << ')';
	    if (seerp->pid < maxNames  &&  procNames[seerp->pid] != NULL)
		{
		free((void*)procNames[seerp->pid]);
		procNames[seerp->pid] = NULL;
		}
	    break;
	    }
	case SEER_SYS_fork:
	    {
	    cout << "fork(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_open:
	    {
	    cout << "open(\"";
	    int argno = show_pn(seerp, 0);
	    cout << "\", ";
	    bool needBar = false;
	    if (seerp->args[argno] & SEER_O_APPEND)
		{
		cout << "O_APPEND";
		needBar = true;
		}
	    if (seerp->args[argno] & SEER_O_CREAT)
		{
		if (needBar)
		    cout << '|';
		cout << "O_CREAT";
		needBar = true;
		}
	    if (seerp->args[argno] & SEER_O_TRUNC)
		{
		if (needBar)
		    cout << '|';
		cout << "O_TRUNC";
		needBar = true;
		}
	    if (seerp->args[argno] & SEER_O_EXCL)
		{
		if (needBar)
		    cout << '|';
		cout << "O_EXCL";
		needBar = true;
		}
	    if (needBar)
		cout << '|';
	    switch (seerp->args[argno] & SEER_ACCMODE)
		{
		case 0:
		    {
		    /* ( */ cout << "O_RDONLY)";
		    break;
		    }
		case 1:
		    {
		    /* ( */ cout << "O_WRONLY)";
		    break;
		    }
		case 2:
		    {
		    /* ( */ cout << "O_RDWR)";
		    break;
		    }
		case 3:
		    {
		    /* ( */ cout << "3)";
		    break;
		    }
		default:
		    panic("showent: weird access mode!");
		    break;
		}
	    break;
	    }
	case SEER_SYS_close:
	    {
	    cout << "close(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_creat:
	    {
	    cout << "creat(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_link:
	    {
	    cout << "link(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\", \"";
	    (void) show_pn(seerp, 1);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_unlink:
	    {
	    cout << "unlink(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_execv:
	    {
	    cout << "execv(\"";
	    if (seerp->flags & SEER_BEFORE)
		{
		(void) show_pn(seerp, 0);
		if (seerp->pid < maxNames)
		    {
		    if (procNames[seerp->pid] != NULL)
			free((void*)procNames[seerp->pid]);
		    procNames[seerp->pid] = strdup((char*)&seerp->args[0], 0);
		    }
		}
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_chdir:
	    {
	    cout << "chdir(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_mknod:
	    {
	    cout << "mknod(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_oldstat:
	    {
	    cout << "oldstat(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_utime:
	    {
	    cout << "utime(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_access:
	    {
	    cout << "access(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_stat:
	    {
	    cout << "stat(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_lstat:
	    {
	    cout << "lstat(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_dup:
	    {
	    cout << "dup(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_symlink:
	    {
	    cout << "symlink(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\", \"";
	    (void) show_pn(seerp, 1);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_readlink:
	    {
	    cout << "readlink(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_execve:
	    {
	    cout << "execve(\"";
	    if (seerp->flags & SEER_BEFORE)
		{
		(void) show_pn(seerp, 0);
		if (seerp->pid < maxNames)
		    {
		    if (procNames[seerp->pid] != NULL)
			free((void*)procNames[seerp->pid]);
		    procNames[seerp->pid] = strdup((char*)&seerp->args[0], 0);
		    }
		}
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_chroot:
	    {
	    cout << "chroot(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_vfork:
	    {
	    cout << "vfork(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_dup2:
	    {
	    cout << "dup2(" << seerp->args[0] << ", " << seerp->args[1]
	      << ')';
	    break;
	    }
	case SEER_SYS_fcntl:
	    {
	    cout << "fcntl(" << seerp->args[0];
	    switch (seerp->args[1])
		{
		case F_DUPFD:	/* Duplicate fildes */
		    {
		    cout << ", F_DUPFD, " << seerp->args[2];
		    break;
		    }
		case F_GETFD:	/* Get fildes flags (close on exec) */
		    {
		    cout << ", F_GETFD, " << seerp->args[2];
		    break;
		    }
		case F_SETFD:	/* Set fildes flags (close on exec) */
		    {
		    cout << ", F_SETFD, " << seerp->args[2];
		    break;
		    }
		case F_GETFL:	/* Get file flags */
		    {
		    cout << ", F_GETFL, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_SETFL:	/* Set file flags */
		    {
		    cout << ", F_SETFL, " << printInHex(seerp->args[2]);
		    break;
		    }
#ifdef F_GETOWN
		case F_GETOWN:	/* Get owner - for ASYNC */
		    {
		    cout << ", F_GETOWN, " << seerp->args[2];
		    break;
		    }
		case F_SETOWN:	/* Set owner - for ASYNC */
		    {
		    cout << ", F_SETOWN, " << seerp->args[2];
		    break;
		    }
#endif // F_GETOWN
		case F_GETLK:	/* Get record-locking information */
		    {
		    cout << ", F_GETLK, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_SETLK:	/* Set or Clear a record-lock (Non-Blocking) */
		    {
		    cout << ", F_SETLK, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_SETLKW:	/* Set or Clear a record-lock (Blocking) */
		    {
		    cout << ", F_SETLKW, " << printInHex(seerp->args[2]);
		    break;
		    }
#ifdef F_RGETLK
		case F_RGETLK:	/* Test remote lock to see if it is blocked */
		    {
		    cout << ", F_RGETLK, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_RSETLK:	/* Set or unlock a remote lock */
		    {
		    cout << ", F_RSETLK, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_CNVT:	/* Convert a fhandle to an open fd */
		    {
		    cout << ", F_CNVT, " << printInHex(seerp->args[2]);
		    break;
		    }
		case F_RSETLKW:	/* Set or Clear remote record-lock(Blocking) */
		    {
		    cout << ", F_RSETLKW, " << printInHex(seerp->args[2]);
		    break;
		    }
#endif // F_RGETLK
		default:
		    {
		    cout << ", " << seerp->args[1] << ", "
		      << printInHex(seerp->args[2]);
		    break;
		    }
		}
	    cout << ')';
	    break;
	    }
	case SEER_SYS_rename:
	    {
	    cout << "rename(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\", \"";
	    (void) show_pn(seerp, 1);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_truncate:
	    {
	    cout << "truncate(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_mkdir:
	    {
	    cout << "mkdir(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_rmdir:
	    {
	    cout << "rmdir(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_utimes:
	    {
	    cout << "utimes(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_fchdir:
	    {
	    cout << "fchdir(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_fchroot:
	    {
	    cout << "fchroot(" << seerp->args[0] << ')';
	    break;
	    }
	case SEER_SYS_oldlstat:
	    {
	    cout << "oldlstat(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_uselib:
	    {
	    cout << "uselib(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	case SEER_SYS_mkfifo:
	    {
	    cout << "mkfifo(\"";
	    (void) show_pn(seerp, 0);
	    cout << "\")";
	    break;
	    }
	default:
	    {
	    cout << "??" << (int)seerp->callcode << "??";
	    break;
	    }
	}
    if (seerp->errcode)
	cout << " = -1 (" << (unsigned int) seerp->errcode << ")\n";
    else
	cout << " = " << seerp->retval << '\n';
    }

static int show_pn(			// Display a pathname from trace
    seerbuf_t*		seerp,		// Trace buffer to display from
    int			pathno)		// Which pathname to display
    {
    char*		pathp;		// Pointer to pathname to print

    pathp = (char*) &seerp->args[0];
    while (--pathno >= 0)
	pathp += (strlen(pathp) + 1 + sizeof seerp->args[0] - 1)
	  / sizeof seerp->args[0] * sizeof seerp->args[0];
    cout << pathp;
    pathp += (strlen(pathp) + 1 + sizeof seerp->args[0] - 1)
      / sizeof seerp->args[0] * sizeof seerp->args[0];
    return (int*) pathp - &seerp->args[0];
    }
