#ifndef lint
static const char Rcs_Id[] =
    "$Id: strdup.cc,v 1.4 1996/02/02 18:29:25 geoff Exp $";
#endif

/*
 * $Id: strdup.cc,v 1.4 1996/02/02 18:29:25 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: strdup.cc,v $
 * Revision 1.4  1996/02/02 18:29:25  geoff
 * Add a really ugly kludge to get around the Sparc strncpy bug.
 *
 * Revision 1.3  1995/11/10  03:00:27  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.2  1995/09/15  00:06:39  geoff
 * Change the RCS ID to be a const.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "zzinlinefunctions.hh"
#include <stdlib.h>

#ifndef NULL
#define NULL		0
#endif // NULL

char*			strdup(const char* string, int size);

extern void		nomem(const char* where, int line, unsigned int size);

char*			strdup(		// Allocate and copy a string
    const char*		string,		// String to be copied
    int			size)		// Optional size of the string
    {
    char*		newstr;		// String to be returned

    if (size == 0)
	size = strlen(string);
    size++;
    newstr = (char*)malloc(size);
#ifdef __sparc__
    /*
     * The SunOS 4.1.1 version of strncpy has a bug which can cause it
     * to over-read the source string if it is over 8 bytes long and
     * misaligned with the destination.  This can cause a segmentation
     * violation if the source string happens to end on a page
     * boundary.  To work around this, if our newly-allocated memory
     * happens to end on or near a page boundary, we re-allocate it
     * four bytes larger.  UGH!  GHK, 1-3-96.
     */
    while (newstr != NULL
      &&  ((unsigned long)(newstr + size + 3) & 0xffc) == 0)
	{
	free(newstr);
	size += 4;
	newstr = (char*)malloc(size);
	}
#endif // __sparc__
    if (newstr == NULL)
	nomem("strdup", __LINE__, size);
    return strcpy(newstr, string);
    }
