#ifndef _STREAMSET_HH
#define _STREAMSET_HH
/*
 * $Id: streamset.hh,v 1.7 1998/10/23 23:15:56 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: streamset.hh,v $
 * Revision 1.7  1998/10/23 23:15:56  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.6  1996/09/11 23:00:23  geoff
 * Convert from dlists to slists.
 *
 * Revision 1.5  1996/05/03  01:13:09  geoff
 * Update the copyright.
 *
 * Revision 1.4  1996/02/02  18:29:27  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.3  1995/11/10  03:00:30  geoff
 * Don't include the icc file, since zzinlinefunctions now does that.
 *
 * Revision 1.2  1995/03/29  21:05:13  geoff
 * Switch to byte-swapped streams.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "cdecls.hh"
#include "list.hh"
#include <sys/types.h>
#include <sys/socket.h>

/*
 * Class used to describe a group of open streams
 */

#ifndef NULL
#define NULL		0
#endif // NULL

class StreamSet
    {
    public:
	inline		StreamSet();
			~StreamSet();
	void		parseFds(const char* descriptor);
					// Parse ASCII into a stream set
	void		addFd(int fd);
					// Add a newly-opened FD to set
	int		removeFd(int fd);
					// Remove an FD from the set
	char*		saveFds() const; // Turn set into an ASCII string
	inline bool	hasFd(int fd) const;
					// See if fd is in stream's fd_set
	inline bool	isEmpty() const; // Test for empty stream list
	class ByteSwappedStream& streamFor(int fd) const;
					// Return stream for given FD
    private:
	SList		streams;	// List of all streams
	fd_set		fds;		// Fd list for selecting
    };

#define _STREAMSET_ICC_NEEDED

#endif // _STREAMSET_HH
