/*
 * $Id: zinlinefunctions.hh,v 1.7 1998/10/23 23:15:57 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: zinlinefunctions.hh,v $
 * Revision 1.7  1998/10/23 23:15:57  geoff
 * Rewrite as a small prototype prefix for the real header file, which is
 * now generated automatically.
 *
 * Revision 1.6  1998/02/19 02:10:14  geoff
 * Add support for having FileState in a separate module.
 *
 * Revision 1.5  1996/05/03 01:13:11  geoff
 * Update the copyright.
 *
 * Revision 1.4  1996/04/10  06:09:39  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.3  1996/02/02  18:29:29  geoff
 * Add support for the modified LRU managers.
 *
 * Revision 1.2  1995/12/19  00:35:09  geoff
 * Add filehistory.icc and process.icc.  Don't conditionalize based on
 * whether the file has been previously included, since in some cases it
 * *needs* to be included twice, and all of its children have
 * double-inclusion protection.
 *
 * Revision 1.1  1995/11/10  02:09:40  geoff
 * Initial revision
 *
 */

/*
 * Because of the complexity of inter-structure references, it is
 * almost impossible to arrange include-file and declaration orders on
 * a per-header-file basis such that all declarations have been seen
 * by the compiler before the inline functions that need them.  In
 * other words, if foo.hh includes foo.icc, there is a strong chance
 * (i.e., a certainty) that foo.icc will use some structure "bar" that
 * hasn't yet been elaborated, but if foo.hh includes bar.hh, then
 * bar.icc will refer to something in foo.hh that hasn't yet been
 * declared.
 *
 * The solution to this problem is to simply defer the declaration of
 * all inline functions until after all .hh files have been included.
 * This is the purpose of the file zzinlinesfunctions.hh:  if it is
 * the last include file named, it will define all necessary inline
 * functions, and there will be no circular dependencies because all
 * necessary structures will also have been defined.
 *
 * A more difficult problem is that the order of inline declarations
 * themselves is fragile.  An ANSI-conforming compiler is allowed to
 * insist on having seen the inline declaration for a function before
 * seeing a call to that function.  With multiple inlines, this
 * becomes a serious problem; a file "foo.icc" may require that some
 * functions from "bar.icc" precede it, and others follow it.
 *
 * To solve the second problem, we dynamically generate
 * zzinlinefunctions.hh, carefully arranging the order of all inline
 * functions so that these dependencies can be satisfied.  To do so,
 * we use "tsort" to generate an ordered list of inlines, extract the
 * various functions from the files in which they are defined, and
 * create zzinlinesfunctions.hh dynamically.  This results in a very
 * large include file, but it guarantees that there will be no
 * dependency problems.  See the make_inlines script for more
 * information on how the file is actually generated.
 */
 
